IMPLEMENTATION MODULE BoyerMoore;

(*$R-,S-,J+*)
(*==============================================================*
 * Modul:               Stringsuchen nach Boyer-Moore           *
 * Autor:               Johannes Gttker-Schnetmann             *
 * erstellt am:         18.04.1992                              *
 * letzte nderung am:  19.04.1992                              *
 * Version:             1.0                                     *
 * Interne Version:     V#0001                                  *
 *==============================================================*

 *----------------------------------------------------------------------------
 * Datum    Vers. Autor  nderung (Arbeitsbericht)                            
 *----------------------------------------------------------------------------
 *----------------------------------------------------------------------------
 *)

(*
TYPE LongString = ARRAY[0..MAX(LONGINT)] OF CHAR;
TYPE longPtr    = POINTER TO longString;

TYPE tableType  = ARRAY[0..377C] OF INTEGER;
    (* Tabelle der Zeichenpositionen im String (von hinten) *)
*)

PROCEDURE LCAP (ch: CHAR; gross : BOOLEAN): CHAR;
BEGIN
  IF gross THEN
    RETURN CAP (ch)
  ELSE RETURN ch;
  END;
END LCAP;

PROCEDURE LCAPS (VAR s: ARRAY OF CHAR; L: INTEGER);
VAR i:      INTEGER;
BEGIN
  IF L > 0 THEN
    FOR i := 0 TO L - 1 DO
      s [i] := LCAP (s [i], TRUE);
    END (* FOR *);
  END (* IF *);
END LCAPS;

PROCEDURE InitTable(VAR table  : tableType;      (* Tabelle fr Boyer-Moore-Suche *)
                    VAR subStr : ARRAY OF CHAR;  (* Zu suchender String           *)
                        len    : CARDINAL;       (* Lnge dieses Strings          *)
                        gross  : BOOLEAN;        (* gro-klein unterscheiden      *)
                        reverse: BOOLEAN);       (* Suche von hinten nach vorne   *)
(* len mu grer als 0 sein! *)
(* Wenn gross TRUE ist, wird der String entsprechend gewandelt *)
VAR c : CHAR; j : INTEGER;
    l : INTEGER;
BEGIN
  IF ~gross THEN LCAPS(subStr, len); END;
  FOR c := 0C TO 377C DO
    table [c] := len; (* Tabelle vor-initalisieren *)
  END (* FOR *);
  IF reverse 
  THEN
    (* subStr swappen *)
    l := len - 1;
    FOR j := 0 TO (l DIV 2) DO
      c := subStr[j];
      subStr[j] := subStr[l-j];
      subStr[l-j] := c;
    END;
  END;
  (* Fuer jede Position im Suchwort subStr wird unter dem Index des
     jten Zeichens in die Tabelle der Abstand vom Ende des Such-
     wortes eingetragen. *)
  FOR j := 0 TO INTEGER(len)-2 DO
  (* bis len-1 wren alle Arrayelemente, die letzte stimmt aber auch so schon *)
    table [subStr [j]] := INTEGER(len) - j - 1;
  END (* FOR *);
  IF reverse 
  THEN
    (* subStr swappen *)
    l := len - 1;
    FOR j := 0 TO (l DIV 2) DO
      c := subStr[j];
      subStr[j] := subStr[l-j];
      subStr[l-j] := c;
    END;
  END;
END InitTable;

PROCEDURE Pos (start  : LONGINT;         (* Startposition im verwendeten Puffer *)
               str    : longPtr;         (* Zeiger auf den Pufferbereich        *)
               pLen   : LONGINT;         (* Lnge dieses Bereiches              *)
           REF substr : ARRAY OF CHAR;   (* gesuchter String                    *)
               len    : INTEGER;         (* Lnge dieses Strings                *)
               gross  : BOOLEAN;         (* gross-klein unterscheiden?          *)
           REF table  : tableType): LONGINT; (* Mit InitTable angelegte Tabelle *)
(* Da die zu verarbeitenden Speicherblcke auch recht gro werden knnen sollen :-) *)
(* hier die Deklaration als LONGINT. Man knnte die Suche notfalls auch stckeln,   *)
(* dann knnte es aber Probleme geben, wenn der zu suchende String gerade auf der   *)
(* Blockgrenze steht. Lsung dazu: Einfach um 32k - Stringlnge nur vorrcken, das  *)
(* wre mir aber trotzdem im Moment etwas zuviel Aufwand.                           *)
VAR (*$Reg *) index, k, j : LONGINT;
    ch:             CHAR;
    res : LONGINT;
        
BEGIN
  IF LONGINT(LONG(len)) > pLen THEN RETURN pLen+1 END;
  index := LONGINT(LONG(len)) + start;
  (* Index, bei dem die Suche startet (Wortende) *)
  REPEAT
    j := len; k := index; (* Laufindizes setzen *)
    REPEAT
      DEC (k); DEC (j);
    UNTIL (j < 0) OR (substr[SHORT(j)] # LCAP(str^[k], ~gross));
      (* vergleicht substr[j] mit str[k] bis Ungleichheit oder
         Suchstring erschoepft *)
    IF (j >= 0) THEN INC(index, LONGINT(LONG(table[LCAP(str^[index-1], ~gross)])));
    END (* IF nicht gefunden *);
  UNTIL (j < 0) OR (index > pLen);
  IF (j < 0) & (index <= pLen) THEN
    RETURN index - LONGINT(LONG(len));
  ELSE
    RETURN pLen + 1;
  END (* IF *);
END Pos;

PROCEDURE BackPos (start  : LONGINT;         (* Startposition im verwendeten Puffer *)
                   str    : longPtr;         (* Zeiger auf den Pufferbereich        *)
                   pLen   : LONGINT;         (* Lnge dieses Bereiches              *)
               REF substr : ARRAY OF CHAR;   (* gesuchter String                    *)
                   len    : INTEGER;        (* Lnge dieses Strings                *)
                   gross  : BOOLEAN;         (* gross-klein unterscheiden?          *)
               REF table  : tableType): LONGINT; (* Mit InitTable angelegte Tabelle *)
(* Da die zu verarbeitenden Speicherblcke auch recht gro werden knnen sollen :-) *)
(* hier die Deklaration als LONGINT. Man knnte die Suche notfalls auch stckeln,   *)
(* dann knnte es aber Probleme geben, wenn der zu suchende String gerade auf der   *)
(* Blockgrenze steht. Lsung dazu: Einfach um 32k - Stringlnge nur vorrcken, das  *)
(* wre mir aber trotzdem im Moment etwas zuviel Aufwand.                           *)
VAR (*$Reg*)index, k, j : LONGINT;
    (*$Reg*)ch:             CHAR;
BEGIN
  IF LONGINT(LONG(len)) > pLen THEN RETURN pLen+1 END;
  index := start;
      (* Index, bei dem die Suche startet (Wortende) *)
  REPEAT
    j := 0; k := index; (* Laufindizes setzen *)
    WHILE (j < LONG(len)) & (substr[SHORT(j)] = LCAP(str^[k], ~gross)) DO
      INC (k); INC(j);
    END;
      (* vergleicht substr[j] mit str[k] bis Ungleichheit oder
         Suchstring erschoepft *)
    IF (j < LONG(len)) THEN DEC(index, LONGINT(LONG(table[LCAP(str^[index], ~gross)])));
    END (* IF nicht gefunden *);
  UNTIL (j >= LONG(len)) OR (index < 0);
  IF (j >= LONG(len)) & (index >= 0) THEN
    RETURN index;
  ELSE
    RETURN pLen + 1;
  END (* IF *);
END BackPos;

END BoyerMoore.
