IMPLEMENTATION MODULE Clip;

(*==============================================================*
 * Modul:               Clipboard-Modul, nach Atari-Doku        *
 * Autor:               Johannes Gttker-Schnetmann             *
 * erstellt am:         10.02.1992                              *
 * letzte nderung am:  09.03.1992                              *
 * Version:             1.1                                     *
 * Interne Version:     V#0003                                  *
 *==============================================================*

 *----------------------------------------------------------------------------
 * Datum    Vers. Autor  nderung (Arbeitsbericht)
 *----------------------------------------------------------------------------
 * 10.02.92 0001  JGS    Erstellen
 * 08.03.92 0002  JGS    Environment bercksichtigen 
 * 09.03.92 0003  JGS    ScrapClear/ScrapDirty
 * 07.09.92 0004  DS     Bei GetScrap wird ein \CLIPBRD\ genommen, falls auf einem
 *                       Laufwerk vorhanden. Erst danach wird beim User angefragt.
 * 08.12.92 0005  DS     Bei folderExists wird auch Rcksicht auf gelockte Laufwerke
 *                       genommen, d.h. bei ACCDN wird nicht nachgefragt
 * 11.09.94 0006  DS     Wenn wir etwas ins Clipboard schreiben, wird ein Broadcast 
 *                       mit der Message SCCHANGED gemacht.
 *----------------------------------------------------------------------------
 *)

FROM SYSTEM IMPORT ADR, ADDRESS;
FROM Void IMPORT v;
IMPORT CatFiles;
IMPORT MagicAES;
IMPORT MagicDOS;
IMPORT mtDir;
IMPORT mtAppl;
IMPORT mtAlerts;
IMPORT MagicStrings;
IMPORT mtCommand;
IMPORT Mintbind;
IMPORT CatTypes;
IMPORT CatGlobal;

VAR scrapDirty : BOOLEAN;

PROCEDURE ScrapClear(REF f, not : ARRAY OF CHAR);
(* leert das Clipboard falls Pfad leer, sonst den bergebenen Pfad *)
VAR path : CatTypes.pathStrType;
BEGIN
  MagicStrings.Assign(f, path);
  (* Falls kein Pfad bergeben wurde, dann mal anfragen *)
  IF (path[0] = 0C) & ~GetScrap(path) THEN RETURN END;
  CatFiles.ClearPath (path, 'SCRAP.*', not);
  scrapDirty := FALSE;
END ScrapClear;

PROCEDURE folderExist(l : CARDINAL; REF pfad : ARRAY OF CHAR):BOOLEAN;
VAR dta : MagicDOS.DTA;
    oDta: ADDRESS;
    res : INTEGER;
    path: CatTypes.pathStrType;
    dHandle: LONGINT;
BEGIN
  MagicStrings.Assign (pfad, path);
  IF l = 0 THEN RETURN FALSE END;
  path[l-1] := 0C; (* Den backslash am Ende weglassen *)
  IF ~CatGlobal.isMintDomain
  THEN
    oDta := MagicDOS.Fgetdta ();
    MagicDOS.Fsetdta (ADR(dta));
    res := MagicDOS.Fsfirst (path, {MagicDOS.Folder});
    MagicDOS.Fsetdta (oDta);
    RETURN (res = 0) OR (res = MagicDOS.EAccDn)
  ELSE
    dHandle := Mintbind.Dopendir (path, 0);
    IF dHandle >= 0
    THEN
      v.lint := Mintbind.Dclosedir (dHandle);
      RETURN TRUE;
    END;
    RETURN FALSE;
  END;
  (*
  mtDir.SearchParas(pfad, {MagicDOS.Folder}, ADR(dta), TRUE);
  RETURN mtDir.Found();
  *)
END folderExist;

PROCEDURE GetScrap(VAR f : ARRAY OF CHAR):BOOLEAN;
(* Liefert den aktuellen Clipboard-Ordner *)
VAR s, scrap : CatTypes.pathStrType;
    name, n  : CatTypes.nameStrType;
    suff     : CatTypes.extStrType;
    drv      : CARDINAL;
    z        : CARDINAL;
    b        : BITSET;
    setNew   : BOOLEAN;

BEGIN
  setNew := FALSE; (* vorlufig nicht neu setzen *)
  f[0]   := 0C;    (* Erstmal entwerten *)

  b := MagicAES.ScrpRead(scrap);

  IF b = {} THEN RETURN FALSE END;
  (* if scrp_read() returns an error code of zero, then the 
     application should act as though nothing is available
     on the clipboard *)

  (* Jetzt noch im Environment nachsehen, falls bisher noch nichts gefunden *)
  IF scrap[0] = 0C THEN
    IF ~mtCommand.EnvVar(mtCommand.EnvClipbrd, scrap) THEN
      scrap[0] := 0C;
    END;
    MagicStrings.CAPS(scrap);
  END;

  IF scrap[0] # 0C THEN
    IF scrap[MagicStrings.Length(scrap)-1] # '\'
    THEN
      MagicStrings.Append('\', scrap);
    END;
      
    mtDir.SplitPath(scrap, s,n,suff); (* In die Bestandteile aufspalten *)
    z := MagicStrings.Length(s);

    setNew := (n[0] # 0C) OR (suff[0] # 0C) OR (z = 0) OR (s[z-1] # '\');
    (* In all diesen Fllen mu das Clipboard neu gesetzt werden *)

    IF (z > 0) & (s[z-1] # '\') THEN
      MagicStrings.Append('\', s);
      INC(z); (* Fr den Test unten *)
    END;

    IF ~folderExist(z, s) THEN
      (* Dann probieren, ob das ganze evtl. schon ein
        Pfad ist und nur der Backslash vergessen wurde *)
      z := LENGTH(scrap);
      MagicStrings.Append('\', scrap);
      IF ~folderExist(z+1, scrap) THEN
        scrap[0] := 0C; (* Dann User fragen  *)
      END;
    ELSE
      MagicStrings.Assign(s, scrap);
    END;
  END;

  IF (scrap[0] = 0C) THEN (* User fragen *)
    (* Zunchst mal bei den existierenden Laufwerken nach dem Rechten sehen..*)
    MagicDOS.Dsetdrv(MagicDOS.Dgetdrv(),v.lbset);
    IF     2 IN v.lbset THEN drv := 2 (* erstmal auf C:\ *)
    ELSIF  0 IN v.lbset THEN drv := 0 (* dann auf A:\ *)
    ELSE (* und jetzt mal suchen, ob sonst vielleicht noch eins zu finden ist *)
      z := 0;
      WHILE ~(z IN v.lbset) & (z < 31) DO INC(z) END;
      IF ~(z IN v.lbset) THEN (* Da kann man nichts machen.. *)
        v.int := mtAlerts.Alert(1, '[3][|Kein Laufwerk gefunden!][[Abbruch]');
        RETURN FALSE
      END;
    END;

    scrap := CatTypes.rawClip;
    scrap[0] := CHR(ORD('A')+drv); (* passendes Laufwerk aussuchen *)
    (* mal nachsehen, ob's dort ein CLIPBRD gibt *)

    IF ~folderExist(LENGTH(CatTypes.rawClip), scrap) 
    THEN
      IF mtAlerts.Alert(1, '[7][|Klemmbrett-Ordner|nicht festgelegt!][[Suchen|:[Abbruch]') = 1
      THEN 
        (* nur das Laufwerk vorschlagen *)
      
        scrap[3] := 0C;
        MagicStrings.Append('*.*', scrap);
        name := '';
        MagicAES.WindUpdate(MagicAES.BEGMCTRL);
        IF ~mtDir.GetDir(scrap, name, 'Clipboard-Ordner aussuchen') THEN (* Mal beim User fragen *)
          MagicAES.WindUpdate(MagicAES.BEGMCTRL);
          RETURN FALSE
        END;
        MagicAES.WindUpdate(MagicAES.ENDMCTRL);
  
        mtDir.SplitPath (scrap, s, n, suff); (* Zerlegt Pfad in seine Bestandteile *)
        MagicStrings.Assign(s, scrap);
        (* Doppelte Deklaration, das sonst in SplitPath Probleme auftreten knnen *)
  
        IF name[0] # 0C THEN(* Es wurde ein Dateiname eingegeben *)
          IF (mtAlerts.Alert(1, '[3][|Clipboard-Ordner anlegen?][[Ok|:[Abbruch]') = 1)
          THEN (* Versuchen, diesen anzulegen *)
            IF ~CatFiles.SetPath(scrap) OR (MagicDOS.Dcreate(name) < 0) THEN
              v.int := mtAlerts.Alert(1, '[3][|Das Klemmbrett konnte|nicht angelegt werden!][[Abbruch]');
              RETURN FALSE
            END;
            MagicStrings.Append('\', name);
            MagicStrings.Append(name, scrap);
          ELSE
            RETURN FALSE
          END;
        END;
        setNew := TRUE; (* Jetzt in jedem Fall setzen *)
      ELSE
        RETURN FALSE
      END;
    ELSE
      (* folder existiert, neu setzen darauf! *)
      setNew := TRUE;
    END;
  END;
  IF setNew THEN MagicAES.ScrpWrite(scrap); (* und mal alles anmelden *) END;
  MagicStrings.Assign(scrap, f); (* .. und am Ende das Resultat zuweisen *)
  RETURN TRUE
END GetScrap;

PROCEDURE ScrapDirty(REF f : ARRAY OF CHAR):BOOLEAN;
(* Ist etwas im Clipboard? *)
(* d.h. sieht entweder im bergebenen Pfad oder im Clipboard nach *)
(* ob ein SCRAP.TXT vorhanden ist *)
VAR path : CatTypes.pathStrType;
BEGIN
  MagicStrings.Assign(f, path);
  (* Falls kein Pfad bergeben wurde, dann mal anfragen *)
  IF (path[0] = 0C) & ~GetScrap(path) THEN
    scrapDirty := FALSE
  ELSE
    MagicStrings.Append(CatTypes.scrapFile, path);
    scrapDirty := CatFiles.Exists (path);
  END;
  RETURN scrapDirty;
END ScrapDirty;

PROCEDURE ScrapWritten (type: BITSET; defExt: ARRAY OF CHAR);
(* Wir haben selber etwas in's Clipboard geschrieben *)
  VAR msg : ARRAY [0..7] OF INTEGER;
      strPtr : CatTypes.Str255Ptr;
      appName : CatTypes.String127;
      i       : INTEGER;
      otherId : INTEGER;
      pType   : INTEGER;
BEGIN
  scrapDirty := TRUE;
  (* Bei MultiTOS oder Magix >= 2.0 *)
  IF CatGlobal.multiTOS OR
     (CatGlobal.magIx & (CatGlobal.magIxVer >= $200))
  THEN
    (* Nachricht basteln *)
    msg[0] := MagicAES.SCCHANGED;
    msg[1] := mtAppl.ApplIdent;
    msg[2] := 0;
    msg[3] := INTEGER(type);
    strPtr := ADR (msg[4]);
    IF HIGH (defExt) >= 4 THEN defExt [4] := 0C; END;
    MagicStrings.Assign (defExt, strPtr^);
    msg[6] := 0;
    msg[7] := 0;
    IF CatGlobal.multiTOS
    THEN
      v.int := MagicAES.ExtShelWrite (MagicAES.BroadcastMessage, 0, 0, ADR(msg), "");
    ELSE
      (* Mittels ApplSearch alle Applikationen durchgehen 
       *)
      FOR i := 0 TO 12 DO appName[i] := 0C; END;
      IF MagicAES.ApplSearch (MagicAES.APFIRST, appName, pType, otherId)
      THEN
        REPEAT 
          IF (pType # MagicAES.APTSYSTEM) & 
             (otherId # mtAppl.ApplIdent)
          THEN
            (* Message verschicken *)
            MagicAES.ApplWrite (otherId, 16, msg);
          END;
          FOR i := 0 TO 12 DO appName[i] := 0C; END;
        UNTIL ~MagicAES.ApplSearch (MagicAES.APNEXT, appName, pType, otherId);
      END;
    END;
  END;
END ScrapWritten;

PROCEDURE SomethingInScrap (): BOOLEAN;
(* Sieht intern in einer Variablen nach. Im Zweifelsfall eher 
 * ScrapDirty benutzen
 *)
BEGIN
  RETURN scrapDirty
END SomethingInScrap;

PROCEDURE ScrapIsDirty();
BEGIN
  scrapDirty := TRUE;
END ScrapIsDirty;

BEGIN
  scrapDirty := FALSE;
END Clip.
