DEFINITION MODULE ConfVars;

(* Dieses Modul verwaltet Konfigurationsvariablen innerhalb einer Liste.
 * Fr jede Variable wird ein Eintrag angelegt, in dem der Name, der Typ
 * und der Wert festgehalten wird.
 * Der Zugriff ist jederzeit ber die Funktionen dieses Moduls mglich.
 * Beim Schreiben einer Variable wird der alte Inhalt ersetzt. Eine Typwandlung
 * einer Variablen ist nicht mglich!
 *)

FROM GrafBase    IMPORT Rectangle;
FROM Lists       IMPORT List;
FROM mtTextfiles IMPORT TEXTFILE;

(*------------- Variablen lesen ohne Default-Wert -------------------*)

PROCEDURE GetConfigString (REF name : ARRAY OF CHAR; VAR val : ARRAY OF CHAR) : BOOLEAN;
(* Liest eine String-Variable. Liefert bei TRUE den Wert in val zurck 
 *)

PROCEDURE GetConfigInt (REF name : ARRAY OF CHAR; VAR val : INTEGER): BOOLEAN;
(* Liest eine Integer-Variable. Liefert bei TRUE den Wert in val zurck 
 *)

PROCEDURE GetConfigLongInt (REF name : ARRAY OF CHAR; VAR val : LONGINT): BOOLEAN;
(* Liest eine LongInteger-Variable. Liefert bei TRUE den Wert in val zurck 
 *)

PROCEDURE GetConfigBool (REF name : ARRAY OF CHAR; VAR val : BOOLEAN): BOOLEAN;
(* Liest eine Bool'sche Variable. Liefert bei TRUE den Wert in val zurck 
 *)

PROCEDURE GetConfigRect (REF name : ARRAY OF CHAR; VAR val : Rectangle): BOOLEAN;
(* Liest eine Rechteck-Variable. Liefert bei TRUE den Wert in val zurck 
 *)

(*------------- Variablen lesen mit Default-Wert -------------------*)

PROCEDURE GetConfDefStr (REF name : ARRAY OF CHAR; VAR val : ARRAY OF CHAR;
                           REF default : ARRAY OF CHAR);
(* Liest eine String-Variable. Wenn die Variable nicht gefunden wird,
 * dann wird der Default-Wert zurckgegeben.
 *)

PROCEDURE GetConfDefInt (REF name : ARRAY OF CHAR; VAR val : INTEGER;
                         default : INTEGER);
(* Liest eine Integer-Variable. Wenn die Variable nicht gefunden wird,
 * dann wird der Default-Wert zurckgegeben.
 *)

PROCEDURE GetConfDefLongInt (REF name : ARRAY OF CHAR; VAR val : LONGINT;
                             default : LONGINT);
(* Liest eine Integer-Variable. Wenn die Variable nicht gefunden wird,
 * dann wird der Default-Wert zurckgegeben.
 *)

PROCEDURE GetConfDefBool (REF name : ARRAY OF CHAR; VAR val : BOOLEAN;
                          default : BOOLEAN);
(* Liest eine Bool'sche Variable. Wenn die Variable nicht gefunden wird,
 * dann wird der Default-Wert zurckgegeben.
 *)

PROCEDURE GetConfDefRect (REF name : ARRAY OF CHAR; VAR val : Rectangle;
                          default : Rectangle);
(* Liest eine Rechteck-Variable. Wenn die Variable nicht gefunden wird,
 * dann wird der Default-Wert zurckgegeben.
 *)

(*------------- Variablen schreiben -------------------*)

PROCEDURE SetConfigString (REF name : ARRAY OF CHAR; REF val : ARRAY OF CHAR) : BOOLEAN;
(* Schreibt eine String-Variable. Bei TRUE konnte die Variable angelegt werden.
 *)

PROCEDURE SetConfigInt (REF name : ARRAY OF CHAR; val : INTEGER): BOOLEAN;
(* Schreibt eine Integer-Variable. Bei TRUE konnte die Variable angelegt werden.
 *)

PROCEDURE SetConfigLongInt (REF name : ARRAY OF CHAR; val : LONGINT): BOOLEAN;
(* Schreibt eine Integer-Variable. Bei TRUE konnte die Variable angelegt werden.
 *)

PROCEDURE SetConfigBool (REF name : ARRAY OF CHAR; val : BOOLEAN): BOOLEAN;
(* Schreibt eine Bool'sche Variable. Bei TRUE konnte die Variable angelegt werden.
 *)

PROCEDURE SetConfigRect (REF name : ARRAY OF CHAR; val : Rectangle ): BOOLEAN;
(* Schreibt eine Rechteck-Variable. Bei TRUE konnte die Variable angelegt werden.
 *)

(*------------- Variable lschen -------------------*)

PROCEDURE DeleteConfigVar (REF name : ARRAY OF CHAR);
(* Lscht die Config-Variable name, falls vorhanden, 
 * und dealloziert smtlichen bentigten Speicher
 *)

PROCEDURE DeleteAllVars ();
(* Lscht alle Variablen der aktuellen Variablenliste
 *)

PROCEDURE CloseConfig ();
(* Wird aufgerufen, wenn das Programm beendet wird.
 * Dadurch wird ein eventuell vorhandenes Request-File
 * geschlossen
 *)

(*------------- File I/O -------------------*)

PROCEDURE ReadConfigFile (REF filename : ARRAY OF CHAR) : BOOLEAN;
(* Liest ein Config-File ein. Danach stehen alle Variablen aus dem File
 * in der internen Liste, falls das Lesen erfolgreich war.
 *)

TYPE infoProc = PROCEDURE (REF (* info *) TEXTFILE);

PROCEDURE emptyInf( REF info: TEXTFILE);
(* Leere Infoproc, gibt nichts aus *)

PROCEDURE WriteConfigFile (REF filename : ARRAY OF CHAR; info: infoProc) : BOOLEAN;
(* Schreibt alle Config-Variablen in eine Datei. Diese Datei wird 
 * dabei angelegt.
 * ber die infoProc knnen am Anfang der Datei noch zustzliche
 * Informationen abgespeichert werden. Dazu mssen die Ausgabeprozeduren von
 * mtTextfiles benutzt werden.
 *)


(*------------- Modulkonfiguration ---------------*)

PROCEDURE ConfigConfigModule (useTmpVars : BOOLEAN);
(* Konfiguriert das Modul um, welche Variablenliste benutzt
 * wird. Default sind die normalen Variablen.
 *)

PROCEDURE SetConfigVarList (l : List);
(* Konfiguriert das Modul um, welche Variablenliste benutzt
 * wird. 
 *)

PROCEDURE ResetConfigVarList();
(* Setzt die interne Variablenliste wieder auf die 
 * normalen Configvariablen zurck. 
 *)

END ConfVars.
