DEFINITION MODULE data;

(*==============================================================*
 * Modul:               CAT-Datenbank, neue Version             *
 * Autor:               Johannes Gttker-Schnetmann             *
 * erstellt am:         26.07.1991                              *
 * letzte nderung am:  13.08.1993                              *
 * Version:             1.0                                     *
 * Interne Version:     V#0018                                  *
 *==============================================================*

  Das Modul implementiert eine nach Gruppen getrennte Datenbank fr die 
  Nachrichten der Maus. Weiterhin werden pro Gruppe drei Positionen 
  verwaltet: letzte gelesene Position, Position der ersten neuen Msg und 
  letzte Msg der Datenbank.

  Bei ID's, Namen, und Betreff vorgesehene max. Stringslnge ist im
  Moment 1024. (String255)

  Achtung, die GruppenNummer ist ab jetzt undefiniert und mu ber GroupNumber
  erfragt werden. Naja vielleicht kann man auch einfach auf die neue Nummerierung
  Rcksicht nehmen, aber besser wre undefiniert!

 *----------------------------------------------------------------------------
 * Datum    Vers. Autor  nderung (Arbeitsbericht)
 *----------------------------------------------------------------------------
 * 26.07.91 0001  JGS    Erstellen, erster Vorschlag
 * 14.09.91 0002  JGS    GroupHandles, statt interner Verwaltung
 * 02.10.91       JGS    Doku verbessert
 * 03.10.91 0003  JGS    Kleinere nderungen an den Defintionen
 * 06.10.91 0004  JGS    dito
 * 11.10.91 0005  JGS    longString auf 1024 Zeichen verlngert
 *                       erste Definition fr AppendMessage
 * 19.10.91 0006  JGS    Doku verbessert, Stringnamen gendert
 * 27.12.91 0007  JGS    Definition AppendMessage/AppendPersonal
 *                       private = 0, seit neuestem! Die entsprechenden Proc's
 *                       knnen jetzt damit umgehen.
 * 29.12.91 0008  JGS    Neue Definition der Schreibproc's, Gruppenhandle mu
 *                       bergeben werden.
 * 10.03.92 0009  JGS    Einige Sachen nach CatTypes verlagert sowie
 *                       EstimateNecessaryMemory
 * 19.03.92 0010  JGS    Listen von auen erweiterbar, fr parser
 * 26.03.92 0011  JGS    Umstellung (mtTextfiles9 fr Behandlung der eigenen,
 *                       persnlichen Nachrichten
 * 02.04.92       JGS    nderung bei der Definition von "FirstNewMsg"
 * 04.04.92 0012  JGS    Deklaration der Listen gendert, damit sie auch
 *                       im Hauptmodul direkt verwendet werden knnen
 * 15.04.92 0013  JGS    Neue Proc PreCheck fr bessere Strukturierung
 * 19.04.92 0014  JGS    Suchen-Deklaration
 * 04.10.92 0015  JGS    Strings in "MessageType" jetzt als Pointer
 * 20.11.92 0016  JGS    Totalabbruch bei Schreibfehler, neues Flag..
 * 23.12.92 0017  JGS    Behandlung der ungelesenen Msgs begonnen
 * 31.12.92       JGS    Baum durchlaufen, Baum lschen..
 * 09.01.93 0018  JGS    Vererben-Flag, Gefiltert-Flag rein; Gefunden-Flag raus
 * 13.08.93       JGS    Suchroutinendefinition erweitert.
 * 17.08.93       DS     MessageType um Datum im MausTauschformat erweitert
 *----------------------------------------------------------------------------
 *)

FROM SYSTEM IMPORT ADDRESS;
FROM Lists    IMPORT List;
FROM CatTypes IMPORT String255, Str255Ptr, DateType, BigTextPtr;
FROM dataSys  IMPORT pBlockPtr, onePos;
IMPORT mtTextfiles;

TYPE PosType     = (nextMess, previousMess, nextKom, prevKom, downKom, upKom);

TYPE OneGroupHandle;

(* Achtung, nderungen hier betreffen auch das Acc-Protokoll!! *)
TYPE tDistribution = (dNone, dLokal, dMausNet, dNet);

     MessageType = RECORD
       MailNr,
       MailAnz   : CARDINAL;
       KommentierteID : String255;
       fromOther : BOOLEAN; (* Wurde die ID von der anderen Msg gelesen? *)
       MailID,
       Betreff,
       Absender,
       Empfaenger,
       mid,
       rid,
       box,
       gate,
       mime,
       followupTo,
       replyTo,
       sender,
       name     : Str255Ptr; (* zeigen auf Positionen in "InfoStrings" *)
       Gruppe   : CARDINAL;
       Datum,
       StatusDatum            (* nur persnliche *)
                : DateType;
       infoLen  : CARDINAL;   (* letztes Byte in <InfoStrings> *)
       textLen  : CARDINAL;   (* letztes Byte im <Text> *)

       StatusBits : BITSET;

       up,
       down,
       left,                  
       right  : CARDINAL;
       KommentarAnzahl
                : CARDINAL;
       statusDate      : LONGCARD;     (* Bei persnlichen: Statusdatum im MT-Format *)
       tauschDate      : LONGCARD;     (* Datum im MausTauschformat *)
       EigeneNachricht : BOOLEAN;
       Status : CHAR;
       Text,
       InfoStrings  : BigTextPtr;    (* Freigeben nicht vergessen.. :-) *)
       distribution : tDistribution;
     END;

    listEntryType = RECORD
                       len      : CARDINAL;
                       number   : CARDINAL;
                       selected : BOOLEAN;
                       gName    : String255;
                     END;
    listEntryPtr  = POINTER TO listEntryType;

    errorType = (noError, notFound, IOError, noMemErr, crcError, fileError);

VAR names  : List; (* beide readOnly! *)

    error  : errorType; (* Fehlerrckgabe *)

PROCEDURE InitDataBase():BOOLEAN;
(* Initialisieren, Fehlermeldung wird selber ausgegeben *)
PROCEDURE ResetDataBase();

PROCEDURE CloseBase():BOOLEAN;
(* Datenbank schlieen, abmelden am Ende *)

PROCEDURE PrepareToWrite();
(* setzt die Positionen so, da nach dem Einfgen etwas hbsches rauskommt :-) *)

PROCEDURE OpenOneGroup(group, add : CARDINAL; wannaWrite : BOOLEAN; VAR handle 
: OneGroupHandle):BOOLEAN;
(* FALSE, wenn nicht geklappt *)
(* add gibt an, fr wieviele neue Nachrichten Platz bentigt wird wenn wannaWrite *)
(* TRUE ist, dann wird auch versucht ein Schreibpuffer fr die DAT anzulegen *)
PROCEDURE CloseOneGroup(VAR handle : OneGroupHandle);
(* Gruppe, die zum Lesen geffnet war schlieen *)
PROCEDURE CloseOneWriteGroup(VAR handle : OneGroupHandle;
                                 newMsg : CARDINAL;
                                 write  : BOOLEAN; 
                             VAR abort  : BOOLEAN);
(* Gruppe, die zum schreiben geffnet war schlieen *)
(* newMsg -> Anzahl der neuen (d.h. nicht gefilterten oder ignorierten) Msgs *)
(* write  -> Puffer wirklich zurckschreiben? Normalerweise nicht bei Abbruch *)
(* abort  -> etwas ist beim Schlieen schiefgelaufen *)

PROCEDURE SetLogDate(VAR date : ARRAY OF CHAR);
(* Datum aus dem Logfile bernehmen *)

PROCEDURE SaveNames ();
(* Sichert die Liste der Namen komplett 
 *)

PROCEDURE AppendName(VAR name : ARRAY OF CHAR);
(* (VAR wg. interner Probleme mit mtTextfiles)
   Neuen Absender bernehmen, meckert selber, falls es nicht klappt *)

(*
PROCEDURE AppendGroup(VAR group : ARRAY OF CHAR):BOOLEAN;
(* Neuen Gruppennamen bernehmen, meckert selber, falls es nicht klappt *)
*)

PROCEDURE SortList (VAR l : List);
(* Gruppen- oder Namensliste sortieren *)


PROCEDURE ReadPos():BOOLEAN;
PROCEDURE WritePos():BOOLEAN;
(* Positionstabelle einlesen und schreiben, Fehlermeldung wird ausgegeben *)

(*
PROCEDURE GroupNumber(REF gName : ARRAY OF CHAR; VAR nr : CARDINAL):BOOLEAN;
(* Gruppennummer zu einem String feststellen *)
(* Wenn FALSE, dann kommt Gruppenanzahl in nr zurck, Anzahl      *)
(* inklusive persnliche                                          *)
PROCEDURE NextGroupNumber(old : CARDINAL):CARDINAL;
(* Nummer der nchsten Gruppe in der Reihenfolge der momentanen Sortierung *)
(* feststellen; 0FFFF = empty falls keine weiter Gruppe vorhanden ist      *)
PROCEDURE GroupName(nr : CARDINAL; VAR name : ARRAY OF CHAR);
(* Gruppenname zu einer Nummer feststellen, liefert im Fehlerfall *)
(* den letzten Eintrag                                            *)
*)

PROCEDURE GetCheckArea (VAR area: onePos);
(* Holt die Prfarea fr das Datum aus dem GRUPPEN.POS
 *)

PROCEDURE SetCheckArea (area: onePos);
(* Setzt die Prfarea fr das Datum in dem GRUPPEN.POS
 *)

PROCEDURE FirstNewMsg(group : CARDINAL):CARDINAL;
(* erste Msg aus dem letzten Outfile in dieser Gruppe *)
(* "empty", wenn keine neue in dieser Gruppe ist      *)
PROCEDURE LastMsgOfGroup(group : CARDINAL):CARDINAL;
(* letzte Msg der Gruppe *)
PROCEDURE lastReadMsgOfGroup(group : CARDINAL):CARDINAL;
(* letzte vom User gelesene Msg dieser Gruppe *)
PROCEDURE unreadMsgCount(group : CARDINAL):CARDINAL;
(* Anzahl der ungelesenen Msgs in der Gruppe *)
PROCEDURE unreadMsgPos(group : CARDINAL):CARDINAL;
(* Position der ersten ungelesenen Msg in der Gruppe *)

PROCEDURE forceUnreadRefresh(handle : OneGroupHandle);
(* erste Ungelesene suchen und durchzhlen *)
PROCEDURE SetLastReadMsg(group, nr : CARDINAL);
(* neue letzte vom User gelesene Msg setzen *)

(* Fehlerrckgabe der nchsten 2 Positionsfunktionen: noError/notFound *)

PROCEDURE NumberOfID(handle : OneGroupHandle; REF ID : ARRAY OF CHAR):CARDINAL;
(* MsgNummer zu einer ID feststellen *)
PROCEDURE NumberOfDate(handle : OneGroupHandle; date : LONGCARD):CARDINAL; 
(* date im Cat-Format, umwandeln mit ConvertDate.StrToDate *)
(* MsgNummer zu einem Datum suchen *)

(* Mgliche Fehler: noError, notFound, IOError *)

PROCEDURE ReadMessage(handle : OneGroupHandle; nr : CARDINAL; VAR mess : MessageType);
(* Eine Msg aus der Datenbank lesen *)
PROCEDURE ReadSmallHeader(handle : OneGroupHandle; nr : CARDINAL; VAR mess : MessageType);
(* Liest nur die Headerinformationen aus der .PAR-Datei. Alle anderen Daten 
 * sind *nicht* initialisiert
 *)
PROCEDURE ReadHeader(handle : OneGroupHandle; nr : CARDINAL; VAR mess : MessageType);
(* Nur einen Teil der Msg lesen, evtl. weitere Funktionen *)
PROCEDURE ReadState(handle : OneGroupHandle; nr : CARDINAL; VAR flags : BITSET);
(* Nur die Flags lesen, fr "Gelesene ignorieren" *)
PROCEDURE ReadPersState(handle : OneGroupHandle; nr : CARDINAL; VAR state : CHAR; VAR own : BOOLEAN; VAR bits : BITSET);
(* Nur den Status einer persnlichen Msg lesen *)
PROCEDURE ReadRightNumber(handle : OneGroupHandle; nr : CARDINAL; VAR right : CARDINAL);
(* Nummer der Nachricht in der Kommentarverkettung nach rechts erfragen *)
PROCEDURE ReadText(handle : OneGroupHandle; nr : CARDINAL; adr : ADDRESS);
(* den Text zur msg in den Puffer lesen, mu natrlich gro genug sein! *)
PROCEDURE SetBits(handle : OneGroupHandle; nr : CARDINAL; bits : BITSET);
(* Statusbits setzen *)
PROCEDURE ChangeState(handle : OneGroupHandle; nr : CARDINAL;NewState : CHAR);
(* StatusChar setzen *)
PROCEDURE HasAnswer(handle : OneGroupHandle; nr : CARDINAL; VAR bits : BITSET):BOOLEAN;
(* Testet, ob diese Msg noch eine Antwort hat, fr grin.grinResetState *)

(* Idee: Bei Fehler Einfgen abbrechen und Fehlerdatei schreiben; 
 * Fehlerbehandlung ber eigenes Programm
 *)


TYPE SearchType = (inBits, inText, changeBits, dontShow, ask, reverse);
(* Die ersten drei geben je an, wo gesucht werden soll. changeBits sagt, *)
(* da die Bits der gefundenen Msg gem der Setzmaske gesetzt werden    *)
(* sollen. Bei dontShow werden alle Msgs bis zum Ende behandlet. Mit     *)
(* reverse kann man die Suchrichtung auf rckwrtes stellen              *)
TYPE SearchSet  = SET OF SearchType;

PROCEDURE ComplexSearch(handle  : OneGroupHandle;
                        start   : CARDINAL;         (* StartNachricht         *)
                        setBits,
                        clearBits  : BITSET;        (* Zusammen Suchmaske     *)
                    VAR str, str2, 
                        str3, str4 : ARRAY OF CHAR; (* Zu suchender String    *)
                        wo1, wo2, 
                        wo3, wo4   : INTEGER;       (* Wo sollen sie stehen?  *)
                        verkn1, verkn2,
                        verkn3     : INTEGER;       (* logische Verknpfung   *)
                        gross      : BOOLEAN;       (* gro=klein?            *)
                        toSetBits,
                        toClearBits  : BITSET;      (* Zusammen Setzmaske     *)
                        what         : SearchSet;   (* Was machen?            *)
                    VAR break : BOOLEAN;            (* Abbruch durch esc?     *)
                    VAR nr : CARDINAL):BOOLEAN;     (* Nummer, falls gefunden *)
(* Sucht eine Nachricht nach den angegebenen Werten, falls sie gefunden wird *)
(* dann ist found TRUE und es wird die Nummer dieser Nachricht zurckgegeben *)
(* str wird evtl. wegen gro-klein-Unterscheidung gendert, deswegen         *)
(* vielleicht einen Backup-String bergeben.                                 *)
(* wo1..wo4, verkn1..verkn3 wie in SearchHelp.. *)

(* bergabetypen fr Appendmessage *)

TYPE stringTypes = (bId, bEZeit, bGruppe, bBSZeit,
                    bWegen, bVon, bAn, bMId, bRId, bBox, bName, bRefNr, 
                    bDistribution, bGate, bMime, bFollowupTo, bReplyTo, bSender, (* Reihenfolge wichtig! *)
                    bText, bTextDatei);
TYPE stringSet = SET OF stringTypes;

(* Neue Reihenfolge fr WatchDog-Protokoll, wegen Kompatibilitt *)
(* WatchDog darf nur gefragt werden, wenn es keine persnliche   *)
(* Msg ist, sonst ist evtl. z.B. pText NIL                       *)
TYPE PtrRecord =
       RECORD
         pGruppe, pVon,         (* Absender,          Gruppe            *)
         pWegen  : Str255Ptr;   (* Betreff,                             *)
         pText   : BigTextPtr;  (* MsgText                              *)
         TextMax : CARDINAL;
         pId,                   (* Maus-ID der Mitteilung               *)
         pRefNr,   pAn,         (* kommentierte Msg,  Empfnger         *)
         pEZeit,   pBSZeit,     (* Eingabezeit,       Bearb.status+Zeit *)
         pMId,     pRId,        (* MessageID,         RId gem. Def      *)
         pBox,                  (* Box gem. Def.                        *)
         pName,                 (* Name gem. Def falls # Absenderangabe *)
         pGate,                 (* Gateway *)
         pMime,                 (* MIME-Zeile *)
         pFollowupTo,           (* FollowUp-To - Zeile *)
         pReplyTo,              (* Reply-To - Zeile *)
         pSender : Str255Ptr;   (* Sender - Zeile *)
         (* Achtung, die Gre dieses Strings wird bei der Verkettung *)
         (* per RId bentigt -> bei nderungen bercksichtigen *)
         txt     : mtTextfiles.TEXTFILE;
         whatsThere : stringSet;
         pDistribution : Str255Ptr; (* Distribution *)
       END;

(* Falls das Bit bTextDatei gesetzt ist, dann soll der Text aus der bergebenen *)
(* Datei bernommen werden, er steht dann nicht im Speicher                     *)
(* !!Achtung: Weiterhin ist dieses Bit der Anzeiger dafr, da es sich bei      *)
(* dieser  Message um eine eigene handelt, die gerade eingefgt werden soll.    *)

TYPE WhatsThat = (own, personal, status, normal, garbage);

PROCEDURE PreCheck(VAR mess : PtrRecord):WhatsThat;
(* Testet die Nachricht einmal durch, damit der parser schon einen Teil der *)
(* Fehlerbehandlung sowie die WatchDog-Behandlung machen kann               *)

PROCEDURE AppendMessage(handle : OneGroupHandle;
                        type   : WhatsThat;
                        bitWishes : BITSET;
    (* Bits, die beim Einfgen schon gesetzt werden sollen; bFiltered bisher *)
                    VAR mess   : PtrRecord;
                    VAR usenetChain, usnetOk : BOOLEAN;
                    VAR isOneDupe: BOOLEAN;
                    VAR abort  : BOOLEAN):BOOLEAN;
(* Eine Message bearbeiten/ abort -> sofort abbrechen, schwerer Fehler *)
(* Returnwert zeigt an, ob die Nachricht geschrieben wurde *)
(* Es knnen natrlich nur die Typen own, personal, status und normal geschrieben *)
(* werden. Bei Mibrauch: Garbage in possible crash out! *)

PROCEDURE dumpMess(mess : PtrRecord);
(* Fehlermeldung zu dieser Nachricht in catlog.txt ausgeben *)

PROCEDURE EstimateNecessaryMemory():LONGCARD;
(* Versucht den fr den nchsten Einfgevorgang ntigen Platz zu schtzen *)

(*$H+*)
TYPE treeProc = PROCEDURE ((* handle: *) OneGroupHandle,
                           (* msgIdx: *) CARDINAL,
                           (* mess:   *) pBlockPtr);

PROCEDURE WalkTree (handle : OneGroupHandle; mess: CARDINAL; fromTop : BOOLEAN; proc: treeProc);
(* Allgemeine Prozedur, um einen Baum zu durchlaufen *)

PROCEDURE WalkTree2 (handle : OneGroupHandle; mess: CARDINAL; maxUp: INTEGER; maxDepth: INTEGER; proc: treeProc);
(* Etwas weniger allgemeine Prozedur, um einen Baum zu durchlaufen 
 * Es wird maximal um maxUp Ebenen nach oben gegangen, und von dort aus 
 * maximal maxDepth Ebenen nach unten 
 *)

TYPE updateProc = PROCEDURE((* msgIndex *) CARDINAL, (* newFlags *) BITSET);
TYPE flagchangeProc = PROCEDURE( (* oldFlags *) BITSET ) : BITSET;

PROCEDURE TreeFlags(handle : OneGroupHandle; mess : CARDINAL; fromTop : BOOLEAN; 
                    updt : updateProc; flagchange : flagchangeProc);
(* Flags eines Baumes bearbeiten *)

END data.

