IMPLEMENTATION MODULE EditBase;

FROM SYSTEM    IMPORT ADDRESS, ADR, TSIZE, CALLSYS, CADR, CODE;
FROM Storage   IMPORT ALLOCATE, DEALLOCATE;
IMPORT Block;

IMPORT BinOps;
IMPORT Storage;

IMPORT AssFuncs;
IMPORT EditTypes;
FROM EditTypes IMPORT EDITPTR;

FROM Void    IMPORT v;

CONST   Debug = FALSE;

CONST   CR      = 15C;
        LF      = 12C;

TYPE copyFunction = PROCEDURE (ADDRESS, ADDRESS, LONGCARD);

VAR CopyBlock : copyFunction;

PROCEDURE blockCopy (from, to : ADDRESS; len : LONGCARD);
BEGIN
  Block.Copy (from, len, to);
END blockCopy;

(*$? Debug:
(* Debug - Funktionen, fliegen in der endgltigen Version raus *)

PROCEDURE checkBuffer (VAR buff : EditTypes.aBufferPtr);
  VAR used : LONGINT; 
      i    : INTEGER;
BEGIN
  used := 0;
  WITH buff^ DO
    FOR i := 0 TO SHORT(toLine - fromLine) DO
      IF lines # NIL THEN INC (used, lines^[i]); END;
    END;
      CALLSYS (1, 4201, CADR("used gezhlt: %ld, eingetragen: %d"), used, usedMem);
      CALLSYS (1, 4201, CADR("MaxLines: %d"), maxLine);
      CALLSYS (1, 4201, CADR("fromLine: %ld"), fromLine);
      CALLSYS (1, 4201, CADR("toLine: %ld"), toLine);
      CALLSYS (1, 4201, CADR("usedLines: %ld"), toLine - fromLine + 1);
    IF used # LONG(usedMem) THEN
      HALT
    END; 
  END;
END checkBuffer;

PROCEDURE TestStructure (ed : EDITPTR);
  VAR buf : EditTypes.aBufferPtr;
      bufNum : INTEGER;
      used   : INTEGER;
      i      : INTEGER;
BEGIN
  WITH ed^ DO
    CALLSYS (1, 4201, CADR("Teste Struktur und Fllungsgrad ..."));
    buf := firstBuff;
    bufNum := 1;
    WHILE buf # NIL DO
      WITH buf^ DO
        used := 0;
        FOR i := 0 TO SHORT(toLine - fromLine) DO
          INC (used, lines^[i]);
        END;
        CALLSYS (1, 4201, CADR("Buffer Nummer: %d"), bufNum);
        CALLSYS (1, 4201, CADR("usedMem: %d"), usedMem);
        CALLSYS (1, 4201, CADR("used by count: %d"), used);
        CALLSYS (1, 4201, CADR("MaxLines: %d"), maxLine);
        CALLSYS (1, 4201, CADR("fromLine: %ld"), fromLine);
        CALLSYS (1, 4201, CADR("toLine: %ld"), toLine);
        CALLSYS (1, 4201, CADR("usedLines: %ld"), toLine - fromLine + 1);
      END;
      buf := buf^.next;
      INC (bufNum);
    END;
    CALLSYS (1, 4201, CADR("totalLineNr: %ld"), totalLineNr);
  END;
END TestStructure;
*)


PROCEDURE findLineNum (VAR ed : EDITPTR; lineNum : LONGINT): BOOLEAN;
BEGIN
  WITH ed^ DO
    (* Erst mal Buffer suchen und CurrentBuff entsprechend setzen *)
    IF firstBuff = NIL THEN RETURN FALSE END;
    IF currentBuff = NIL THEN currentBuff := firstBuff END;
    IF (lineNum > currentBuff^.toLine)
    THEN 
      WHILE (currentBuff # NIL) & (lineNum > currentBuff^.toLine) DO
        currentBuff := currentBuff^.next;
      END;
    ELSIF (lineNum < currentBuff^.fromLine) 
    THEN
      (* Nach vorne suchen (rckwrts) *)
      WHILE (currentBuff # NIL) & (lineNum < currentBuff^.fromLine) DO
        currentBuff := currentBuff^.prev;
      END;
    END;
    IF currentBuff = NIL
    THEN
      (*$? Debug:
      IF lineNum # totalLineNr
      THEN
        CALLSYS (1, 4201, CADR("Fehler bei Anforderung von Zeilenadresse"));
        CALLSYS (1, 4201, CADR("Angeforderte Zeile: %ld, maxZeile: %ld"), lineNum, totalLineNr);
        HALT;
      END;
      *)
      currentBuff := firstBuff;
      RETURN FALSE
    ELSE
      RETURN TRUE;
    END;
  END;
END findLineNum;

PROCEDURE createBuffer (defSize : INTEGER) : EditTypes.aBufferPtr;
  VAR newBuff : EditTypes.aBufferPtr;
BEGIN
  (* Neuen Verwaltungsblock anlegen *)
  IF defSize = 0 THEN defSize := memBlockSize; END;
  NEW (newBuff);
  (* Wenn wir jetzt nicht genug Speicher haben, dann ist eh alles vorbei *)
  IF newBuff = NIL THEN RETURN NIL END;
  (* Textspeicher allozieren *)
  ALLOCATE (newBuff^.buffer, defSize);
  IF newBuff^.buffer = NIL THEN DISPOSE (newBuff); RETURN NIL END;
  Block.Clear (newBuff^.buffer, defSize);
  newBuff^.bufSize := defSize;
  newBuff^.fromLine := 0;
  newBuff^.toLine := 0;
  newBuff^.usedMem := 0;
  newBuff^.maxLine := 0;
  newBuff^.lines := NIL;
  RETURN newBuff;
END createBuffer;

PROCEDURE copyBuffer (VAR ed : EDITPTR; line : LONGINT; VAR newBuff : EditTypes.aBufferPtr; 
                      VAR copyLines : INTEGER; copyOnly : BOOLEAN) : BOOLEAN;
(* Kopiert aus dem aktuellen Buffer ab Zeile line in neuen Buffer newBuff.
 * Wenn line < 0 ist, dann wird nach Prozenten gesplittet. 
 *)
 VAR i       : INTEGER;
     offSet2 : INTEGER;
     usedLines : INTEGER;
BEGIN
  WITH ed^.currentBuff^ DO
    newBuff := createBuffer(0);
    IF newBuff = NIL THEN RETURN FALSE END;
    (* Nun Kopierposition feststellen *)
    offSet2 := 0;
    usedLines := SHORT (toLine - fromLine);
    IF ~(copyOnly OR (line >= 0))
    THEN
      (* Jetzt Bytes und Zeilen zum Splitten feststellen *)
      i := 0;
      WHILE (offSet2 < usedMem DIV 2) & (i <= usedLines) DO
        INC (offSet2, lines^[i]);
        INC (i);
      END;
    ELSE
      IF line > toLine THEN
        eReason := notFound;
        RETURN FALSE
      END;
      FOR i := 0 TO SHORT(line-fromLine) DO 
        INC (offSet2, lines^[i]);
      END;
      i := SHORT(line - fromLine)+1;
    END;
    copyLines := i;
    (* In i steht die relative Zeilennummer im Buffer, ab der in den 
     * neuen Buffer kopiert wird.
     *)
    newBuff^.maxLine := BinOps.HigherInt(200, SHORT(toLine-fromLine-LONG(i))*2);
    ALLOCATE (newBuff^.lines, LONG(newBuff^.maxLine) * TSIZE (INTEGER));
    IF newBuff^.lines = NIL 
    THEN 
      DEALLOCATE (newBuff^.buffer, 0);
      DISPOSE (newBuff);
      RETURN FALSE 
    END;
    (* So, jetzt knnen wir auch kopieren *)
    (* Text kopieren *)
    CopyBlock (ADR (buffer^[offSet2]), newBuff^.buffer, usedMem - offSet2);
    newBuff^.usedMem := usedMem - offSet2;
    IF ~copyOnly THEN usedMem := offSet2; END;
    (* Zeilenarray kopieren *)
    CopyBlock (ADR(lines^[i]), newBuff^.lines, (toLine - fromLine - LONG(i)+1) * TSIZE (INTEGER));
    (* Zeilennummern in neuen Buffer schreiben *)
    newBuff^.fromLine := fromLine + LONG(i);
    newBuff^.toLine   := toLine;
  END;
  RETURN TRUE;
END copyBuffer;

PROCEDURE splitBuffer (VAR ed : EDITPTR; line: LONGINT) : BOOLEAN;
(* Splittet den aktuellen Buffer auf und hngt einen neuen 
 * Buffer hinter den aktuellen. 
 * Falls line >= 0 ist, wird hinter line gesplittet. Line ist die 
 * Zeilennummer in diesem Buffer!
 * FALSE: Nicht genug Speicher, eReason wird gesetzt
 *)
 VAR newBuff : EditTypes.aBufferPtr;
     i       : INTEGER;
BEGIN
  WITH ed^.currentBuff^ DO
    IF ~copyBuffer (ed, line, newBuff, i, FALSE)
    THEN 
      RETURN FALSE
    END;
    (* Jetzt haben wir alles alloziert, jetzt knnen wir den neuen 
     * Buffer in die Liste einhngen 
     *)
    newBuff^.prev := ed^.currentBuff;
    newBuff^.next := next;
    IF newBuff^.next # NIL
    THEN
      newBuff^.next^.prev := newBuff;
    END;
    ed^.currentBuff^.next := newBuff;
    newBuff^.fromLine := fromLine + LONG(i);
    newBuff^.toLine := toLine;
    toLine := fromLine + LONG(i-1);
    (* Jetzt noch die Buffer berprfen *)
  END (* WITH ed^.currentBuff^ *);
  RETURN TRUE;
END splitBuffer;

PROCEDURE compactTwo (first, second : EditTypes.aBufferPtr; VAR freedSecond : BOOLEAN) : BOOLEAN;
(* Bringt Daten aus dem zweiten Buffer in den ersten, so da dieser zu
 * ca. fillPercent gefllt wird
 *)
 CONST fillPercent = 95L;
 VAR   usedInSecond : INTEGER;
       freeInFirst,
       secIdx,
       i            : INTEGER;
       newLines     : INTEGER;
       newArray     : EditTypes.lineArrayPtr;
BEGIN
  freedSecond := FALSE;
  IF (LONG(first^.usedMem) * 100L) DIV LONG(first^.bufSize) < fillPercent
  THEN
    freeInFirst := SHORT((LONG(first^.bufSize) * fillPercent) DIV 100L) - first^.usedMem;
    (* Jetzt in zweitem Buffer suchen, bis freeInFirst Bytes berschritten *)
    usedInSecond := 0;
    secIdx := 0;
    WITH second^ DO
      WHILE (secIdx <= SHORT(toLine - fromLine)) & (usedInSecond + lines^[secIdx] < freeInFirst) DO
        INC (usedInSecond, lines^[secIdx]);
        INC (secIdx);
      END;
    END;
    (* Jetzt haben wir die Anzahl Bytes, die zu verschieben sind *)
    IF (usedInSecond = 0) OR (freeInFirst = 0) THEN RETURN TRUE END;
    (* Jetzt nachsehen, ob das LineArray in first gro genug ist *)
    WITH first^ DO
      newLines := SHORT(toLine-fromLine) + secIdx;
      IF newLines >= maxLine 
      THEN
        (* lineArray zu klein, neu allozieren! *)
        IF newLines < $4000 
        THEN
          newLines := BinOps.HigherInt (200, newLines * 2);
        ELSIF newLines < bufSize-100
        THEN
          INC (newLines,100);
        ELSE
          (* Zu viele Zeilen in einem Block! *)
          HALT
        END;
        ALLOCATE (newArray, LONG(newLines) * TSIZE (INTEGER));
        IF newArray = NIL THEN RETURN FALSE END;
        (* Ok, wir haben ein neues lineArray *)
        FOR i := 0 TO SHORT (toLine - fromLine) DO
          newArray^[i] := lines^[i]
        END;
        maxLine := newLines;
        DEALLOCATE (lines, 0);
        lines := newArray;
      END;
      (* Jetzt lineArray aus zweitem Buffer kopieren *)
      FOR i := 0 TO secIdx-1 DO
        lines^[SHORT (toLine-fromLine) + i + 1] := second^.lines^[i];
      END;
      (* Jetzt Speicherbereich umkopieren *)
      CopyBlock (second^.buffer,ADR(first^.buffer^[usedMem]), usedInSecond);
    END;
    WITH second^ DO
      (* Jetzt lineArray in zweitem Buffer umkopieren *)
      FOR i := secIdx TO SHORT(toLine-fromLine) DO
        lines^[i - secIdx] := lines^[i];
      END;
      (* Jetzt Speicher intern umkopieren *)
      CopyBlock (ADR(buffer^[usedInSecond]),buffer, usedMem - usedInSecond);
    END; 
    (* Jetzt noch die usedMems und die Zeilen anpassen *)
    INC (first^.usedMem, usedInSecond);
    DEC (second^.usedMem, usedInSecond);
    INC (first^.toLine, secIdx);
    INC (second^.fromLine, secIdx);
    (* Jetzt noch testen, ob evtl. der zweite Buffer leer geworden ist: *)
    IF second^.usedMem = 0
    THEN
      (* Buffer komplett freigeben und aus Liste aushngen *)
      first^.next := second^.next;
      IF second^.next # NIL THEN
        second^.next^.prev := second^.prev;
      END;
      DEALLOCATE (second^.buffer, 0);
      DEALLOCATE (second^.lines, 0);
      DISPOSE (second);
      freedSecond := TRUE
    END;
  END;
  RETURN TRUE;
END compactTwo;

PROCEDURE deleteLineBlock (VAR ed : EDITPTR; lineFrom, lineTo : LONGINT) : BOOLEAN;
  VAR i, lineIdx, lineCount,
      len, offset : INTEGER;
BEGIN
  WITH ed^.currentBuff^ DO 
    IF (lineFrom < fromLine) OR (lineTo < fromLine) OR (lineTo > toLine) OR (lineFrom > toLine)
    THEN 
      RETURN FALSE
    END;
    offset := 0;
    FOR i := 0 TO SHORT(lineFrom - fromLine) - 1 DO
      INC (offset, lines^[i]);
    END;
    lineCount := SHORT(lineTo - lineFrom) + 1;
    lineIdx := SHORT(lineFrom - fromLine);
    len := 0;
    FOR i := SHORT(lineFrom - fromLine) TO SHORT(lineFrom - fromLine) + lineCount  - 1 DO
      INC( len, lines^[i]);
    END;
    (* LineArray anpassen *)
    CopyBlock (ADR(lines^[lineIdx+lineCount]), ADR(lines^[lineIdx]), (SHORT (toLine - lineTo) + 1) * TSIZE (INTEGER)); 
    (* Im Speicher verschieben *)
    CopyBlock (ADR (buffer^[offset+len]), ADR (buffer^[offset]), usedMem - offset - len);
    (* usedMem anpassen *)
    DEC (usedMem, len);
  END; (* WITH *)
  RETURN TRUE;
END deleteLineBlock;

PROCEDURE countLines (buff: EditTypes.aBufferPtr) : INTEGER;
(* Zhlt die Zeilen in einem Buffer *)
  VAR lineCount : INTEGER;
      idx2      : INTEGER;
      ch        : CHAR;
BEGIN
  WITH buff^ DO
    (* Zeilen zhlen *)
    lineCount := 0;
    idx2 := 0;
    WHILE idx2 < usedMem DO
      idx2 := SHORT(AssFuncs.findLineEnd (buffer, idx2, usedMem));
      ch := buffer^[idx2];
      IF (ch = LF) OR (ch = CR) OR (ch = 0C) OR (idx2 > usedMem)
      THEN
        INC (lineCount);
        INC(idx2);
        IF ((buffer^[idx2] = LF) OR (buffer^[idx2] = CR)) & (buffer^[idx2] # ch)  THEN INC (idx2) END;
      END;
    END;
  END (* WITH buff^ DO *);
  RETURN lineCount;
END countLines;

PROCEDURE fillLineArray(buff: EditTypes.aBufferPtr);
(* Fllt das Zeilenarray zu einem Buffer *)
  VAR lineCount : INTEGER;
      idx2      : INTEGER;
      lastIdx   : INTEGER;
      lineLen   : INTEGER;
      ch        : CHAR;
BEGIN
  WITH buff^ DO
    (* Jetzt nur noch das ZeilenArray fllen *)
    lineCount := 0;
    idx2 := 0;
    lastIdx := 0;
    lineLen := 0;
    WHILE idx2 < usedMem DO
      idx2 := SHORT(AssFuncs.findLineEnd (buffer, idx2, usedMem));
      (*
      WHILE (idx2 < usedMem) & (ch # CR) & (ch # LF) DO
        INC (idx2); INC (lineLen); ch := text^[idx2];
      END;
      *)
      ch := buffer^[idx2];
      IF (ch = LF) OR (ch = CR) OR (ch = 0C) OR (idx2 > usedMem)
      THEN
        INC (idx2);
        IF idx2 > usedMem THEN idx2 := usedMem END;
        lineLen := idx2 - lastIdx;
        IF ((buffer^[idx2] = LF) OR (buffer^[idx2] = CR)) & (buffer^[idx2] # ch)  THEN INC (idx2); INC (lineLen); END;
        lines^[lineCount] := lineLen;
        INC (lineCount);
        (* Notbremse *)
        IF lineCount >= maxLine THEN RETURN END;
        lineLen := 0;
        lastIdx := idx2;
      END;
    END;
  END (* WITH buff^ DO *)
END fillLineArray;

PROCEDURE FixLine (ed : EDITPTR; line: LONGINT);
  VAR len, l : INTEGER;
      txt    : EditTypes.textPtr;
BEGIN
  WITH ed^ DO
    v.bool := GetLineLength (ed, line, len);
    ALLOCATE (txt, len+4);
    IF txt = NIL
    THEN
      RETURN
    END;
    v.bool := GetLine (ed, line, txt^, len);
    (* Jetzt Zeile testen *)
    l := LENGTH (txt^);
    IF (txt^[l] # CR) & (txt^[l] # LF)
    THEN
      (* CR/LF anhngen *)
      txt^[l] := CR;
      txt^[l+1] := LF;
      txt^[l+2] := 0C;
      l := LENGTH (txt^);
      v.bool := PutLine (ed, line, txt^, l);
    END;
    (* Speicher wieder freigeben *)
    DEALLOCATE (txt, 0);
  END;
END FixLine;

PROCEDURE FixBuffer (ed: EDITPTR);
  VAR oldLines,
      lineCount : INTEGER;
      
      addLines,
      oldTo     : LONGINT;
      tBuf      : EditTypes.aBufferPtr;
      doHalt    : BOOLEAN;  (* Halt am Ende? *)
BEGIN
  doHalt := FALSE;
  WITH ed^ DO
    WITH currentBuff^ DO
      (* Erstmal letzte Zeile im Buffer holen und testen 
      FixLine (ed, toLine);
      *)
      (* So, und jetzt die Struktur des Buffers nochmal neu aufbauen *)
      lineCount := countLines (currentBuff);
      (* Wir haben also jetzt lineCount Zeilen in dem Buffer.
       * Jetzt nachsehen, ob wir das Zeilenarray neu allozieren mssen 
       *)
      IF lineCount >= maxLine
      THEN
        oldLines := maxLine;
        maxLine := BinOps.HigherInt (100, (lineCount * 3) DIV 2);
        DEALLOCATE (lines, 0);
        ALLOCATE (lines, LONG(maxLine) * TSIZE (INTEGER));
        IF lines = NIL 
        THEN 
          (* Shit! Kein Speicher mehr! *)
          maxLine := lineCount + 1;
          ALLOCATE (lines, LONG(maxLine) * TSIZE (INTEGER));
          IF lines = NIL
          THEN
            (* Wir haben ein echtes Problem: Das alte Zeilenarray ist futsch, 
             * und fr das neue ist nicht genug Platz da!
             * Da kann ich nur hoffen, da das nie auftritt!
             * Erstmal alten Speicher wieder allozieren und dann
             * das Zeilenarray soweit fllen, wie es geht, und dann 
             * die Notbremse treten.
             *)
            ALLOCATE (lines, LONG(oldLines) * TSIZE (INTEGER));
            IF lines = NIL
            THEN
              (* Auch fr das alte Array kein Speicher mehr?
               * Dann ist jetzt alles vorbei, es hat auch  keinen
               * Zweck mehr, irgendwelche Rettungsversuche zu 
               * unternehmen.
               *)
              HALT;
            ELSE
              doHalt := TRUE;
            END;
          END;
        END;
      END;
      oldTo := toLine;
      toLine := fromLine + LONG(lineCount) - 1;
      (* neue toLine berechnen *)
      addLines := toLine - oldTo;
      
      fillLineArray (currentBuff);

    END (* WITH currentBuff^ *);
    (* Jetzt noch Zeilennummern in folgenden Buffern 
     * korrigieren.
     *)
    tBuf := currentBuff^.next;
    WHILE tBuf # NIL DO
      INC (tBuf^.fromLine, addLines);
      INC (tBuf^.toLine, addLines);
      tBuf := tBuf^.next;
    END;
    INC (totalLineNr, addLines);

    (* Jetzt die Notbremse ziehen *)
    IF doHalt THEN HALT END;
  END (* WITH ed^ DO *)
END FixBuffer;

(* Exportierte Prozeduren *)

PROCEDURE InsertBuffer (VAR ed : EDITPTR; VAR buff : ADDRESS; len : LONGCARD; freeBuffer: BOOLEAN): BOOLEAN;
  (* Es wird ein Text aus einem Buffer in ed an der aktuellen Stelle eingefgt.
   * Vorgehensweise:
   * Die Einfgeposition wird gemerkt, ebenso der Zeilenindex davon. Dann wird 
   * die Einfgeposition im Buffer berechnet und nachgesehen, wieviel Platz
   * noch im Buffer ist. Wenn noch gengend Platz in dem Buffer ist, wird
   * der Teil hinter der Einfgeposition um len Bytes nach hinten verschoben
   * und danach das Zeilenarray einfach komplett neu aufgebaut. 
   * Sollte nicht gengend Platz vorhanden sein, dann wird hinter der Einfgestelle
   * der Buffer gesplittet und solange neue Buffer angelegt, bis der Text komplett hinein-
   * pat. Danach wird dann der buffer einkopiert und das Zeilenarray ab dem ersten Buffer,
   * in den eingefgt wurde, neu aufgebaut. 
   * Es wird auch noch festgestellt, wieviele Zeilen eingefgt wurden, so da bei Blcken
   * die Endposition bestimmt werden kann.
   * Umbruch wird keiner vorgenommen.
   *)
  PROCEDURE appendToBuffer (buf : ADDRESS; len : LONGCARD; offset: INTEGER; startByte, freeBytes : LONGCARD): BOOLEAN;
    VAR ptr     : POINTER TO ARRAY [0..$FFFFFFFF] OF CHAR;
        idx     : LONGCARD;
        i       : LONGCARD;
        first   : BOOLEAN;
        ch      : CHAR;
  BEGIN 
    (* Jetzt ersten Teil aus Buffer einkopieren *)
    first := TRUE;
    ptr := buf;
    idx := 0;   (* Index in Buffer *)
    REPEAT
      IF ~first
      THEN
        IF ~GetBuffer (ed, 0) THEN
          RETURN FALSE
        END;
        startByte := (LONGCARD(LONG(ed^.currentBuff^.bufSize)) * 19) DIV 20; (* Wir wollen den Buffer zu 95% fllen *)
        freeBytes := LONGCARD(LONG(ed^.currentBuff^.bufSize));
      END;
      IF idx + startByte < len
      THEN
        (* Von Index idx+26000 vorwrts nach letztem Zeilenende suchen *)
        i := AssFuncs.findLineEnd (buf, idx+startByte, idx + freeBytes);
        ch := ptr^[i];
        (*
        i := idx + 26000L;
        ch := ptr^[i];
        WHILE (ch # CR) & (ch # LF) & (i < idx + $8000) DO
          INC (i); ch := ptr^[i];
        END;
        *)
        IF i >= idx + freeBytes - 2 (* -2 fr CR/LF! *)
        THEN
          (* Shit, keines mehr gefunden *)
          i := idx + startByte;
          ch := ptr^[i];
          WHILE (ch # CR) & (ch # LF) & (i > idx) DO
            DEC (i); ch := ptr^[i];
          END;
          IF i = idx THEN RETURN FALSE END;
        ELSE
          IF ((ptr^[i+1] = CR) OR (ptr^[i+1] = LF)) & (ptr^[i+1] # ch) THEN INC (i) END;
        END;
        INC(i);
      ELSE
        i := len;
      END;
      IF first
      THEN
        CopyBlock (ADR(ptr^[idx]), ADR(ed^.currentBuff^.buffer^[offset]), i - idx);
        INC (ed^.currentBuff^.usedMem, INTEGER(SHORT(i-idx)));
      ELSE
        CopyBlock (ADR(ptr^[idx]), ed^.currentBuff^.buffer, i - idx);
        ed^.currentBuff^.usedMem := INTEGER(SHORT(i - idx)); 
      END;
      idx := i;
      first := FALSE;
    UNTIL idx >= len;
    RETURN TRUE;
  END appendToBuffer;    
  
  VAR 
      lineCount : INTEGER;
      
      addLines,
      oldTo,
      insertLine: LONGINT;
      insPos,
      freeInBuff,
      insertRow : INTEGER;
      bBuf,
      cBuf,
      cNext,
      tBuf      : EditTypes.aBufferPtr;
      bUsed,
      startByte : INTEGER;
      tmpBuf    : ADDRESS;
      tLen      : INTEGER;
      first     : BOOLEAN;
      startCount: LONGINT;
      
BEGIN
  tmpBuf := NIL;
  WITH ed^ DO
    cBuf := currentBuff;
    (* Einfgeposition merken *)
    insertLine := currLineNr;
    insertRow  := currRow;
    (* Jetzt Offset in Buffer bestimmen und currentBuff setzen *)
    insPos := GetOffset (ed);
    IF insPos < 0 THEN 
      (* Zeile nicht gefunden, am Ende anfgen *)
      currentBuff := firstBuff;
      WHILE currentBuff^.next # NIL DO 
        currentBuff := currentBuff^.next;
      END;
      insPos := currentBuff^.usedMem;
    END;
    (* Jetzt nachsehen, wieviel Platz in currentBuff noch drin ist *)
    freeInBuff := currentBuff^.bufSize - currentBuff^.usedMem;
    IF len < VAL(LONGCARD,freeInBuff)
    THEN
      (* Gesamter Buffer pat noch in aktuellen Buffer, nur ein
       * paar Zeilen verschieben
       *)
      WITH currentBuff^ DO
        (* Text hinter Einfgeposition verschieben *)
        tLen := usedMem - insPos;
        CopyBlock (ADR(buffer^[insPos]), ADR(buffer^[insPos+VAL(INTEGER,len)]), tLen);
        (* Text aus buff einfgen *)
        CopyBlock (buff, ADR(buffer^[insPos]), len);
        (* Jetzt Speicher direkt freigeben *)

        IF freeBuffer THEN DEALLOCATE (buff, 0); buff := NIL; END;
        
        INC (usedMem, len);
        (* Jetzt Zeilenarray neu berechnen.
         * Erst mal Zeilen zhlen in Buffer 
         *)

        lineCount := countLines (currentBuff);
        (* Wir haben also jetzt lineCount Zeilen in dem Buffer.
         * Jetzt nachsehen, ob wir das Zeilenarray neu allozieren mssen 
         *)
        IF lineCount >= maxLine
        THEN
          maxLine := BinOps.HigherInt (100, SHORT(BinOps.LowerLInt (32000L, (LONG(lineCount) * 3))) DIV 2);
          DEALLOCATE (lines, 0);
          ALLOCATE (lines, LONG(maxLine) * TSIZE (INTEGER));
          IF lines = NIL 
          THEN 
            (* Shit! Kein Speicher mehr! *)
            (* Also wieder zurckkopieren und zurck *)
            CopyBlock (ADR(buffer^[insPos+VAL(INTEGER,len)]), ADR(buffer^[insPos]), tLen);
            DEC (usedMem, len);
            RETURN FALSE
          END;
        END;
        oldTo := toLine;
        toLine := fromLine + LONG(lineCount) - 1;
        (* neue toLine berechnen *)
        addLines := toLine - oldTo;
        
        fillLineArray (currentBuff);

      END (* WITH currentBuff^ *);
      (* Jetzt noch Blockende ggf. feststellen und Zeilennummern in folgenden Buffern 
       * korrigieren.
       *)
      tBuf := currentBuff^.next;
      WHILE tBuf # NIL DO
        INC (tBuf^.fromLine, addLines);
        INC (tBuf^.toLine, addLines);
        tBuf := tBuf^.next;
      END;
      INC (totalLineNr, addLines);
      
      (* Jetzt Blockende holen *)
      IF block THEN 
        GetPos (ed, insPos+VAL(INTEGER,len), blocks[0].blockEnd.line, blocks[0].blockEnd.row); 
      END;
      currentBuff := firstBuff;
      v.bool := findLineNum (ed, currLineNr);
    ELSE
      (* Buffer pat nicht in currentBuff.
       * Jetzt machen wir das folgendermaen: Wir kopieren alles hinter der Einfgepositon 
       * aus dem aktuellen Buffer in einen temporren Buffer, dann kopieren wir den ersten 
       * Teil des Textes in den aktuellen Buffer. Danach hngen wir weitere Buffer an, bis der
       * Text komplett drin ist und hngen dann noch den temporren Buffer an.
       *)
      WITH currentBuff^ DO
        tLen := usedMem - insPos;
        IF tLen > 0 THEN 
          ALLOCATE (tmpBuf, LONG(usedMem-insPos));
          IF tmpBuf = NIL THEN RETURN FALSE END;
          (* Jetzt ist ein temporrer Buffer da. *)
          CopyBlock (ADR(buffer^[insPos]), tmpBuf, usedMem-insPos);
          usedMem := insPos;
        ELSE
          tmpBuf := NIL;
        END;
        cBuf := currentBuff;
        cNext := currentBuff^.next;
        Block.Clear (ADR(buffer^[insPos]), currentBuff^.bufSize - insPos);

      END;
      startByte := currentBuff^.bufSize - insPos-100;
      IF startByte < 0 THEN startByte := currentBuff^.bufSize - insPos END;

      v.bool := appendToBuffer (buff, len, insPos, startByte, currentBuff^.bufSize - insPos);
      (* Jetzt ist der Text aufgesplittet in mehrere Buffer. 
       * Nun hngen wir noch den temporren Buffer wieder dran, 
       * Vorgehensweise ist die gleiche wie gerade, nur die Startposition ist 
       * leicht verndert, nmlich jetzt das letzte Byte im aktuellen Buffer
       * (usedMem).
       *)
      IF freeBuffer THEN DEALLOCATE (buff, 0); buff := NIL; END;

      (* Hier diese Position mu ich mir noch merken, um das Blockende festzustellen 
       *)
      bBuf := currentBuff;
      bUsed := currentBuff^.usedMem;
      
      IF tLen > 0 THEN
        startByte := currentBuff^.bufSize - currentBuff^.usedMem - 1000;
        IF startByte < 0 THEN startByte := currentBuff^.bufSize - currentBuff^.usedMem-100 END;
        IF startByte < 0 THEN startByte := currentBuff^.bufSize - currentBuff^.usedMem END;
     
        v.bool := appendToBuffer (tmpBuf, tLen, currentBuff^.usedMem, startByte, 
                                  currentBuff^.bufSize - currentBuff^.usedMem);
      END;
      
      (* So, jetzt setzen wir den currentBuff wieder auf den ersten, in
       * den eingefgt wurde, und bauen die Zeilenstruktur neu auf!
       *)
      
      currentBuff := cBuf;
      first := TRUE;
      WHILE currentBuff # cNext DO
        WITH currentBuff^ DO
          lineCount := countLines (currentBuff);
          (* Wir haben also jetzt lineCount Zeilen in dem Buffer.
           * Jetzt nachsehen, ob wir das Zeilenarray neu allozieren mssen 
           *)
          IF (lineCount >= maxLine) OR (lines = NIL) 
          THEN
            maxLine := BinOps.HigherInt (200, SHORT(BinOps.LowerLInt (32000L, (LONG(lineCount) * 3))) DIV 2);
            IF lines # NIL THEN DEALLOCATE (lines, 0); END;
            ALLOCATE (lines, LONG(maxLine) * TSIZE (INTEGER));
            IF lines = NIL 
            THEN 
              (* Shit! Kein Speicher mehr! *)
              (* Also wieder zurckkopieren und zurck *)
              CopyBlock (tmpBuf, ADR(cBuf^.buffer^[insPos]), tLen);
              cBuf^.usedMem := insPos + tLen;
              (* Jetzt noch neu angehngte Blcke lschen *)
              (* YYYY Hier mssen alle neuen Blcke freigegeben werden und
               * noch was zurck kopiert werden. 
               *)
               
              RETURN FALSE
            END;
          END;
          IF first THEN
            (* neue toLine berechnen *)
            oldTo := toLine;
            toLine := fromLine + LONG(lineCount) - 1;
            startCount := toLine;
            addLines := toLine - oldTo;
          ELSE
            fromLine := startCount + 1;
            toLine := startCount + LONG(lineCount);
            INC (addLines, lineCount);
            INC (startCount, lineCount);
          END;
          fillLineArray (currentBuff);
          first := FALSE;
        END (* WITH *);
        currentBuff := currentBuff^.next;
      END; (* WHILE *)
      
      IF tmpBuf # NIL THEN DEALLOCATE (tmpBuf, 0); END;
      
      (* So, jetzt alle weiteren Zeilennummern anpassen *)
      tBuf := cNext;
      WHILE tBuf # NIL DO
        INC (tBuf^.fromLine, addLines);
        INC (tBuf^.toLine, addLines);
        tBuf := tBuf^.next;
      END;
      INC (totalLineNr, addLines);
      
      (* Jetzt noch Position vom Blockende feststellen *) 
      currentBuff := bBuf;
  
      IF block 
      THEN 
        GetPos (ed, bUsed, blocks[0].blockEnd.line, blocks[0].blockEnd.row); 
      END;

      (* Und jetzt noch wieder komprimieren *)
      v.bool := CompactBuffer (ed);
    END;
    (* Und jetzt noch aktuellen Buffer restaurieren *)
    (* Das kann evtl. bse in die Hose gehen, falls der Buffer wegkomprimiert wurde *)
(*    currentBuff := cBuf;       *)
    (*$? Debug:
    TestStructure (ed);
    *)
  END (* WITH ed^ *);
  RETURN TRUE;
END InsertBuffer;

(*
PROCEDURE CopyToBuff(VAR ed : EDITPTR; tStartLine, tEndLine : LONGINT; VAR addBuf : EditTypes.aBufferPtr) : BOOLEAN;
(* Kopieren von einem EditBuffer in einen anderen.
 * Es wird aus dem Editor in addBuf der Bereich von 
 * tStartLine bis tEndLine kopiert.
 *)
VAR tBuf,
    nBuf, 
    listBuf : EditTypes.aBufferPtr;
    i       : INTEGER;
    
 PROCEDURE freeTmpBuff();
 BEGIN
   (* Speicher freigeben *)
   tBuf := addBuf;
   WHILE tBuf # NIL DO
     DEALLOCATE (tBuf^.buffer, 0);
     IF tBuf^.lines # NIL THEN DEALLOCATE (tBuf^.lines, 0); END;
     nBuf := tBuf^.next;
     DISPOSE (tBuf);
     tBuf := nBuf;
   END;
 END freeTmpBuff;
 
BEGIN
  (* Startzeile suchen *)
  IF tStartLine > 0 
  THEN
    IF ~findLineNum (ed, tStartLine-1) THEN RETURN FALSE END;
  ELSE
    ed^.currentBuff := ed^.firstBuff
  END;
  WITH ed^ DO 
    (* Erstmal aus erstem Buffer kopieren *)
    IF ~copyBuffer (ed, tStartLine-1, addBuf, v.int, TRUE)
    THEN
      RETURN FALSE
    END;
    addBuf^.next := NIL;
    (* Jetzt ggf. weitere Buffer anhngen *)
    IF tEndLine > addBuf^.toLine
    THEN
      (* weiter kopieren *)
      listBuf := addBuf;
      tBuf := currentBuff^.next;
      REPEAT
        IF tBuf # NIL
        THEN
          (* Einen kompletten Buffer kopieren und anhngen an addBuf *)
          nBuf := createBuffer (0);
          IF nBuf = NIL 
          THEN
            freeTmpBuff();
            RETURN FALSE
          END;
          (* Speicher fr Zeilenarray anlegen *)
          ALLOCATE (nBuf^.lines, LONG(tBuf^.maxLine) * TSIZE (INTEGER));
          IF nBuf^.lines = NIL
          THEN
            freeTmpBuff();
            RETURN FALSE;
          END;
          (* Jetzt kompletten Buffer kopieren *)
          CopyBlock (tBuf^.buffer, nBuf^.buffer, currentBuff^.bufSize);
          CopyBlock (tBuf^.lines, nBuf^.lines, LONG(tBuf^.maxLine)*TSIZE (INTEGER));
          (* Struktur kopieren *)
          nBuf^.usedMem  := tBuf^.usedMem;
          nBuf^.maxLine  := tBuf^.maxLine;
          nBuf^.fromLine := tBuf^.fromLine;
          nBuf^.toLine   := tBuf^.toLine;
          (* Jetzt an addBuff-Kette anhngen *)
          listBuf^.next := nBuf;
          nBuf^.prev :=listBuf; 
          nBuf^.next := NIL;
          listBuf := nBuf;
          (* Jetzt nchsten Block untersuchen *)
          tBuf := tBuf^.next;
        END (* IF *);
      UNTIL (tBuf = NIL) OR (tBuf^.fromLine > tEndLine);
      (* Jetzt noch letzte Endzeilennummer anpassen *)
      listBuf^.toLine := tEndLine;
    ELSE
      WITH addBuf^ DO
        toLine := tEndLine;
        usedMem := 0;
        FOR i := 0 TO SHORT(toLine-fromLine) DO 
          INC (usedMem, lines^[i]);
        END;
      END;
    END (* IF *);
    (*$? Debug:
    TestStructure (ed);
    *)
  END (* WITH *);
  RETURN TRUE;
END CopyToBuff;
*)

PROCEDURE CopyToBuff(VAR ed : EDITPTR; tStartLine, tEndLine : LONGINT; 
                     VAR addBuf : ADDRESS; VAR len : LONGCARD) : BOOLEAN;
(* Kopieren von einem EditBuffer in einen Speicherbereich.
 * Es wird aus dem Editor in addBuf der Bereich von 
 * tStartLine bis tEndLine kopiert.
 * Neue Fassung, die alles in einen Block kopiert und nicht in n kleine Blcke.
 *)
VAR tBuf    : EditTypes.aBufferPtr;
    i       : INTEGER;
    tBytes,
    cBytes,
    bytes   : LONGINT;
    offset  : INTEGER;
    fromAddr,
    toAddr  : ADDRESS;
    
BEGIN
  (* Startzeile suchen *)
  IF tStartLine > 0 
  THEN
    IF ~findLineNum (ed, tStartLine-1) THEN RETURN FALSE END;
  ELSE
    ed^.currentBuff := ed^.firstBuff
  END;
  WITH ed^ DO 
    IF tEndLine > totalLineNr - 1 THEN tEndLine := totalLineNr - 1 END;
    (* Bytes zum Allozieren zhlen *)
    bytes := currentBuff^.usedMem;
    offset := 0;
    FOR i := 0 TO SHORT(tStartLine - currentBuff^.fromLine - 1) DO
      DEC (bytes, currentBuff^.lines^[i]);
      INC (offset, currentBuff^.lines^[i]);
    END;
    tBuf := currentBuff;
    WHILE (tBuf # NIL) & (tEndLine > tBuf^.toLine) DO
      tBuf := tBuf^.next;
      IF tBuf # NIL
      THEN
        INC (bytes, tBuf^.usedMem);
      END;
    END;
    FOR i := SHORT(tEndLine - tBuf^.fromLine + 1) TO SHORT(tBuf^.toLine - tBuf^.fromLine) DO
      DEC (bytes, tBuf^.lines^[i]);
    END;
    IF bytes < 0 THEN HALT END;
    INC (bytes, 10);        (* ein bichen Sicherheit *)
    ALLOCATE (addBuf, bytes);
    IF addBuf = NIL THEN RETURN FALSE END;
    
    (* Jetzt den ganzen Kram in den Buffer kopieren *)
    tBuf := currentBuff;
    tBytes := bytes-10;
    toAddr := addBuf;
    fromAddr := ADR(tBuf^.buffer^[offset]);
    IF tBytes < LONG(tBuf^.usedMem - offset)
    THEN
      CopyBlock (fromAddr, toAddr, tBytes);
    ELSE
      CopyBlock (fromAddr, toAddr, tBuf^.usedMem - offset);
      DEC (tBytes, tBuf^.usedMem - offset);
      INC (toAddr, tBuf^.usedMem - offset);
      REPEAT
        tBuf := tBuf^.next;
        IF tBuf # NIL
        THEN
          cBytes := BinOps.LowerLInt (tBytes, tBuf^.usedMem);
          CopyBlock (tBuf^.buffer, toAddr, cBytes);
          INC (toAddr, cBytes);
          DEC (tBytes, cBytes);
        END;
      UNTIL (tBuf = NIL) OR (tBytes <= 0);
    END;
    len := bytes-10;
  END (* WITH *);
  RETURN TRUE;
END CopyToBuff;

(*
PROCEDURE CopyFromBuff(VAR ed : EDITPTR; tStartLine : LONGINT; addBuf : EditTypes.aBufferPtr; prev : BOOLEAN) : BOOLEAN;
(* Kopieren von einem EditBuffer in einen anderen.
 * Es wird der Sourcebuffer hinter oder vor die Zeile 
 * tStartLine angehngt.
 * Vorgehensweise:
 *  - Zielbuffer hinter tStartLine splitten
 *  - Zeilen aus Source dahinter hngen (buffer einhngen)
 *  - Zeilennummern anpassen
 *  - compactBuffer laufen lassen, um Speicherbedarf einzuschrnken
 *)
 VAR    tBuf    : EditTypes.aBufferPtr;
        lines   : LONGINT;
        i       : INTEGER;
        lineStart : LONGINT;
        
BEGIN
  IF prev THEN 
    DEC (tStartLine);
  END;
  IF tStartLine >= 0
  THEN
    (* Zeile suchen *)
    IF ~findLineNum (ed, tStartLine) THEN RETURN FALSE END;
    (* Buffer gezielt splitten *)
    IF ~splitBuffer (ed, tStartLine) THEN RETURN FALSE END;
  ELSE
    ed^.currentBuff := ed^.firstBuff;
  END;
  WITH ed^ DO
    (* Zeilennummern in angehngten Buffern korrigieren *)
    tBuf := addBuf;
    lineStart := tStartLine + 1;
    lines := 0;
    WHILE tBuf # NIL DO
      INC (lines, tBuf^.toLine-tBuf^.fromLine+1);
      tBuf^.toLine := lineStart + tBuf^.toLine-tBuf^.fromLine;
      tBuf^.fromLine := lineStart;
      lineStart := tBuf^.toLine + 1;
      tBuf := tBuf^.next;
    END;
    (* Zeilennummern in Editorbuffern ndern *)
    IF tStartLine = -1 THEN
      tBuf := currentBuff
    ELSE
      tBuf := currentBuff^.next;
    END;
    WHILE tBuf # NIL DO
      INC (tBuf^.toLine, lines);
      INC (tBuf^.fromLine, lines);
      tBuf := tBuf^.next;
    END;
    (* Zeilennummer fr Editor ndern *)
    INC (totalLineNr, lines);
    (* anderen Buffer anhngen *)
    IF tStartLine = -1 THEN
      tBuf := currentBuff;
      firstBuff := addBuf;
      currentBuff := addBuf;
      addBuf^.prev := NIL;
    ELSE
      tBuf := currentBuff^.next;
      currentBuff^.next := addBuf;
      addBuf^.prev := currentBuff;
    END;
    WHILE addBuf^.next # NIL DO addBuf := addBuf^.next END;
    addBuf^.next := tBuf;
    tBuf^.prev := addBuf;
  END(*WITH*);
  (*$? Debug:
  TestStructure (ed);
  *)
  (* Jetzt Speicher kompaktieren *)
  v.bool := CompactBuffer (ed);
  RETURN TRUE
END CopyFromBuff;
*)

PROCEDURE GetOffset (VAR ed : EDITPTR) : INTEGER;
(* Setzt Buffer auf CurrentLine und berechnet Offset in 
 * currentBuffer fr currLineNr und currRow
 *)
  VAR offSet,
      i : INTEGER;
BEGIN
  WITH ed^ DO
    IF ~findLineNum (ed, currLineNr) THEN RETURN -1 END;
    (* Aktueller Buffer enthlt lineNum *)
    WITH currentBuff^ DO
      offSet := 0;
      FOR i := 0 TO SHORT (currLineNr - fromLine) - 1 DO
        INC (offSet, lines^[i]);
      END;
    END;
    INC (offSet, currRow);
    RETURN offSet;
  END;
END GetOffset;

PROCEDURE GetPos (VAR ed : EDITPTR; offset : INTEGER; VAR line : LONGINT; VAR row : INTEGER);
(* Berechnet aus Offset in currentBuff die aktuelle
 * Zeile und Spalte
 *)
  VAR offset2,
      i         : INTEGER;
BEGIN
  WITH ed^ DO
    (* Aktueller Buffer enthlt lineNum *)
    WITH currentBuff^ DO
      i := 0;
      offset2 := 0;
      WHILE (i <= SHORT(toLine - fromLine)) & (offset2 + lines^[i] <= offset) DO 
        INC (offset2, lines^[i]);
        INC (i);
      END;
      line := fromLine + LONG(i);
      row := offset - offset2;
    END;
  END;
END GetPos;

PROCEDURE GetLineLength (VAR ed : EDITPTR; lineNum : LONGINT; VAR len : INTEGER) : BOOLEAN;
(* Liefert die Lnge der Zeile lineNum
 * FALSE: Zeile nicht vorhanden.
 *)
BEGIN
  WITH ed^ DO
    IF (lineNum >= totalLineNr) OR (lineNum < 0) 
    THEN 
      len := 0; 
      eReason := notFound; 
      RETURN FALSE 
    END;
    (* Erst mal Buffer suchen und CurrentBuff entsprechend setzen *)
    IF ~findLineNum (ed, lineNum) THEN RETURN FALSE END;
    len := currentBuff^.lines^[SHORT(lineNum-currentBuff^.fromLine)];
    IF len = 0 THEN 
      (* Sollte nicht passieren, also Struktur korrigieren *)
      FixBuffer (ed);
      RETURN GetLineLength (ed, lineNum, len);
    END;
    IF lineNum = currLineNr THEN len := editLen END;
    RETURN TRUE;
  END;
END GetLineLength;

PROCEDURE GetLineAdr (VAR ed : EDITPTR; lineNum : LONGINT; VAR text : ADDRESS; VAR len : INTEGER) : BOOLEAN;
(* Liefert die Adresse der Zeile und die Lnge.
 * Nur fr Lesezugriff!
 * FALSE: Zeile nicht vorhanden
 *)
  VAR i   : INTEGER;
      offSet : INTEGER;
BEGIN
  WITH ed^ DO
    IF (lineNum >= totalLineNr) OR (lineNum < 0) 
    THEN 
      len := 0; 
      eReason := notFound; 
      RETURN FALSE 
    END;
    IF (lineNum = currLineNr) & editChanged
    THEN
      text := editLine.text;
      len := editLen;
      RETURN TRUE
    END;
    (* Erst mal Buffer suchen und CurrentBuff entsprechend setzen *)
    IF ~findLineNum (ed, lineNum) THEN RETURN FALSE END;
    (* Aktueller Buffer enthlt lineNum *)
    offSet := 0;
    WITH currentBuff^ DO
      FOR i := 0 TO SHORT (lineNum - fromLine) - 1 DO
        INC (offSet, lines^[i]);
      END;
      len := lines^[SHORT(lineNum-fromLine)];
      text := ADR (buffer^[offSet]);
    END;
    RETURN TRUE;
  END;
END GetLineAdr;

PROCEDURE GetLine (VAR ed : EDITPTR; lineNum : LONGINT; VAR str : ARRAY OF CHAR; VAR len : INTEGER) : BOOLEAN;
  VAR text : ADDRESS;
      i    : INTEGER;
      found: BOOLEAN;
BEGIN
  WITH ed^ DO
    IF (lineNum = currLineNr) & editChanged THEN 
      CopyBlock (editLine.text, ADR (str), editLen);
      len := editLen;
      str[len] := 0C;
      RETURN TRUE
    END;
    IF ~GetLineAdr (ed, lineNum, text, len)
    THEN 
      str[0] := "";
      len := 0;
      RETURN FALSE
    END;
    eReason := lineToLong;
    IF len > INTEGER(HIGH (str)) THEN 
      CopyBlock (text, ADR(str), HIGH(str)-1);
      RETURN FALSE 
    END;
    CopyBlock (text, ADR(str), len);
    found := FALSE;
    FOR i := len-2 TO len-1 DO 
      IF (i >= 0) & ((str[i] = LF) OR (str[i] = CR)) THEN found := TRUE END;
    END;
    IF ~found THEN
      str[len] := CR;
      str[len+1] := LF;
      INC (len,2);
    END;
    str[len] := 0C;
    RETURN TRUE;
  END;
END GetLine;

PROCEDURE PutLine (VAR ed : EDITPTR; lineNum : LONGINT; VAR str : ARRAY OF CHAR; len : INTEGER) : BOOLEAN;
  VAR oldLen : INTEGER;
      offSet : INTEGER;
      i      : INTEGER;
BEGIN
  WITH ed^ DO
    IF (lineNum > totalLineNr) OR (lineNum < 0) 
    THEN 
      len := 0; 
      eReason := notFound;
      RETURN FALSE 
    END;
    (* Erst mal Buffer suchen und CurrentBuff entsprechend setzen *)
    IF ~findLineNum (ed, lineNum) THEN RETURN FALSE END;
    (* Aktueller Buffer enthlt lineNum *)
    WITH currentBuff^ DO
      offSet := 0;
      FOR i := 0 TO SHORT (lineNum - fromLine) - 1 DO
        INC (offSet, lines^[i]);
      END;
      oldLen := lines^[SHORT(lineNum-fromLine)];
      IF len <= oldLen
      THEN 
        (* Easy, Zeile nur kopieren, und evtl. was nachschieben von hinten *)
        CopyBlock (ADR(str),ADR(buffer^[offSet]), len);
        IF len < oldLen
        THEN
          CopyBlock (ADR(buffer^[offSet+oldLen]),ADR(buffer^[offSet+len]), usedMem - offSet - oldLen); 
          (* Benutzen Speicher updaten *)
          DEC (usedMem, oldLen-len);
        END;
        lines^[SHORT(lineNum-fromLine)] := len;
        RETURN TRUE;
      ELSE
        (* Jetzt wird es echt schwierig: Ich mu erst nachsehen, ob 
         * noch genug Platz im aktuellen Buffer ist. Wenn das nicht 
         * der Fall ist, dann mu ich einen neuen Buffer allozieren
         * und ca. 50% des Blocks auslagern. Wenn das auch nicht mehr 
         * geht (kein Speicher mehr fr Buffer), dann mu ich eine 
         * Garbage Collection machen (Alle Blcke zusammenschieben)
         * und dann das nochmal versuchen.
         *)
        eReason := notEnoughMem;
        IF usedMem - oldLen  < currentBuff^.bufSize - len
        THEN 
          (* Ok, pat noch rein *)
          (* Bytes nach hinten schieben *)
          CopyBlock (ADR (buffer^[offSet+oldLen]),ADR(buffer^[offSet+len]), usedMem - offSet - oldLen); 
          (* Zeile einfgen *)
          CopyBlock (ADR(str), ADR(buffer^[offSet]), len);
          (* Benutzen Speicher im Block hochzhlen *)
          INC (usedMem, len - oldLen);
          lines^[SHORT(lineNum-fromLine)] := len;
          RETURN TRUE;
        ELSE
          (* Nicht genug Platz, also neuen Buffer anhngen *)
          IF ~splitBuffer (ed,-1)
          THEN 
            RETURN FALSE 
          END;
          (* Und jetzt mssen wir nur noch die eigentliche Zeile einfgen.
           * Da ich faul bin, mache ich das einfach durch eine Rekursion 
           *)
          RETURN PutLine (ed, lineNum, str, len);
        END;
      END;
    END (* WITH currentBuff^ DO *);
  END (* WITH ed^ DO *)
END PutLine;

PROCEDURE InsertCr (VAR ed : EDITPTR; lineNum : LONGINT; row : INTEGER) : BOOLEAN;
(* Trennt die Zeile lineNum an der Stelle row auf.
 * FALSE: Kein Speicherplatz mehr frei.
 *)
  VAR lineIdx : INTEGER;
      newArray: EditTypes.lineArrayPtr;
      buff    : EditTypes.aBufferPtr;
      doFixBuf: BOOLEAN;
BEGIN
  WITH ed^ DO
    IF (lineNum > totalLineNr) OR (lineNum < 0) 
    THEN 
      eReason := notFound; 
      RETURN FALSE 
    END;
    WITH currentBuff^ DO
      (* Wir suchen einfach nach der Zeile und fgen 
       * noch einen Eintrag im LineArray an.
       *)
      eReason := notEnoughMem;
      lineIdx := SHORT(lineNum - fromLine);
      IF SHORT (toLine - fromLine) + 1 >= maxLine 
      THEN
        (* LineArray neu allozieren *)
        ALLOCATE (newArray, LONG(maxLine + 20) * TSIZE (INTEGER));
        IF newArray = NIL THEN RETURN FALSE END;
        CopyBlock (lines, newArray, maxLine * TSIZE (INTEGER));
        DEALLOCATE (lines, 0);
        lines := newArray;
        INC (maxLine, 20);
      END;
      (* Platz im LineArray ist jetzt garantiert da! *)
      (* Zeilenindex einfgen *)
      doFixBuf := FALSE;
      (* Prfen, ob eine Zeile 0 Byte lang werden wrde *)
      IF (lines^[lineIdx] = row) OR (row = 0) THEN 
      (*$? Debug:
        TestStructure (ed);
      *)
        FixLine (ed, lineNum);
        RETURN InsertCr (ed, lineNum, row);
      END;
      CopyBlock (ADR(lines^[lineIdx]), ADR(lines^[lineIdx+1]), (SHORT(toLine - fromLine) - lineIdx + 1) * TSIZE (INTEGER)); 
      lines^[lineIdx+1] := lines^[lineIdx] - row; 
      IF lines^[lineIdx+1] = 0 THEN doFixBuf := TRUE; END;
      lines^[lineIdx] := row; 
      IF lines^[lineIdx] = 0 THEN doFixBuf := TRUE; END;
      INC (toLine);
      INC (totalLineNr);
      buff := currentBuff^.next;
      WHILE (buff # NIL) DO
        INC (buff^.fromLine);
        INC (buff^.toLine);
        buff := buff^.next;
      END;
      IF doFixBuf
      THEN
        FixBuffer (ed);
      END;
    END;
  END;
  RETURN TRUE
END InsertCr; 

PROCEDURE InsertLine (VAR ed : EDITPTR; lineNum : LONGINT; VAR str : ARRAY OF CHAR; pre : BOOLEAN; len : INTEGER) : BOOLEAN;
(* Fgt die Zeile lineNum im Buffer ein.
 * Wenn pre = TRUE ist, dann wird die Zeile vor die alte Zeile lineNum gesetzt,
 * ansonsten hinter die Zeile lineNum.
 * Ergebnis: FALSE: Nicht genug Speicher mehr frei
 *) 
 VAR buff : EditTypes.aBufferPtr;
     newArray : EditTypes.lineArrayPtr;
     offset,
     lineIdx,
     i        : INTEGER;
BEGIN
  WITH ed^ DO
    IF (lineNum > totalLineNr) OR (lineNum < 0) 
    THEN 
      eReason := notFound; 
      RETURN FALSE 
    END;
    (* Wir fgen immer hinter der Zeile ein. Bei pre wird
     * einfach lineNum dekrementiert 
     *)
    eReason := notEnoughMem;
    IF pre THEN DEC (lineNum) END;
    IF lineNum = -1
    THEN
      currentBuff := firstBuff;
      offset := 0;
      lineIdx := 0;
    ELSE
      IF ~findLineNum (ed, lineNum) THEN RETURN FALSE END;
      WITH currentBuff^ DO
        offset := 0;
        FOR i := 0 TO SHORT(lineNum - fromLine) DO
          INC (offset, lines^[i]);
        END;
        lineIdx := SHORT(lineNum - fromLine)+1;
      END;
    END;
    (* Jetzt sind offset und Buffer bestimmt *)
    WITH currentBuff^ DO
      IF usedMem < currentBuff^.bufSize - len
      THEN
        (* Ok, pat noch rein *)
        IF SHORT (toLine - fromLine) + 4 >= maxLine - 1
        THEN
          (* LineArray neu allozieren *)
          ALLOCATE (newArray, LONG(maxLine + 20) * TSIZE (INTEGER));
          IF newArray = NIL THEN RETURN FALSE END;
          CopyBlock (lines, newArray, maxLine * TSIZE (INTEGER));
          DEALLOCATE (lines, 0);
          lines := newArray;
          INC (maxLine, 20);
        END;
        (* Platz im LineArray ist jetzt garantiert da! *)
        (* Zeilenindex einfgen *)
        (*
        CopyBlock (ADR(lines^[lineIdx]), ADR(lines^[lineIdx+1]), (SHORT(toLine - fromLine + 1) - lineIdx + 1) * TSIZE (INTEGER));
        *)
        FOR i := SHORT(toLine - fromLine + 1) TO lineIdx BY -1 DO
          lines^[i+1] :=  lines^[i];
        END;
        lines^[lineIdx] := len;
        CopyBlock (ADR (buffer^[offset]), ADR(buffer^[offset+len]), usedMem - offset); 
        CopyBlock (ADR (str), ADR (buffer^[offset]), len);
        INC (usedMem, len);
        (* Jetzt Zeilenverweise updaten in allen folgenden Buffern *)
        INC (toLine);
        INC (totalLineNr);
        buff := currentBuff^.next;
        WHILE (buff # NIL) DO
          INC (buff^.fromLine);
          INC (buff^.toLine);
          buff := buff^.next;
        END;
      ELSE
        (* Nicht genug Platz im aktuellen Buffer! Also das alte Spiel:
         * Splitten und dann nochmal versuchen 
         *)
        IF ~splitBuffer (ed, -1)
        THEN RETURN FALSE END;
        RETURN InsertLine (ed, lineNum, str, FALSE, len);
      END;
    END;
  END;
  RETURN TRUE
END InsertLine;

PROCEDURE DeleteSomeLines (VAR ed : EDITPTR; from, to : LONGINT) : BOOLEAN;
(* Lscht einen Bereich von Zeilen ber mehrere Blcke hinweg 
 * Dabei wird mittels deleteLineBlock in den ueren Blcken 
 * gelscht und die inneren Blcke entfernt.
 *)
  VAR zBuf, tBuf : EditTypes.aBufferPtr;
      lineCount : INTEGER;
BEGIN
  (* Buffer suchen zu Zeile *)
  IF ~findLineNum (ed, from) THEN RETURN FALSE END;
  WITH ed^ DO
    to := BinOps.LowerLInt (to, totalLineNr - 1);
    IF currentBuff^.toLine >= to
    THEN
      IF ~deleteLineBlock (ed, from, to)
      THEN
        RETURN FALSE
      END;
    ELSE
      IF from = currentBuff^.fromLine
      THEN
        (* currentBuff komplett lschen *)
        tBuf := currentBuff;
        DEALLOCATE (tBuf^.buffer, 0);
        DEALLOCATE (tBuf^.lines, 0);
        IF tBuf^.prev # NIL
        THEN
          tBuf^.prev^.next := tBuf^.next;
        END;
        IF tBuf^.next # NIL
        THEN
          tBuf^.next^.prev := tBuf^.prev;
        END;
        IF firstBuff = currentBuff
        THEN
          firstBuff := currentBuff^.next;
        END;
        currentBuff := currentBuff^.next;
        DISPOSE (tBuf);
      ELSE
        IF ~deleteLineBlock (ed, from, currentBuff^.toLine)
        THEN RETURN FALSE END;
        currentBuff^.toLine := from-1;
        currentBuff := currentBuff^.next;
      END;
      WHILE (currentBuff # NIL) & (to > currentBuff^.toLine) DO
        (* Buffer komplett freigeben *)
        tBuf := currentBuff;
        IF tBuf^.prev # NIL
        THEN
          tBuf^.prev^.next := tBuf^.next;
        END;
        IF tBuf^.next # NIL
        THEN
          tBuf^.next^.prev := tBuf^.prev;
        END;
        IF firstBuff = currentBuff
        THEN
          firstBuff := currentBuff^.next;
        END;
        currentBuff := currentBuff^.next;
        DEALLOCATE (tBuf^.buffer, 0);
        DEALLOCATE (tBuf^.lines, 0);
        DISPOSE (tBuf);
      END;
      (* Jetzt noch restliche Zeilen lschen *)
      IF ~deleteLineBlock (ed, currentBuff^.fromLine, to)
      THEN 
        RETURN FALSE
      END;
      currentBuff^.fromLine := from;
    END;
    WITH currentBuff^ DO
      lineCount := SHORT(to - from) + 1;
      (* Zeilenverweise anpassen *)
      DEC (ed^.totalLineNr, lineCount);
      DEC (toLine, lineCount);
      tBuf := next;
      WHILE tBuf # NIL DO 
        DEC (tBuf^.fromLine, lineCount);
        DEC (tBuf^.toLine, lineCount);
        tBuf := tBuf^.next;  
      END;
    END;
    (*$? Debug:
    TestStructure (ed);
    *)
  END;
  RETURN TRUE;
END DeleteSomeLines;

PROCEDURE DeleteLine (VAR ed : EDITPTR; lineNum : LONGINT) : BOOLEAN;
(* Lscht die Zeile lineNum im Buffer
 * FALSE: Zeile nicht vorhanden 
 *)
 VAR i,
     lineIdx,
     len,
     offset     : INTEGER;
     lineCount  : INTEGER;
     buff       : EditTypes.aBufferPtr;
BEGIN
  WITH ed^ DO
    IF (lineNum > totalLineNr) OR (lineNum < 0) 
    THEN 
      eReason := notFound; 
      RETURN FALSE 
    END;
    IF ~findLineNum (ed, lineNum) THEN RETURN FALSE END;
    WITH currentBuff^ DO
      (* Zeile lschen *)
      IF ~deleteLineBlock (ed, lineNum, lineNum) 
      THEN 
        RETURN FALSE
      END;
    (*
      offset := 0;
      FOR i := 0 TO SHORT(lineNum - fromLine) - 1 DO
        INC (offset, lines^[i]);
      END;
      lineIdx := SHORT(lineNum - fromLine);
      len := lines^[lineIdx];
      (* LineArray anpassen *)
      CopyBlock (ADR(lines^[lineIdx+1]), ADR(lines^[lineIdx]), SHORT (toLine - lineNum + 1) * TSIZE (INTEGER)); 
      (* Im Speicher verschieben *)
      CopyBlock (ADR (buffer^[offset+len]), ADR (buffer^[offset]), usedMem - offset - len); 
      (* usedMem anpassen *)
      DEC (usedMem, len);
      (* Zeilenverweise anpassen *)
      DEC (totalLineNr);
      DEC (toLine);
      buff := next;
      WHILE buff # NIL DO 
        DEC (buff^.fromLine);
        DEC (buff^.toLine);
        buff := buff^.next;
      END;
    *)
      lineCount := 1;
      (* Zeilenverweise anpassen *)
      DEC (ed^.totalLineNr, lineCount);
      IF totalLineNr = 0 
      THEN
        totalLineNr := 1;
        lineCount := SHORT(toLine);
      END;
      DEC (toLine, lineCount);
      buff := next;
      WHILE buff # NIL DO 
        DEC (buff^.fromLine, lineCount);
        DEC (buff^.toLine, lineCount);
        buff := buff^.next;
      END;
      IF usedMem < currentBuff^.bufSize DIV 4
      THEN 
        (* weniger als 25 % belegt,
         * Buffer mit folgendem oder vorherigem zusammenlegen
         *)
        IF next # NIL
        THEN
          v.bool := compactTwo (currentBuff, currentBuff^.next, v.bool);
        ELSIF prev # NIL
        THEN
          v.bool := compactTwo (currentBuff^.prev, currentBuff, v.bool);
        END;
      END;
      IF currLineNr = editLineNr
      THEN 
        editChanged := FALSE;
      END;
      IF currLineNr >= totalLineNr THEN
        currLineNr := totalLineNr - 1
      END;
    END;
  END;
  RETURN TRUE;
END DeleteLine;

PROCEDURE CompactBuffer (VAR ed : EDITPTR) : BOOLEAN;
(* Kompaktiert den Bufferspeicher. FALSE: Nicht mehr mglich, da
 * gar kein Speicher mehr frei war.
 *)
 VAR buffOne : EditTypes.aBufferPtr;
     freedSecond,
     success : BOOLEAN;
BEGIN
  WITH ed^ DO
    buffOne := firstBuff;
    success := TRUE;
    WHILE (buffOne # NIL) & (buffOne^.next # NIL) DO
      v.bool := compactTwo (buffOne, buffOne^.next, freedSecond);
      success := success AND v.bool;
      IF ~freedSecond
      THEN
        buffOne := buffOne^.next;
      END;
    END;
    currentBuff := firstBuff;
  END;
  RETURN success;
END CompactBuffer;

PROCEDURE GetCurrLine (VAR ed : EDITPTR; VAR str : ARRAY OF CHAR; VAR len : INTEGER);
(* Liefert die aktuelle Zeile, in der der Cursor steht.
 *)
BEGIN
  v.bool := GetLine (ed, ed^.currLineNr, str, len);
END GetCurrLine;

PROCEDURE PutCurrLine (VAR ed : EDITPTR): BOOLEAN;
(* Schiebt die aktuelle Zeile, in der der Cursor steht, wieder in den Buffer
 *)
  VAR res : BOOLEAN;
BEGIN
  WITH ed^ DO
    IF editChanged
    THEN
      (*$? Debug:
      IF editLen # INTEGER((*SYSTEM.*)LENGTH (editLine.text^))
      THEN HALT END;
      *)
      IF editLen=0
      THEN
        editLine.text^ := CR+LF+0C;
        editLen := 2;
      END;
      res := PutLine (ed, currLineNr, editLine.text^, editLen);
      IF ~res & ~CompactBuffer (ed)
      THEN 
        RETURN FALSE;
      ELSIF ~res
      THEN
        IF ~PutLine (ed, currLineNr, editLine.text^, editLen)
        THEN 
          RETURN FALSE
        END;
      END;
      editChanged := FALSE;
    END
  END;
  RETURN TRUE;
END PutCurrLine;

PROCEDURE GetBuffer (VAR ed : EDITPTR; defSize : INTEGER) : BOOLEAN;
(* Erzeugt einen neuen Buffer hinter currentBuff
 * Es wird nur der Textbuffer und der Descriptor alloziert,
 * das LineArray noch nicht.
 *)
  VAR newBuff : EditTypes.aBufferPtr;
BEGIN
  WITH ed^ DO
    newBuff := createBuffer (defSize);
    IF newBuff = NIL THEN RETURN FALSE END;
    (* Jetzt kommt das Einhngen in die Liste *)
    IF firstBuff = NIL
    THEN
      newBuff^.prev := NIL;
      newBuff^.next := NIL;
      firstBuff := newBuff;
      currentBuff := firstBuff;
    ELSE
      IF currentBuff = NIL
      THEN
        currentBuff := firstBuff;
      END;
      newBuff^.prev := currentBuff;
      newBuff^.next := currentBuff^.next;
      IF newBuff^.next # NIL
      THEN
        newBuff^.next^.prev := newBuff;
      END;
      currentBuff^.next := newBuff;
      currentBuff := newBuff;
    END;
  END;
  RETURN TRUE;
END GetBuffer;

PROCEDURE FreeBuffer (VAR ed : EDITPTR);
  VAR buff : EditTypes.aBufferPtr;
BEGIN
  WITH ed^ DO
    IF currentBuff = NIL THEN RETURN END;
    IF currentBuff^.prev # NIL
    THEN
      currentBuff^.prev^.next := currentBuff^.next;
    END;
    IF currentBuff^.next # NIL
    THEN
      currentBuff^.next^.prev := currentBuff^.prev;
    END;
    IF firstBuff = currentBuff
    THEN
      firstBuff := currentBuff^.next;
    END;
    buff := currentBuff;
    currentBuff := currentBuff^.next;
    IF buff^.buffer # NIL THEN DEALLOCATE (buff^.buffer, 0); END;
    IF buff^.lines # NIL THEN DEALLOCATE (buff^.lines, 0); END;
    DISPOSE (buff);
  END;
END FreeBuffer;

PROCEDURE CountBytes (VAR ed: EDITPTR; lineFrom : LONGINT; rowFrom: INTEGER;
                      lineTo: LONGINT; rowTo: INTEGER; VAR bytes: LONGCARD);
(* Zhlt die Bytes, die von dem Text zwischen den beiden Textmarken belegt werden
 *)
  VAR buff : EditTypes.aBufferPtr;
      offset: INTEGER;
      i     : INTEGER;
BEGIN
  WITH ed^ DO
    IF ~findLineNum (ed, lineTo) 
    OR ~findLineNum (ed, lineFrom) 
    THEN 
      bytes := 0;
      RETURN
    END;
    WITH currentBuff^ DO 
      offset := 0;
      FOR i := 0 TO SHORT(lineFrom - fromLine) - 1 DO
        INC (offset, lines^[i]);
      END;
      INC (offset, rowFrom);
      bytes := usedMem - offset;
    END;
    (* Jetzt Endebuffer suchen *)
    WHILE (currentBuff # NIL) & (lineTo > currentBuff^.toLine) DO
      INC (bytes, currentBuff^.usedMem);
      currentBuff := currentBuff^.next;
    END;
    IF currentBuff # NIL
    THEN
      offset := 0;
      WITH currentBuff^ DO
        FOR i := 0 TO SHORT(lineTo - fromLine) - 1 DO
          INC (offset, lines^[i]);
        END;
      END;
      INC (offset, rowTo);
      INC (bytes, offset);
    ELSE
      bytes := 0;
      RETURN
    END;
  END;
END CountBytes;

BEGIN
  CopyBlock := blockCopy;
END EditBase.
