IMPLEMENTATION MODULE FredProtokoll;

(*==============================================================*
 * Modul:               Protokollmodul fr Fred                 *
 * Autor:               Johannes Gttker-Schnetmann             *
 * Autor:               Dirk Steins                             *
 * erstellt am:         Juni 1993                               *
 * letzte nderung am:  12.06.1993                              *
 * Version:             1.0                                     *
 * Interne Version:     V#0002                                  *
 *==============================================================*

 Dieses Modul macht Fred zu einem Xacc-Hauptprogramm der Stufe 1. Auerdem 
 werden vom AV/VA-Protokoll die sinnvollen Msgs untersttzt.
 
 Entstanden ist dieses Modul aus dem Protokoll-Modul von CAT.

 Die Erluterung des erweiterten Protokolls steht weiter unten. "HelloWatchDog"
 hat jetzt die allgemeinere Bedeutung "CatMsg".
 Insbesondere die Erkennung des WatchDogs soll in einer der nchsten Versionen
 nach Mglichkeit auf die Methode des erweiterten Protokolls umgestellt werden.

 *----------------------------------------------------------------------------
 * Datum    Vers. Autor  nderung (Arbeitsbericht)                            
 *----------------------------------------------------------------------------
 * 01.02.92 0002  JGS    Protokollerweiterung fr Steueraccs
 * 24.05.93 0003  DS     Erweiterung um MultiTasking-Untersttzung
 *                       CAT XACC-Versionsnummer gendert auf 20H
 *----------------------------------------------------------------------------
 *)

FROM SYSTEM          IMPORT ADR, ADDRESS, WORD, CADR, CALLSYS, TSIZE;
FROM WdwManager      IMPORT NewWindowIsTop, WindowIsClosed;
FROM Void            IMPORT v;
FROM CatTypes        IMPORT Str255Ptr, String255;
FROM CatGlobal       IMPORT multiTask, multiTOS, magIx, magIxVer;
FROM Storage         IMPORT ALLOCATE;
IMPORT BinOps;
IMPORT FileNames;

IMPORT MagicAES;
IMPORT MagicStrings;
IMPORT mtAppl;
IMPORT mtAlerts;
IMPORT MagicDOS;
IMPORT MagicCookie;
IMPORT mtCommand;

FROM SEProto IMPORT tShellSet, tEditSet, tEditCommands, tShellCommands;
IMPORT SEProto;

CONST

  appName    = 'FRED    '+0C+'XDSC'+0C+'1Texteditor'+0C+'2ED'+0C+0C;
  
  supportedSEProtoVersion = 0100H;
  

(* XAcc: *)

  ACC_ID    = 400H;
  ACC_OPEN  = 401H;
  ACC_CLOSE = 402H;
  ACC_ACC   = 403H;
  ACC_EXIT  = 404H;
  ACC_ACK   = 500H;
  ACC_TEXT  = 501H;
  ACC_KEY   = 502H;
  
(* PC-Help Protokoll *)
  AC_HELP   = 1025;         (* message to DA *)
  AC_REPLY  = 1026;         (* message from DA *)
  AC_VERSION= 1027;         (* Ask DA for version *)
  AC_COPY   = 1028;         (* Ask DA to copy screen to clipboard *)

(* AV/VA-Protokoll *)

  AV_PROTOKOLL     = 4700H;
  VA_PROTOSTATUS   = 4701H;

(* not used *)
  AV_GETSTATUS     = 4703H;
  AV_STATUS        = 4704H;
  VA_SETSTATUS     = 4705H;
(*----------*)

  AV_SENDKEY       = 4710H;

(* not used *)
  VA_START         = 4711H;
  AV_ASKFILEFONT   = 4712H;
  VA_FILEFONT      = 4713H;
  AV_ASKCONFONT    = 4714H;
  VA_CONFONT       = 4715H;
  AV_ASKOBJECT     = 4716H;
  VA_OBJECT        = 4717H;
  AV_OPENCONSOLE   = 4718H;
  VA_CONSOLEOPEN   = 4719H;
(*----------*)

  AV_OPENWIND      = 4720H;
  VA_WINDOPEN      = 4721H;
  AV_STARTPROG     = 4722H;
  VA_PROGSTART     = 4723H;
  AV_ACCWINDOPEN   = 4724H;

(* not used *)
  VA_DRAGACCWIND   = 4725H; (* NOCH, wichtig fr Kommandozeile an Takeoff *)
(*----------*)

  AV_ACCWINDCLOSED = 4726H;
  AV_PATH_UPDATE   = 4730H;

(* Bits *)

  xAV_SENDKEY       = 0;
  xAV_ASKFILEFONT   = 1;
  xAV_ASKCONFONT    = 2; (* auch xAV_OPENCONSOLE *)
  xAV_ASKOBJECT     = 3;
  xAV_OPENWIND      = 4;
  xAV_STARTPROG     = 5;
  xAV_ACCWINDOPEN   = 6; (* auch xAV_ACCWINDCLOSED *)
  xAV_STATUS        = 7; (* auch AV_GETSTATUS *)


(* Um sich alle installierten Accs beim XAcc-Protokoll zu merken wird hier ein *) 
(* kleines Array angelegt. Das ist natrlich statisch, aber darber wrde ich  *)
(* mir erstmal noch keine Sorgen machen.. :-)                                  *)

CONST maxAcc = 40;      (* Wegen MTOS von 6 auf 40 erweitert. Zwar kann 
                         * man unter MTOS auch mehr als 40 ACCs und Applikationen 
                         * haben, aber das wre ziemlich krank *)

CONST maxBufLen = 128;  (* Lnge des statischen Textbuffers *)

      memDescrSize  = 2048; (* 1 kB globaler Speicher *)

TYPE        (* for systems with memory protection (MultiTOS) *)
        memDescr        = RECORD
                            CASE : BOOLEAN OF
                            TRUE:
                              apName  : ARRAY [0..49] OF CHAR; (* our extended application name *)
                              textBuf : ARRAY [0..maxBufLen-1] OF CHAR;
                              mPath   : String255;
                              mName   : String255; 
                              mRec    : SEProto.tMessRec; |
                            FALSE:
                              allInOne: ARRAY [0..memDescrSize-1] OF CHAR;
                            END;
                          END;

TYPE sePartner  = RECORD
                    apId    : INTEGER;
                    protoVersion: CARDINAL;
                    shell   : tShellSet;
                    editor  : tEditSet;
                  END;

CONST   maxSe   = 3;

VAR accCount : CARDINAL;
    acc      : ARRAY[1..maxAcc] OF INTEGER;
    messBuff : ARRAY [0..31] OF INTEGER;    (* Platz fr Messagebuffer *)
    startApId: INTEGER;     (* Von wem kam die letzte AV_STARTPROG *)
    memPtr   : POINTER TO memDescr;         (* global buffer *)
    seList   : ARRAY [1..maxSe] OF sePartner;
    seCount  : INTEGER;


PROCEDURE send(msg : CARDINAL; dest : INTEGER; REF data : ARRAY OF WORD);
(* Anpassen, damit man damit demnchst lngere Msgs verschicken kann *)
VAR mess : POINTER TO RECORD
             msg : CARDINAL;
             id  : INTEGER;
             overLength : INTEGER;
             w : ARRAY[0..4] OF WORD
           END;
   z : CARDINAL;
BEGIN
  mess := ADR(messBuff);
  FOR z := 0 TO 4 DO mess^.w[z] := data[z] END;
  mess^.msg := msg;
  mess^.id  := mtAppl.ApplIdent;
  mess^.overLength := 0;
  MagicAES.ApplWrite(dest, 16, mess^);
  MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Um es auch mal abzuschicken.. *)
  MagicAES.WindUpdate(MagicAES.BEGUPDATE);
END send;

PROCEDURE SendMess(msg : CARDINAL; catType : INTEGER; l : LONGCARD; i2, i3, dest : INTEGER);
VAR data : RECORD catType : INTEGER; l : LONGCARD; i2, i3 : INTEGER END;
BEGIN
  data.catType := catType; data.l := l; data.i2 := i2; data.i3 := i3;
  send(msg, dest, data);
END SendMess;

PROCEDURE SendMessInt(msg : CARDINAL; catType, i1, i2, i3, i4, dest : INTEGER);
VAR data : RECORD catType, i1, i2, i3, i4 : INTEGER END;
BEGIN
  data.catType := catType; data.i1 := i1; data.i2 := i2; data.i3 := i3; data.i4 := i4;
  send(msg, dest, data);
END SendMessInt;

PROCEDURE SendAdr(msg : CARDINAL; catType, i1 : INTEGER; l : LONGCARD; i2, dest : INTEGER);
VAR data : RECORD catType : INTEGER; i1 : INTEGER; l : LONGCARD; i2 : INTEGER END;
BEGIN
  data.catType := catType; data.i1 := i1; data.l := l; data.i2 := i2;
  send(msg, dest, data);
END SendAdr;

PROCEDURE SendTerminate();
(* Nachricht an den WatchDog schicken, da Cat beendet wurde *)
  VAR m : POINTER TO RECORD
            messId : CARDINAL;
            apId   : INTEGER;
            over   : INTEGER;
            r      : ARRAY[3..7] OF INTEGER;
          END;
      i : CARDINAL;
      mess: SEProto.tPtrMessRec;
BEGIN
  (* Und unter MultiTOS auch noch ACC_EXIT verschicken an alle 
   * anderen Applikationen, die sich gemeldet haben
   *)
  IF multiTask
  THEN
    m := ADR(messBuff);
    WITH m^ DO
      messId := ACC_EXIT;
      apId   := mtAppl.ApplIdent;
      over   := 0;
    END;
    FOR i := 1 TO accCount - 1 DO
      MagicAES.ApplWrite(acc[i], 16, m^);
      MagicAES.WindUpdate(MagicAES.ENDUPDATE);
      MagicAES.WindUpdate(MagicAES.BEGUPDATE);
    END;
    (* Jetzt an alle SE-Applikationen noch ein ES_QUIT schicken *)
    mess := ADR(messBuff);
    WITH mess^ DO
      msg := SEProto.ES_QUIT;
      apId := mtAppl.ApplIdent;
      overLen := 0;
      FOR i := 0 TO 4 DO
        data[i] := 0;
      END;
    END;
    i := 1; 
    WHILE i <= CARDINAL(seCount) DO
      MagicAES.ApplWrite(seList[i].apId, 16, mess^);
      MagicAES.WindUpdate(MagicAES.ENDUPDATE);
      MagicAES.WindUpdate(MagicAES.BEGUPDATE);
      INC (i);
    END;
  ELSE
    accCount := 1;
  END;
END SendTerminate;

PROCEDURE SendProgstart(error : BOOLEAN);
(* dem Acc mitteilen, ob der Programmstart geklappt hat, not yet implementet *)
  VAR val : INTEGER;
BEGIN
  IF error THEN val := 0 ELSE val := 1 END;
  SendMess(VA_PROGSTART, val, 0,0,0, startApId);
END SendProgstart;

PROCEDURE SendAcknowledge (apId : INTEGER; succ: BOOLEAN);
  VAR val : INTEGER;
BEGIN
  IF succ THEN val := 1 ELSE val := 0 END;
  SendMess (ACC_ACK, val, 0,0,0, apId);
END SendAcknowledge;

PROCEDURE SendVaWindopen (apId : INTEGER; error : BOOLEAN);
(* Das ist unbenutzt, ist wohl auch nur in Gemini sinnvoll *)
BEGIN
  IF error 
  THEN
    SendMess(VA_WINDOPEN, 0, 0,0,0, apId);
  ELSE
    SendMess(VA_WINDOPEN, 1, 0,0,0, apId);
  END;
END SendVaWindopen;

PROCEDURE SendOpenwind (id : INTEGER; path, name : ADDRESS);
(* einer anderen Applikation sagen, das sie ein Fenster ffnen soll
 * ist fr den Fall, das CAT oder Eddix schon geffnet sind, dann wird 
 * diese Message verschickt.
 *)
  VAR mess : POINTER TO RECORD
               messId : INTEGER;
               apId   : INTEGER;
               over   : INTEGER;
               fPath              : ADDRESS;
               fName              : ADDRESS;
               e15                : INTEGER;
             END;
     fPtr  : Str255Ptr;
BEGIN
  mess := ADR (messBuff);
  (* Filenamen und Pfad kopieren *)
  fPtr := path;
  MagicStrings.Assign (fPtr^, memPtr^.mPath);
  fPtr := name;
  MagicStrings.Assign (fPtr^, memPtr^.mName);
  WITH mess^ DO
    messId := AV_OPENWIND;
    apId   := mtAppl.ApplIdent;
    over   := 0;
    fPath  := ADR (memPtr^.mPath);
    fName  := ADR (memPtr^.mName);
    e15    := 0;
  END;
  MagicAES.ApplWrite(id, 16, mess^);
  MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Um es auch mal abzuschicken.. *)
  MagicAES.WindUpdate(MagicAES.BEGUPDATE);
END SendOpenwind;

PROCEDURE SendHelp (VAR str : ARRAY OF CHAR);
  (* Request Help via PC_HELP Protokoll *)
  VAR accId : INTEGER;
      mess  : POINTER TO RECORD
               messId : INTEGER;
               apId   : INTEGER;
               over   : INTEGER;
               data   : ADDRESS;
               fName  : ADDRESS;
               e15    : INTEGER;
             END;
     helpApp: ARRAY [0..9] OF CHAR; 
     i      : INTEGER;
BEGIN
  mess := ADR(messBuff);
  (* String kopieren in statischen Buffer (igitt) *)
  i := 0;
  WHILE (i < maxBufLen-1) & (str[i] # 0C) DO
    memPtr^.textBuf[i] := str[i];
    INC (i);
  END;
  memPtr^.textBuf [i] := 0C;
  (* Search for PC/TC-HELP Accessory *)
  MagicStrings.Assign ('ST-GUIDE', helpApp);
  accId := MagicStrings.Length (helpApp);
  (* Extend name with Spaces *)
  WHILE accId < 8 DO 
    helpApp[accId] := ' '; 
    helpApp[accId+1] := ''; 
    INC (accId);
  END;
  accId := MagicAES.ApplFind (ADR(helpApp));
  IF accId >= 0
  THEN
    WITH mess^ DO
      messId := AC_HELP;
      apId   := mtAppl.ApplIdent;
      over   := 0;
      data   := ADR(memPtr^.textBuf);
    END;
    MagicAES.ApplWrite (accId, 16, mess^);
  END;
END SendHelp;

PROCEDURE SendSeAcknowledge (apId: INTEGER; res: BOOLEAN);
  VAR mess  : SEProto.tPtrMessRec;
      i     : INTEGER;
BEGIN
  mess := ADR (messBuff);
  WITH mess^ DO
    msg := SEProto.ES_ACK;
    apId  := mtAppl.ApplIdent;
    overLen := 0;
    FOR i := 0 TO 4 DO data[i] := 0 END;
    acknowledge := res;
  END;
  MagicAES.ApplWrite (apId, 16, mess^);
END SendSeAcknowledge;

PROCEDURE SendEsMessage (messNum: INTEGER; adr: ADDRESS);
  VAR mess  : SEProto.tPtrMessRec;
      fPtr  : Str255Ptr;
      i     : INTEGER;
BEGIN
  IF seCount < 1 THEN RETURN END;
  mess := ADR (messBuff);
  fPtr := adr;
  IF fPtr # NIL
  THEN
    MagicStrings.Assign (fPtr^, memPtr^.mPath);
    fPtr := ADR(memPtr^.mPath);
  END;
  IF ~(tEditCommands(messNum-SEProto.ES_INIT) IN seList[1].editor)
  THEN RETURN END;
  WITH mess^ DO
    msg := messNum;
    apId  := mtAppl.ApplIdent;
    overLen := 0;
    FOR i := 0 TO 4 DO data[i] := 0 END;
    filePtr := ADDRESS(fPtr);
  END;
  MagicAES.ApplWrite (seList[1].apId, 16, mess^);
END SendEsMessage;

PROCEDURE InitSeProto (adr: ADDRESS);
  VAR mess : SEProto.tPtrMessRec;
      newMess: SEProto.tPtrMessRec;
      i     : INTEGER;
BEGIN
  mess := adr;
  newMess := ADR (messBuff);
  (* Eintragen in SeProtoliste *)
  WITH newMess^ DO
    msg := SEProto.ES_OK;
    apId  := mtAppl.ApplIdent;
    overLen := 0;
    FOR i := 0 TO 4 DO data[i] := 0 END;
    okShellBits := tShellSet {seInit, seOk, seAck, seOpen, seError, seErrfile, seQuit,
                    seTerminate, seClose};
    okEditBits  := tEditSet {esInit, esOk, esAck, esCompile, esMake, esMakeall, 
                    esLink, esExec, esMakeexec, esQuit};
    okVersion   := supportedSEProtoVersion;
    otherId := mess^.apId;
  END;
  (* Antwort zusammenbasteln *)
  IF seCount < maxSe 
  THEN
    INC (seCount);
    WITH seList[seCount] DO
      apId := mess^.apId;
      (* Minimale Daten erstellen *)
      protoVersion := BinOps.LowerCard (supportedSEProtoVersion, mess^.version);
      shell := mess^.shellBits * newMess^.okShellBits;
      editor := mess^.editBits * newMess^.okEditBits;
    END;
  END;
  (* Jetzt Antwort schicken *)
  MagicAES.ApplWrite (mess^.apId, 16, newMess^);
END InitSeProto;

PROCEDURE ActivateSeProto (adr: ADDRESS);
  VAR mess : SEProto.tPtrMessRec;
BEGIN
  mess := adr;
  IF mess^.otherId # mtAppl.ApplIdent THEN RETURN END;
  (* Eintragen in Liste *)
  IF seCount < maxSe 
  THEN
    INC (seCount);
    WITH seList[seCount] DO
      apId := mess^.apId;
      (* Minimale Daten erstellen *)
      protoVersion := BinOps.LowerCard (supportedSEProtoVersion, mess^.okVersion);
      shell := mess^.okShellBits * tShellSet {seInit, seOk, seAck, seOpen, seError, seErrfile, seQuit,
                    seTerminate, seClose};
      editor := mess^.okEditBits * tEditSet {esInit, esOk, esAck, esCompile, esMake, esMakeall, 
                    esLink, esExec, esMakeexec, esQuit};
    END;
  END;
END ActivateSeProto;

PROCEDURE RemoveSeProto (apId : INTEGER);
  VAR i : INTEGER;
BEGIN
  i := 1;
  WHILE (i <= seCount) & (seList[i].apId # apId) DO INC (i) END;
  IF (i <= seCount) & (i <= maxSe) THEN
    FOR i := i+1 TO seCount DO
      acc[i-1] := acc[i];
    END;
    DEC (seCount);
  END;
END RemoveSeProto;

PROCEDURE IsProtokoll(mess : ADDRESS;
                     VAR kReturn : INTEGER; VAR mokState : BITSET;
                     VAR pName, pCmd : Str255Ptr):React;

  PROCEDURE accInit(mess : ADDRESS);
  (* XAcc-Protokoll, accInit behandeln *)
  VAR i : CARDINAL;
      access : POINTER TO RECORD
                 messId : INTEGER;
                 apId   : INTEGER;
                 over   : INTEGER;
                 version, protoStep : CHAR;
                 name               : Str255Ptr;
                 menuId             : INTEGER;
                 accId              : INTEGER
               END;
    PROCEDURE checkId (id : INTEGER): BOOLEAN;
    (* Prft, ob schon ein Prozess mit der ID registriert ist *)
      VAR i : CARDINAL;
    BEGIN
      FOR i := 1 TO accCount - 1 DO 
        IF acc[i] = id THEN RETURN FALSE END;
      END;
      RETURN TRUE;
    END checkId;
    
  BEGIN
    access := mess;
    IF multiTask
    THEN
      IF (accCount < maxAcc) & checkId (access^.apId)
      THEN
        acc[accCount]     := access^.apId; (* ID des Acc, das sich gerade anmeldet *)
        access^.messId    := ACC_ACC;
        access^.apId      := mtAppl.ApplIdent;
        access^.version   := CHR(20H);
        access^.protoStep := 1C;
        access^.name      := ADR(memPtr^.apName);
        access^.menuId    := 0;
        access^.accId     := 0;

        MagicAES.ApplWrite(acc[accCount], 16, access^);
        MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Abschicken *)
        MagicAES.WindUpdate(MagicAES.BEGUPDATE);
        INC(accCount);
      END;
    ELSE (* not MultiTOS *)
      IF (accCount < maxAcc) & checkId (access^.apId) THEN
  
        acc[accCount]     := access^.apId; (* ID des Acc, das sich gerade anmeldet *)
        access^.messId    := ACC_ACC;
        access^.accId     := access^.apId;
        access^.apId      := mtAppl.ApplIdent;
        access^.over      := 0;
        (* Die anderen Daten stimmen zu diesem Zeitpunkt schon.. *)
  
        FOR i := 1 TO accCount-1 DO
          MagicAES.ApplWrite(acc[i], 16, access^);
          MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Abschicken *)
          MagicAES.WindUpdate(MagicAES.BEGUPDATE);
        END; (* Alle anderen Accs benachrichtigen *)
                                   
        access^.messId    := ACC_ID;
        access^.version   := CHR(20H);
        access^.protoStep := 1C;
        access^.name      := ADR(memPtr^.apName);
        access^.menuId    := 0;
        access^.accId     := 0;
  
        MagicAES.ApplWrite(acc[accCount], 16, access^);
        INC(accCount);
  
        MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Abschicken *)
        MagicAES.WindUpdate(MagicAES.BEGUPDATE);
      END; 
    END;
  END accInit;

  PROCEDURE SendVaProtostatus(mess : ADDRESS);
  VAR id : INTEGER;
      access : POINTER TO RECORD
                 messId : INTEGER;
                 apId   : INTEGER;
                 over   : INTEGER;
                 b21, b22, b23      : BITSET;
                 prgName            : ADDRESS;
               END;
  BEGIN
    access := mess;
    id             := access^.apId;
    access^.messId := VA_PROTOSTATUS;
    access^.apId   := mtAppl.ApplIdent;
    access^.over   := 0;
    access^.b21    := {xAV_SENDKEY, xAV_ACCWINDOPEN, xAV_OPENWIND};
    access^.b22    := {};
    access^.b23    := {};
    access^.prgName:= ADR(memPtr^.apName);
    MagicAES.ApplWrite(id, 16, access^);
    MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Abschicken *)
    MagicAES.WindUpdate(MagicAES.BEGUPDATE);
  END SendVaProtostatus;
  
  PROCEDURE RemoveAcc (apId : INTEGER);
    VAR i : CARDINAL;
  BEGIN
    i := 1;
    WHILE (i < accCount) & (acc[i] # apId) DO INC (i) END;
    IF i < accCount THEN
      FOR i := i TO accCount - 2 DO
        acc[i] := acc[i+1];
      END;
      DEC (accCount);
    END;
  END RemoveAcc;

VAR access : POINTER TO RECORD
               messId : INTEGER;
               apId   : INTEGER;
               over   : INTEGER;
               CASE :CARDINAL OF
               (* AV_SENDKEY *)
                 1 : mokState       : INTEGER;
                     kReturn        : INTEGER;
                     e11, e12, e13  : INTEGER|
               (* AV_STARTPROG *)
                 2 : pName, pCmd    : ADDRESS;
                     e14            : INTEGER|
               (* AV_OPENWIND  *)
                 3 : fPath          : ADDRESS;
                     fName          : ADDRESS;
                     e15            : INTEGER |
                 4 : i              : ARRAY[0..4] OF INTEGER|
                 5 : e16            : INTEGER;
                     text           : ADDRESS; |
               END;
             END;

(* TYPE React        = (rNoproto, rNone, rChar, rText); *)
BEGIN
  access := mess;
  (* Zunchst einmal die Initialisierungsmsgs fr XAcc und VA *)
  CASE access^.messId OF
    ACC_ID      : accInit(mess); |     (* Xacc-Protokoll behandeln *)
  (* AccOpen und AccClose werden einfach berhrt! *)
    ACC_ACC     : RETURN rNone; |
    ACC_EXIT    : (* XACC Acc_Exit *)
                  (* remove Accessorie from acc list *)
                  RemoveAcc (access^.apId); |
    ACC_KEY     : (* Tastendruck durchreichen *)
                  (* Genau umgedreht zu AV_SENDKEY *)
                  kReturn := access^.mokState;
                  mokState:= BITSET(access^.kReturn);
                  SendAcknowledge (access^.apId, TRUE);
                  RETURN rChar; |
    ACC_TEXT    : (* get text via XACC *)
                  kReturn := access^.apId;    (* apId merken fr Acknowledge *)
                  pName   := access^.text;    (* Pointer auf Text *)
                  RETURN rText; |
  ELSE
  END;
  CASE access^.messId OF
    AV_PROTOKOLL: (* hier behandelt, WatchDog suchen *)
                  SendVaProtostatus(mess); |  (* .. und dem Acc sagen, was wir knnen *)
    (* AV-Msgs, die direkt hier behandelt werden *)
    AV_SENDKEY  : (* Tastendruck durchreichen *)
                  kReturn := access^.kReturn;
                  mokState:= BITSET(access^.mokState);
                  RETURN rChar; |
    AV_ACCWINDOPEN: (* AccFenster auf -> rein in die Queue *)
                  NewWindowIsTop(access^.i[0], access^.apId); |
    AV_ACCWINDCLOSED:  (* AccFenster zu -> raus aus der Queue *)
                  WindowIsClosed(access^.i[0]); |
    AV_STARTPROG: (* Acc sagt: Folgendes Programm starten! *)
                  pName := access^.pName; pCmd := access^.pCmd;
                  startApId := access^.apId;
                  RETURN rStart; |
    AV_OPENWIND:  (* Woasn doas? *)
                  pName := access^.fPath; pCmd := access^.fName;
                  SendVaWindopen(access^.apId, FALSE);
                  RETURN rOpen; |
    VA_START:     (* Unter Mag!X: Neue Dokumente auf CAT gezogen oder so *)
                  pName := NIL; pCmd := access^.fPath;
                  (* Besttigung gibt es keine *)
                  RETURN rOpen; |
  ELSE
  END;
  CASE access^.messId OF
    SEProto.SE_INIT: InitSeProto (access); |
    SEProto.SE_OK:   ActivateSeProto (access); |
    SEProto.SE_ACK:  (* Nichts weiter machen *) |
    SEProto.SE_OPEN:      
                  (* Analog zu VA_START behandeln *)
                  pName := NIL; pCmd := access^.fPath;
                  startApId := access^.apId;
                  SendSeAcknowledge (access^.apId, TRUE);
                  RETURN rOpen; |
    SEProto.SE_ERROR,
    SEProto.SE_ERRFILE,
    SEProto.SE_TERMINATE,
    SEProto.SE_CLOSE:     
                  (* Einfach Pointer auf Messagebuffer zurckgeben
                   *)
                  pCmd := ADDRESS(access);
                  SendSeAcknowledge (access^.apId, TRUE); 
                  RETURN rSeMess; |
    SEProto.SE_QUIT:      
                  RemoveSeProto (access^.apId); |
  ELSE
  END;
  RETURN rNoproto;

(*
  IF    access^.messId = ACC_ID   THEN             (* hier behandelt, WatchDog suchen *)
    accInit(mess);      (* Xacc-Protokoll behandeln *)
    RETURN rNone;
  (* AccOpen und AccClose werden einfach berhrt! *)
  ELSIF (access^.messId = ACC_ACC) & multiTask THEN  (* hier behandelt, WatchDog suchen *)
    RETURN rNone;
  (* AccOpen und AccClose werden einfach berhrt! *)
  ELSIF access^.messId = AV_PROTOKOLL THEN         (* hier behandelt, WatchDog suchen *)
    SendVaProtostatus(mess);                  (* .. und dem Acc sagen, was wir knnen *)
    RETURN rNone;
  
  ELSIF access^.messId = ACC_EXIT THEN                    (* XACC Acc_Exit *)
    (* remove Accessorie from acc list *)
    RemoveAcc (access^.apId);
    RETURN rNone;
  ELSIF access^.messId = ACC_KEY THEN                     (* Tastendruck durchreichen *)
    (* Genau umgedreht zu AV_SENDKEY *)
    kReturn := access^.mokState;
    mokState:= BITSET(access^.kReturn);
    SendAcknowledge (access^.apId, TRUE);
    RETURN rChar;
  ELSIF access^.messId = ACC_TEXT THEN
    (* get text via XACC *)
    kReturn := access^.apId;    (* apId merken fr Acknowledge *)
    pName   := access^.text;    (* Pointer auf Text *)
    RETURN rText;
    
  (* AV-Msgs, die direkt hier behandelt werden *)
  ELSIF access^.messId = AV_SENDKEY   THEN                (* Tastendruck durchreichen *)
    kReturn := access^.kReturn;
    mokState:= BITSET(access^.mokState);
    RETURN rChar;
  ELSIF access^.messId = AV_ACCWINDOPEN  THEN  (* AccFenster auf -> rein in die Queue *)
    NewWindowIsTop(access^.i[0], access^.apId);
    RETURN rNone;
  ELSIF access^.messId = AV_ACCWINDCLOSED THEN (* AccFenster zu -> raus aus der Queue *)
    WindowIsClosed(access^.i[0]);
    RETURN rNone;
  ELSIF access^.messId = AV_STARTPROG THEN   (* Acc sagt: Folgendes Programm starten! *)
    pName := access^.pName; pCmd := access^.pCmd;
    startApId := access^.apId;
    RETURN rStart;
  ELSIF access^.messId = AV_OPENWIND THEN                              (* Woasn doas? *)
    pName := access^.fPath; pCmd := access^.fName;
    SendVaWindopen(access^.apId, FALSE);
    RETURN rOpen;
  ELSIF access^.messId = VA_START THEN
    (* Unter Mag!X: Neue Dokumente auf CAT gezogen oder so *)
    pName := NIL; pCmd := access^.fPath;
    (* Besttigung gibt es keine *)
    RETURN rOpen;
  ELSE (* Kein bekanntes Protokoll *)
    RETURN rNoproto;
  END;
*)
END IsProtokoll;

PROCEDURE ProtoInit ();
(* XACC-Initialisierung unter MTOS *)
  VAR i : INTEGER;
      access : POINTER TO RECORD
                 messId : INTEGER;
                 apId   : INTEGER;
                 over   : INTEGER;
                 version, protoStep : CHAR;
                 name               : Str255Ptr;
                 menuId             : INTEGER;
                 accId              : INTEGER
               END;
     apName : String255;
     apMode,
     apType,
     apId   : INTEGER;
BEGIN
  access := ADR(messBuff);
  access^.messId    := ACC_ID;
  access^.apId      := mtAppl.ApplIdent;
  access^.version   := CHR(20H);
  access^.protoStep := 1C;
  access^.name      := ADR(memPtr^.apName);
  access^.menuId    := 0;
  access^.accId     := 0;
  IF multiTOS OR (magIx & (magIxVer >= $200))
  THEN
    apMode := MagicAES.APFIRST;
    WHILE MagicAES.ApplSearch (apMode, apName, apType, apId) DO
      apMode := MagicAES.APNEXT;
      IF (apType # 1) & (apId # mtAppl.ApplIdent)
      THEN
        MagicAES.ApplWrite(apId, 16, access^);
      END;
    END;
  ELSIF magIx & (magIxVer < $200)
  THEN
    FOR i := 0 TO MagicAES.AESGlobal.apCount - 1 DO
      apName[0] := '?';
      apName[1] := 0C;
      apName[2] := CHR(i);
      apName[3] := 0C;
      apId := MagicAES.ApplFind (ADR(apName));
      IF (apId # 0) & (apId # mtAppl.ApplIdent)
      THEN 
        MagicAES.ApplWrite(apId, 16, access^);
      END;
    END;
  END;
END ProtoInit;

PROCEDURE SeProtoInit ();
(* SE-Protokoll Initialisierung unter MTOS *)
  VAR i : INTEGER;
      access : SEProto.tPtrMessRec;
      apName : String255;
      apMode,
      apType,
      apId   : INTEGER;
BEGIN
  access := ADR(messBuff);
  access^.msg       := SEProto.ES_INIT;
  access^.apId      := mtAppl.ApplIdent;
  access^.overLen   := 0;
  access^.shellBits := tShellSet {seInit, seOk, seAck, seOpen, seError, seErrfile, seQuit,
                               seTerminate, seClose};
  access^.editBits  := tEditSet {esInit, esOk, esAck, esCompile, esMake, esMakeall, 
                                   esLink, esExec, esMakeexec, esQuit};
  access^.version   := supportedSEProtoVersion;

  IF multiTOS OR (magIx & (magIxVer >= $200))
  THEN
    apMode := MagicAES.APFIRST;
    WHILE MagicAES.ApplSearch (apMode, apName, apType, apId) DO
      apMode := MagicAES.APNEXT;
      IF (apType # 1) & (apId # mtAppl.ApplIdent)
      THEN
        MagicAES.ApplWrite(apId, 16, access^);
      END;
    END;
  ELSIF magIx & (magIxVer < $200)
  THEN
    FOR i := 0 TO MagicAES.AESGlobal.apCount - 1 DO
      apName[0] := '?';
      apName[1] := 0C;
      apName[2] := CHR(i);
      apName[3] := 0C;
      apId := MagicAES.ApplFind (ADR(apName));
      IF (apId # 0) & (apId # mtAppl.ApplIdent)
      THEN 
        MagicAES.ApplWrite(apId, 16, access^);
      END;
    END;
  END;
END SeProtoInit;

PROCEDURE SEProtoActiv (): BOOLEAN;
(* Gibt zurck, ob das SE-Protokoll aktiv ist
 *)
BEGIN
  RETURN seCount >= 1;
END SEProtoActiv;

PROCEDURE FindAVServer (VAR id: INTEGER): BOOLEAN;
  CONST cAVServer = "AVSERVER=";
  VAR avServer : ARRAY [0..79] OF CHAR;
      i        : INTEGER;
      apType   : INTEGER;
BEGIN
  FOR i := 0 TO 79 DO avServer[i] := 0C; END;
  id := -1;
  IF mtCommand.EnvVar (cAVServer, avServer)
  THEN
    FOR i := 0 TO 7 DO 
      IF avServer[i] = 0C
      THEN
        avServer[i] := ' ';
      END;
    END;
    avServer[8] := 0C;
    id := MagicAES.ApplFind (ADR(avServer));
  END;
  IF id < 0
  THEN
    (* Nach Applikation AVSERVER suchen *)
    MagicStrings.Assign (cAVServer, avServer);
    avServer[8] := 0C;
    id := MagicAES.ApplFind (ADR(avServer));
  END;
  IF id < 0
  THEN
    (* Shell mit appl_search suchen *)
    IF multiTOS OR (magIx & (magIxVer >= $200))
    THEN
      IF MagicAES.ApplSearch (MagicAES.APSHELL, avServer, apType, id)
      THEN
        IF apType # MagicAES.APTAPP
        THEN
          id := -1;
        END;
      END;
    END;
  END;
  RETURN id >= 0;
END FindAVServer;

PROCEDURE SendPathUpdate (REF path: ARRAY OF CHAR);
(* Schickt ein AV_PATH_UPDATE an den AV_SERVER *)
  VAR mess : POINTER TO RECORD
               messId : INTEGER;
               apId   : INTEGER;
               over   : INTEGER;
               fPath              : ADDRESS;
               fName              : ADDRESS;
               e15                : INTEGER;
             END;
     avId : INTEGER;
BEGIN
  IF ~FindAVServer (avId)
  THEN
    (* Wir schicken das an ID 0 *)
    avId := 0;
  END;
  (* Wir schicken nichts an uns selbst *)
  IF (avId = mtAppl.ApplIdent) THEN RETURN END;
  mess := ADR (messBuff);
  MagicStrings.Assign (FileNames.FilePath(path), memPtr^.mPath);
  WITH mess^ DO
    messId := AV_PATH_UPDATE;
    apId   := mtAppl.ApplIdent;
    over   := 0;
    fPath  := ADR (memPtr^.mPath);
    fName  := NIL;
    e15    := 0;
  END;
  MagicAES.ApplWrite(avId, 16, mess^);
END SendPathUpdate;

VAR hasMxAlloc  : BOOLEAN;
    i           : CARDINAL;
    tPtr        : Str255Ptr;

BEGIN
  accCount := 1;
  seCount  := 0;
  hasMxAlloc := MagicDOS.Mxalloc ($FFFFFFFF,3) # ADDRESS(-32L);
  IF multiTask & hasMxAlloc
  THEN
    (* Mxalloc: Readable, TT-RAM preferred *)
    memPtr := MagicDOS.Mxalloc (TSIZE (memDescr), $43);
  ELSE
    ALLOCATE (memPtr, TSIZE (memDescr));
  END;
  IF memPtr = NIL THEN HALT END;
  tPtr := CADR (appName);
  FOR i := 0 TO 49 DO
    memPtr^.apName[i]   := tPtr^[i];
  END;
END FredProtokoll.
