(* Konzept fr neue Gruppenauswahl und neue interne Gruppenverwaltung 
 * 
 * Es gibt ein Modul, das eine Gruppenliste verwaltet. Diese Gruppenliste
 * wird aus einer eigenen Datei aufgebaut, die von CAT erstellt wird.
 * Die Datei beinhaltet folgendes:
 *
 * Lokalen Gruppennamen
 * Netzgruppennamen
 * Chef der Gruppe
 * Beschreibung der Gruppe
 * Follow-Up to (erst ab Maus 9 im ITG)
 * Sysop-Info fr die Gruppe
 * Status der Gruppe (aus ITG)
 * Lesereihenfolge (diese Gruppe an x-ter Stelle zu lesen)
 * interne Gruppennummer
 *
 * Diese Datei wird von CAT aus dem ITG und der Gruppen.INF erstellt. 
 * Wenn die Datei nicht existiert, dann wird sie neu erstellt, wobei dann 
 * die Lesereihenfolge nicht festgelegt ist. 
 * Wenn ein ITG neu kommt, dann wird die Datei mit der neuen ITG abgeglichen 
 * und ggf. korrigiert. Ebenso, wenn neue Gruppen angelegt werden.
 * 
 * Das Modul, da diese Liste verwaltet, stellt Routinen zur Verfgung, um
 * auf der Liste zu arbeiten und Teillisten fr verschiedene Funktionen
 * zur Verfgung zu haben. Es stellt mindestens folgende Listen zur Verfgung:
 *
 * Komplette Gruppenliste fr Gruppenauswahl beim An-/Abstellen 
 * Gruppenliste der in CAT vorhandenen Gruppen
 * Gruppenliste mit Chefstatus
 * Gruppenliste mit allen Gruppen mit Schreiberlaubnis (fr neue Nachrichten)
 * 
 * Datenkonsistenz:
 *
 * Die Datei mu immer dann bearbeitet werden, wenn eine neue Gruppe in CAT
 * ankommt, ein neues ITG (oder IGK!) kommt oder die Lesereihenfolge gendert
 * wird. Es ist darauf zu achten, da auch CatPutz diese Datei beim Lschen
 * einer Gruppe mit pflegt. 
 *
 * Damit die Gruppenliste nur in diesem Modul verwaltet wird, mssen Funktionen
 * fr alle mglichen Operationen vorhanden sein, auch zum Lschen von Eintrgen.
 * Auch das Umbenennen von Gruppen sollte mglich sein. Auch die Datei GRUPPEN.INF
 * sollte danach nur noch von diesem Modul behandelt werden. Sollte keine ITG 
 * von der Box kommen, so mu das Modul auch mit der Datei IGK arbeiten knnen,
 * allerdings nur eingeschrnkt.
 *
 * Definitiv durch dieses Modul beeinflut wird data, da dort bis jetzt die
 * Gruppenliste verwaltet wurde. Ebenfalls gendert werden mu das Modul
 * GruppenAuswahl, da diese natrlich die Informationen aus dem ITG untersttzen
 * soll.
 *
 * Ggf. weitere Anforderungen stellen sich dann whrend der Implementation 
 * heraus.
 *)
IMPLEMENTATION MODULE GroupSelect;

(* Aus Platzgrnden werden bei den Namen nur Pointer gespeichert und ein 
 * Block alloziert, der alle Namen enthlt.
 *)

FROM SYSTEM     IMPORT ADDRESS, ADR, TSIZE, CADR, ASSEMBLER, CALLSYS;

(* MM2-Module *)
FROM Storage    IMPORT ALLOCATE, DEALLOCATE;
IMPORT Block, BinOps, Lists, Strings, StrConv;
FROM Characters         IMPORT CR, LF, SUB;

(* Magic-Module *)
IMPORT MagicStrings, mtTextfiles, mtAppl, MagicDOS;

(* CAT-Module *)
FROM CatTypes   IMPORT Str255Ptr;
FROM Void       IMPORT v;
FROM MTPaths    IMPORT DataPath;

IMPORT CatTypes, Infofiles, MTE, QuickSort, CatFiles, dataSys, ConfVars, 
       Varnames, AssFuncs, CatLog, data, Protokoll;


CONST   (* Magics fr Header *)
        groupCatMagic     = 43415447H;  (* "CATG"                    *)
        groupVersion      = 4;          (* Versionsnummer der CATGROUP.DAT *)
        groupVersionMagic = 0104H;      (* Versionsmagic                   *)


CONST   groupFileName   = 'catgroup.inf';
        groupDataName   = 'catgroup.dat';
        renameFileName  = 'REN.txt';

TYPE
    groupHdr = RECORD
                name,                    (* Offsets in Data hinein, der  *)
                netname,                 (* erste ist immer 0. Ansonsten heit *)
                chef,                    (* ein offset von 0, da das Feld nicht *)
                kurztext,                (* vorhanden ist *)
                sysop,
                followup,
                herkunft,
                sprache,
                alias     : CARDINAL;
                readNumber: INTEGER;
                catNumber : INTEGER;
                readState : rwState;      (* F-Zeile *)
                writeState: rwState;      (* F-Zeile *)
                userState : uState;       (* F-Zeile *)
                isNet     : netState;     (* F-Zeile *)
                hasProg   : BOOLEAN;      (* F-Zeile *)
                withDollar: BOOLEAN;      (* F-Zeile *)
                selected  : BOOLEAN;      (* Fr Darstellung in Liste, wird beim Einlesen auf FALSE gesetzt *)
                ordered   : orderState;   (* Fr An-/Abbestellen von Gruppen *)
                orderDays : INTEGER;      (* Madness-Erweiterung: Gruppe soviel Tage zurck bestellen *)
                dataLength: CARDINAL;     (* Die Lnge des folgenden Datenblocks *)
               END;

TYPE mayCopyProc = PROCEDURE ( (* entry: *) groupPtr): BOOLEAN;

     groupArrayType = ARRAY [0..$7FFF] OF group;
     
     groupArrayPtr  = POINTER TO groupArrayType;

VAR listModified: BOOLEAN;
    
    groupArray  : groupArrayPtr;
    maxGroups   : INTEGER;
    groups      : INTEGER;
    
    fullList    : BOOLEAN;
    isSysop     : BOOLEAN;

PROCEDURE MakeNewGroupArray (force : BOOLEAN; new : CARDINAL) : BOOLEAN;
(* Alloziert ein neues Gruppenarray und kopiert das alte hinein.
 * FALSE: Nicht genug Speicher!
 *)
 VAR i :  INTEGER;
     newMax : CARDINAL;
     newArray : groupArrayPtr;
BEGIN
  IF (maxGroups = 0) OR (groups >= maxGroups - 1) OR force
  THEN
    IF force
    THEN
      newMax := new + 50;
    ELSE
      newMax := maxGroups + 50;
    END;
    ALLOCATE (newArray, LONG(newMax) * TSIZE (group));
    IF newArray = NIL
    THEN RETURN FALSE END;
    Block.Clear (newArray, LONG(newMax)*TSIZE(group));
    IF groupArray # NIL 
    THEN
      FOR i := 0 TO INTEGER(maxGroups) - 1 DO newArray^[i] := groupArray^[i] END;
    END;
    DEALLOCATE (groupArray, 0);
    maxGroups := newMax; 
    groupArray := newArray;
  END;
  RETURN TRUE;
END MakeNewGroupArray;

PROCEDURE ParseFlags (VAR entry: group; REF flags: ARRAY OF CHAR);
  VAR i, l : INTEGER;
      rwEntry : rwState;
BEGIN
  l := LENGTH (flags);
  i := 0;
  WHILE i < l DO
    CASE flags [i] OF 
      'L', 
      'S'  : CASE flags[i+1] OF
                '+' : rwEntry := rwOn; |
                'B' : rwEntry := rwPossible; |
                'C' : rwEntry := rwAskChef; |
                '-' : rwEntry := rwImpossible; |
                'P' : rwEntry := rwDefault; |
                'S' : rwEntry := rwAskSysop; |
(*                'N' : rwEntry := rwNoPermission; | *)
             ELSE
               rwEntry := rwNA;
             END;
             IF flags[i] = 'L'
             THEN
               entry.readState := rwEntry;
             ELSE
               entry.writeState := rwEntry;
             END; |
      'V'  : IF flags[i+1] = '+' THEN
               entry.isNet := nNet;
             ELSIF flags[i+1] = '-' THEN
               entry.isNet := nLokal;
             ELSIF flags[i+1] = '=' THEN
               entry.isNet := nNetLokal;
             ELSIF flags[i+1] = '!' THEN
               entry.isNet := nNetForbid;
             ELSE
               entry.isNet := nNA;
             END |
      'G'  : IF flags[i+1] = '*' THEN
               entry.userState := uIsChef;
             ELSIF flags[i+1] = '+' THEN
               entry.userState := uIsUser;
             ELSIF flags[i+1] = '-' THEN
               entry.userState := uNoUser;
             ELSE
               entry.userState := uNA;
             END; |
      'P'  : entry.hasProg := flags[i+1] = '+'; |
      '$'  : entry.withDollar := flags[i+1] = '+'; |
    ELSE
    END;
    INC (i, 2);
  END;
END ParseFlags;

  TYPE lineEntry = (gL, nL, uL, cL, fL, sL, rL, hL, pL, aL, cN, rN, oN, dN);
       lineSet   = SET OF lineEntry;

VAR compName : CatTypes.String255;

PROCEDURE findEntry (REF name: ARRAY OF CHAR; VAR idx: INTEGER): BOOLEAN;
(* Binre Suche im Array *)
  VAR left, right, mid : INTEGER;
      found : BOOLEAN;
      res   : MagicStrings.Relation;
BEGIN
  found := FALSE;
  left := 0; right := groups-1;
  WHILE (left <= right) & ~found DO
    (* Beware of overflows! *)
    mid := SHORT((LONG(left) + LONG(right)) DIV 2);
    res := AssFuncs.StrICompare (name, groupArray^[mid].name^);
    IF res = MagicStrings.equal
    THEN
      found := TRUE; idx := mid;
    ELSIF res = MagicStrings.greater
    THEN
      left := mid + 1
    ELSE
      right := mid - 1
    END;
  END;
  IF ~found
  THEN
    (* den kleineren Wert zuweisen, dahinter wird dann ggf. eingefgt *)
    idx := right;
  END;
  RETURN found;
END findEntry;

PROCEDURE findNetEntry (REF name: ARRAY OF CHAR; VAR idx: INTEGER): BOOLEAN;
  VAR i : INTEGER;
BEGIN
  FOR i := 0 TO groups - 1 DO
    IF groupArray^[i].catNumber > 0
    THEN
      IF (groupArray^[i].netname # NIL) &
         AssFuncs.StrIequal (name, groupArray^[i].netname^)
      THEN
        idx := i;
        RETURN TRUE
      END;
    END;
  END;
  RETURN FALSE
END findNetEntry;

PROCEDURE countCatEntries (VAR count: INTEGER);
  VAR i : INTEGER;
BEGIN
  count := 0;
  FOR i := 0 TO groups - 1 DO 
    IF groupArray^[i].catNumber >= 0 THEN INC (count) END;
  END;
END countCatEntries;

PROCEDURE findNumberEntry (number : CARDINAL; VAR idx: CARDINAL): BOOLEAN;
  VAR i : INTEGER;
BEGIN
  FOR i := 0 TO groups - 1 DO
    IF groupArray^[i].catNumber = INTEGER(number)
    THEN
      idx := i;
      RETURN TRUE
    END;
  END;
  RETURN FALSE
END findNumberEntry;

PROCEDURE findReadNumberEntry (number : CARDINAL; VAR idx: CARDINAL): BOOLEAN;
  VAR i : INTEGER;
BEGIN
  FOR i := 0 TO groups - 1 DO
    IF groupArray^[i].readNumber = INTEGER(number)
    THEN
      idx := i;
      RETURN TRUE
    END;
  END;
  RETURN FALSE
END findReadNumberEntry;

PROCEDURE MakeEntry (gr: groupPtr; VAR gName, nName, uLine, cLine, fLine, sLine, rLine,
                     hLine, pLine, aLine : ARRAY OF CHAR;
                     readNum, catNum : INTEGER; oState: orderState; days : INTEGER; 
                     defaultFlags: BOOLEAN; avail : lineSet): BOOLEAN;
(* Fllt den Eintrag gr mit den Daten 
 *)
VAR gLen,
    nLen,
    uLen,
    sLen,
    rLen,
    cLen,
    hLen,
    pLen,
    aLen    : INTEGER;
    addLen  : INTEGER;
    target  : ADDRESS;
BEGIN
  (* Lngen herausfinden *)
  gLen := (*SYSTEM.*)LENGTH (gName);
  nLen := (*SYSTEM.*)LENGTH (nName);
  uLen := (*SYSTEM.*)LENGTH (uLine);
  cLen := (*SYSTEM.*)LENGTH (cLine);
  sLen := (*SYSTEM.*)LENGTH (sLine);
  rLen := (*SYSTEM.*)LENGTH (rLine);
  hLen := (*SYSTEM.*)LENGTH (hLine);
  pLen := (*SYSTEM.*)LENGTH (pLine);
  addLen := gLen + nLen + uLen + cLen + sLen + rLen + hLen + pLen + 12;
  (* Jetzt Speicher allozieren *)
  IF gr^.data # NIL THEN DEALLOCATE (gr^.data, 0) END;
  gr^.dataLength := 0;
  ALLOCATE (gr^.data, addLen);
  IF gr^.data = NIL THEN RETURN FALSE END;
  gr^.dataLength := addLen;
  (* Speicher vorhanden, jetzt Eintrge setzen *)
  target := gr^.data;
  INC (gLen);
  Block.Copy (ADR(gName), gLen, target);
  gr^.name := target;
  INC (target, gLen);
  IF nLen > 0 THEN
    INC (nLen);
    Block.Copy (ADR(nName), nLen, target);
    gr^.netname := target;
    INC (target, nLen);
  ELSE
    gr^.netname := NIL
  END;
  IF cLen > 0 THEN
    INC (cLen);
    Block.Copy (ADR(cLine), cLen, target);
    gr^.chef := target;
    INC (target, cLen);
  ELSE
    gr^.chef := NIL
  END;
  IF uLen > 0 THEN
    INC (uLen);
    Block.Copy (ADR(uLine), uLen, target);
    gr^.kurztext := target;
    INC (target, uLen);
  ELSE
    gr^.kurztext := NIL
  END;
  IF sLen > 0 THEN
    INC (sLen);
    Block.Copy (ADR(sLine), sLen, target);
    gr^.sysop := target;
    INC (target, sLen);
  ELSE
    gr^.sysop := NIL
  END;
  IF rLen > 0 THEN
    INC (rLen);
    Block.Copy (ADR(rLine), rLen, target);
    gr^.followup := target;
    INC (target, rLen);
  ELSE
    gr^.followup := NIL
  END;
  IF hLen > 0 THEN
    INC (hLen);
    Block.Copy (ADR(hLine), hLen, target);
    gr^.herkunft := target;
    INC (target, hLen);
  ELSE
    gr^.herkunft := NIL
  END;
  IF pLen > 0 THEN
    INC (pLen);
    Block.Copy (ADR(pLine), pLen, target);
    gr^.sprache := target;
    INC (target, pLen);
  ELSE
    gr^.sprache := NIL
  END;
  IF aLen > 0 THEN
    INC (aLen);
    Block.Copy (ADR(aLine), aLen, target);
    gr^.alias := target;
    INC (target, aLen);
  ELSE
    gr^.alias := NIL
  END;
  IF rN IN avail
  THEN
    gr^.readNumber := readNum;
  END;
  IF cN IN avail
  THEN
    gr^.catNumber := catNum
  END;
  IF oN IN avail
  THEN
    gr^.ordered := oState
  END;
  IF dN IN avail
  THEN
    gr^.orderDays := days
  END;
  (* Okay, Texte sind gesetzt, jetzt Flags einsetzen, falls vorhanden *)
  (* Default-Einstellung herstellen fr Flags *)
  IF defaultFlags
  THEN
    gr^.readState := rwNA;
    gr^.writeState := rwNA;
    gr^.userState := uNA;
    gr^.isNet     := nNA;
    gr^.hasProg   := FALSE;
    gr^.withDollar:= FALSE;
    gr^.selected  := FALSE;
  END;
  IF fL IN avail
  THEN
    ParseFlags (gr^, fLine);
  END;
  RETURN TRUE;
END MakeEntry;

PROCEDURE UpdateEntry (VAR gName, nName, uLine, cLine, fLine, sLine, rLine, hLine, pLine, aLine : ARRAY OF CHAR;
                       readNum, catNum : INTEGER; oState: orderState; days : INTEGER; 
                       avail : lineSet;
                       keepOld: BOOLEAN;
                       preferLocal: BOOLEAN): BOOLEAN;

VAR
    found   : BOOLEAN;
    gr      : groupPtr;
    grIdx   : INTEGER;
    newGroup: group;
    i       : INTEGER;

BEGIN
  (* So, und jetzt suchen wir nur nach dem Namen *)
  found := findEntry (gName, grIdx);
  IF found
  THEN
    gr := ADR (groupArray^[grIdx]);
    (* Jetzt das raussuchen, was ggf. neu gesetzt werden mu 
     * Dabei umgedrehte Logik: Wenn etwas nicht in den neu zu setzenden
     * Sachen ist, aber in den alten, dann wird der alte Eintrag einfach
     * fr den neuen bernommen, wenn keepOld TRUE ist 
     * (bei GRUPPEN.INF oder IGK der Fall)
     *)
    IF preferLocal
    THEN
      (* Lokal vorhandene Gruppen bevorzugen *)
      IF gr^.isUpdated
      THEN
        IF (fL IN avail) 
        THEN
          ParseFlags (newGroup, fLine);
          IF (newGroup.isNet = nNetLokal) OR (newGroup.isNet = nNetForbid)
          THEN
            (* Neuer Eintrag ist lokal nicht vorhandene Netzgruppe,
             * wir behalten einfach den alten
             *)
            RETURN TRUE;
          END;
        ELSIF (gr^.isNet # nNetLokal) & (gr^.isNet # nNetForbid) & (gr^.isNet # nNA)
        THEN
          (* Der alte Eintrag bezeichnet schon eine vorhandene Gruppe *)
          RETURN TRUE;
        END;
      END;
    END;
    gr^.isUpdated := TRUE;
    IF keepOld
    THEN
      IF ~(nL IN avail) & (gr^.netname # NIL)
      THEN
        (* Netname zuweisen *)
        MagicStrings.Assign (gr^.netname^, nName);
      END;
      IF ~(uL IN avail) & (gr^.kurztext # NIL)
      THEN
        (* Kurztext zuweisen *)
        MagicStrings.Assign (gr^.kurztext^, uLine);
      END;
      IF ~(cL IN avail) & (gr^.chef # NIL)
      THEN
        (* Chef zuweisen *)
        MagicStrings.Assign (gr^.chef^, cLine);
      END;
      IF ~(sL IN avail) & (gr^.sysop # NIL)
      THEN
        (* Sysopinfo zuweisen *)
        MagicStrings.Assign (gr^.sysop^, sLine);
      END;
    END;
    IF ~(rL IN avail) & (gr^.followup # NIL)
    THEN
      (* Followup zuweisen *)
      MagicStrings.Assign (gr^.followup^, rLine);
    END;
    IF ~(hL IN avail) & (gr^.herkunft # NIL)
    THEN
      (* Herkunft zuweisen *)
      MagicStrings.Assign (gr^.herkunft^, hLine);
    END;
    IF ~(pL IN avail) & (gr^.sprache # NIL)
    THEN
      (* Sprache zuweisen *)
      MagicStrings.Assign (gr^.sprache^, pLine);
    END;
    IF ~(aL IN avail) & (gr^.alias # NIL)
    THEN
      (* Alias zuweisen *)
      MagicStrings.Assign (gr^.alias^, aLine);
    END;
    IF gr^.data # NIL THEN DEALLOCATE (gr^.data, 0) END;
    gr^.dataLength := 0;
  ELSE
    (* Eintrag lschen *)
    newGroup := group{NIL,NIL,NIL,NIL,NIL,NIL,NIL,NIL,NIL,-1,-1,rwNA,rwNA,uNA,nNA,
                      FALSE,FALSE,FALSE,oNA,0,TRUE,0,NIL};
    gr := ADR (newGroup);
  END;
  IF ~MakeEntry (gr, gName, nName, uLine, cLine, fLine, sLine, rLine, hLine, pLine, aLine, readNum, catNum, oState, days, TRUE, avail)
  THEN 
    RETURN FALSE
  END;
  IF found
  THEN
    (* Ist nichts mehr zu tun, da direkt die Adresse des Eintrags bergeben wurde
     *)
  ELSE
    (* neuen Eintrag anlegen *)
    IF ~MakeNewGroupArray (FALSE, 0)
    THEN RETURN FALSE END;
    (* Alle nachfolgenden verschieben *)
    IF groups > 0
    THEN
      FOR i := groups TO grIdx + 2 BY -1 DO
        groupArray^[i] := groupArray^[i-1]
      END;
    END;
    (* Eintrag bernehmen *)
    groupArray^[grIdx+1] := gr^;
    (* Gruppenanzahl ndern *)
    INC (groups);
  END;
  (* Eintrge jetzt wieder lschen *)
  MagicStrings.Assign ('', gName);
  MagicStrings.Assign ('', nName);
  MagicStrings.Assign ('', uLine);
  MagicStrings.Assign ('', cLine);
  MagicStrings.Assign ('', fLine);
  MagicStrings.Assign ('', sLine);
  MagicStrings.Assign ('', rLine);
  MagicStrings.Assign ('', hLine);
  MagicStrings.Assign ('', pLine);
  MagicStrings.Assign ('', aLine);
  listModified := TRUE;
  RETURN TRUE;
END UpdateEntry;

TYPE    lSortType   = (sByName, sByReadNum, sByNum, sByOrder, sByNet, sByLang);

PROCEDURE readNumComp (a1, a2 : ADDRESS) : BOOLEAN;
  VAR p1, p2 : POINTER TO groupPtr;
      str1, str2 : CatTypes.String255;
BEGIN
  p1 := a1;
  p2 := a2;
  IF p1^^.readNumber = p2^^.readNumber
  THEN
    RETURN AssFuncs.StrICompare (p1^^.name^, p2^^.name^) = MagicStrings.less;
    (*
    MagicStrings.Assign (p1^^.name^, str1);
    MagicStrings.Assign (p2^^.name^, str2);
    MagicStrings.CAPS (str1);
    MagicStrings.CAPS (str2);
    RETURN MagicStrings.Compare (str1, str2) = MagicStrings.less;
    *)
  ELSE 
    RETURN p1^^.readNumber < p2^^.readNumber
  END;
END readNumComp;

PROCEDURE orderComp (a1, a2 : ADDRESS) : BOOLEAN;
  VAR p1, p2 : POINTER TO groupPtr;
      str1, str2 : CatTypes.String255;
      on1, on2 : BOOLEAN;
BEGIN
  p1 := a1;
  p2 := a2;
  on1 := (p1^^.readState = rwOn) OR (p1^^.readState = rwDefault);
  on2 := (p2^^.readState = rwOn) OR (p2^^.readState = rwDefault);
  IF on1 = on2
  THEN
    RETURN AssFuncs.StrICompare (p1^^.name^, p2^^.name^) = MagicStrings.less;
    (*
    MagicStrings.Assign (p1^^.name^, str1);
    MagicStrings.Assign (p2^^.name^, str2);
    MagicStrings.CAPS (str1);
    MagicStrings.CAPS (str2);
    RETURN MagicStrings.Compare (str1, str2) = MagicStrings.less;
    *)
  ELSE 
    RETURN on1
  END;
END orderComp;

PROCEDURE nameComp (a1, a2 : ADDRESS) : BOOLEAN;
  VAR p1, p2 : POINTER TO groupPtr;
      str1, str2 : CatTypes.String255;
BEGIN
  p1 := a1;
  p2 := a2;
  RETURN AssFuncs.StrICompare (p1^^.name^, p2^^.name^) = MagicStrings.less;
  (*
  MagicStrings.Assign (p1^^.name^, str1);
  MagicStrings.Assign (p2^^.name^, str2);
  MagicStrings.CAPS (str1);
  MagicStrings.CAPS (str2);
  RETURN MagicStrings.Compare (str1, str2) = MagicStrings.less;
  *)
END nameComp;

PROCEDURE ExtraktString (REF source: ARRAY OF CHAR; VAR dest: ARRAY OF CHAR);
  VAR p: CARDINAL;
BEGIN
  MagicStrings.Assign ("", dest);
  p := MagicStrings.Pos (":", source, 0, FALSE);
  IF p <= LENGTH (source)
  THEN
    MagicStrings.Copy (source, 0, p, dest);
  END;
END ExtraktString;

PROCEDURE netComp (a1, a2 : ADDRESS) : BOOLEAN;
  VAR p1, p2 : POINTER TO groupPtr;
      str1, str2 : CatTypes.String255;
      compRes: MagicStrings.Relation;
BEGIN
  p1 := a1;
  p2 := a2;
  IF (p1^^.herkunft # NIL) & (p2^^.herkunft # NIL)
  THEN
    (* Herkunft extrahieren *)
    ExtraktString (p1^^.herkunft^, str1);
    ExtraktString (p2^^.herkunft^, str2);
    compRes := AssFuncs.StrICompare (str1, str2);
    IF compRes = MagicStrings.equal
    THEN
      RETURN AssFuncs.StrICompare (p1^^.name^, p2^^.name^) = MagicStrings.less;
    END;
    RETURN compRes = MagicStrings.less;
  ELSIF (p1^^.herkunft # NIL)
  THEN
    RETURN TRUE;
  ELSIF (p2^^.herkunft # NIL)
  THEN
    RETURN FALSE;
  END;
  RETURN AssFuncs.StrICompare (p1^^.name^, p2^^.name^) = MagicStrings.less;
  (*
  MagicStrings.Assign (p1^^.name^, str1);
  MagicStrings.Assign (p2^^.name^, str2);
  MagicStrings.CAPS (str1);
  MagicStrings.CAPS (str2);
  RETURN MagicStrings.Compare (str1, str2) = MagicStrings.less;
  *)
END netComp;

PROCEDURE langComp (a1, a2 : ADDRESS) : BOOLEAN;
  VAR p1, p2 : POINTER TO groupPtr;
      str1, str2 : CatTypes.String255;
      compRes: MagicStrings.Relation;
BEGIN
  p1 := a1;
  p2 := a2;
  IF (p1^^.sprache # NIL) & (p2^^.sprache # NIL)
  THEN
    (* sprache extrahieren *)
    ExtraktString (p1^^.sprache^, str1);
    ExtraktString (p2^^.sprache^, str2);
    compRes := AssFuncs.StrICompare (str1, str2);
    IF compRes = MagicStrings.equal
    THEN
      RETURN AssFuncs.StrICompare (p1^^.name^, p2^^.name^) = MagicStrings.less;
    END;
    RETURN compRes = MagicStrings.less;
  ELSIF (p1^^.sprache # NIL)
  THEN
    RETURN TRUE;
  ELSIF (p2^^.sprache # NIL)
  THEN
    RETURN FALSE;
  END;
  RETURN AssFuncs.StrICompare (p1^^.name^, p2^^.name^) = MagicStrings.less;
  (*
  MagicStrings.Assign (p1^^.name^, str1);
  MagicStrings.Assign (p2^^.name^, str2);
  MagicStrings.CAPS (str1);
  MagicStrings.CAPS (str2);
  RETURN MagicStrings.Compare (str1, str2) = MagicStrings.less;
  *)
END langComp;

PROCEDURE SortList (VAR l : Lists.List; sortType :lSortType);
(* sortiert eine Gruppenliste *)

  VAR count : CARDINAL;
      sort  : POINTER TO ARRAY [0..$FFFF] OF ADDRESS;
      i     : CARDINAL;
      adr   : ADDRESS;

BEGIN
  count := Lists.NoOfEntries (l);
  IF count = 0 THEN RETURN END;
  ALLOCATE (sort, LONG(count) * TSIZE (ADDRESS));
  IF sort = NIL THEN RETURN END;
  Lists.ResetList (l);
  FOR i := 0 TO count-1 DO
    sort^[i] := Lists.NextEntry (l);
  END;
  CASE sortType OF
    sByName : v.bool := QuickSort.sortIt (0, count-1, sort^, nameComp, TSIZE (ADDRESS), QuickSort.noBreak); |
    sByReadNum : v.bool := QuickSort.sortIt (0, count-1, sort^, readNumComp, TSIZE (ADDRESS), QuickSort.noBreak); |
    sByOrder : v.bool := QuickSort.sortIt (0, count-1, sort^, orderComp, TSIZE (ADDRESS), QuickSort.noBreak); |
    sByNet   : v.bool := QuickSort.sortIt (0, count-1, sort^, netComp, TSIZE (ADDRESS), QuickSort.noBreak); |
    sByLang  : v.bool := QuickSort.sortIt (0, count-1, sort^, langComp, TSIZE (ADDRESS), QuickSort.noBreak); |
  ELSE
  END;
  (* Liste wieder zurckbernehmen *)
  Lists.ResetList (l);
  FOR i := 0 TO count-1 DO
    adr := Lists.NextEntry (l);
    Lists.RemoveEntry (l, v.bool);
  END;
  Lists.ResetList (l);
  FOR i := 0 TO count - 1 DO 
    Lists.AppendEntry (l, sort^[i], v.bool);
  END;
  DEALLOCATE (sort, 0);
END SortList;

PROCEDURE ParseITG(file : mtTextfiles.TEXTFILE);
  VAR gName, nName, uLine, cLine, fLine, 
      sLine, rLine, hLine, aLine, pLine : CatTypes.String255;
      scrap : CatTypes.String255;
      infoWritten: BOOLEAN;
      linesAvail : lineSet;
      idx, i     : INTEGER;
      p0         : CARDINAL;
BEGIN
  (* Erstmal alle Eintrge lschen, die keine CAT-Gruppe sind!
   *)
  idx := 0;
  WHILE idx < groups DO
    WITH groupArray^[idx] DO
      IF catNumber >= 0 THEN
        groupArray^[idx].isUpdated := FALSE; 
        INC (idx)
      ELSE
        FOR i := idx+1 TO groups-1 DO
          (* Eintrge runterkopieren *)
          groupArray^[i-1] := groupArray^[i];
        END;
        DEC (groups);
      END;
    END;
  END;
  infoWritten := TRUE;
  linesAvail := lineSet{oN, dN};
  MagicStrings.Assign ('', gName);
  MagicStrings.Assign ('', nName);
  MagicStrings.Assign ('', uLine);
  MagicStrings.Assign ('', cLine);
  MagicStrings.Assign ('', aLine);
  MagicStrings.Assign ('', fLine);
  MagicStrings.Assign ('', sLine);
  MagicStrings.Assign ('', rLine);
  WHILE ~mtTextfiles.EndofText (file) DO
    mtTextfiles.ReadLine (file, scrap);
    mtTextfiles.ReadLn (file);
    CASE scrap[0] OF
      'G' : IF ~infoWritten
            THEN
              v.bool := UpdateEntry (gName, nName, uLine, cLine, fLine, sLine, rLine, hLine, pLine, aLine, -1, -1, oNA, 0, linesAvail, FALSE, TRUE);
              linesAvail := lineSet{oN, dN};
            END;
            MagicStrings.Delete (scrap, 0, 1);
            MagicStrings.Assign (scrap, gName); 
            INCL (linesAvail, gL);
            infoWritten := FALSE; |
      'N' : MagicStrings.Delete (scrap, 0, 1);
            MagicStrings.Assign (scrap, nName); 
            INCL (linesAvail, nL); |
      'U' : MagicStrings.Delete (scrap, 0, 1);
            MagicStrings.Assign (scrap, uLine); 
            INCL (linesAvail, uL); |
      'C' : MagicStrings.Delete (scrap, 0, 1);
            MagicStrings.Assign (scrap, cLine); 
            INCL (linesAvail, cL); |
      'F' : MagicStrings.Delete (scrap, 0, 1);
            MagicStrings.Assign (scrap, fLine); 
            INCL (linesAvail, fL); |
      'S' : MagicStrings.Delete (scrap, 0, 1);
            MagicStrings.Assign (scrap, sLine); 
            INCL (linesAvail, sL); |
      'R' : MagicStrings.Delete (scrap, 0, 1);
            MagicStrings.Assign (scrap, rLine); 
            INCL (linesAvail, rL); |
      'H' : MagicStrings.Delete (scrap, 0, 1);
            MagicStrings.Assign (scrap, hLine); 
            INCL (linesAvail, hL); |
      'P' : MagicStrings.Delete (scrap, 0, 1);
            MagicStrings.Assign (scrap, pLine); 
            INCL (linesAvail, pL); |
      'A' : MagicStrings.Delete (scrap, 0, 1);
            MagicStrings.Assign (scrap, aLine); 
            INCL (linesAvail, aL); |
    ELSE
    END;
  END;
  IF ~infoWritten 
  THEN
    v.bool := UpdateEntry (gName, nName, uLine, cLine, fLine, sLine, rLine, hLine, pLine, aLine, -1, -1, oNA, 0, linesAvail, FALSE, TRUE);
  END;
END ParseITG;

PROCEDURE ParseIGK (file: mtTextfiles.TEXTFILE);
  VAR gName, uLine, fLine : CatTypes.String255;
      scrap : CatTypes.String255;
      linesAvail : lineSet;
      p     : CARDINAL;

  PROCEDURE isGroupName (REF s : ARRAY OF CHAR) : BOOLEAN;
    VAR valid : BOOLEAN;
  BEGIN
    (* 1. Zeichen: Darf Space, + oder - sein. *)
    valid := (s[0] = '+') OR (s[0] = '-') OR (s[0] = ' ');
    (* 2. Zeichen: Darf '*' oder Space sein *)
    valid := valid & ((s[1] = ' ') OR (s[1] = '*'));
    (* 3. Zeichen: Darf nicht Space sein! *)
    valid := valid & ~(s[2] = ' ');
    RETURN valid;
  END isGroupName;

BEGIN
  linesAvail := lineSet{gL, uL, fL, oN, dN};
  MagicStrings.Assign ('', gName);
  MagicStrings.Assign ('', uLine);
  MagicStrings.Assign ('', fLine);
  WHILE ~mtTextfiles.EndofText (file) DO
    mtTextfiles.ReadLine (file, scrap);
    mtTextfiles.ReadLn(file);
    IF isGroupName (scrap)
    THEN
      IF scrap[0] = '+'
      THEN
        MagicStrings.Append ('L+S+', fLine);
      ELSIF scrap[0] = '-'
      THEN
        MagicStrings.Append ('LBSB', fLine);
      END;
      IF scrap[1] = '*'
      THEN
        MagicStrings.Append ('G*', fLine);
      ELSIF scrap[0] = '-'
      THEN
        MagicStrings.Append ('G+', fLine);
      END;
      (* Flagzeile ist zusammengebaut *)
      IF scrap[2] = '?'
      THEN
        (* Fragezeichen entfernen *)
        MagicStrings.Delete (scrap, 2, 1);
      END;
      (* Jetzt Gruppennamen rauskopieren. Gruppenname geht bis zum ersten Space
       * ab der 3. Spalte, aber mindestens 10 Zeichen weit!
       *)
      p := MagicStrings.Pos (' ', scrap, 2, FALSE);
      IF (p < (*SYSTEM.*)LENGTH (scrap)) 
      THEN
        (* Space gefunden *)
        IF p < 13
        THEN
          (* Mindestens bis zur 13 Stelle kopieren und dann Spaces 
           * am Ende abschneiden 
           *)
          (* Erstmal nachsehen, ob nach dem Space noch ein Zeichen kommt
           *)
          IF scrap[p+1] # ' '
          THEN
            p := MagicStrings.Pos (' ', scrap, p+1, FALSE);
            IF (p < (*SYSTEM.*)LENGTH (scrap)) 
            THEN
              (* Space gefunden *)
              IF p < 13
              THEN
                p := 13;
              END;
            END;
          ELSE
            p := 13;
          END;
        END;
      ELSE
        p := 13;
      END;
      (* Gruppenname kopieren *)
      MagicStrings.Copy (scrap, 2, p-2, gName);
      (* Rest ist Beschreibung *)
      MagicStrings.Delete (scrap, 0, p);
      MagicStrings.Assign (scrap, uLine);
      (* Jetzt Spaces lschen *)
      Strings.DelTrailingBlanks (gName);
      Strings.DelTrailingBlanks (uLine);
      Strings.DelLeadingBlanks (gName);
      Strings.DelLeadingBlanks (uLine);
      (* Und jetzt Eintrag updaten *)
      MagicStrings.Assign ('', scrap);
      v.bool := UpdateEntry (gName, scrap, uLine, scrap, fLine, scrap, scrap, scrap, scrap, scrap, -1, -1, oNA, 0, linesAvail, TRUE, FALSE);
    END;
  END;  
END ParseIGK;

PROCEDURE compGrName (a1, a2 : ADDRESS) : BOOLEAN;
  VAR p1, p2 : groupPtr;
      str1, str2 : CatTypes.String255;
BEGIN
  p1 := a1;
  p2 := a2;
  (*
  IF p1^.name = NIL THEN RETURN TRUE END;
  IF p2^.name = NIL THEN RETURN FALSE END;
  *)
  MagicStrings.Assign (p1^.name^, str1);
  MagicStrings.Assign (p2^.name^, str2);
  RETURN AssFuncs.StrICompare (str1, str2) = MagicStrings.less;
END compGrName;

PROCEDURE SortGroupArray();
BEGIN
  v.bool := QuickSort.sortIt (0, groups-1, groupArray^, compGrName, TSIZE (group), QuickSort.noBreak);
END SortGroupArray;

PROCEDURE ParseRenamefile ();
  VAR 
      fname     : CatTypes.String255;
      file      : mtTextfiles.TEXTFILE;
      first,
      success,
      newSet,
      oldSet    : BOOLEAN;
      scrap,
      newName,
      oldName   : CatTypes.String255;
BEGIN
  first  := TRUE;
  newSet := FALSE;
  oldSet := FALSE;
  MagicStrings.Assign (DataPath, fname);
  MagicStrings.Append (renameFileName, fname);
  IF mtTextfiles.OpenTextfile (fname, mtTextfiles.READ, 16384, file)
  THEN
    WHILE ~mtTextfiles.EndofText (file) DO
      mtTextfiles.ReadLine (file, scrap);
      mtTextfiles.ReadLn (file);
      CASE scrap[0] OF
        'O': MagicStrings.Delete (scrap, 0, 1);
             MagicStrings.Assign (scrap, oldName); 
             oldSet := TRUE; |
        'N': MagicStrings.Delete (scrap, 0, 1);
             MagicStrings.Assign (scrap, newName); 
             newSet := TRUE; |
      ELSE
      END (* CASE *);
      IF newSet & oldSet
      THEN
        success := RenameGroup (oldName, newName, FALSE);
        IF first 
        THEN
          CatLog.WriteLn ();
          CatLog.WriteLn ();
          first := FALSE;
        END;
        IF success
        THEN
          CatLog.putTime ();
          CatLog.WriteString ("Gruppe ");
          CatLog.WriteLine (oldName);
          CatLog.WriteString ("         umbenannt nach ");
          CatLog.WriteLine (newName);
        ELSE
          CatLog.putTime ();
          CatLog.WriteString ("FEHLER! Gruppe ");
          CatLog.WriteLine (oldName);
          CatLog.WriteString ("         konnte nicht umbenannt werden nach ");
          CatLog.WriteLine (newName);
        END;
        oldSet := FALSE;
        newSet := FALSE;
      END;
    END; (* WHILE *)
    mtTextfiles.CloseTextfile (file);
  END;
END ParseRenamefile;

PROCEDURE ParseInfofile (): BOOLEAN;
  VAR parseItg  : BOOLEAN;
      fname     : CatTypes.String255;
      file      : mtTextfiles.TEXTFILE;
BEGIN
  parseItg := Infofiles.GetInfoFilename ('ITG',fname);
  IF parseItg
  THEN
    IF ~mtTextfiles.OpenTextfile (fname, mtTextfiles.READ, 32767, file)
    THEN
      parseItg := FALSE;
    ELSE
      ParseITG (file);
      mtTextfiles.CloseTextfile (file);
      (* Die Liste mu nicht mehr sortiert werden, da die automatisch sortiert
       * eingefgt werden
       *)
    END;
  END;
  IF ~parseItg
  THEN
    (* mit IGK versuchen *)
    IF Infofiles.GetInfoFilename ('IGK', fname)
     & mtTextfiles.OpenTextfile (fname, mtTextfiles.READ, 32767, file)
    THEN
      ParseIGK (file);
      mtTextfiles.CloseTextfile (file);
      (* Die Liste mu nicht mehr sortiert werden, da die automatisch sortiert
       * eingefgt werden
       *)
    ELSE
      RETURN FALSE
    END;
  END;
  RETURN TRUE;
END ParseInfofile;

PROCEDURE CountInfofileLines (): CARDINAL;
  VAR parseItg  : BOOLEAN;
      fname     : CatTypes.String255;
      file      : mtTextfiles.TEXTFILE;
      lines     : CARDINAL;
      ch        : CHAR;
BEGIN
  lines := 0;
  parseItg := Infofiles.GetInfoFilename ('ITG',fname);
  IF parseItg
  THEN
    IF ~mtTextfiles.OpenTextfile (fname, mtTextfiles.READ, 8192, file)
    THEN
      parseItg := FALSE;
    ELSE
      WHILE ~mtTextfiles.EndofText (file) DO
        mtTextfiles.ReadChar (file, ch);
        mtTextfiles.ReadLn (file);
        IF (ch = 'G')
        THEN
          (* Nur Zeilen mit Gruppennamen zhlen bei der ITG *)
          INC (lines);
        END;
      END;
      mtTextfiles.CloseTextfile (file);
    END;
  END;
  IF ~parseItg
  THEN
    (* mit IGK versuchen *)
    IF Infofiles.GetInfoFilename ('IGK', fname)
     & mtTextfiles.OpenTextfile (fname, mtTextfiles.READ, 8192, file)
    THEN
      WHILE ~mtTextfiles.EndofText (file) DO
        mtTextfiles.ReadLn (file);
        INC (lines);
      END;
      mtTextfiles.CloseTextfile (file);
    END;
  END;
  RETURN lines;
END CountInfofileLines;
      
CONST   catGroupInf = 'gruppen.inf';

PROCEDURE ParseGruppenInf(file : mtTextfiles.TEXTFILE): BOOLEAN;
  VAR gName, void : CatTypes.String255;
      linesAvail : lineSet;
      lineNum    : INTEGER;
      i          : INTEGER;
      gr         : groupPtr;
BEGIN
  linesAvail := lineSet{gL, cN, rN, oN, dN};
  MagicStrings.Assign ('', gName);
  MagicStrings.Assign ('', void);
  lineNum := 1;
  WHILE ~mtTextfiles.EndofText (file) DO
    mtTextfiles.ReadLine (file, gName);
    mtTextfiles.ReadLn(file);
    IF gName[0] # ''
    THEN
      IF ~UpdateEntry (gName, void, void, void, void, void, void, void, void, void, lineNum, lineNum, oNA, 0, linesAvail, TRUE, FALSE)
      THEN
        RETURN FALSE
      END;
      INC (lineNum);
    END;
  END;
  ConfVars.GetConfDefBool (cSortGroups, v.bool, TRUE);
  IF v.bool
  THEN
    (* Einfach die Lesenummern neu vergeben, da die Liste schon beim
     * erstellen sortiert wird
     *)
    lineNum := 1;
    FOR i := 0 TO groups - 1 DO
      IF groupArray^[i].catNumber >= 0 THEN 
        groupArray^[i].readNumber := lineNum;
        INC (lineNum);
      END;
    END;
  END; 
  RETURN TRUE;
END ParseGruppenInf;

(*----- Eigene Liste schreiben und lesen ------------------------------*)

(* Format der Datei, in der die Liste abgespeichert wird: 
 * Pro Eintrag eine Zeile, die einzelnen Elemente sind getrennt durch ein 
 * TAB-Zeichen (das kommt definitiv nie in Gruppennamen vor).
 * <name>TAB<netname>TAB<chef>TAB<kurztext>TAB<sysopinfo>TAB<followupto>TAB<readNumber>TAB<catNumber>TAB<flags>
 * Dabei werden die Flags wie im ITG gespeichert.
 *)

PROCEDURE SaveGroupList(): BOOLEAN;
  VAR fname : CatTypes.String255;
      file  : mtTextfiles.TEXTFILE;
      i     : INTEGER;
      dbHdr : dataSys.FileHeaderType;
      fhdl  : INTEGER;
  
  PROCEDURE saveEntry (e, i : ADDRESS);
    TYPE  shortStr   = ARRAY [0..7] OF CHAR;
    CONST rwStateStr = shortStr(' +BCS-P ');
          uStateStr  = shortStr(' *+-    ');
          nStateStr  = shortStr(' +-=!   ');
          TAB        = 11C;
    VAR flags : CatTypes.String127;
        gr    : groupPtr;
  BEGIN
    gr := groupPtr (e);
    WITH gr^ DO
      IF name # NIL THEN mtTextfiles.WriteLine (file, name^); END;
      mtTextfiles.WriteChar (file, TAB);
      IF netname # NIL THEN mtTextfiles.WriteLine (file, netname^); END;
      mtTextfiles.WriteChar (file, TAB);
      IF chef # NIL THEN mtTextfiles.WriteLine (file, chef^); END;
      mtTextfiles.WriteChar (file, TAB);
      IF kurztext # NIL THEN mtTextfiles.WriteLine (file, kurztext^); END;
      mtTextfiles.WriteChar (file, TAB);
      IF sysop # NIL THEN mtTextfiles.WriteLine (file, sysop^); END;
      mtTextfiles.WriteChar (file, TAB);
      IF followup # NIL THEN mtTextfiles.WriteLine (file, followup^); END;
      mtTextfiles.WriteChar (file, TAB);
      mtTextfiles.WriteInt (file, readNumber, 0);
      mtTextfiles.WriteChar (file, TAB);
      mtTextfiles.WriteInt (file, catNumber, 0);
      mtTextfiles.WriteChar (file, TAB);
      MagicStrings.Assign ('', flags);
      IF readState # rwNA THEN 
        MagicStrings.Append ('L', flags); 
        MagicStrings.Append (rwStateStr[ORD(readState)], flags);
      END;
      IF writeState # rwNA THEN 
        MagicStrings.Append ('S', flags); 
        MagicStrings.Append (rwStateStr[ORD(writeState)], flags);
      END;
      IF userState # uNA THEN 
        MagicStrings.Append ('G', flags); 
        MagicStrings.Append (uStateStr[ORD(userState)], flags);
      END;
      IF isNet # nNA THEN 
        MagicStrings.Append ('V', flags); 
        MagicStrings.Append (nStateStr[ORD(isNet)], flags);
      END;
      IF hasProg THEN 
        MagicStrings.Append ('P+', flags);
      ELSE
        MagicStrings.Append ('P-', flags);
      END;
      IF withDollar THEN 
        MagicStrings.Append ('$+', flags);
      ELSE
        MagicStrings.Append ('$-', flags);
      END;
      mtTextfiles.WriteLine (file, flags);
      mtTextfiles.WriteChar (file, TAB);
      mtTextfiles.WriteInt (file, ORD(ordered), 0);
      mtTextfiles.WriteChar (file, TAB);
      mtTextfiles.WriteInt (file, orderDays, 0);
      mtTextfiles.WriteChar (file, TAB);
      IF herkunft # NIL THEN mtTextfiles.WriteLine (file, herkunft^); END;
      mtTextfiles.WriteChar (file, TAB);
      IF sprache  # NIL THEN mtTextfiles.WriteLine (file, sprache^); END;
      mtTextfiles.WriteChar (file, TAB);
      IF alias    # NIL THEN mtTextfiles.WriteLine (file, alias^); END;
      mtTextfiles.WriteLn(file);
    END;
  END saveEntry;
  
  PROCEDURE saveDataEntry (e, i : ADDRESS): BOOLEAN;
    VAR gr    : groupPtr;
        hdr   : groupHdr;
        fPtr  : POINTER TO INTEGER;
  BEGIN
    gr := groupPtr (e);
    fPtr := i;
    WITH hdr DO
      name := 0;
      netname       := 0;
      chef          := 0;
      kurztext      := 0;
      sysop         := 0;
      followup      := 0;
      herkunft      := 0;
      sprache       := 0;
      alias         := 0;
      IF gr^.netname  # NIL THEN netname  := SHORT(LONGCARD(gr^.netname - gr^.data)) ; END;
      IF gr^.chef     # NIL THEN chef     := SHORT(LONGCARD(gr^.chef - gr^.data)); END;
      IF gr^.kurztext # NIL THEN kurztext := SHORT(LONGCARD(gr^.kurztext - gr^.data)); END;
      IF gr^.sysop    # NIL THEN sysop    := SHORT(LONGCARD(gr^.sysop - gr^.data)); END;
      IF gr^.followup # NIL THEN followup := SHORT(LONGCARD(gr^.followup - gr^.data)); END;
      IF gr^.herkunft # NIL THEN herkunft := SHORT(LONGCARD(gr^.herkunft - gr^.data)); END;
      IF gr^.sprache  # NIL THEN sprache  := SHORT(LONGCARD(gr^.sprache  - gr^.data)); END;
      IF gr^.alias    # NIL THEN alias    := SHORT(LONGCARD(gr^.alias    - gr^.data)); END;
      (* Jetzt die brigen Daten setzen *)
      readNumber := gr^.readNumber;
      catNumber  := gr^.catNumber ;
      readState  := gr^.readState ;   (* F-Zeile *)
      writeState := gr^.writeState;   (* F-Zeile *)
      userState  := gr^.userState ;   (* F-Zeile *)
      isNet      := gr^.isNet     ;   (* F-Zeile *)
      hasProg    := gr^.hasProg   ;   (* F-Zeile *)
      withDollar := gr^.withDollar;   (* F-Zeile *)
      selected   := gr^.selected  ;   (* Fr Darstellung in Liste, wird beim Einlesen auf FALSE gesetzt *)
      ordered    := gr^.ordered   ;   (* Fr An-/Abbestellen von Gruppen *)
      orderDays  := gr^.orderDays ;   (* Madness-Erweiterung: Gruppe soviel Tage zurck bestellen *)
      dataLength := gr^.dataLength;   (* Die Lnge des folgenden Datenblocks *)
    END;
    CatFiles.WriteMuch (TSIZE (groupHdr), fPtr^, ADR(hdr));
    IF CatFiles.FileError # CatFiles.NoError 
    THEN
      RETURN FALSE
    END;
    (* Jetzt die Daten schreiben *)
    CatFiles.WriteMuch (hdr.dataLength, fPtr^, gr^.data);
    RETURN CatFiles.FileError = CatFiles.NoError;
  END saveDataEntry;
  
BEGIN
  MagicStrings.Assign (DataPath, fname);
  MagicStrings.Append (groupFileName, fname);
  IF mtTextfiles.OpenTextfile (fname, mtTextfiles.WRITE, 16384, file)
  THEN
    FOR i := 0 TO groups - 1 DO
      saveEntry (ADR(groupArray^[i]), ADR(file));
    END;
    mtTextfiles.CloseTextfile (file);
    listModified := FALSE;
    (* Jetzt noch das Binrfile schreiben *)
    fhdl := CatFiles.CreateFile (DataPath, groupDataName);
    IF fhdl > 0
    THEN
      (* Header schreiben *)
      dbHdr.CatMagic := groupCatMagic;
      dbHdr.Version  := groupVersion;
      dbHdr.VersionMagic := groupVersionMagic;
      CatFiles.WriteMuch (dataSys.dbHeaderLength, fhdl, ADR(dbHdr));
      IF CatFiles.FileError # CatFiles.NoError
      THEN
        CatFiles.CloseFile (fhdl);
        RETURN TRUE;
      END;
      i := 0;
      WHILE (i < groups) & 
            saveDataEntry (ADR(groupArray^[i]), ADR(fhdl)) DO INC (i) END;
      CatFiles.CloseFile (fhdl);
    END;
    Protokoll.SendPathUpdate (DataPath);
    RETURN TRUE;
  ELSE
    (* Fehlermeldung rauswerfen *)
    MTE.InfoAlert (MTE.noFile1, groupFileName, MTE.noFile4);
    RETURN FALSE;
  END;
END SaveGroupList;

PROCEDURE SaveChangedGroupList(): BOOLEAN;
(* Sichert die interne Gruppenliste in die Datei CATGROUP.INF, 
 * wenn sie seit dem letzten Einlesen irgendwie gendert wurde
 *)
BEGIN
  IF listModified THEN RETURN SaveGroupList(); END;
  RETURN TRUE
END SaveChangedGroupList;

PROCEDURE GetString (VAR str : ARRAY OF CHAR; p0 : CARDINAL;
                     VAR str2 : ARRAY OF CHAR; VAR p1 : CARDINAL);
  CONST TAB        = 11C;
BEGIN
  p1 := Strings.PosLen (TAB, str, p0);
  IF p1 > p0
  THEN
    Strings.Copy (str, p0, p1-p0, str2, v.bool);
  ELSE
    Strings.Assign ("", str2, v.bool);
  END;
END GetString;

PROCEDURE ReadGroupList(): BOOLEAN;
  VAR fname : CatTypes.String255;
      file  : mtTextfiles.TEXTFILE;
      fhdl  : INTEGER;
      scrap : CatTypes.String255;
      lc    : CARDINAL;

  VAR  
      p0, p1  : CARDINAL;
      ordDays,
      catNum,
      readNum : INTEGER;
      oState  : orderState;
      avail   : lineSet;
      fLine,
      oLine,
      mNumber,
      oNumber,
      rNumber,
      cNumber : ARRAY [0..30] OF CHAR;
      gName, nName, uLine, cLine, sLine, rLine, hLine, pLine, aLine : CatTypes.String255;
  
  PROCEDURE parseLines (): BOOLEAN;
    VAR gr : groupPtr;
        lines : CARDINAL;
  BEGIN
    (* Erstmal Zeilen zhlen *)
    lines := 0;
    WHILE ~mtTextfiles.EndofText (file) DO
      mtTextfiles.ReadLn (file);
      INC (lines);
    END;
    (* Jetzt wieder an den Anfang gehen *)
    mtTextfiles.Reset (file);
    (* Jetzt allozieren *)
    IF ~MakeNewGroupArray (TRUE, lines)
    THEN 
      RETURN FALSE
    END;
    lines := 0;
    WHILE ~mtTextfiles.EndofText (file) DO
      p0 := 0;
      p1 := 0;
      (* Eine Zeile lesen und parsen *)
      mtTextfiles.ReadLine (file, scrap);
      mtTextfiles.ReadLn(file);
      GetString (scrap, p0, gName, p1);
      p0 := p1+1;
      GetString (scrap, p0, nName, p1);
      p0 := p1+1;
      GetString (scrap, p0, cLine, p1);
      p0 := p1+1;
      GetString (scrap, p0, uLine, p1);
      p0 := p1+1;
      GetString (scrap, p0, sLine, p1);
      p0 := p1+1;
      GetString (scrap, p0, rLine, p1);
      p0 := p1+1;
      GetString (scrap, p0, rNumber, p1);
      p0 := p1+1;
      GetString (scrap, p0, cNumber, p1);
      p0 := p1+1;
      GetString (scrap, p0, fLine, p1);
      p0 := p1+1;
      GetString (scrap, p0, oLine, p1);
      p0 := p1+1;
      GetString (scrap, p0, oNumber, p1);
      p0 := p1+1;
      GetString (scrap, p0, hLine, p1);
      p0 := p1+1;
      GetString (scrap, p0, pLine, p1);
      p0 := p1+1;
      GetString (scrap, p0, aLine, p1);
      avail := lineSet {gL..dN};
      (* Orderstatus herausfinden *)
      IF oLine[0] # ''
      THEN
        p0 := 0;
        catNum := StrConv.StrToInt (oLine, p0, v.bool);
        oState := orderState (catNum);
      ELSE
        oState := oNA;
      END;
      IF oNumber[0] # ''
      THEN
        p0 := 0;
        ordDays := StrConv.StrToInt (oNumber, p0, v.bool);
      ELSE
        ordDays := 0;
      END;
      (* Jetzt die Lesenummern rausfinden *)
      IF cNumber[0] # ''
      THEN
        p0 := 0;
        catNum := StrConv.StrToInt (cNumber, p0, v.bool);
      ELSE
        catNum := -1;
      END;
      IF rNumber[0] # ''
      THEN
        p0 := 0;
        readNum := StrConv.StrToInt (rNumber, p0, v.bool);
      ELSE
        readNum := -1;
      END;
      (* Jetzt Eintrag setzen *)
      gr := ADR (groupArray^[lines]);
      gr^.data := NIL;
      gr^.dataLength := 0;
      IF ~MakeEntry (gr, gName, nName, uLine, cLine, fLine, sLine, rLine, hLine, pLine, aLine, readNum, catNum, oState, ordDays, TRUE, avail)
      THEN
        RETURN FALSE
      END;
      (* Ok, Eintrag jetzt vollstndig. Jetzt anhngen *)
      (* Ist durch Array schon angehngt :-) *)
      INC (lines);
      INC (groups);
    END;
    RETURN TRUE;
  END parseLines;
  
  PROCEDURE parseData(): BOOLEAN;
    VAR dbHdr   : dataSys.FileHeaderType;
        hdr     : groupHdr;
        gr      : groupPtr;
  BEGIN
    (* Aufbau der CATGROUP.DAT: 
     * fileHeader
     * pro Gruppe:
     * Header
     * Textdaten (variabel)
     *)
    CatFiles.ReadMuch (dataSys.dbHeaderLength, fhdl, ADR(dbHdr));
    IF CatFiles.FileError # CatFiles.NoError
    THEN
      RETURN FALSE
    END;
    (* Jetzt berprfen des Headers *)
    IF (dbHdr.CatMagic # groupCatMagic)
    OR (dbHdr.Version  # groupVersion)
    OR (dbHdr.VersionMagic # groupVersionMagic)
    THEN
      (* Abbrechen *)
      RETURN FALSE
    END;
    (* hier eventuell mal zhlen, wie viele es denn sind *)
    IF ~MakeNewGroupArray (TRUE, 600)
    THEN 
      RETURN FALSE
    END;
    groups := 0;
    REPEAT
      (* Header lesen *)
      CatFiles.ReadMuch (TSIZE (groupHdr), fhdl, ADR(hdr));
      IF CatFiles.FileError = CatFiles.NoError
      THEN
        (* Ok, Header gelesen *)
        IF ~MakeNewGroupArray (FALSE, 0)
        THEN
          RETURN FALSE
        END;
        (* Jetzt Eintrag setzen *)
        gr := ADR (groupArray^[groups]);
        ALLOCATE (gr^.data, hdr.dataLength);
        IF gr^.data = NIL THEN RETURN FALSE END;
        (* Jetzt Daten kopieren *)
        WITH gr^ DO
          (* Pointer setzen *)
          name := data;
          netname       := NIL;
          chef          := NIL;
          kurztext      := NIL;
          sysop         := NIL;
          followup      := NIL;
          herkunft      := NIL;
          sprache       := NIL;
          alias         := NIL;
          IF hdr.netname  # 0 THEN netname  := data + ADDRESS(LONG(hdr.netname)); END;
          IF hdr.chef     # 0 THEN chef     := data + ADDRESS(LONG(hdr.chef)); END;
          IF hdr.kurztext # 0 THEN kurztext := data + ADDRESS(LONG(hdr.kurztext)); END;
          IF hdr.sysop    # 0 THEN sysop    := data + ADDRESS(LONG(hdr.sysop)); END;
          IF hdr.followup # 0 THEN followup := data + ADDRESS(LONG(hdr.followup)); END;
          IF hdr.herkunft # 0 THEN herkunft := data + ADDRESS(LONG(hdr.herkunft)); END;
          IF hdr.sprache  # 0 THEN sprache  := data + ADDRESS(LONG(hdr.sprache)); END;
          IF hdr.alias    # 0 THEN alias    := data + ADDRESS(LONG(hdr.alias)); END;
          (* Jetzt die brigen Daten setzen *)
          readNumber := hdr.readNumber;
          catNumber  := hdr.catNumber ;
          readState  := hdr.readState ;   (* F-Zeile *)
          writeState := hdr.writeState;   (* F-Zeile *)
          userState  := hdr.userState ;   (* F-Zeile *)
          isNet      := hdr.isNet     ;   (* F-Zeile *)
          hasProg    := hdr.hasProg   ;   (* F-Zeile *)
          withDollar := hdr.withDollar;   (* F-Zeile *)
          selected   := FALSE  ;          (* Fr Darstellung in Liste, wird beim Einlesen auf FALSE gesetzt *)
          ordered    := hdr.ordered   ;   (* Fr An-/Abbestellen von Gruppen *)
          orderDays  := hdr.orderDays ;   (* Madness-Erweiterung: Gruppe soviel Tage zurck bestellen *)
          dataLength := hdr.dataLength;   (* Die Lnge des folgenden Datenblocks *)
          isUpdated  := FALSE;            (* Eintrag wurde noch nicht upgedatet *)
        END;
        (* Jetzt Daten lesen *)
        CatFiles.ReadMuch (hdr.dataLength, fhdl, gr^.data);
        IF CatFiles.FileError # CatFiles.NoError
        THEN
          (* Lesefehler, abbrechen *)
          RETURN FALSE
        END;
        INC (groups);
      END;
    UNTIL CatFiles.FileError # CatFiles.NoError;
    RETURN TRUE;
  END parseData;

  VAR size  : LONGCARD;
      exists: BOOLEAN;
      
      i     : INTEGER;

BEGIN
  mtAppl.MouseOn();
  mtAppl.MouseBusy();
  DEALLOCATE (groupArray, 0);
  groups := 0;
  maxGroups := 0;
  MagicStrings.Assign (DataPath, fname);
  MagicStrings.Append (groupDataName, fname);
  size := CatFiles.FileSize (fname, exists);
  IF exists & (size > 0) 
  THEN
    fhdl:= CatFiles.OpenFile (DataPath, groupDataName, CatFiles.readFile);
    IF fhdl > 0
    THEN
      IF ~parseData()
      THEN
        (* Einfach danach das CATGROUP.INF parsen, 
         * deshalb alles wieder freigeben 
         *)
        DEALLOCATE (groupArray, 0);
        groups := 0;
        maxGroups := 0;
        CatFiles.CloseFile (fhdl);
      ELSE
        CatFiles.CloseFile (fhdl);
        listModified := FALSE;
        mtAppl.MouseArrow();
        RETURN TRUE;
      END;
    END;
  END;
  MagicStrings.Assign (DataPath, fname);
  MagicStrings.Append (groupFileName, fname);
  size := CatFiles.FileSize (fname, exists);
  IF exists & (size > 0) & mtTextfiles.OpenTextfile (fname, mtTextfiles.READ, 32000, file)
  THEN
    IF ~parseLines()
    THEN
      mtAppl.MouseArrow();
      MTE.noMemAlert();
      mtTextfiles.CloseTextfile (file);
      RETURN FALSE;
    END;
    mtTextfiles.CloseTextfile (file);
    listModified := TRUE;
    mtAppl.MouseArrow();
    RETURN TRUE;
  ELSE
    (* File kann nicht geffnet werden. Dann versuchen wir es mal 
     * anders und bauen das neu auf aus der GRUPPEN.INF und der 
     * ITG bzw. IGK
     *)
    MagicStrings.Assign (DataPath, fname);
    MagicStrings.Append (catGroupInf, fname);
    size := CatFiles.FileSize (fname, exists);
    IF exists & (size > 0) & 
       mtTextfiles.OpenTextfile (fname, mtTextfiles.READ, 8192, file)
    THEN
      (* Erstmal hier die Zeilen zhlen *)
      lc := 0;
      WHILE ~mtTextfiles.EndofText (file) DO
        mtTextfiles.ReadLn (file);
        INC (lc);
      END;
      (* Jetzt wieder an den Anfang gehen *)
      mtTextfiles.Reset (file);
      (* Jetzt die Zeilen im Infofile zhlen *)
      INC (lc, CountInfofileLines ());
      (* Und jetzt allozieren wir den Speicher *)
      IF ~MakeNewGroupArray (TRUE, lc)
      THEN  
        mtTextfiles.CloseTextfile (file);
        mtAppl.MouseArrow();
        MTE.noMemAlert();
        RETURN FALSE;
      END;
      v.bool := ParseGruppenInf (file);
      listModified := TRUE;
      mtTextfiles.CloseTextfile (file);
      IF v.bool
      THEN
        (* Ok, Gruppen.inf geparst, jetzt noch ITG bzw. IGK parsen *)
        v.bool := ParseInfofile();
        v.bool := SaveGroupList();
        mtAppl.MouseArrow();
        RETURN TRUE;
      ELSE
        mtAppl.MouseArrow();
        MTE.noMemAlert();
        RETURN FALSE;
      END;
    ELSIF ~exists OR (size = 0)
    THEN
      (* Keine GRUPPEN.INF da, wahrscheinlich leere Database. 
       * Deshalb jetzt nur die Infofiles parsen 
       *)
      lc := CountInfofileLines ();
      (* Und jetzt allozieren wir den Speicher *)
      IF ~MakeNewGroupArray (TRUE, lc)
      THEN  
        mtTextfiles.CloseTextfile (file);
        mtAppl.MouseArrow();
        MTE.noMemAlert();
        RETURN FALSE;
      END;
      v.bool := ParseInfofile();
      v.bool := SaveGroupList();
      mtAppl.MouseArrow();
      RETURN TRUE;
    ELSE
      (* Fehlermeldung rauswerfen *)
      mtAppl.MouseArrow();
      MTE.noMemAlert ();
      RETURN FALSE;
    END;
  END;
  mtAppl.MouseArrow();
  RETURN FALSE;
END ReadGroupList;

(* Funktionen, die fr das Zusammenstellen der neuen Listen bentigt werden *)

PROCEDURE CopyEntry (entry: groupPtr; VAR l : Lists.List): BOOLEAN;
(* kopiert einen Eintrag in die Liste l 
 *)
  VAR new : groupPtr;
      fail : BOOLEAN;
BEGIN
  NEW (new); 
  IF new = NIL THEN RETURN FALSE END;
  new^ := entry^;
  new^.data := NIL;
  WITH new^ DO
    ALLOCATE (data, entry^.dataLength);
    IF data = NIL THEN DISPOSE (new); RETURN FALSE END;
    Block.Copy (entry^.data, entry^.dataLength, data);
    IF entry^.name # NIL THEN  name := data + entry^.name - entry^.data; ELSE name := NIL; END;
    IF entry^.netname # NIL THEN  netname := data + entry^.netname - entry^.data; ELSE netname := NIL; END;
    IF entry^.chef # NIL THEN  chef := data + entry^.chef - entry^.data; ELSE chef := NIL; END;
    IF entry^.kurztext # NIL THEN  kurztext := data + entry^.kurztext - entry^.data; ELSE kurztext := NIL; END;
    IF entry^.sysop  # NIL THEN  sysop := data + entry^.sysop - entry^.data; ELSE sysop := NIL; END;
    IF entry^.followup # NIL THEN  followup := data + entry^.followup - entry^.data; ELSE followup := NIL; END;
    IF entry^.herkunft # NIL THEN  herkunft := data + entry^.herkunft - entry^.data; ELSE herkunft := NIL; END;
    IF entry^.sprache # NIL THEN  sprache := data + entry^.sprache - entry^.data; ELSE sprache := NIL; END;
    IF entry^.alias   # NIL THEN  alias   := data + entry^.alias   - entry^.data; ELSE alias   := NIL; END;
    dataLength := entry^.dataLength;
  END;
  Lists.AppendEntry (l, new, fail);
  RETURN ~fail;
END CopyEntry;

PROCEDURE CopyList (VAR list: Lists.List; doCopy : mayCopyProc): BOOLEAN;
(* Kopiert die Gruppenliste in eine neue Liste. Dabei wird ein Eintrag
 * genau dann kopiert, wenn doCopy TRUE liefert.
 *)
 VAR fail : BOOLEAN;
     run  : groupPtr;
     idx  : INTEGER;
BEGIN
  Lists.CreateList (list, fail);
  IF fail THEN RETURN FALSE END;
  FOR idx := 0 TO groups - 1 DO
    run := ADR (groupArray^[idx]);
    (* Testen, ob es eine CAT-Gruppe ist *)
    IF doCopy (run)
    THEN
      (* Eintrag kopieren(!!) und anhngen an neue Liste *)
      IF ~CopyEntry (run, list)
      THEN
        DeleteList (list);
        RETURN FALSE
      END;
    END;
  END;
  RETURN TRUE;
END CopyList;

(* hier kommen jetzt die einzelnen Vergleichsfunktionen fr CopyList
 *)
PROCEDURE GetNew (catNum: INTEGER): CARDINAL;
  VAR newIdx,
      lastIdx   : CARDINAL;
BEGIN
  IF catNum < 0 
  THEN
    RETURN 0;
  END;
  newIdx := data.FirstNewMsg (catNum);
  lastIdx := data.LastMsgOfGroup (catNum);
  IF (lastIdx # dataSys.empty) & (newIdx # dataSys.empty) & (newIdx <= lastIdx) 
  THEN RETURN lastIdx - newIdx+1; 
  ELSE RETURN 0
  END;
END GetNew;

PROCEDURE copyByNum (entry: groupPtr): BOOLEAN;
BEGIN
  RETURN entry^.catNumber >= 0;
END copyByNum;

PROCEDURE copyNew (entry: groupPtr): BOOLEAN;
 VAR  new : CARDINAL;
      catNum: INTEGER;
BEGIN
  catNum := entry^.catNumber;
  RETURN (GetNew (catNum) > 0);
END copyNew;

PROCEDURE copyUnread (entry: groupPtr): BOOLEAN;
  VAR unrd  : CARDINAL;
      catNum: INTEGER;
BEGIN
  catNum := entry^.catNumber;
  IF catNum < 0 
  THEN
    RETURN FALSE;
  END;
  unrd := data.unreadMsgCount(catNum);
  IF unrd = 0FFFFH THEN unrd := 0 END;
  RETURN (unrd > 0);
END copyUnread;

PROCEDURE copyUser (entry: groupPtr): BOOLEAN;
BEGIN
  RETURN (entry^.userState = uNA) OR (entry^.userState = uIsChef) OR 
         (isSysop & ~((entry^.isNet = nNetLokal) OR (entry^.isNet = nNetForbid)));
END copyUser;

PROCEDURE copyLokal (entry: groupPtr): BOOLEAN;
BEGIN
  RETURN ~((entry^.isNet = nNetLokal) OR (entry^.isNet = nNetForbid));
END copyLokal;

PROCEDURE copyWrite (entry: groupPtr): BOOLEAN;
BEGIN
  IF fullList
  THEN
    RETURN (entry^.writeState = rwNA) OR (entry^.writeState = rwOn) OR (entry^.writeState = rwPossible);
  ELSE
    RETURN (entry^.catNumber > 0) & 
           ((entry^.writeState = rwNA) OR (entry^.writeState = rwOn) OR (entry^.writeState = rwPossible));
  END;
END copyWrite;

PROCEDURE copyAll (entry: groupPtr): BOOLEAN;
BEGIN
  RETURN TRUE;
END copyAll;

(* Hier kommen jetzt Funktionen, die einzelne Listen oder Teillisten 
 * den anderen Modulen zur Verfgung stellen.
 *)

PROCEDURE GetReadList (VAR list : Lists.List; byReadNum: BOOLEAN; getNew, getUnread: BOOLEAN): BOOLEAN;
(* Diese Prozedur liefert eine Liste der Gruppen, die in CAT vorhanden
 * und damit lesbar sind. Die Gruppe 'PERSNLICHE' ist mit enthalten.
 * Die Liste ist nach der Lesereihenfolge sortiert, die Gruppe 'PERSNLICHE'
 * ist immer als erste Gruppe vorhanden (die Lesenummer davon kann nicht 
 * gendert werden).
 * FALSE: Liste konnte nicht erzeugt werden
 *)
 VAR fail : BOOLEAN;
     run,
     new  : groupPtr;
     unrdCnt : CARDINAL;
     copyPers: BOOLEAN;
     
BEGIN
  mtAppl.MouseBusy();
  IF getNew & ~getUnread
  THEN
    IF ~CopyList (list, copyNew) THEN mtAppl.MouseArrow(); RETURN FALSE; END;
  ELSIF ~getNew & getUnread
  THEN
    IF ~CopyList (list, copyUnread) THEN mtAppl.MouseArrow(); RETURN FALSE; END;
  ELSE
    IF ~CopyList (list, copyByNum) THEN mtAppl.MouseArrow(); RETURN FALSE; END;
  END;
  (* Jetzt haben wir eine neue Liste, die wir nur noch nach der 
   * Lesereihenfolge sortieren mssen *)
  IF byReadNum
  THEN
    SortList (list, sByReadNum);  (* Hier eventuell noch Sortierprozedur bergeben *)
(*  ELSE
    SortList (list, sByName);  (* Hier eventuell noch Sortierprozedur bergeben *)
*)  END;
  (* So, jetzt haben wir eine sortierte Liste. Jetzt mssen wir nur 
   * noch die persnliche Gruppe evtl. dazu eintragen und vorne einfgen.
   *)
  IF getNew & ~getUnread
  THEN
    copyPers := GetNew (dataSys.private) > 0;
  ELSIF ~getNew & getUnread
  THEN
    unrdCnt := data.unreadMsgCount(dataSys.private);
    copyPers := (unrdCnt > 0) & (unrdCnt < 0FFFFH);
  ELSE
    copyPers := TRUE;
  END;
  IF copyPers
  THEN
    NEW (new);
    IF new = NIL THEN 
      DeleteList (list);
      mtAppl.MouseArrow();
      RETURN FALSE 
    END;
    WITH new^ DO
      ALLOCATE (data, LENGTH (dataSys.personalName)+2);
      IF data = NIL THEN 
        DeleteList (list);
        mtAppl.MouseArrow();
        RETURN FALSE 
      END;
      dataLength := LENGTH (dataSys.personalName) +2;
      name := data;
      Strings.Assign (dataSys.personalName, name^, fail);
      netname := NIL;
      chef    := NIL;
      kurztext  := NIL;
      sysop     := NIL;
      followup  := NIL;
      herkunft  := NIL;
      sprache   := NIL;
      alias     := NIL;
      readNumber:= 0;
      catNumber := dataSys.private;
      readState := rwOn;
      writeState:= rwImpossible;
      userState := uNA;
      isNet     := nNA;
      ordered   := oNA;
      orderDays := 0;
      hasProg   := FALSE;
      withDollar:= FALSE;
      selected  := FALSE;
    END;
    Lists.ResetList (list);
    Lists.InsertEntry (list, new, fail);
    IF fail THEN 
      DEALLOCATE (new^.data, 0);
      DISPOSE (new);
      DeleteList (list);
      mtAppl.MouseArrow();
      RETURN FALSE 
    END;  
  END;
  mtAppl.MouseArrow();
  RETURN TRUE;
END GetReadList;

PROCEDURE UpdateReadList (list: Lists.List): BOOLEAN;
(* Hiermit wird das Modul aufgefordert, die Lesereihenfolge zu ndern. 
 * Es mu als Parameter eine Liste bergeben werden, die eine genderte 
 * Lesereihenfolge hat und die ber GetReadList angefordert wurde.
 * FALSE: Interne Liste konnte nicht upgedatet werden
 *)
  VAR gr, 
      run   : groupPtr;
      grIdx : INTEGER;
BEGIN
  mtAppl.MouseBusy();
  Lists.ResetList (list);
  run := Lists.NextEntry (list);
  WHILE run # NIL DO
    IF run^.name # NIL
    THEN
      IF findEntry (run^.name^, grIdx)
      THEN
        IF groupArray^[grIdx].readNumber # run^.readNumber
        THEN
          groupArray^[grIdx].readNumber := run^.readNumber;
          listModified := TRUE;
        END;
      END;
    END;
    run := Lists.NextEntry (list);
  END;
  mtAppl.MouseArrow();
  RETURN TRUE;
END UpdateReadList;

PROCEDURE GetUserList (VAR list: Lists.List): BOOLEAN;
(* Diese Prozedur liefert eine Liste der Gruppen, in denen man Chef ist.
 * Diese wird bentigt, um User in andere Gruppen einzutragen.
 * FALSE: Liste konnte nicht erzeugt werden
 *)
 VAR fail : BOOLEAN;
BEGIN
  mtAppl.MouseBusy();
  (* Sysop feststellen *)
  isSysop := Infofiles.IsInfoFile ('ISB', TRUE);
  fail := CopyList (list, copyUser);
  mtAppl.MouseArrow();
  RETURN fail;
END GetUserList;

PROCEDURE GetWriteList (VAR list: Lists.List; full: BOOLEAN): BOOLEAN;
(* Diese Prozedur liefert eine Liste der Gruppen, in denen man schreiben 
 * kann. Diese Liste wird dann benutzt, wenn man eine neue Nachricht in 
 * eine Gruppe schreiben will.
 * FALSE: Liste konnte nicht erzeugt werden
 *)
 VAR fail : BOOLEAN;
     new  : groupPtr;
BEGIN
  mtAppl.MouseBusy();
  fullList := full;
  fail := CopyList (list, copyWrite);
  IF ~fail THEN mtAppl.MouseArrow(); RETURN FALSE END;
  NEW (new);
  IF new = NIL THEN 
    DeleteList (list);
    mtAppl.MouseArrow();
    RETURN FALSE 
  END;
  WITH new^ DO
    ALLOCATE (data, LENGTH (dataSys.personalName)+2);
    IF data = NIL THEN 
      DeleteList (list);
      mtAppl.MouseArrow();
      RETURN FALSE 
    END;
    dataLength := LENGTH (dataSys.personalName) +2;
    name := data;
    Strings.Assign (dataSys.personalName, name^, fail);
    netname := NIL;
    chef    := NIL;
    kurztext  := NIL;
    sysop     := NIL;
    followup  := NIL;
    herkunft  := NIL;
    sprache   := NIL;
    alias     := NIL;
    readNumber:= 0;
    catNumber := dataSys.private;
    readState := rwOn;
    writeState:= rwImpossible;
    userState := uNA;
    isNet     := nNA;
    ordered   := oNA;
    orderDays := 0;
    hasProg   := FALSE;
    withDollar:= FALSE;
    selected  := FALSE;
  END;
  Lists.ResetList (list);
  Lists.InsertEntry (list, new, fail);
  mtAppl.MouseArrow();
  IF fail THEN 
    DEALLOCATE (new^.data, 0);
    DISPOSE (new);
    DeleteList (list);
    RETURN FALSE 
  END;  
  RETURN TRUE;
END GetWriteList;

PROCEDURE GetSelectList (VAR list: Lists.List; lokal: BOOLEAN;
                         sortBy: selSortType): BOOLEAN;
(* Diese Prozedur liefert eine Liste der Gruppen, die in der Maus vorhanden
 * sind bzw. alle Gruppen aus dem ITG. Benutzt wird diese Liste fr die 
 * Gruppenauswahl.
 * lokal: Es werden nur die lokal vorhandenen Gruppen geliefert
 * FALSE: Liste konnte nicht erzeugt werden
 *)
BEGIN
  mtAppl.MouseBusy();
  IF lokal
  THEN
    IF ~CopyList (list, copyLokal) THEN mtAppl.MouseArrow(); RETURN FALSE END;
  ELSE
    IF ~CopyList (list, copyAll) THEN mtAppl.MouseArrow(); RETURN FALSE END;
  END;
  (* Nun noch sortieren *)
  CASE sortBy OF
    sName : (* Ist per Default nach Namen sortiert *) |
    sNet  : SortList (list, sByNet);  |
    sLang : SortList (list, sByLang); |
  ELSE
  END;
  mtAppl.MouseArrow();
  RETURN TRUE;
END GetSelectList;

PROCEDURE UpdateSelectList (list: Lists.List; resort: BOOLEAN): BOOLEAN;
(* Hiermit wird das Modul aufgefordert, sich zu merken, welche Gruppen
 * beim nchsten Tausch bestellt oder abbestellt werden sollen. 
 * Es mu als Parameter eine Liste bergeben werden,
 * die ber GetSelectList angefordert wurde 
 * FALSE: Interne Liste konnte nicht upgedatet werden
 *)
  VAR gr, 
      run   : groupPtr;
      grIdx : INTEGER;
BEGIN
  mtAppl.MouseBusy();
  Lists.ResetList (list);
  run := Lists.NextEntry (list);
  WHILE run # NIL DO
    IF run^.name # NIL
    THEN
      IF findEntry (run^.name^, grIdx)
      THEN
        gr := ADR (groupArray^[grIdx]);
        IF (gr^.ordered # run^.ordered) OR (gr^.orderDays # run^.orderDays)
        THEN
          gr^.ordered := run^.ordered;
          gr^.orderDays := run^.orderDays;
          listModified := TRUE;
        END;
        IF (run^.followup # NIL)
        THEN
          IF ((gr^.followup # NIL) & ~AssFuncs.StrIequal (run^.followup^, gr^.followup^))
            OR (gr^.followup = NIL)
          THEN
            (* Followup setzen *)
            SetFollowUp (gr, run^.followup^);
            listModified := TRUE;
          END;
        ELSIF (gr^.followup # NIL)
        THEN
          (* Followup wurde gelscht *)
          SetFollowUp (gr, "");
          listModified := TRUE;
        END;
      END;
    END;
    run := Lists.NextEntry (list);
  END;
  mtAppl.MouseArrow();
  RETURN TRUE;
END UpdateSelectList;

PROCEDURE DeleteList (VAR list: Lists.List);
(* Lscht eine Liste vollstndig, also auch die Wurzel davon. Danach darf
 * nicht mehr auf die Liste zugegriffen werden.
 *)
  VAR entry: groupPtr;
      fail : BOOLEAN;
BEGIN
  mtAppl.MouseBusy();
  Lists.ResetList (list);
  entry := Lists.PrevEntry (list);
  WHILE entry # NIL DO
    Lists.RemoveEntry (list, v.bool);
    DEALLOCATE (entry^.data, 0L);
    DEALLOCATE (entry, 0L);
    entry := Lists.CurrentEntry (list);
  END;
  Lists.DeleteList (list, fail);
  mtAppl.MouseArrow();
END DeleteList;

(* Jetzt noch ein paar Prozeduren fr data *)
PROCEDURE NameToFile (REF name : ARRAY OF CHAR; REF path, fname : ARRAY OF CHAR):BOOLEAN;
(*$W-*)
(* wg. mtTextfiles.WriteLine der erste Parameter als VAR.. *)
(* 'name' in angegebene Datei schreiben, fr Absender- und Gruppennamen *)
VAR out : INTEGER;
BEGIN
  out := CatFiles.OpenFile(path, fname, CatFiles.writeFile);
  IF out > 0 THEN
    CatFiles.Seek(0, out, CatFiles.end);
    CatFiles.WriteMuch(LONG(LENGTH(name)), out, ADR(name));
    CatFiles.WriteFile(CR, out); CatFiles.WriteFile(LF, out);
    CatFiles.CloseFile(out);
    RETURN CatFiles.FileError = 0
  ELSE
    MTE.info(MTE.nameNotWritten);
    RETURN FALSE
  END;
END NameToFile;
(*$W=*)

PROCEDURE SaveGruppenInf ();
(* Sichert die Liste der Gruppen komplett 
 *)
 VAR file : CatTypes.String255;
     out  : mtTextfiles.TEXTFILE;
     num, 
     i    : INTEGER;
     idx  : CARDINAL;
     tmp  : CatTypes.String255;
BEGIN
  MagicStrings.Assign (DataPath, file);
  MagicStrings.Append (catGroupInf, file);
  IF mtTextfiles.OpenTextfile (file, mtTextfiles.WRITE, 4096, out)
  THEN
    countCatEntries (num);
    FOR i := 1 TO num  DO
      IF findNumberEntry (i, idx)
      THEN
        IF (groupArray^[idx].name # NIL)
        THEN
          MagicStrings.Assign (groupArray^[idx].name^, tmp);
          MagicStrings.CAPS (tmp);
          mtTextfiles.WriteLine (out, tmp);
          mtTextfiles.WriteLn(out);
        END;
      END;
    END;
    mtTextfiles.CloseTextfile (out);
    Protokoll.SendPathUpdate (DataPath);
  END;
END SaveGruppenInf;

PROCEDURE RenameGroup (REF oldName, newName : ARRAY OF CHAR; extern: BOOLEAN): BOOLEAN;
(* Benennt eine Gruppe um. Die GRUPPEN.INF wird danach direkt neu 
 * geschrieben, die nderungen in den anderen Daten werden 
 * am Programmende gesichert
 *)
  VAR found : BOOLEAN;
      idx   : INTEGER;
      idx2  : INTEGER;
      avail : lineSet;
      void,
      gName, nName, uLine, cLine, sLine, rLine, hLine, pLine, aLine : CatTypes.String255;
      gr    : groupPtr;
BEGIN
  (* Persnliche lt sich nicht umbenennen *)
  IF AssFuncs.StrIequal (oldName, newName) THEN RETURN TRUE; END;
  IF AssFuncs.StrIequal(oldName, dataSys.personalName) THEN RETURN FALSE END;
  found := findEntry (oldName, idx);
  IF found
  THEN
    IF findEntry (newName, idx2)
    THEN
      IF groupArray^[idx2].catNumber >= 0
      THEN
        (* Gruppe ist in CAT schon vorhanden *)
        RETURN FALSE
      END;
      (* Es sind beide Gruppennamen vorhanden. D.h. der Eintrag newName
       * erhlt einfach die CAT-Infos zugewiesen, am Eintrag selber
       * wird aber nichts gendert.
       *)
      gr := ADR(groupArray^[idx]);
      WITH groupArray^[idx2] DO
        readNumber := gr^.readNumber;
        catNumber := gr^.catNumber;
        gr^.readNumber := -1;
        gr^.catNumber := -1;
      END;
      listModified := TRUE;
      (* Jetzt die genderte GRUPPEN.INF speichern *)
      SaveGruppenInf();
      (* Jetzt das Array neu sortieren! *)
      SortGroupArray();
      (* Und jetzt raus hier *)
      RETURN TRUE;
    END;
    gr := ADR(groupArray^[idx]);
    avail := lineSet {gL, cN, rN, oN, dN};
    MagicStrings.Assign (newName, gName);
    MagicStrings.Assign ('', nName);
    MagicStrings.Assign ('', uLine);
    MagicStrings.Assign ('', cLine);
    MagicStrings.Assign ('', sLine);
    MagicStrings.Assign ('', rLine);
    MagicStrings.Assign ('', hLine);
    MagicStrings.Assign ('', pLine);
    MagicStrings.Assign ('', aLine);
    IF gr^.netname # NIL
    THEN
      (* Netname zuweisen *)
      MagicStrings.Assign (gr^.netname^, nName);
      INCL (avail, nL);
    END;
    IF (gr^.kurztext # NIL)
    THEN
      (* Netname zuweisen *)
      MagicStrings.Assign (gr^.kurztext^, uLine);
      INCL (avail, uL);
    END;
    IF (gr^.chef # NIL)
    THEN
      (* Netname zuweisen *)
      MagicStrings.Assign (gr^.chef^, cLine);
      INCL (avail, cL);
    END;
    IF (gr^.sysop # NIL)
    THEN
      (* Netname zuweisen *)
      MagicStrings.Assign (gr^.sysop^, sLine);
      INCL (avail, sL);
    END;
    IF (gr^.followup # NIL)
    THEN
      (* Followup zuweisen *)
      MagicStrings.Assign (gr^.followup^, rLine);
      INCL (avail, rL);
    END;
    IF (gr^.herkunft # NIL)
    THEN
      (* Herkunft zuweisen *)
      MagicStrings.Assign (gr^.herkunft^, hLine);
      INCL (avail, hL);
    END;
    IF (gr^.sprache # NIL)
    THEN
      (* Sprache zuweisen *)
      MagicStrings.Assign (gr^.sprache^, pLine);
      INCL (avail, pL);
    END;
    IF (gr^.alias  # NIL)
    THEN
      (* Alias zuweisen *)
      MagicStrings.Assign (gr^.alias^, aLine);
      INCL (avail, aL);
    END;
    IF ~MakeEntry (gr, gName, nName, uLine, cLine, void, sLine, rLine, hLine, pLine, aLine, gr^.readNumber, gr^.catNumber, gr^.ordered, gr^.orderDays, FALSE, avail)
    THEN 
      RETURN FALSE
    END;
    listModified := TRUE;
    (* Jetzt die genderte GRUPPEN.INF speichern *)
    SaveGruppenInf();
    (* Jetzt das Array neu sortieren! *)
    SortGroupArray();
  END;
  RETURN found;
END RenameGroup;

PROCEDURE AppendGroup (VAR groupName : ARRAY OF CHAR):BOOLEAN;
(* Neuen Gruppennamen bernehmen, meckert selber, falls es nicht klappt *)
  VAR new, 
      gr    : groupPtr;
      fail,
      found : BOOLEAN;
      catNums : INTEGER;
      void     : CatTypes.String127;
      testName : CatTypes.String255;
      linesAvail : lineSet;
      lastReadNum: INTEGER;
      i,
      idx        : INTEGER;
      newGroup   : group;
      compRes    : MagicStrings.Relation;
      
BEGIN
  catNums := 0;
  countCatEntries (catNums);
  catNums := catNums + 1;
  lastReadNum := 0;
  idx := 0;
  found := FALSE;
  REPEAT
    gr := ADR(groupArray^[idx]);
    IF (gr^.name # NIL)
    THEN
      IF ~found
      THEN
        compRes := AssFuncs.StrICompare (gr^.name^, groupName);
        IF compRes = MagicStrings.equal
        THEN
          found := TRUE;
          gr^.catNumber := catNums;
          gr^.readNumber := lastReadNum + 1;
        ELSIF compRes = MagicStrings.greater
        THEN
          (* neuen Eintrag anlegen *)
          IF ~MakeNewGroupArray (FALSE, 0)
          THEN RETURN FALSE END;
          newGroup := group{NIL,NIL,NIL,NIL,NIL,NIL,NIL,NIL,NIL,-1,-1,rwNA,rwNA,uNA,nNA,
                            FALSE,FALSE,FALSE,oNA,0,TRUE,0,NIL};
          gr := ADR (newGroup);

          (* Hier jetzt einfgen, da er hierhin pat! *)
          linesAvail := lineSet{gL, cN, rN, oN};
          MagicStrings.Assign ('', void);
          IF ~MakeEntry (gr, groupName, void, void, void, void, void, void, void, void, void, lastReadNum + 1, catNums, oNA, 0, TRUE, linesAvail)
          THEN
            RETURN FALSE
          END;
          IF groups > 0
          THEN
            (* Alle nachfolgenden verschieben *)
            FOR i := groups TO idx + 1 BY -1 DO
              groupArray^[i] := groupArray^[i-1]
            END;
          END;
          (* Eintrag jetzt bernehmen *)
          groupArray^[idx] := gr^;
          INC (groups);
          found := TRUE;
        ELSIF gr^.catNumber >= 0
        THEN
          lastReadNum := gr^.readNumber
        END;
      END;
    END;
    INC (idx);
  UNTIL (idx >= groups) OR (found);
  IF ~found
  THEN
    (* Wahrscheinlich leere Liste, Eintrag bei Index groups anlegen 
     *)
    (* neuen Eintrag anlegen *)
    IF ~MakeNewGroupArray (FALSE, 0)
    THEN RETURN FALSE END;
    newGroup := group{NIL,NIL,NIL,NIL,NIL,NIL,NIL,NIL,NIL,-1,-1,rwNA,rwNA,uNA,nNA,
                      FALSE,FALSE,FALSE,oNA,0,TRUE,0,NIL};
    gr := ADR (newGroup);

    (* Hier jetzt einfgen, da er hierhin pat! *)
    linesAvail := lineSet{gL, cN, rN, oN};
    MagicStrings.Assign ('', void);
    IF ~MakeEntry (gr, groupName, void, void, void, void, void, void, void, void, void, lastReadNum + 1, catNums, oNA, 0, TRUE, linesAvail)
    THEN
      RETURN FALSE
    END;
    (* Eintrag jetzt bernehmen *)
    groupArray^[groups] := gr^;
    INC (groups);
  END;
  (* Jetzt die Lesenummern anpassen *)
  FOR idx := 0 TO groups - 1 DO
    gr := ADR (groupArray^[idx]);
    IF (gr^.catNumber >= 0) & (gr^.readNumber > lastReadNum) & (gr^.catNumber # catNums)
    THEN
      INC (gr^.readNumber);
    END;
  END;
  listModified := TRUE;
  (* Jetzt Namen an GRUPPEN.INF anhngen, dort immer gro geschrieben *)
  MagicStrings.CAPS (groupName);
  RETURN NameToFile(groupName, DataPath, catGroupInf);
END AppendGroup;

PROCEDURE GroupNumber (REF gName : ARRAY OF CHAR; VAR nr : CARDINAL):BOOLEAN;
(* Gruppennummer zu einem String feststellen *)
(* Wenn FALSE, dann kommt Gruppenanzahl in nr zurck, Anzahl      *)
(* inklusive persnliche                                          *)
  VAR found : BOOLEAN;
      idx   : INTEGER;
      iNr   : INTEGER;
BEGIN
  IF AssFuncs.StrIequal(gName, dataSys.personalName) THEN nr := 0; RETURN TRUE END;
  found := findEntry (gName, idx);
  IF found
  THEN
    IF groupArray^[idx].catNumber < 0
    THEN
      found := FALSE;
    ELSE
      nr := CARDINAL (groupArray^[idx].catNumber);
      RETURN TRUE
    END;
  END;
  IF ~found
  THEN
    countCatEntries (iNr);
    nr := iNr + 1;
  END;
  RETURN FALSE
END GroupNumber;

PROCEDURE CountCatGroups(): CARDINAL;
(* Liefert die Zahl der Gruppen in CAT zurck, inklusive der persnlichen Gruppe 
 *)
 VAR iNr : INTEGER;
BEGIN
  countCatEntries (iNr);
  RETURN iNr + 1;
END CountCatGroups;

PROCEDURE NextGroupNumber (old : CARDINAL):CARDINAL;
(* Nummer der nchsten Gruppe in der Reihenfolge der momentanen Sortierung *)
(* feststellen; 0FFFF = empty falls keine weiter Gruppe vorhanden ist      *)
  VAR new : INTEGER;
      idx  : CARDINAL;
BEGIN
  new := -1;
  IF old # dataSys.private
  THEN
    IF findNumberEntry (old, idx)
    THEN
      old := groupArray^[idx].readNumber;
    END;
  END;
  IF findReadNumberEntry (old+1, idx)
  THEN 
    IF groupArray^[idx].catNumber >= 0
    THEN 
      new := groupArray^[idx].catNumber
    END;
  END;
  RETURN CARDINAL (new);
END NextGroupNumber;

PROCEDURE PreviousGroupNumber (old : CARDINAL):CARDINAL;
(* Nummer der vorhergehenden Gruppe in der Reihenfolge der momentanen Sortierung *)
(* feststellen; 0FFFF = empty falls keine weiter Gruppe vorhanden ist      *)
  VAR new : INTEGER;
      idx  : CARDINAL;
BEGIN
  new := -1;
  IF old # dataSys.private
  THEN
    IF findNumberEntry (old, idx)
    THEN
      old := groupArray^[idx].readNumber;
    END;
  ELSE
    RETURN 0FFFFH;
  END;
  IF findReadNumberEntry (old-1, idx)
  THEN 
    IF groupArray^[idx].catNumber >= 0
    THEN 
      new := groupArray^[idx].catNumber
    END;
  END;
  RETURN CARDINAL (new);
END PreviousGroupNumber;

PROCEDURE NetGroupNumber(REF gName : ARRAY OF CHAR; VAR nr : CARDINAL):BOOLEAN;
(* Gruppennummer zu einem String mit einem Netzgruppennamen feststellen *)
(* Wenn FALSE, dann kommt Gruppenanzahl in nr zurck, Anzahl      *)
(* inklusive persnliche                                          *)
  VAR found : BOOLEAN;
      idx   : INTEGER;
      iNr   : INTEGER;
BEGIN
  IF AssFuncs.StrIequal(gName, dataSys.personalName) THEN nr := 0; RETURN TRUE END;
  found := findNetEntry (gName, idx);
  IF found
  THEN
    IF groupArray^[idx].catNumber < 0
    THEN
      found := FALSE;
    ELSE
      nr := CARDINAL (groupArray^[idx].catNumber);
      RETURN TRUE
    END;
  END;
  IF ~found
  THEN
    countCatEntries (iNr);
    nr := iNr + 1;
  END;
  RETURN FALSE
END NetGroupNumber;

PROCEDURE GroupName (nr : CARDINAL; VAR name : ARRAY OF CHAR);
(* Gruppenname zu einer Nummer feststellen, liefert im Fehlerfall *)
(* keinen Eintrag *)
  VAR gr    : groupPtr;
      iNr   : INTEGER;
      idx   : CARDINAL;
BEGIN
  iNr := INTEGER (nr);
  MagicStrings.Assign ("", name);
  IF iNr = dataSys.private
  THEN
    MagicStrings.Assign (dataSys.personalName, name)
  ELSIF (iNr >= 0)
  THEN
    IF findNumberEntry (nr, idx)
    THEN
      IF (groupArray^[idx].name # NIL)
      THEN
        MagicStrings.Assign (groupArray^[idx].name^, name);
      END;
    END;
  END;
END GroupName;

VAR groupChangeIdx : INTEGER;

PROCEDURE GetGroupChange (VAR str : ARRAY OF CHAR; mode: gcMode): BOOLEAN;
(* Liefert bei TRUE in str einen String mit +gname oder -gname, 
 * je nachdem, ob eine Gruppe angestellt oder abgestellt werden soll
 *)
  VAR gr : groupPtr;
BEGIN
  MagicStrings.Assign ("", str);
  IF mode = GFIRST
  THEN
    groupChangeIdx := 0;
  ELSE
    INC (groupChangeIdx);
  END;
  WHILE (groupChangeIdx < groups) DO
    gr := ADR (groupArray^[groupChangeIdx]);
    IF (gr^.ordered # oNA)
     & (gr^.name # NIL)
    THEN
      IF gr^.ordered = oOn
      THEN
        MagicStrings.Assign ('+', str);
      ELSE
        MagicStrings.Assign ('-', str);
      END;
      MagicStrings.Append (gr^.name^, str);
      RETURN TRUE;
    END;
    INC (groupChangeIdx);
  END;
  RETURN FALSE;
END GetGroupChange;

PROCEDURE ClearGroupChange (VAR name : ARRAY OF CHAR);
(* Lscht eine Gruppennderung fr die Gruppe name
 *)
  VAR idx : INTEGER;
BEGIN
  IF findEntry (name, idx)
  THEN
    IF groupArray^[idx].ordered = oOn
    THEN
      groupArray^[idx].readState := rwOn;
    ELSIF groupArray^[idx].ordered = oOff
    THEN
      groupArray^[idx].readState := rwPossible;
    END;
    groupArray^[idx].ordered := oNA;
    groupArray^[idx].orderDays := 0;
    listModified := TRUE;
  END;
END ClearGroupChange;

PROCEDURE SetFollowUp (gr : groupPtr; REF name : ARRAY OF CHAR);
(* Setzt das Follow-Up fr die Gruppe group
 *)
  VAR avail : lineSet;
      void,
      gName, nName, uLine, cLine, sLine, rLine, hLine, pLine, aLine : CatTypes.String255;
BEGIN
  avail := lineSet {rL, cN, rN, oN, dN};
  MagicStrings.Assign ('', gName);
  MagicStrings.Assign ('', nName);
  MagicStrings.Assign ('', uLine);
  MagicStrings.Assign ('', cLine);
  MagicStrings.Assign ('', sLine);
  MagicStrings.Assign ('', hLine);
  MagicStrings.Assign ('', pLine);
  MagicStrings.Assign ('', aLine);
  IF gr^.name # NIL
  THEN
    MagicStrings.Assign (gr^.name^, gName);
    INCL (avail, gL);
  END;
  IF gr^.netname # NIL
  THEN
    (* Netname zuweisen *)
    MagicStrings.Assign (gr^.netname^, nName);
    INCL (avail, nL);
  END;
  IF (gr^.kurztext # NIL)
  THEN
    (* Netname zuweisen *)
    MagicStrings.Assign (gr^.kurztext^, uLine);
    INCL (avail, uL);
  END;
  IF (gr^.chef # NIL)
  THEN
    (* Netname zuweisen *)
    MagicStrings.Assign (gr^.chef^, cLine);
    INCL (avail, cL);
  END;
  IF (gr^.sysop # NIL)
  THEN
    (* Netname zuweisen *)
    MagicStrings.Assign (gr^.sysop^, sLine);
    INCL (avail, sL);
  END;
  IF (gr^.herkunft # NIL)
  THEN
    (* Netname zuweisen *)
    MagicStrings.Assign (gr^.herkunft^, hLine);
    INCL (avail, hL);
  END;
  IF (gr^.sprache # NIL)
  THEN
    (* Netname zuweisen *)
    MagicStrings.Assign (gr^.sprache^, pLine);
    INCL (avail, pL);
  END;
  IF (gr^.alias # NIL)
  THEN
    (* Alias zuweisen *)
    MagicStrings.Assign (gr^.alias^, aLine);
    INCL (avail, aL);
  END;
  MagicStrings.Assign (name, rLine);
  IF ~MakeEntry (gr, gName, nName, uLine, cLine, void, sLine, rLine, hLine, pLine, aLine, gr^.readNumber, gr^.catNumber, gr^.ordered, gr^.orderDays, FALSE, avail)
  THEN 
  END;
END SetFollowUp;

PROCEDURE GetFollowUp (REF gName: ARRAY OF CHAR; VAR followupTo : ARRAY OF CHAR): BOOLEAN;
(* Liefert das Followup fr eine Gruppe 
 *)
  VAR idx : INTEGER;
BEGIN
  IF findEntry (gName, idx)
  THEN
    IF groupArray^[idx].followup # NIL
    THEN
      MagicStrings.Assign (groupArray^[idx].followup^, followupTo);
      RETURN TRUE
    END;
  END;
  RETURN FALSE;
END GetFollowUp;

PROCEDURE GetNetName (VAR name: ARRAY OF CHAR);
(* Liefert zu einer Gruppe den Netzgruppennamen 
 *)
 VAR idx : INTEGER;
BEGIN
  IF findEntry (name, idx)
  THEN
    IF groupArray^[idx].netname # NIL
    THEN
      MagicStrings.Assign (groupArray^[idx].netname^, name);
    END;
  END;
END GetNetName;

PROCEDURE GetWriteAccess (REF name: ARRAY OF CHAR): BOOLEAN;
(* Liefert TRUE zurck, falls die Gruppe existiert und der 
 * Benutzer laut ITG schreiben darf
 *)
 VAR idx : INTEGER;
BEGIN
  IF findEntry (name, idx)
  THEN
    RETURN (groupArray^[idx].writeState = rwNA)
        OR (groupArray^[idx].writeState = rwOn)
        OR (groupArray^[idx].writeState = rwPossible);
  END;
  RETURN FALSE;
END GetWriteAccess;

(* Funktionen fr CatPutz: *)

PROCEDURE DeleteGroup (REF gName: ARRAY OF CHAR);
(* Lscht die Gruppe aus der Liste der in CAT verfgbaren Gruppen und 
 * pat die Lesenummern an. Die Filenummern fr die Datenbank
 * mssen anders angepat werden (z.B. mit SetNewCatNumber).
 * Ebenso mu das Gruppen.Pos extern angepat werden.
 *)
  VAR idx : INTEGER;
      readnum : INTEGER;
      i     : INTEGER;
BEGIN
  IF findEntry (gName, idx)
  THEN
    readnum := groupArray^[idx].readNumber;
    groupArray^[idx].catNumber := -1;
    groupArray^[idx].readNumber := -1;
    listModified := TRUE;
    FOR i := 0 TO groups - 1 DO
      IF groupArray^[i].readNumber > readnum
      THEN
        DEC (groupArray^[i].readNumber);
      END;
    END;
  END; 
END DeleteGroup;

PROCEDURE SetNewCatNumber (REF gName: ARRAY OF CHAR; newNumber: INTEGER);
(* Setzt eine neue Gruppennummer fr eine Gruppe. 
 * Das GRUPPEN.INF wird danach neu geschrieben 
 *)
  VAR idx : INTEGER;
BEGIN
  IF findEntry (gName, idx)
  THEN
    groupArray^[idx].catNumber := newNumber;
    listModified := TRUE;
    (* Jetzt neues GRUPPEN.INF schreiben *)
    SaveGruppenInf();
  END;
END SetNewCatNumber;

BEGIN
  listModified := FALSE;
  maxGroups := 0;
  groups    := 0;
  groupArray := NIL;
  isSysop   := FALSE;
END GroupSelect.
