IMPLEMENTATION MODULE GruppenAuswahl;

(*============================================================================
 * Modul:               GruppenAuswahl
 * Autor:               Johannes Gttker-Schnetmann
 * Beschreibung:        Dialoghandling fr Auswahl der angestellten Gruppen
 *                      Wird gelesen aus kurzer Gruppenliste (IGK.TXT)
 * Version:             2.0
 * letzte nderung:     26.7.93
 * ============================================================================
 * Autor Datum     Version  nderung
 *  DS   10.6.91   1.2      groe Teile neu geschrieben. Gruppen werden jetzt
 *                          sortiert angezeigt. Umgestellt auf mtTextfiles.
 *                          Teilweise Code auch gestrafft!
 *  DS   25.10.91  1.2x     Anpassung an neues Format von IGK mit Erluterungszeilen
 *                          am Ende.
 *  DS   26.07.93  2.0      Komplett neu geschrieben fr ITG und interne Gruppenliste
 *)

FROM SYSTEM       IMPORT ADDRESS, TSIZE, ADR, CADR;
IMPORT GrafBase;
IMPORT BinOps;
IMPORT Lists;

IMPORT MagicAES;
IMPORT MagicDOS;
IMPORT MausTauschrsc;
IMPORT MagicStrings;
IMPORT mtAppl;
IMPORT WinDials;
IMPORT mtUtils;
IMPORT mtAlerts;
IMPORT mtPopups;

IMPORT CatTypes;
IMPORT CatFiles;
IMPORT CatHelp;
IMPORT CatGlobal;
IMPORT MTE;
IMPORT grinTools;
IMPORT GroupSelect;
IMPORT ListDl;
IMPORT ListHelp;
IMPORT Messages;
IMPORT MTPaths;
IMPORT ConfVars;
IMPORT Infofiles;
IMPORT VDIUtil;
IMPORT Varnames;
FROM ListDl       IMPORT ldElems;
FROM Void         IMPORT v;

FROM GroupSelect IMPORT rwState, uState, netState, orderState;

CONST   cSelFollowup    = 'Followup auswhlen';
        cSelGroup       = 'Gruppen auswhlen';

TYPE Trick =  POINTER TO RECORD
                CASE : BOOLEAN OF
                  TRUE:  wert: INTEGER;|
                  FALSE: hi: CHAR;
                         lo: CHAR;|
                END;
              END;

VAR grList  : Lists.List;
    groups  : ADDRESS;
    globalLength: INTEGER;
    maxGroupNameLength : INTEGER;
    isSysop : BOOLEAN;

PROCEDURE ExtendStr (VAR str : ARRAY OF CHAR; len : CARDINAL);
  VAR (*$Reg*) max, 
      (*$Reg*) s, 
      (*$Reg*) i : CARDINAL;
BEGIN
  IF len > HIGH (str) THEN max := HIGH(str) ELSE max := len END;
  s := LENGTH (str);
  IF s < len
  THEN
    FOR i := s TO max DO str[i] := ' '; END;
    IF max < HIGH (str) THEN str[max+1] := 0C; END;
  ELSE
    (* Nur ein Space anhngen *)
    str[s] := ' ';
    str[s+1] := 0C;
  END;
END ExtendStr;

PROCEDURE selGrEntry (entry, env : ADDRESS; line : INTEGER): BOOLEAN;
  VAR gr : GroupSelect.groupPtr;
      str: ARRAY [0..255] OF CHAR;
BEGIN
  gr := entry;
  IF gr = NIL THEN RETURN FALSE END;
  gr^.selected := ~gr^.selected;
  IF gr^.selected & (line >= 0) THEN
    mtUtils.SetObjcString (groups, MausTauschrsc.grname, gr^.name^); 
    WinDials.WinDialDraw (groups, MausTauschrsc.grname, 0, v.r, FALSE);
  END;
  RETURN FALSE
END selGrEntry;

PROCEDURE groupToStr (entry, env : ADDRESS; VAR str : ARRAY OF CHAR);
  VAR gr : GroupSelect.groupPtr;
BEGIN
  gr := entry;
  IF gr = NIL THEN RETURN END;
  MagicStrings.Assign (gr^.name^,str);
END groupToStr;

PROCEDURE buildGroup (entry, env : ADDRESS; VAR str : ARRAY OF CHAR);
  VAR gr : GroupSelect.groupPtr;
BEGIN
  gr := GroupSelect.groupPtr (entry);
  IF ((gr^.readState = rwOn) OR (gr^.readState = rwDefault)) & (gr^.ordered # oOff)
  THEN
    MagicStrings.Assign (10C+' ', str);
  ELSIF (gr^.ordered = oOff)
  THEN
    MagicStrings.Assign ('- ', str);
  ELSIF (gr^.ordered = oOn)
  THEN
    MagicStrings.Assign ('+ ', str);
  ELSE
    MagicStrings.Assign ('  ', str);
  END;
  MagicStrings.Append (gr^.name^,str);
  IF gr^.followup # NIL
  THEN
    MagicStrings.Append (' '+3C+0C, str);
  END;
  ExtendStr (str, maxGroupNameLength+5);
  IF gr^.kurztext # NIL THEN MagicStrings.Append (gr^.kurztext^, str); END;
  globalLength := BinOps.HigherInt (globalLength, LENGTH (str)+1);
END buildGroup;

(*$Z-*)  
PROCEDURE findSel (e, i : ADDRESS): BOOLEAN;
  VAR entry : GroupSelect.groupPtr;
BEGIN
  entry := e;
  RETURN entry^.selected;
END findSel;
(*$Z=*)

(*
CONST   isUnOrdered = '[2][CAT:|Diese Gruppe ist bereits abgestellt.|Doch nicht abstellen?][[Ja|:[Nein]';
        isOrdered   = '[2][CAT:|Diese Gruppe ist bereits angestellt.|Doch nicht bestellen?][[Ja|:[Nein]';
*)

VAR     Adressat    : CatTypes.String255;
        Wegen       : CatTypes.String255;

PROCEDURE orderGroup (obj: INTEGER; env, info: ADDRESS; VAR draw : BOOLEAN; VAR exit : BOOLEAN);
  VAR oneSel : BOOLEAN;
      gr     : GroupSelect.groupPtr;
BEGIN
  exit := FALSE;
  draw := TRUE;
  Lists.ResetList (grList);
  Lists.ScanEntries (grList, Lists.forward, findSel, NIL, oneSel);
  IF oneSel
  THEN
    (* Gruppe an oder abbestellen *)
    gr := Lists.CurrentEntry (grList);
    CASE gr^.readState OF
      rwOn        : IF gr^.ordered # oNA
                    THEN
                      gr^.ordered := oNA
                      (* 
                      v.int := mtAlerts.Alert (1, isUnOrdered);
                      IF v.int = 1
                      THEN
                        gr^.ordered := oNA
                      ELSE
                        draw := FALSE;
                      END;
                      *)
                    ELSE
                      gr^.ordered := oOff;
                    END; |
      rwNA,
      rwPossible  : IF gr^.ordered # oNA
                    THEN 
                      gr^.ordered := oNA;
                      (*
                      v.int := mtAlerts.Alert (1, isOrdered);
                      IF v.int = 1
                      THEN
                        gr^.ordered := oNA
                      ELSE
                        draw := FALSE;
                      END;
                      *)
                    ELSE
                      gr^.ordered := oOn; 
                    END; |
      rwAskChef   : v.int := mtAlerts.Alert (1, MTE.orderChef);
                    IF v.int = 1
                    THEN
                      MagicStrings.Assign ('Chef ', Adressat);
                      MagicStrings.Append (gr^.name^, Adressat);
                      MagicStrings.Assign ('Gruppe ', Wegen);
                      MagicStrings.Append (gr^.name^, Wegen);
                      exit := TRUE;
                      draw := FALSE;
                    END; |
      rwAskSysop  : v.int := mtAlerts.Alert (1, MTE.orderSysop);
                    IF v.int = 1
                    THEN
                      MagicStrings.Assign ('Sysop', Adressat);
                      MagicStrings.Assign ('Gruppe ', Wegen);
                      MagicStrings.Append (gr^.name^, Wegen);
                      draw := FALSE;
                      exit := TRUE;
                    END; |
      rwImpossible: draw := FALSE; MTE.info (MTE.orderImpossible); |
      rwDefault   : draw := FALSE; MTE.info (MTE.unOrderImpossible); |
      rwNoPermission: draw := FALSE; MTE.info (MTE.orderNoPermission); |
    ELSE
    END;
  END;
  mtUtils.ExclState (groups, obj, MagicAES.SELECTED);
  WinDials.WinDialDraw (groups, obj, 1, v.r, FALSE);
END orderGroup;

PROCEDURE doChef (obj: INTEGER; env, info: ADDRESS; VAR draw, exit: BOOLEAN);
  CONST title = "Gruppenchef setzen";
        cname = "Chef:";
        CR      = 15C;
        LF      = 12C;
VAR cmd,
    uName: CatTypes.String255;
    out  : INTEGER;
    but  : INTEGER;
    wegen: ADDRESS;
    gr      : GroupSelect.groupPtr;
    oneSel  : BOOLEAN;
BEGIN
  (* Neuen Dialog aufmachen und nach Usernamen fragen
   *)
  Lists.ResetList (grList);
  Lists.ScanEntries (grList, Lists.forward, findSel, NIL, oneSel);
  IF oneSel
  THEN
    (* Followup fr Gruppe einstellen *)
    gr := Lists.CurrentEntry (grList);
    
    wegen := MausTauschrsc.TreeAddr^[MausTauschrsc.wegen];
    mtUtils.SetObjcStringAdr (wegen, MausTauschrsc.wgruppe, gr^.name);
    mtUtils.SetObjcStringAdr (wegen, MausTauschrsc.optstr,  CADR(title));
    mtUtils.SetObjcStringAdr (wegen, MausTauschrsc.prestr,  CADR(cname));
    mtUtils.SetObjcString (wegen, MausTauschrsc.wegenstr, "");
    but := VDIUtil.FormHandle (wegen);
    IF but = MausTauschrsc.wegenok
    THEN
      mtUtils.ObjcString (wegen, MausTauschrsc.wegenstr, uName);
      IF LENGTH (uName) = 0
      THEN
        MTE.info (MTE.noChefSelected);
      ELSE
        (* User als Chef setzen *)
        out := CatFiles.OpenFile(MTPaths.MessagePath, 'grselect.inf', CatFiles.writeFile);
        IF out < 0 THEN
          MTE.info (MTE.changeNoSave);
        ELSE
          CatFiles.Seek(0, out, CatFiles.end);
        END;
        IF out > 0 THEN
          (* Kommandostring bauen *)
          MagicStrings.Assign(":GU", cmd);
          MagicStrings.Append(uName, cmd);
          MagicStrings.Append("=", cmd);
          MagicStrings.Append(gr^.name^, cmd);
          CatFiles.WriteMuch(LONG(MagicStrings.Length(cmd)), out, ADR(cmd));
          CatFiles.WriteFile(CR, out); 
          CatFiles.WriteFile(LF, out);
          CatFiles.CloseFile(out);
        END;
      END;
    END;
  END;
  mtUtils.ExclState (groups, obj, MagicAES.SELECTED);
  WinDials.WinDialDraw (groups, obj, 1, v.r, FALSE);
  draw := FALSE;
  exit := FALSE;
END doChef;

PROCEDURE doHelp (obj: INTEGER; env, info: ADDRESS; VAR draw, exit: BOOLEAN);
BEGIN
  CatHelp.DoHelp (CatHelp.selectGroup);
  mtUtils.ExclState (groups, obj, MagicAES.SELECTED);
  WinDials.WinDialDraw (groups, obj, 1, v.r, FALSE); 
  draw := FALSE;
  exit := FALSE;
END doHelp;

PROCEDURE switchSort  (obj: INTEGER; env, info: ADDRESS; VAR draw, exit: BOOLEAN);
  VAR typeSel : mtUtils.tObjcTree;
      tree    : mtUtils.tObjcTree;
      x, y          : INTEGER;
      select        : INTEGER;
      t             : Trick;
      sort          : GroupSelect.selSortType;
      but           : INTEGER;
BEGIN
  (* Liste aktualisieren *)
  GroupSelect.DeleteList (grList);

  tree := groups;
  typeSel := MausTauschrsc.TreeAddr^[MausTauschrsc.sorttype];
  but := obj;
  CASE obj OF 
    MausTauschrsc.grsort0,
    MausTauschrsc.grsort1: 
           but := MausTauschrsc.grsort1;
           MagicAES.ObjcOffset (tree, but, x, y);
           t := ADR (tree^[but].obState);
           select := mtPopups.TreePopup (typeSel, x, y, ORD(t^.hi))-1;
           IF select >= 0 THEN
             t^.hi := CHR(select);
             VDIUtil.SetXString (tree, but, typeSel, select);
             WinDials.WinDialDraw (tree, but, 0, v.r, FALSE);
           END;|
    MausTauschrsc.grsort2:
           t := ADR (tree^[but-1].obState);
           t^.hi := CHR((ORD(t^.hi)+1) MOD 3);
           select := ORD (t^.hi);
           VDIUtil.SetXString (tree, but-1, typeSel, select);
           WinDials.WinDialDraw (tree, but-1, 0, v.r, FALSE); |
  ELSE
  END;
  sort := GroupSelect.selSortType(VDIUtil.GetXState (tree, MausTauschrsc.grsort1));
  v.bool := ConfVars.SetConfigInt (cSortSelectList, VDIUtil.GetXState (tree, MausTauschrsc.grsort1));
  
  IF mtUtils.InState (groups, MausTauschrsc.grlokal, MagicAES.SELECTED)
  THEN
    v.bool := GroupSelect.GetSelectList (grList, TRUE, sort);
  ELSE
    (* Nur Liste der lokal vorhandenen Gruppen holen *)
    v.bool := GroupSelect.GetSelectList (grList, FALSE, sort);
  END;
  draw := TRUE;
  exit := FALSE;
END switchSort;

PROCEDURE switchLokal (obj: INTEGER; env, info: ADDRESS; VAR draw, exit: BOOLEAN);
  VAR sort          : GroupSelect.selSortType;
BEGIN
  sort := GroupSelect.selSortType(VDIUtil.GetXState (groups, MausTauschrsc.grsort1));
  GroupSelect.DeleteList (grList);
  IF mtUtils.InState (groups, obj, MagicAES.SELECTED)
  THEN
    v.bool := GroupSelect.GetSelectList (grList, TRUE, GroupSelect.sName);
    v.bool := ConfVars.SetConfigBool (cLokalSelectList, TRUE);
  ELSE
    (* Nur Liste der lokal vorhandenen Gruppen holen *)
    v.bool := GroupSelect.GetSelectList (grList, FALSE, GroupSelect.sName);
    v.bool := ConfVars.SetConfigBool (cLokalSelectList, FALSE);
  END;
  draw := TRUE;
  exit := FALSE;
END switchLokal;

VAR followUpSelect  : BOOLEAN;
    followUpEntry   : GroupSelect.groupPtr;

PROCEDURE followUpTo (obj: INTEGER; env, info: ADDRESS; VAR draw : BOOLEAN; VAR exit : BOOLEAN);
  VAR gr : GroupSelect.groupPtr;
      oneSel : BOOLEAN;
      newFollowup: BOOLEAN;
      r     : mtUtils.tRect;
  
  PROCEDURE drawBack (obj: INTEGER);
  BEGIN
    mtUtils.CalcArea (groups, obj, r);
    INC (r.w, 2); INC (r.h, 2);
    WinDials.WinDialDraw (groups, 0, 8, r, TRUE);
  END drawBack;
  
BEGIN
  exit := FALSE;
  draw := FALSE;
  Lists.ResetList (grList);
  Lists.ScanEntries (grList, Lists.forward, findSel, NIL, oneSel);
  IF oneSel
  THEN
    (* Followup fr Gruppe einstellen *)
    gr := Lists.CurrentEntry (grList);
    IF gr^.followup # NIL
    THEN
      v.int := mtAlerts.Alert (2, MTE.followUpDefined);
      newFollowup := v.int = 2;
      IF v.int = 1
      THEN
        (* Followup lschen *)
        GroupSelect.SetFollowUp (gr, "");
        draw := TRUE;
      END;
    ELSE
      v.int := mtAlerts.Alert (1, MTE.noFollowUp);
      newFollowup := v.int = 1;
    END;
    IF newFollowup
    THEN
      (* Neuer Modus! *)
      (* Jetzt die beiden Buttons verstecken *)
      mtUtils.SetFlag (groups, MausTauschrsc.grorder, MagicAES.HIDETREE, TRUE);
      mtUtils.SetFlag (groups, MausTauschrsc.grchef, MagicAES.HIDETREE, TRUE);
      mtUtils.SetFlag (groups, MausTauschrsc.grfollow, MagicAES.HIDETREE, TRUE);
      drawBack (MausTauschrsc.grchef);
      drawBack (MausTauschrsc.grorder);
      drawBack (MausTauschrsc.grfollow);
      (* Neuen Titel setzen *)
      mtUtils.SetObjcString (groups, MausTauschrsc.grtitle, cSelFollowup);
      WinDials.WinDialDraw (groups, MausTauschrsc.grtitle, 8, v.r, FALSE);
      WinDials.WinUpdateTitle (groups);
      MTE.info (MTE.selFollowUp);
      followUpSelect := TRUE;
      followUpEntry := gr;
    END;
  END;
  mtUtils.ExclState (groups, obj, MagicAES.SELECTED);
  WinDials.WinDialDraw (groups, obj, 1, v.r, FALSE);
END followUpTo;

VAR buttonOrder : BOOLEAN;

PROCEDURE ExtraktValue (REF source: ARRAY OF CHAR; VAR dest: ARRAY OF CHAR);
  VAR p: CARDINAL;
BEGIN
  MagicStrings.Assign ("", dest);
  p := MagicStrings.Pos (":", source, 0, FALSE);
  IF p <= LENGTH (source)
  THEN
    MagicStrings.Copy (source, p+1, LENGTH (source) - p - 1, dest);
  END;
END ExtraktValue;

PROCEDURE updateButton (entry, env : ADDRESS);
  VAR oneSel : BOOLEAN;
      gr     : GroupSelect.groupPtr;
      str    : CatTypes.String255;
      str2   : CatTypes.String255;
      r1, r2 : GrafBase.Rectangle;
      grIsOn : BOOLEAN;
      sort   : GroupSelect.selSortType;
BEGIN
  Lists.ResetList (grList);
  Lists.ScanEntries (grList, Lists.forward, findSel, NIL, oneSel);
  IF oneSel
  THEN
    (* Die Textzeilen unten mit Zusatzinfos fllen *)
    gr := Lists.CurrentEntry (grList);
    (* Erste Zeile: Textinfos *)
    MagicStrings.Assign ('', str);
    IF gr^.chef # NIL
    THEN
      MagicStrings.Append ('Chef: ', str);
      MagicStrings.Append (gr^.chef^, str);
      MagicStrings.Append ('  ', str);
    END;
    IF gr^.netname # NIL
    THEN
      MagicStrings.Append ('Netzname: ', str);
      MagicStrings.Append (gr^.netname^, str);
      MagicStrings.Append ('  ', str);
    END;
    IF gr^.followup # NIL
    THEN
      MagicStrings.Append ('Follow-Up: ', str);
      MagicStrings.Append (gr^.followup^, str);
      MagicStrings.Append ('  ', str);
    END;
    IF gr^.alias # NIL
    THEN
      MagicStrings.Append ('Alias: ', str);
      MagicStrings.Append (gr^.alias^, str);
      MagicStrings.Append ('  ', str);
    END;
    sort := GroupSelect.selSortType(VDIUtil.GetXState (groups, MausTauschrsc.grsort1));
    IF (sort = GroupSelect.sNet) 
    THEN
      IF (gr^.herkunft # NIL)
      THEN
        MagicStrings.Append ('Netz: ', str);
        ExtraktValue (gr^.herkunft^, str2);
        MagicStrings.Append (str2, str);
      END;
    ELSIF (sort = GroupSelect.sLang) 
    THEN
      IF (gr^.sprache # NIL)
      THEN
        MagicStrings.Append ('Sprache: ', str);
        ExtraktValue (gr^.sprache^, str2);
        MagicStrings.Append (str2, str);
      END;
    END;
    mtUtils.SetObjcString (groups, MausTauschrsc.groupinf1, str);
    (* Zweite Zeile: Flags auswerten *)
    MagicStrings.Assign ('', str);
    IF (gr^.isNet # nNetForbid)
    THEN
      IF (gr^.readState # rwNA)
      THEN
        MagicStrings.Append ('Lesen: ', str);
        CASE gr^.readState OF
          rwOn    : MagicStrings.Append ('ist an', str) |
          rwPossible  : MagicStrings.Append ('anstellbar', str) |
          rwAskChef   : MagicStrings.Append ('Chef fragen', str) |
          rwAskSysop  : MagicStrings.Append ('Sysop fragen', str) |
          rwImpossible: MagicStrings.Append ('unmglich', str) |
          rwDefault   : MagicStrings.Append ('immer an', str) |
          rwNoPermission  : MagicStrings.Append ('nicht erlaubt', str) |
        ELSE
        END;
        MagicStrings.Append ('  ', str);
      END;
      IF (gr^.writeState # rwNA) 
      THEN
        IF gr^.writeState # rwNoPermission
        THEN
          MagicStrings.Append ('Schreiben: ', str);
        END;
        CASE gr^.writeState OF
          rwOn    : MagicStrings.Append ('ist mglich', str) |
          rwPossible  : MagicStrings.Append ('anstellbar', str) |
          rwAskChef   : MagicStrings.Append ('Chef fragen', str) |
          rwAskSysop  : MagicStrings.Append ('Sysop fragen', str) |
          rwImpossible: MagicStrings.Append ('unmglich', str) |
        ELSE
        END;
        MagicStrings.Append ('  ', str);
      END;
    END;
    IF gr^.isNet # nNA
    THEN
      CASE gr^.isNet OF
        nNetLokal,
        nNet    : MagicStrings.Append ('Vernetzt', str); |
        nLokal  : MagicStrings.Append ('Lokal', str); |
        nNetForbid: MagicStrings.Append ('Diese Gruppe ist fr diese Box nicht bestellbar.', str); |
      ELSE
      END;
      MagicStrings.Append ('  ', str);
    END;
    (*
    IF gr^.userState # uNA
    THEN
      MagicStrings.Append ('Sie sind ', str);
      CASE gr^.userState OF
        uIsChef : MagicStrings.Append ('der Chef', str) |
        uIsUser : MagicStrings.Append ('Mitglied', str) |
        uNoUser : MagicStrings.Append ('kein Mitglied', str) |
      ELSE
      END;
      MagicStrings.Append (' der Gruppe', str);
    END;
    *)
    mtUtils.SetObjcString (groups, MausTauschrsc.groupinf2, str);
    (* Button richtig setzen *)
    IF ~followUpSelect & mtUtils.InState (groups, MausTauschrsc.grorder, MagicAES.DISABLED)
    THEN
      mtUtils.SetState (groups, MausTauschrsc.grchef, MagicAES.DISABLED, ~isSysop);
      mtUtils.SetState (groups, MausTauschrsc.grorder, MagicAES.DISABLED, FALSE);
      mtUtils.SetState (groups, MausTauschrsc.grfollow, MagicAES.DISABLED, FALSE);
      WinDials.WinDialDraw (groups, MausTauschrsc.grchef, 0, v.r, FALSE);
      WinDials.WinDialDraw (groups, MausTauschrsc.grorder, 0, v.r, FALSE);
      WinDials.WinDialDraw (groups, MausTauschrsc.grfollow, 0, v.r, FALSE);
    ELSIF followUpSelect & ~mtUtils.InState (groups, MausTauschrsc.grorder, MagicAES.DISABLED)
    THEN
      mtUtils.SetState (groups, MausTauschrsc.grchef, MagicAES.DISABLED, TRUE);
      mtUtils.SetState (groups, MausTauschrsc.grorder, MagicAES.DISABLED, TRUE);
      mtUtils.SetState (groups, MausTauschrsc.grfollow, MagicAES.DISABLED, TRUE);
      WinDials.WinDialDraw (groups, MausTauschrsc.grchef, 0, v.r, FALSE);
      WinDials.WinDialDraw (groups, MausTauschrsc.grorder, 0, v.r, FALSE);
      WinDials.WinDialDraw (groups, MausTauschrsc.grfollow, 0, v.r, FALSE);
    END;
    IF ~followUpSelect
    THEN
      grIsOn := ((gr^.readState = rwOn) OR (gr^.readState = rwDefault) OR (gr^.ordered = oOn)) & ~(gr^.ordered = oOff);
      IF ~grIsOn & buttonOrder
      THEN
        (* Buttontext auf bestellen setzen *)
        mtUtils.SetObjcString (groups, MausTauschrsc.grorder, '[Bestellen');
        buttonOrder := FALSE;
        WinDials.WinDialDraw (groups, MausTauschrsc.grorder, 0, v.r, FALSE);
      ELSIF grIsOn & ~buttonOrder
      THEN
        (* Buttontext auf Abstellen setzen *)
        mtUtils.SetObjcString (groups, MausTauschrsc.grorder, 'A[bstellen');
        buttonOrder := TRUE;
        WinDials.WinDialDraw (groups, MausTauschrsc.grorder, 0, v.r, FALSE);
      END;
    END;
    (* Und jetzt alles zeichnen *)
    mtUtils.CalcArea (groups, MausTauschrsc.groupinf1, r1);
    mtUtils.CalcArea (groups, MausTauschrsc.groupinf2, r2);
    r1 := GrafBase.FrameRects (r1, r2);
    WinDials.WinDialDraw (groups, 0, 8, r1, TRUE);
  ELSE
    (* Textzeilen unten lschen *)
    MagicStrings.Assign ('', str);
    mtUtils.SetObjcString (groups, MausTauschrsc.groupinf1, str);
    mtUtils.SetObjcString (groups, MausTauschrsc.groupinf2, str);
    (* Buttons disablen *)
    mtUtils.SetState (groups, MausTauschrsc.grorder, MagicAES.DISABLED, TRUE);
    mtUtils.SetState (groups, MausTauschrsc.grfollow, MagicAES.DISABLED, TRUE);
    mtUtils.SetState (groups, MausTauschrsc.grchef, MagicAES.DISABLED, TRUE);
    (* Und jetzt alles zeichnen *)
    mtUtils.CalcArea (groups, MausTauschrsc.groupinf1, v.r);
    WinDials.WinDialDraw (groups, 0, 8, v.r, TRUE);
    mtUtils.CalcArea (groups, MausTauschrsc.groupinf2, v.r);
    WinDials.WinDialDraw (groups, 0, 8, v.r, TRUE);
    WinDials.WinDialDraw (groups, MausTauschrsc.grorder, 0, v.r, FALSE);
    WinDials.WinDialDraw (groups, MausTauschrsc.grfollow, 0, v.r, FALSE);
    WinDials.WinDialDraw (groups, MausTauschrsc.grchef, 0, v.r, FALSE);
  END;
END updateButton;

(* List-Zwischen-Prozeduren *)
PROCEDURE groupIsSelected (entry, env : ADDRESS) : BOOLEAN;
  VAR ent : GroupSelect.groupPtr;
BEGIN
  ent := entry;
  RETURN ent^.selected;
END groupIsSelected;

PROCEDURE isEnabled (adr : ADDRESS; env : ADDRESS) : BOOLEAN;
BEGIN
  RETURN TRUE
END isEnabled;

PROCEDURE countEntries ( l : ADDRESS; VAR ll: LONGINT; VAR ww: INTEGER);
 VAR lp : POINTER TO Lists.List;
BEGIN
 lp := l;
 ww := globalLength*mtAppl.CharWidth;
 ll := VAL (LONGINT, Lists.NoOfEntries (lp^));
END countEntries;

VAR  tr     : mtUtils.tObjcTree;

PROCEDURE drawGroupEntry (entry, env : ADDRESS; x, y : INTEGER;
                          offset : INTEGER; clip   : GrafBase.Rectangle);

  CONST spaceString = "          ";
     VAR e   : GroupSelect.groupPtr;
         str : ARRAY [0..255] OF CHAR;
  BEGIN
    e := entry;
    tr^[0].obX := x;
    tr^[0].obY := y;
    tr^[0].obSpec.TedPtr^.teTxtlen := globalLength;
    tr^[0].obWidth := globalLength*mtAppl.CharWidth;
    IF e # NIL
    THEN
      (* Jetzt hier den String fr den Gruppennamen zusammenbauen *)
      buildGroup (e, env,  str);
      mtUtils.SetObjcStringAdr (tr, 0, ADR(str));
      mtUtils.SetState (tr, 0, MagicAES.SELECTED, groupIsSelected (entry, env));
      mtUtils.SetState (tr, 0, MagicAES.DISABLED, ~isEnabled (entry, env));
    ELSE
      mtUtils.SetObjcStringAdr (tr, 0, CADR(spaceString));
      mtUtils.SetState (tr, 0, MagicAES.SELECTED, FALSE);
      mtUtils.SetState (tr, 0, MagicAES.DISABLED, FALSE);
    END;
    MagicAES.ObjcDraw (tr, 0, 8, clip);
  END drawGroupEntry;

PROCEDURE setAndGetGroupValues (tree: ADDRESS; private: ADDRESS; set: BOOLEAN; exitBut: INTEGER);
BEGIN
  IF set
  THEN
  ELSE
    IF exitBut # MausTauschrsc.groupabr
    THEN
      v.bool := GroupSelect.UpdateSelectList (grList, TRUE);
    END;
    (* Gruppenliste wieder freigeben *)
    GroupSelect.DeleteList (grList);
    (* Dialog wieder freigeben *)
    IF exitBut = MausTauschrsc.grorder
    THEN
      IF grinTools.BetreffHolen(-1, Wegen, 'Stichwort fr neue persnliche Nachricht an:',
                                Adressat, 'Wegen:', FALSE) 
      THEN
        Messages.NewPersMsg (Adressat, "", Wegen);
      END;
    END;
  END;
END setAndGetGroupValues;

PROCEDURE checkGroupExit (tree: ADDRESS; private: ADDRESS; button: INTEGER;
                          group: ADDRESS; kState: BITSET): BOOLEAN;
  VAR groupName : ARRAY [0..255] OF CHAR;
      gr        : GroupSelect.groupPtr;
      str1,
      str2      : CatTypes.String255;
      oneSel    : BOOLEAN;
      error     : BOOLEAN;
BEGIN
  error := FALSE;
  IF ~followUpSelect THEN 
    RETURN TRUE 
  ELSIF button = MausTauschrsc.groupok
  THEN
    (* Selektierten Eintrag suchen *)
    Lists.ResetList (grList);
    Lists.ScanEntries (grList, Lists.forward, findSel, NIL, oneSel);
    IF oneSel
    THEN
      (* Eintrag holen *)
      gr := Lists.CurrentEntry (grList);
      (* Jetzt Namen vergleichen *)
      MagicStrings.Assign (followUpEntry^.name^, str1);
      MagicStrings.Assign (gr^.name^, str2);
      MagicStrings.CAPS (str1);
      MagicStrings.CAPS (str2);
      IF MagicStrings.Equal (str1, str2)
      THEN
        MTE.info (MTE.sameGroupError);
        error := TRUE;
      ELSE
        (* Followup neu setzen *)
        GroupSelect.SetFollowUp (followUpEntry, gr^.name^);
        followUpSelect := FALSE;
      END;
    ELSE
      followUpSelect := FALSE;
    END;
  ELSE
    followUpSelect := FALSE;
  END;
  IF ~error
  THEN
    mtUtils.SetFlag (groups, MausTauschrsc.grchef, MagicAES.HIDETREE, FALSE);
    mtUtils.SetFlag (groups, MausTauschrsc.grorder, MagicAES.HIDETREE, FALSE);
    mtUtils.SetFlag (groups, MausTauschrsc.grfollow, MagicAES.HIDETREE, FALSE);
    mtUtils.SetObjcString (groups, MausTauschrsc.grtitle, cSelGroup);
    WinDials.WinUpdateTitle (groups);
    mtUtils.ExclState (groups, button, MagicAES.SELECTED);
    (* Full Redraw ber Dialog *)
    WinDials.WinDialDraw (groups, 0, 8, v.r, FALSE);
  END;
  RETURN FALSE;
END checkGroupExit;

PROCEDURE DoGroupSelect(dial : ADDRESS);
  VAR specials : ARRAY [0..7] OF ListDl.specialButHdler;
      gr,
      group    : GroupSelect.groupPtr;
      maxLength,
      l        : INTEGER;
      str1,
      str2     : CatTypes.String255;
      r        : GrafBase.Rectangle;
      oneSel   : BOOLEAN;
      oldFsel  : BOOLEAN;
      dialHandler: ListDl.ldHandler;
      sort     : INTEGER;
      sortTypeSel: mtUtils.tObjcTree;
BEGIN
  followUpSelect := FALSE;
  (* Baumadresse setzen *)
  groups := dial;
  (* Sysop feststellen *)
  isSysop := Infofiles.IsInfoFile ('ISB', TRUE);
  (* Titel setzen *)
  mtUtils.SetObjcString (groups, MausTauschrsc.grtitle, cSelGroup);
  (* Dial erstellen *)

  (* Sortierung einstellen *)
  ConfVars.GetConfDefInt (cSortSelectList, sort, 0);
  sortTypeSel := MausTauschrsc.TreeAddr^[MausTauschrsc.sorttype];
  VDIUtil.SetXState (groups, MausTauschrsc.grsort1, ORD(sort));
  VDIUtil.SetXString (groups, MausTauschrsc.grsort1, sortTypeSel, ORD(sort));

  ConfVars.GetConfDefBool (cLokalSelectList, oneSel, FALSE);
  IF ~GroupSelect.GetSelectList (grList, oneSel, GroupSelect.selSortType (sort))
  THEN
    MTE.noMemAlert();
    RETURN 
  END;
  mtUtils.SetState (groups, MausTauschrsc.grlokal, MagicAES.SELECTED, oneSel);
  (* Texte unten leeren *)
  MagicStrings.Assign ('', str1);
  mtUtils.SetObjcString (groups, MausTauschrsc.grname, str1);
  mtUtils.SetObjcString (groups, MausTauschrsc.groupinf1, str1);
  mtUtils.SetObjcString (groups, MausTauschrsc.groupinf2, str1);
  (* specials bestimmen *)
  specials[0].objc := MausTauschrsc.grorder;
  specials[0].proc := orderGroup;
  specials[1].objc := MausTauschrsc.grfollow;
  specials[1].proc := followUpTo;
  specials[2].objc := MausTauschrsc.grhelp;
  specials[2].proc := doHelp;
  specials[3].objc := MausTauschrsc.grchef;
  specials[3].proc := doChef;
  specials[4].objc := MausTauschrsc.grlokal;
  specials[4].proc := switchLokal;
  specials[5].objc := MausTauschrsc.grsort0;
  specials[5].proc := switchSort;
  specials[6].objc := MausTauschrsc.grsort1;
  specials[6].proc := switchSort;
  specials[7].objc := MausTauschrsc.grsort2;
  specials[7].proc := switchSort;

  (* Maximale Breite eines Gruppennamens herausfinden *)
  maxGroupNameLength := 0;
  Lists.ResetList (grList);
  group := Lists.NextEntry (grList);
  WHILE group # NIL DO 
    l := LENGTH (group^.name^);
    IF l > maxGroupNameLength 
    THEN
      maxGroupNameLength := l;
    END;
    group := Lists.NextEntry (grList);
  END;

  (* Maximale Breite eines Listeneintrages herausfinden *)
  maxLength := 0;
  Lists.ResetList (grList);
  group := Lists.NextEntry (grList);
  WHILE group # NIL DO 
    buildGroup (group, NIL, str1);
    l := LENGTH (str1);
    IF l > maxLength 
    THEN
      maxLength := l;
    END;
    group := Lists.NextEntry (grList);
  END;
  INC (maxLength);
  
  (* Jetzt das Object zusammenbauen *)
  mtUtils.CalcArea (groups, MausTauschrsc.groupbox, r);
  maxLength := BinOps.HigherInt (maxLength, r.w DIV mtAppl.CharWidth);
  
  globalLength := maxLength;

  (* Objectadresse holen *)
  tr := MausTauschrsc.TreeAddr^[MausTauschrsc.listtxt];

  (* WindowDialog vorbereiten *)
  ListDl.BuildLdHandler (ADR(grList), 
                         ListHelp.resetList, 
                         ListHelp.nextEntry, 
                         ListHelp.prevEntry,
                         countEntries, 
                         isEnabled, 
                         selGrEntry,
                         groupToStr, groupIsSelected, 
                         drawGroupEntry,
                         0, mtAppl.CharHeight, 
                         8, maxLength*mtAppl.CharWidth,
                         dialHandler);
  
  LOOP
  (* ListWinDialog ffnen *)
    oldFsel := followUpSelect;
    IF ListDl.WinListDial (groups, ListDl.ldElemSet{ldSelect, ldArrows, ldAutolocate, ldModal},
                         dialHandler,
                         MausTauschrsc.grordbox,
                         MausTauschrsc.grordback,
                         MausTauschrsc.groupok,
                         MausTauschrsc.groupabr,
                         NIL, 0,
                         MausTauschrsc.grname,
                         updateButton,
                         specials,
                         8,
                         checkGroupExit,
                         setAndGetGroupValues,
                         setAndGetGroupValues,
                         group)
    THEN

    END;
    IF followUpSelect = oldFsel
    THEN
      EXIT;
    END;
  END;
END DoGroupSelect;

END GruppenAuswahl.
