IMPLEMENTATION MODULE Infofiles;
(*$Z+*)

(* Beschreibung des Datenformates: 
 * Die Infofiles im Speicher werden als verkettete Liste 
 * von infEntrys dargestellt. Dabei wird fr einen Eintrag 
 * nur so viel Speicher alloziert, wie unbedingt notwendig.
 * Auf Platte werden die Files als Array [0..#Infofiles] 
 * von infEntrys abgespeichert, wobei diese natrlich den
 * vollen Platz belegen.
 *)

FROM SYSTEM  IMPORT ADDRESS, ADR, TSIZE, CALLSYS, CADR;

(* MM2-Module *)
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
IMPORT Lists, Block, GrafBase;

(* CAT-Module *)
FROM MTPaths IMPORT DataPath;
FROM Void    IMPORT v;
IMPORT CatFiles, MTE, dataSys, CatGlobal, ConvertDate, QuickSort, GroupSelect, 
       CatTypes, Protokoll;

(* Magic-Module *)

IMPORT MagicConvert, MagicStrings, MagicDOS, 
       MagicAES, MagicVDI, mtAlerts, mtUtils, mtPopups, 
       mtAppl, mtTextfiles;

CONST   CreateDefault = FALSE;

CONST   infoInfFile     = 'info.inf';

        (* Magics fr Header *)
        infCatMagic     = 43415449H;  (* "CATI"                    *)
        oldInfVersion   = 1;          (* alte Versionsnummer der Infofile.i *)
        oldInfVersionMagic = 0140H;   (* altes Versionsmagic       *)

        infVersion      = 2;          (* Versionsnummer der Infofile.i *)
        infVersionMagic = 0100H;      (* Versionsmagic             *)

        maxInf          = 1024;

TYPE 
        oldInfEntry = RECORD
                  crc   : CARDINAL;
                  new   : BOOLEAN;
                  avail : BOOLEAN;
                  isMaus: BOOLEAN;
                  ordered: BOOLEAN;
                  orderIntervall: INTEGER; (* -1: Einmalig, 0: Immer, 1: tglich *)
                  lastOrdered: LONGCARD; (* Datum in Tagen seit dem 1.1.1900 *)
                  selected: BOOLEAN;
                  hasCrc: tCrcState;
                  res   : CARDINAL;
                  mausShort : ARRAY [0..9] OF CHAR;
                  mausName  : ARRAY [0..35] OF CHAR;
                  filename  : ARRAY [0..255] OF CHAR;
                END;
     
        oldInfEntryPtr = POINTER TO oldInfEntry;
     
        oldInfArrayType = ARRAY [0..maxInf-1] OF oldInfEntry;
     
        oldInfArrayPtr = POINTER TO oldInfArrayType;
     
     infArrayType = ARRAY [0..maxInf-1] OF infEntry;
     
     infArrayPtr = POINTER TO infArrayType;

VAR infLoaded: BOOLEAN;
    
PROCEDURE InfoLoaded(): BOOLEAN;
BEGIN
  RETURN infLoaded;
END InfoLoaded;

PROCEDURE DeleteSimpleList (VAR l: Lists.List; killCarrier: BOOLEAN);

VAR   entry: ADDRESS;

BEGIN
  Lists.ResetList (l);
  entry := Lists.PrevEntry (l);
  WHILE entry # NIL DO
    Lists.RemoveEntry (l, v.bool);
    DEALLOCATE (entry, 0L);
    entry := Lists.CurrentEntry (l);
  END;
  IF killCarrier THEN Lists.DeleteList (l, v.bool) END;
END DeleteSimpleList;

PROCEDURE SortList (VAR l : Lists.List);                        (* exported *)
(* sortiert die Gruppenliste *)
  VAR count : CARDINAL;
      sort  : POINTER TO ARRAY [0..$FFFF] OF ADDRESS;
      i     : CARDINAL;
      adr   : ADDRESS;

  PROCEDURE grComp (a1, a2 : ADDRESS) : BOOLEAN;
    VAR b1, b2 : POINTER TO infEntryPtr;
        p1, p2 : infEntryPtr;
        str1, str2 : ARRAY [0..9] OF CHAR;
  BEGIN
    b1 := a1; b2 := a2;
    p1 := infEntryPtr (b1^);
    p2 := infEntryPtr (b2^);
    MagicStrings.Assign (p1^.mausShort, str1);
    MagicStrings.Assign (p2^.mausShort, str2);
    MagicStrings.CAPS (str1);
    MagicStrings.CAPS (str2);
    IF p1^.isMaus = p2^.isMaus
    THEN
      RETURN MagicStrings.Compare (str1, str2) = MagicStrings.less;
    ELSE
      RETURN p1^.isMaus
    END;
  END grComp;

BEGIN
  count := Lists.NoOfEntries (l);
  IF count = 0 THEN RETURN END;
  ALLOCATE (sort, LONG(count+1) * TSIZE (ADDRESS));
  IF sort = NIL THEN RETURN END;
  Lists.ResetList (l);
  FOR i := 0 TO count-1 DO
    sort^[i] := Lists.NextEntry (l);
  END;
  v.bool := QuickSort.sortIt (0, count-1, sort^, grComp, TSIZE (ADDRESS), QuickSort.noBreak);
  (* Liste wieder zurckbernehmen *)
  Lists.ResetList (l);
  adr := Lists.PrevEntry (l);
  WHILE adr # NIL DO
    Lists.RemoveEntry (l, v.bool);
    IF v.bool THEN HALT END;
    adr := Lists.CurrentEntry (l);
  END;
  (*
  FOR i := 0 TO count-1 DO
    adr := Lists.NextEntry (l);
    Lists.RemoveEntry (l, v.bool);
    IF v.bool THEN HALT END;
  END;
  *)
  Lists.ResetList (l);
  FOR i := 0 TO count - 1 DO 
    IF sort^[i] # NIL
    THEN
      Lists.AppendEntry (l, sort^[i], v.bool);
      IF v.bool THEN HALT END;
    END;
  END;
  DEALLOCATE (sort, 0);
END SortList;

PROCEDURE IsNeededInfo (REF str: ARRAY OF CHAR): BOOLEAN;
  VAR i: CARDINAL;
BEGIN
  (* Kompliziertere Abfrage, damit man das nicht so leicht 
   * rauspatchen kann!
   * Einfacher wre folgendes: 
   * IF MagicStrings.Equal (mausShort, "ITG")
   * OR MagicStrings.Equal (mausShort, "ITI")
   * OR MagicStrings.Equal (mausShort, "ITK")
   *)
  IF (INTEGER(ORD (str[0])) - INTEGER(ORD('I')) = 0)
   & (INTEGER(ORD (str[1])) - INTEGER(ORD('T')) = 0)
  THEN
    FOR i := 0 TO 4  BY 2 DO
      (* ITG, ITI und ITK werden geprft. 
       * G + 0 = G
       * G + 2 = I
       * G + 4 = K
       *)
      IF ORD(str[2]) = ORD('G') + i
      THEN
        RETURN TRUE;
      END;
    END;
  END;
  RETURN FALSE;
END IsNeededInfo;

PROCEDURE ConvertInfoV1toV2 (REF old: oldInfEntry; VAR new : infEntry);
BEGIN
  WITH new DO
    crc := old.crc;
    new := old.new;
    avail := old.avail;
    isMaus := old.isMaus;
    ordered := old.ordered;
    orderIntervall := old.orderIntervall;
    lastOrdered := old.lastOrdered;
    selected := old.selected;
    hasCrc := old.hasCrc;
    res    := old.res;
    MagicStrings.Assign (old.mausShort, mausShort);
    MagicStrings.Assign (old.mausName, mausName);
    MagicStrings.Assign (old.filename, filename);
  END;
END ConvertInfoV1toV2;

(*$H+*)
PROCEDURE LoadInfoInf(dbChanged, clearNew : BOOLEAN);
 VAR    handle  : INTEGER;
        bSize   : CARDINAL;
        size    : LONGCARD;
        count   : CARDINAL;
        entry   : infEntryPtr;
        infArray: infArrayPtr;
        oldInfArray: oldInfArrayPtr;
        i       : CARDINAL;
        l       : CARDINAL;
        dbHdr   : dataSys.FileHeaderType;
        isOldInf: BOOLEAN;

(*$Z-*)
  PROCEDURE clearNewProc (entry, info : ADDRESS): BOOLEAN;
    VAR e : infEntryPtr;
  BEGIN
    e := infEntryPtr (entry);
    e^.new := FALSE;
    RETURN FALSE;
  END clearNewProc;
(*$Z+*)

BEGIN
  handle := -1;
  handle := CatFiles.OpenFile(DataPath, infoInfFile, CatFiles.readFile);
  IF handle > 0 THEN
    (* Jetzt erst mal Header einlesen *)
    CatFiles.Seek (0, handle, CatFiles.start);
    CatFiles.ReadMuch (dataSys.dbHeaderLength, handle, ADR(dbHdr));
    IF CatFiles.FileError # CatFiles.NoError
    THEN
      CatFiles.ErrorAlert (CatFiles.FileError);
      CatFiles.CloseFile (handle);
      RETURN
    END;
    (* Jetzt berprfen des Headers *)
    IF (dbHdr.CatMagic # infCatMagic)
    OR (dbHdr.Version  # infVersion)
    OR (dbHdr.VersionMagic # infVersionMagic)
    THEN
      IF (dbHdr.CatMagic = infCatMagic)
       & (dbHdr.Version  = oldInfVersion)
       & (dbHdr.VersionMagic  = oldInfVersionMagic)
      THEN
        isOldInf := TRUE;
      ELSE
        MTE.info (MTE.oldVersionAlt);
        CatFiles.CloseFile (handle);
        RETURN
      END;
    ELSE
      isOldInf := FALSE;
    END;
    (* Gre feststellen *)
    CatFiles.Seek (0, handle, CatFiles.end);
    size := CatFiles.FilePos (handle);
    CatFiles.Seek (0, handle, CatFiles.start);
    IF isOldInf
    THEN
      count := SHORT((size-dataSys.dbHeaderLength) DIV TSIZE (oldInfEntry));
      IF count > maxInf THEN count := maxInf END;
      ALLOCATE (oldInfArray, size - dataSys.dbHeaderLength);
      IF oldInfArray = NIL THEN 
        CatFiles.CloseFile (handle);
        MTE.noMemAlert();
        RETURN 
      END;
      (* Neues Array allozieren *)
      ALLOCATE (infArray, LONG(count) * TSIZE (infEntry));
      IF infArray = NIL THEN 
        DEALLOCATE (oldInfArray, 0);
        CatFiles.CloseFile (handle);
        MTE.noMemAlert();
        RETURN 
      END;
      Block.Clear (infArray, LONG (count) * TSIZE (infEntry));

      (* Jetzt Datenblock einlesen *)
      CatFiles.Seek (TSIZE (dataSys.FileHeaderType), handle, CatFiles.start);
      CatFiles.ReadMuch (size - dataSys.dbHeaderLength, handle, oldInfArray);
      IF CatFiles.FileError # CatFiles.NoError
      THEN
        DEALLOCATE (infArray, 0);
        DEALLOCATE (oldInfArray, 0);
        CatFiles.ErrorAlert (CatFiles.FileError);
        CatFiles.CloseFile (handle);
        RETURN
      END;
      (* Jetzt Daten umkopieren *)
      FOR i := 0 TO count - 1 DO
        ConvertInfoV1toV2 (oldInfArray^[i], infArray^[i]);
      END;
      (* Und altes Array freigeben *)
      DEALLOCATE (oldInfArray, 0);
    ELSE
      count := SHORT((size-dataSys.dbHeaderLength) DIV TSIZE (infEntry));
      IF count > maxInf THEN count := maxInf END;
      ALLOCATE (infArray, size - dataSys.dbHeaderLength);
      IF infArray = NIL THEN 
        CatFiles.CloseFile (handle);
        MTE.noMemAlert();
        RETURN 
      END;
      (* Jetzt Datenblock einlesen *)
      CatFiles.Seek (TSIZE (dataSys.FileHeaderType), handle, CatFiles.start);
      CatFiles.ReadMuch (size - dataSys.dbHeaderLength, handle, infArray);
      IF CatFiles.FileError # CatFiles.NoError
      THEN
        CatFiles.ErrorAlert (CatFiles.FileError);
        CatFiles.CloseFile (handle);
        DEALLOCATE (infArray, 0);
        RETURN
      END;
    END;
    CatFiles.CloseFile (handle);
    (* clearNew beachten *)
    IF clearNew
    THEN
      FOR i := 0 TO count-1 DO
        infArray^[i].new := FALSE;
      END;
    END;
    (* Alte Liste lschen, falls nicht leer *)
    DeleteSimpleList (infList, FALSE);
    (* Jetzt umwandeln in Liste *)
    FOR i := 0 TO count-1 DO
      l := (*SYSTEM.*)LENGTH (infArray^[i].filename);
      bSize := TSIZE (infEntry) - 256 + l + 2;
      ALLOCATE (entry, bSize);
      IF entry = NIL
      THEN
        DEALLOCATE (infArray, 0);
        MTE.noMemAlert();
        RETURN 
      END;
      (* Anforderungsintervall fr ITK, ITI und ITG auf 'Immer' setzen,
       * da nicht mehr nderbar
       *)
      WITH infArray^[i] DO
        IF IsNeededInfo (mausShort)
        THEN
          ordered := TRUE;
          orderIntervall := 0;
        END;
      END;
      Block.Copy (ADR(infArray^[i]), bSize, entry);
      entry^.selected := FALSE;
      entry^.res      := 0;
      (*
      WITH infArray^[i] DO
        entry^.crc := crc;
        entry^.new := new;
        entry^.avail := avail;
        entry^.isMaus := isMaus;
        entry^.ordered := ordered;
        entry^.orderIntervall := orderIntervall;
        entry^.lastOrdered := lastOrdered;
        MagicStrings.Assign (mausShort, entry^.mausShort);
        MagicStrings.Assign (mausName, entry^.mausName);
        MagicStrings.Assign (filename, entry^.filename);
      END;
      *)
      Lists.AppendEntry (infList, entry, v.bool);
      IF v.bool
      THEN
        DEALLOCATE (infArray, 0);
        DEALLOCATE (entry, 0);
        MTE.noMemAlert (); 
        RETURN 
      END;
    END; (* FOR *)
    infLoaded := TRUE;
    DEALLOCATE (infArray, 0);
    (* Jetzt sortieren *)
    SortList (infList);
    (* Und wenn es eine alte Liste war, dann direkt mal sichern *)
    IF isOldInf 
    THEN
      SaveInfoInf();
    END;
  ELSE
    IF dbChanged
    THEN
      (* Erstmal Benutzer informieren *)
      MTE.info (MTE.noInfoInf);
      (* Liste komplett leeren *)
      DeleteSimpleList (infList, FALSE);
    END;
    (* Und jetzt Default-Eintrge anlegen. Diese sind:
     * CAT CATLOG.TXT
     * ERR PROTOKOL.TXT
     * LOG Letztes Logfile der Maus
     * ITI Technische Infofileliste
     * ITG Technische Gruppenliste
     * ITK Technische Kommandoliste
     *)
    v.bool := AddEntry (FALSE, FALSE, FALSE, FALSE, 0, 0, 
                        0, 'CAT', 'CAT Einfgeprotokoll', '$Dcatlog.TXT', FALSE, cNoCrc);
    v.bool := AddEntry (FALSE, FALSE, FALSE, FALSE, 0, 0, 
                        0, 'ERR', 'CAT Errorfile', '$Dprotokol.TXT', FALSE, cNoCrc);
    v.bool := AddEntry (FALSE, FALSE, FALSE, FALSE, 0, 0, 
                        0, 'LOG', 'Letztes Logfile', '$DLOG.TXT', FALSE, cNoCrc);
    v.bool := AddEntry (TRUE, TRUE, FALSE, FALSE, 0, 0, 
                        0, 'ITI', 'Maschinenlesbare Infoliste', 'ITI.TXT', TRUE, cWithCrc);
    v.bool := AddEntry (TRUE, TRUE, FALSE, FALSE, 0, 0, 
                        0, 'ITG', 'Maschinenlesbare Gruppenliste', 'ITG.TXT', TRUE, cWithCrc);
    v.bool := AddEntry (TRUE, TRUE, FALSE, FALSE, 0, 0, 
                        0, 'ITK', 'Maschinenlesbare Kommandoliste', 'ITK.TXT', TRUE, cWithCrc);
    (* Jetzt sortieren *)
    SortList (infList);
    IF clearNew
    THEN
      Lists.ResetList (infList);
      Lists.ScanEntries (infList, Lists.forward, 
                         clearNewProc, NIL, v.bool);
    END;
  END;
END LoadInfoInf;

PROCEDURE SaveInfoInf();
  VAR infArray  : infArrayPtr;

  (*$Z-*)
  PROCEDURE copyEntry (e, i : ADDRESS): BOOLEAN;
    VAR entry : infEntryPtr;
        ix    : POINTER TO CARDINAL;
        l, bSize : CARDINAL;
  BEGIN
    ix := i;
    entry := infEntryPtr (e);
    (*$? NOT CreateDefault: 
    l := (*SYSTEM.*)LENGTH (entry^.filename);
    bSize := TSIZE (infEntry) - 256 + l;
    Block.Copy (entry, bSize, ADR(infArray^[ix^]));
    *)
    (*$? CreateDefault:
    WITH infArray^[ix^] DO
      crc    := entry^.crc;
      new    := FALSE;
      avail  := FALSE;
      isMaus := entry^.isMaus;
      ordered := (MagicStrings.Equal (entry^.mausShort, 'ITG') OR;
                  MagicStrings.Equal (entry^.mausShort, 'ITI'));
      orderIntervall := 0;
      lastOrdered    := 0;
      selected       := FALSE;
      hasCrc         := cNA;
      res            := 0;
      MagicStrings.Assign (entry^.mausShort, mausShort);
      MagicStrings.Assign (entry^.mausName, mausName);
      MagicStrings.Assign (entry^.filename, filename);
    END;
    *)
    INC (ix^);
    RETURN FALSE;
  END copyEntry;
  (*$Z+*)
  
  VAR i          : CARDINAL;
      handle     : INTEGER;
      count      : CARDINAL;
      entry      : infEntryPtr;
      dbHdr      : dataSys.FileHeaderType;
BEGIN
  count := Lists.NoOfEntries (infList);
  ALLOCATE (infArray, LONG(count) * TSIZE (infEntry));
  IF infArray = NIL 
  THEN
    MTE.noMemAlert();
    RETURN  
  END;
  Block.Clear (infArray, LONG(count) * TSIZE (infEntry));
  Lists.ResetList (infList);
  i := 0;
  Lists.ScanEntries (infList, Lists.forward, copyEntry, ADR(i), v.bool);
  (* Nun Datei erzeugen *)
  (*$? CreateDefault:
  handle := CatFiles.CreateFile (DataPath, "INFO.NEW");
  *)
  (*$? NOT CreateDefault:
  handle := CatFiles.CreateFile (DataPath, infoInfFile);
  *)
  IF handle < 0 THEN 
    CatFiles.ErrorAlert (handle); 
    DEALLOCATE (infArray, 0);
  END;
  dbHdr.CatMagic := infCatMagic;
  dbHdr.Version  := infVersion;
  dbHdr.VersionMagic := infVersionMagic;
  CatFiles.WriteMuch (dataSys.dbHeaderLength, handle, ADR(dbHdr));
  IF CatFiles.FileError # CatFiles.NoError
  THEN
    CatFiles.ErrorAlert (CatFiles.FileError);
    DEALLOCATE (infArray, 0);
    CatFiles.CloseFile (handle);
    RETURN
  END;
  CatFiles.WriteMuch (LONG(count) * TSIZE (infEntry), handle, infArray);
  IF CatFiles.FileError # CatFiles.NoError
  THEN
    CatFiles.ErrorAlert (CatFiles.FileError);
  END;
  DEALLOCATE (infArray, 0);
  CatFiles.CloseFile (handle);
  Protokoll.SendPathUpdate (DataPath);
END SaveInfoInf;

PROCEDURE ParseFlags (entry: infEntryPtr; REF flags: ARRAY OF CHAR);
  VAR i, l : INTEGER;
BEGIN
  l := LENGTH (flags);
  i := 0;
  entry^.hasCrc := cNa;
  WHILE i < l DO
    CASE flags [i] OF 
      'C'  : CASE flags[i+1] OF
                '+' : entry^.hasCrc  := cWithCrc; |
                '-' : entry^.hasCrc  := cNoCrc; |
             ELSE
               entry^.hasCrc := cNa;
             END; |
    ELSE
    END;
    INC (i, 2);
  END;
END ParseFlags;

(*$Z-*)
PROCEDURE searchName (e, i : ADDRESS) : BOOLEAN;
  VAR entry : infEntryPtr;
      strPtr: POINTER TO ARRAY [0..255] OF CHAR;
BEGIN
  entry := e;
  strPtr := i;
  IF MagicStrings.Equal (entry^.mausShort, strPtr^)
  THEN
    RETURN TRUE
  END;
  RETURN FALSE
END searchName;
(*$Z+*)

PROCEDURE CheckInfos();
(* Formatbeschreibung: 

Ansonsten mach' ich hier gleich mal einen Vorschlag fr das Info-Info, gleich
mal via an die aktuellen Soen angepatem Beispiel:

#ITI
:#IIE
::Einfhrung in die MAUS
:#IIB
::Bedienungsanleitung
:#IIA
::Men-Kurzanleitung
:#IIG
::Gruppen-Anleitung
:#III
::Technische Informationen
:#IIH
::Hardware der MAUS
:#IIM
::MAUS-Beitrag
:#IIL
::LogIn-Zeiten
:#IIT
::MausTausch-Anleitung
:#IIK
::Kommerzielles
:#INA
::MausNet-Anleitung
:#INK
::Kurze Boxen-Liste
:#INL
::Lange Boxen-Liste
:#ING
::Netzgruppen-Liste
:#INP
::Netzplan
:#IGT
::Boxtitel
:#IGE
::Boxabspann
:#IGS
::Spruch des Tages
:#IGK
::Kurze Gruppenliste
:#IGL
::Lange Gruppenliste
:#ISG                                 | erscheint nur,
::SysOp-Netzgruppenliste              | wenn User
:#ISB                                 | berechtigt
::SysOp-Boxenliste                    | zum Abruf
:#ITG
::Maschinenlesbare Gruppenliste
:#ITI
::Maschinenlesbare Infoliste

Geht das so? Wenn ja, bau' ich das gleich ein ...

MfG Kai
*)
  VAR idScrap,
      textScrap,
      flagScrap,
      infoScrap : ARRAY [0..255] OF CHAR;
      name      : ARRAY [0..39] OF CHAR;
      type      : CHAR;
      tf        : mtTextfiles.TEXTFILE;
      found     : BOOLEAN;
      dummy     : infEntry;
      entry     : infEntryPtr;
      infoSet   : BOOLEAN;

BEGIN
  MagicStrings.Assign (DataPath, idScrap);
  MagicStrings.Append ('ITI.TXT', idScrap);
  MagicStrings.Assign ('', infoScrap);
  MagicStrings.Assign ('', flagScrap);
  IF mtTextfiles.OpenTextfile (idScrap, mtTextfiles.READ, 8192, tf)
  THEN
    infoSet := FALSE;
    (* Absolut trivial zu parsen *)
    REPEAT
      mtTextfiles.ReadChar(tf, type);
      mtTextfiles.ReadLine(tf, idScrap);
      mtTextfiles.ReadLn(tf);
      CASE type OF
        '#' : IF infoSet 
              THEN
                (* Eintrag suchen und Infos sichern *)
                (* Beschreibung zu dem Infofile, also einsetzen *)
                Lists.ResetList (infList);
                Lists.ScanEntries (infList, Lists.forward, searchName, ADR(infoScrap), found);
                IF found 
                THEN
                  entry := Lists.CurrentEntry (infList);
                  ParseFlags (entry, flagScrap);
                  IF textScrap[0] # ''
                  THEN
                    MagicStrings.Assign (textScrap, entry^.mausName);
                  END;
                  entry^.res := 1;
                ELSE
                  (* Nicht vorhanden, also neues Infofile anlegen 
                   * und anfordern 
                   *)
                  (* Filenamen zusammensetzen *)
                  MagicStrings.Assign (infoScrap, name);
                  MagicStrings.Append (CatTypes.textExt, name);
                  (* Und zu Liste hinzufgen *)
                  ParseFlags (ADR(dummy), flagScrap);
                  (* Sichern ist nicht ntig, wird am Ende eh gemacht *)
                  v.bool := AddEntry (TRUE, FALSE, FALSE, FALSE, 0, -1,
                                      0, infoScrap, textScrap, name, FALSE, dummy.hasCrc);
                END;
                infoSet := FALSE;
                MagicStrings.Assign ('', infoScrap);
                MagicStrings.Assign ('', flagScrap);
              END;
              IF idScrap[0] # ''
              THEN 
                MagicStrings.Assign (idScrap, infoScrap); 
                infoSet := TRUE;
              END;|
      | ':' : MagicStrings.Assign (idScrap, textScrap);
      | 'F' : MagicStrings.Assign (idScrap, flagScrap); 
      | ELSE
      END;
    UNTIL mtTextfiles.EndofText(tf);
    mtTextfiles.CloseTextfile (tf);
    IF infoSet 
    THEN
      (* Eintrag suchen und Infos sichern *)
      (* Beschreibung zu dem Infofile, also einsetzen *)
      Lists.ResetList (infList);
      Lists.ScanEntries (infList, Lists.forward, searchName, ADR(infoScrap), found);
      IF found 
      THEN
        entry := Lists.CurrentEntry (infList);
        ParseFlags (entry, flagScrap);
        IF textScrap[0] # ''
        THEN
          MagicStrings.Assign (textScrap, entry^.mausName);
        END;
        entry^.res := 1;
      ELSE
        (* Nicht vorhanden, also neues Infofile anlegen 
         * und anfordern 
         *)
        (* Filenamen zusammensetzen *)
        MagicStrings.Assign (infoScrap, name);
        MagicStrings.Append (CatTypes.textExt, name);
        (* Und zu Liste hinzufgen *)
        ParseFlags (ADR(dummy), flagScrap);
        (* Sichern ist nicht ntig, wird am Ende eh gemacht *)
        v.bool := AddEntry (TRUE, FALSE, FALSE, FALSE, 0, -1,
                            0, infoScrap, textScrap, name, FALSE, dummy.hasCrc);
      END;
    END;
    (* Jetzt alle Maus-Infofiles lschen, die nicht gerade kamen *)
    Lists.ResetList (infList);
    REPEAT
      entry := Lists.NextEntry (infList);
      IF (entry # NIL) & (entry^.isMaus) & (entry^.res = 0)
      THEN
        Lists.RemoveEntry (infList, v.bool);
        DEALLOCATE (entry, 0);
        Lists.ResetList (infList);
        entry := Lists.NextEntry (infList);
      END;
    UNTIL entry = NIL;
    (* Jetzt sortieren *)
    SortList (infList);
  END;
END CheckInfos;

PROCEDURE SetInfoCrc (REF inf: ARRAY OF CHAR; VAR str : ARRAY OF CHAR);

  VAR today : LONGCARD; (* Tagesdatum in Tagen seit 1.1.1900 *)
      isNew : BOOLEAN;

  (*$Z-*)
  PROCEDURE setCrc (e, i : ADDRESS) : BOOLEAN;
  (*$Z+*)
    VAR crc : CARDINAL;
        entry : infEntryPtr;
  BEGIN
    entry := infEntryPtr (e);
    IF entry^.isMaus & MagicStrings.Equal (entry^.mausShort, inf)
    THEN
      crc := MagicConvert.StrToCard(str);
      entry^.new := ~entry^.avail OR (crc # entry^.crc) OR (inf[0]='J');
      isNew := entry^.new;
      entry^.avail := TRUE;
      entry^.crc := crc;
      entry^.lastOrdered := today;
      IF entry^.orderIntervall = -1
      THEN
        entry^.orderIntervall := 0;
        entry^.ordered := FALSE;
      END;
      RETURN TRUE
    END;
    RETURN FALSE
  END setCrc;

BEGIN
  today := ConvertDate.DateInNum (MagicDOS.Tgetdate ());
  isNew := FALSE;
  Lists.ResetList (infList);
  Lists.ScanEntries (infList, Lists.forward, setCrc, NIL, v.bool);
  IF MagicStrings.Equal (inf, 'ITI') & isNew
  THEN
    (* Maschinenlesbare Infofileliste ist neu! 
     * Also Abgleich zwischen Infoliste und internen Daten 
     *)
    CheckInfos ();
  END;
  (*
  IF MagicStrings.Equal (inf, 'ITG') & isNew
  THEN
    (* Maschinenlesbare Gruppenliste abgleichen mit
     * interner Gruppenliste
     *)
    v.bool := GroupSelect.ParseInfofile();
  END; 
  *)
END SetInfoCrc;

PROCEDURE GetInfoCrc (VAR str : ARRAY OF CHAR; mode : infMode) : BOOLEAN;
(* Liefert einen Infostring zurck. Bei Mode = IFIRST wird das erste Infofile
 * zurckgeliefert, bei Mode = INEXT wird der nchste zurckgeliefert.
 * Wenn alle Infostrings geliefert wurden, dann wird FALSE zurckgegeben 
 * (kann auch bei IFIRST passieren)
 * Zwischen den Aufrufen sollte nicht anders auf die Liste
 * zugegriffen werden, da ansonsten evtl. nicht alle angefordert
 * werden oder einige doppelt angefordert werden.
 *)

  VAR today : LONGCARD; (* Tagesdatum in Tagen seit 1.1.1900 *)
      stunde: CARDINAL;
      ti    : ConvertDate.Time;

  (*$Z-*)
  PROCEDURE getCrc (e, i : ADDRESS) : BOOLEAN;
  (*$Z+*)
    VAR crc : CARDINAL;
        entry : infEntryPtr;
        numstr: ARRAY [0..9] OF CHAR;
  BEGIN
    entry := infEntryPtr (e);
    IF entry^.isMaus & entry^.ordered 
    THEN
      IF entry^.avail 
      THEN
        IF (entry^.orderIntervall <= 0) 
        OR ((entry^.lastOrdered + LONGCARD(LONG(entry^.orderIntervall)) = today)
            & (stunde >= 3) (* Neues Infofile erst nach dem Nachtnetz anfordern*) )
        OR (entry^.lastOrdered + LONGCARD(LONG(entry^.orderIntervall)) < today) (* auer es ist schon noch etwas lter *)
        OR (entry^.lastOrdered > today)
        THEN
          MagicConvert.CardToStr(entry^.crc, 0, numstr);
        ELSE
          RETURN FALSE
        END;
      ELSE
        MagicConvert.IntToStr (-1, 0, numstr);
      END;
      MagicStrings.Assign (entry^.mausShort, str);
      IF (entry^.hasCrc = cWithCrc) OR 
         ((entry^.hasCrc = cNa) & (str[0] # 'J'))
      THEN
        MagicStrings.Append (" ", str);
        MagicStrings.Append (numstr, str);
      END;
      RETURN TRUE
    END;
    RETURN FALSE
  END getCrc;

BEGIN
  MagicStrings.Assign ("", str);
  today := ConvertDate.DateInNum (MagicDOS.Tgetdate ());
  ti := ConvertDate.UnpackTime(MagicDOS.Tgettime());
  stunde := ti.hour;
  IF mode = IFIRST
  THEN
    Lists.ResetList (infList);
  END;
  Lists.ScanEntries (infList, Lists.forward, getCrc, NIL, v.bool);
  RETURN v.bool;
END GetInfoCrc;

PROCEDURE AddEntry (isMausInf, order, avl, isNew : BOOLEAN; theCrc: CARDINAL; inter : INTEGER;
                    lastOrder : LONGCARD; REF short, name, file : ARRAY OF CHAR;
                    saveInf : BOOLEAN; withCrc: tCrcState): BOOLEAN;
  VAR l, bSize  : CARDINAL;
      entry     : infEntryPtr;

BEGIN
  IF isMausInf THEN
    (* Immer I??.TXT als Filename! *)
    l := 15
  ELSE
    l := (*SYSTEM.*)LENGTH (file);
  END;
  bSize := TSIZE (infEntry) - 256 + l + 2;
  ALLOCATE (entry, bSize);
  IF entry = NIL
  THEN
    MTE.noMemAlert ();
    RETURN FALSE
  END;
  Block.Clear (entry, bSize);
  WITH entry^ DO
    crc := theCrc;
    avail := avl;
    new   := isNew;
    isMaus := isMausInf;
    ordered := order;
    orderIntervall := inter;
    lastOrdered := lastOrder;
    hasCrc := withCrc;
    res := 1;
    MagicStrings.Assign (short, mausShort);
    MagicStrings.Assign (name, mausName);
    IF ~isMausInf
    THEN
      MagicStrings.Assign (file, filename);
    ELSE
      MagicStrings.Assign (short, filename);
      MagicStrings.Append (CatTypes.textExt, filename);
    END;
  END;
  Lists.AppendEntry (infList, entry, v.bool);
  IF v.bool
  THEN
    DEALLOCATE (entry, 0);
    MTE.noMemAlert ();
    RETURN FALSE
  END;
  IF saveInf THEN SaveInfoInf() END;
  RETURN TRUE
END AddEntry;

PROCEDURE buildName (info: infEntryPtr; VAR fname : ARRAY OF CHAR): BOOLEAN;
  VAR filename : ARRAY [0..255] OF CHAR;
      exists   : BOOLEAN;
BEGIN
  IF info^.isMaus
  THEN
    MagicStrings.Assign (DataPath, fname);
    MagicStrings.Assign (info^.mausShort, filename);
    MagicStrings.Append (CatTypes.textExt, filename);
  ELSE
    MagicStrings.Assign ("", fname);
    MagicStrings.Assign (info^.filename, filename);
    CatGlobal.ReplaceStr (filename, '$D', DataPath);
  END;
  MagicStrings.Append (filename, fname);
  (* Jetzt Datei auf Existenz prfen und auf Lnge # 0 *)
  IF CatFiles.FileSize (fname, exists) = 0 THEN 
    (* Erstmal Listeintrag updaten, Infofile nicht da! *)
    IF ~exists
    THEN
      info^.avail := FALSE;
      info^.new := FALSE;
      SaveInfoInf();
    END;
    RETURN FALSE
  END;
  RETURN TRUE
END buildName;

PROCEDURE BuildName (info: ADDRESS; VAR fname : ARRAY OF CHAR): BOOLEAN;
BEGIN
  RETURN buildName (info, fname);
END BuildName;

PROCEDURE GetInfoFilename (REF short: ARRAY OF CHAR; 
                           VAR fname : ARRAY OF CHAR): BOOLEAN;
(* Liefert zum Krzel des Infofiles den Filenamen. 
 * FALSE: Infofile nicht vorhanden 
 *)

VAR found : BOOLEAN;
    info  : infEntryPtr;

BEGIN
  (*$W-*)
  Lists.ResetList (infList);
  Lists.ScanEntries (infList, Lists.forward, searchName, ADR(short), found);
  (*$W=*)
  IF found
  THEN
    info := Lists.CurrentEntry (infList);
    IF (info # NIL) & ~buildName (info, fname)
    THEN
      RETURN FALSE;
    ELSIF (info=NIL)
    THEN
      RETURN FALSE;
    END;
  END;
  RETURN found;
END GetInfoFilename;

PROCEDURE IsInfoFile (REF short: ARRAY OF CHAR; onlyMaus: BOOLEAN): BOOLEAN;
(* Liefert zurck, ob ein Krzel ein Infofile ist oder nicht 
 *)
VAR found : BOOLEAN;
    info  : infEntryPtr;
BEGIN
  (*$W-*)
  Lists.ResetList (infList);
  Lists.ScanEntries (infList, Lists.forward, searchName, ADR(short), found);
  (*$W=*)
  IF found & onlyMaus
  THEN
    info := Lists.CurrentEntry (infList);
    IF info # NIL
    THEN
      found := info^.isMaus
    END;
  END;
  RETURN found
END IsInfoFile;

PROCEDURE TestDirectParse (REF short: ARRAY OF CHAR);
(* Wird vom Parser aufgerufen, wenn ein Infofile geschlossen wurde. Falls es 
 * wichtige Informationen enthlt (ITG), wird es dann direkt geparst
 *)
BEGIN
  IF MagicStrings.Equal (short, 'ITG')
  THEN
    (* Maschinenlesbare Gruppenliste abgleichen mit
     * interner Gruppenliste
     *)
    v.bool := GroupSelect.ParseInfofile();
  ELSIF MagicStrings.Equal (short, 'REN')
  THEN
    (* Gruppennderungen parsen *)
    GroupSelect.ParseRenamefile();
  END;
END TestDirectParse;


(*$H=*)
BEGIN
  Lists.CreateList (infList, v.bool);
END Infofiles.
