IMPLEMENTATION MODULE IoFilter;

(* MM2-Module *)
IMPORT StrConv;

(* Module der MagicLib *)
IMPORT MagicStrings;

(* CAT-Module *)
IMPORT ARCStarter;
IMPORT CatGlobal;
IMPORT CatTypes;
IMPORT ConfVars;
IMPORT Varnames;
IMPORT CatLog;
IMPORT MTE;

FROM MTPaths    IMPORT MessagePath, ImportPath, ExportPath, DataPath, 
                       archiveName;
FROM UserInformation IMPORT batchProcess;

PROCEDURE buildVarName (REF name: ARRAY OF CHAR; number: INTEGER; VAR newName: ARRAY OF CHAR);
BEGIN
  MagicStrings.Assign (name, newName);
  MagicStrings.Append (StrConv.IntToStr (number, 0), newName);
END buildVarName;

PROCEDURE buildCmdLine (REF inCmd: ARRAY OF CHAR; VAR cmdLine: ARRAY OF CHAR);
  VAR mausId: CatTypes.String127;
BEGIN
  ConfVars.GetConfDefStr (cOwnName, mausId, "");
  MagicStrings.Assign (inCmd, cmdLine);
  CatGlobal.ReplaceStr (cmdLine, '$I', ImportPath);
  CatGlobal.ReplaceStr (cmdLine, '$E', ExportPath);
  CatGlobal.ReplaceStr (cmdLine, '$M', MessagePath);
  CatGlobal.ReplaceStr (cmdLine, '$D', DataPath);
  CatGlobal.ReplaceStr (cmdLine, '$F', 'infile.txt');
  CatGlobal.ReplaceStr (cmdLine, '$O', 'outfile.txt');
  CatGlobal.ReplaceStr (cmdLine, '$K', mausId);
  MagicStrings.Insert (' ', cmdLine, 0);
  cmdLine[0] := CHR(LENGTH (cmdLine));
END buildCmdLine;

PROCEDURE ReadFilter (VAR filter: ioFilterStruct; num: INTEGER);
 VAR  varname   : CatTypes.String255;
BEGIN
  WITH filter DO
    (* Filter ist vorhanden, auslesen *)
    buildVarName (cIoFilterInCmd, num, varname);
    ConfVars.GetConfDefStr (varname, inCmd, "");
    buildVarName (cIoFilterOutCmd, num, varname);
    ConfVars.GetConfDefStr (varname, outCmd, ""+0c);
    buildVarName (cIoFilterMsg, num, varname);
    ConfVars.GetConfDefBool (varname, forMsg, FALSE);
    buildVarName (cIoFilterPast, num, varname);
    ConfVars.GetConfDefBool (varname, pastOutfile, FALSE);
    buildVarName (cIoFilterActiv, num, varname);
    ConfVars.GetConfDefBool (varname, activ, TRUE);
    buildVarName (cIoFilterStopAtErr, num, varname);
    ConfVars.GetConfDefBool (varname, stopAtErr, FALSE);
    buildVarName (cIoFilterInfile, num, varname);
    ConfVars.GetConfDefBool (varname, forInfile, FALSE);
    buildVarName (cIoFilterOutfile, num, varname);
    ConfVars.GetConfDefBool (varname, forOutfile, FALSE);
    buildVarName (cIoFilterName, num, varname);
    ConfVars.GetConfDefStr (varname, filtName, "");
  END;
END ReadFilter;

PROCEDURE IoFilterMessage(VAR isError: BOOLEAN): BOOLEAN;
(* Ruft die definierten Filter auf, die vor dem erstellen von 
 * Infile.txt benutzt werden sollen
 *)
  VAR filter: ioFilterStruct;
      res,
      i     : INTEGER;
      cmdLine,
      varname   : CatTypes.String255;
      done      : BOOLEAN;
BEGIN
  isError := FALSE;
  done := FALSE;
  FOR i := 1 TO 10 DO
    (* alle Filter durchgehen
     *)
    WITH filter DO
      buildVarName (cIoFilterProg, i, varname);
      IF ConfVars.GetConfigString (varname, progName)
      THEN
        ReadFilter (filter, i);
        (*
        (* Filter ist vorhanden, auslesen *)
        buildVarName (cIoFilterInCmd, i, varname);
        ConfVars.GetConfDefStr (varname, inCmd, "");
        buildVarName (cIoFilterMsg, i, varname);
        ConfVars.GetConfDefBool (varname, forMsg, FALSE);
        buildVarName (cIoFilterActiv, i, varname);
        ConfVars.GetConfDefBool (varname, activ, TRUE);
        buildVarName (cIoFilterStopAtErr, i, varname);
        ConfVars.GetConfDefBool (varname, stopAtErr, FALSE);
        *)
        (* Jetzt nachsehen, ob wir den Filter auch benutzen 
         * mssen 
         *)
        IF forMsg & activ
        THEN
          (* Kommandozeile zusammenbauen *)
          buildCmdLine (inCmd, cmdLine);
          (* Filter aufrufen *)
          ARCStarter.ExecProgram (progName, cmdLine, FALSE, FALSE, TRUE, res);
          done := TRUE;
          IF (res # 0) & stopAtErr
          THEN
            isError := TRUE;
            buildVarName (cIoFilterName, i, varname);
            IF ~ConfVars.GetConfigString (varname, cmdLine)
            THEN
              MagicStrings.Assign (progName, cmdLine);
            END;
            MagicStrings.Append (", Returncode = ", cmdLine);
            MagicStrings.Append (StrConv.IntToStr (res, 0), cmdLine);
            MagicStrings.Append (".", cmdLine);
            MTE.warnAlert (~batchProcess, "[3][CAT:|Fehler bei IO-Filter fr Messages:|", cmdLine,"][:[Abbruch]");
            RETURN FALSE;
          END;
        END;
      END;
    END;
  END;
  RETURN done;
END IoFilterMessage;

PROCEDURE IoFilterInfile(): BOOLEAN;
(* Ruft die definierten Filter auf, die auf dem Infile arbeiten
 *)
  VAR filter: ioFilterStruct;
      res,
      i     : INTEGER;
      cmdLine,
      varname   : CatTypes.String255;
BEGIN
  FOR i := 1 TO 10 DO
    (* alle Filter durchgehen
     *)
    WITH filter DO
      buildVarName (cIoFilterProg, i, varname);
      IF ConfVars.GetConfigString (varname, progName)
      THEN
        ReadFilter (filter, i);

        (* Filter ist vorhanden, auslesen *)
        buildVarName (cIoFilterInCmd, i, varname);
        ConfVars.GetConfDefStr (varname, inCmd, "");
        buildVarName (cIoFilterMsg, i, varname);
        ConfVars.GetConfDefBool (varname, forMsg, FALSE);
        buildVarName (cIoFilterPast, i, varname);
        ConfVars.GetConfDefBool (varname, pastOutfile, FALSE);
        buildVarName (cIoFilterActiv, i, varname);
        ConfVars.GetConfDefBool (varname, activ, TRUE);
        buildVarName (cIoFilterStopAtErr, i, varname);
        ConfVars.GetConfDefBool (varname, stopAtErr, FALSE);
        (* Jetzt nachsehen, ob wir den Filter auch benutzen 
         * mssen 
         *)
        IF ~forMsg & ~pastOutfile
        THEN
          buildVarName (cIoFilterInfile, i, varname);
          ConfVars.GetConfDefBool (varname, forInfile, FALSE);
          IF forInfile & activ
          THEN
            (* Kommandozeile zusammenbauen *)
            buildCmdLine (inCmd, cmdLine);
            (* Filter aufrufen *)
            ARCStarter.ExecProgram (progName, cmdLine, FALSE, FALSE, TRUE, res);
            IF (res # 0) & stopAtErr
            THEN
              buildVarName (cIoFilterName, i, varname);
              IF ~ConfVars.GetConfigString (varname, cmdLine)
              THEN
                MagicStrings.Assign (progName, cmdLine);
              END;
              MagicStrings.Append (", Returncode = ", cmdLine);
              MagicStrings.Append (StrConv.IntToStr (res, 0), cmdLine);
              MagicStrings.Append (".", cmdLine);
              MTE.warnAlert (~batchProcess, "[3][CAT:|Fehler bei IO-Filter fr Infile:|", cmdLine,"][:[Abbruch]");
              RETURN FALSE;
            END;
          END;
        END;
      END;
    END;
  END;
  RETURN TRUE;
END IoFilterInfile;


PROCEDURE IoFilterOutfile(): BOOLEAN;
(* Ruft die definierten Filter auf, die auf dem Outfile arbeiten
 *)
  VAR filter: ioFilterStruct;
      res,
      i     : INTEGER;
      bool  : BOOLEAN;
      cmdLine,
      varname   : CatTypes.String255;
BEGIN
  FOR i := 1 TO 10 DO
    (* alle Filter durchgehen
     *)
    WITH filter DO
      buildVarName (cIoFilterProg, i, varname);
      IF ConfVars.GetConfigString (varname, progName)
      THEN
        ReadFilter (filter, i);
        (*
        (* Filter ist vorhanden, auslesen *)
        buildVarName (cIoFilterOutCmd, i, varname);
        ConfVars.GetConfDefStr (varname, outCmd, ""+0c);
        buildVarName (cIoFilterMsg, i, varname);
        ConfVars.GetConfDefBool (varname, forMsg, FALSE);
        buildVarName (cIoFilterPast, i, varname);
        ConfVars.GetConfDefBool (varname, pastOutfile, FALSE);
        buildVarName (cIoFilterActiv, i, varname);
        ConfVars.GetConfDefBool (varname, activ, TRUE);
        *)
        (* Jetzt nachsehen, ob wir den Filter auch benutzen 
         * mssen 
         *)
        IF ~forMsg & ~pastOutfile
        THEN
          (*
          buildVarName (cIoFilterOutfile, i, varname);
          ConfVars.GetConfDefBool (varname, forOutfile, FALSE);
          *)
          IF forOutfile & activ
          THEN
            (* Kommandozeile zusammenbauen *)
            buildCmdLine (outCmd, cmdLine);
            (* Filter aufrufen *)
            ConfVars.GetConfDefBool (cRedirectToLog, bool, FALSE);
            CatLog.WriteStringNTime ("Rufe IO-Filter auf: ");
            CatLog.WriteLine (progName);
            CatLog.WriteStringNTime ("Kommandozeile:      ");
            MagicStrings.Assign (cmdLine, varname);
            MagicStrings.Delete (varname, 0, 1);
            CatLog.WriteLine (varname);
            ARCStarter.ExecProgram (progName, cmdLine, bool, FALSE, TRUE, res);
            IF res # 0
            THEN
              CatLog.WriteString ("Returncode ");
              CatLog.WriteInt (res);
              CatLog.WriteString (" bei Aufruf von ");
              CatLog.WriteString (progName);
              CatLog.WriteLn;
              IF stopAtErr THEN 
                buildVarName (cIoFilterName, i, varname);
                IF ~ConfVars.GetConfigString (varname, cmdLine)
                THEN
                  MagicStrings.Assign (progName, cmdLine);
                END;
                MagicStrings.Append (", Returncode = ", cmdLine);
                MagicStrings.Append (StrConv.IntToStr (res, 0), cmdLine);
                MagicStrings.Append (".", cmdLine);
                MTE.warnAlert (~batchProcess, "[3][CAT:|Fehler bei IO-Filter fr Outfile:|", cmdLine,"][:[Abbruch]");
                RETURN FALSE;
              END;
            END;
            CatLog.WriteLineNTime ("IO-Filter beendet");
            CatLog.WriteLn;
          END;
        END;
      END;
    END;
  END;
  RETURN TRUE;
END IoFilterOutfile;

PROCEDURE IoFilterPastOutfile();
(* Ruft die definierten Filter auf, die vor dem erstellen von 
 * Infile.txt benutzt werden sollen
 *)
  VAR filter: ioFilterStruct;
      res,
      i     : INTEGER;
      bool  : BOOLEAN;
      cmdLine,
      varname   : CatTypes.String255;
BEGIN
  FOR i := 1 TO 10 DO
    (* alle Filter durchgehen
     *)
    WITH filter DO
      buildVarName (cIoFilterProg, i, varname);
      IF ConfVars.GetConfigString (varname, progName)
      THEN
        ReadFilter (filter, i);
        (*
        (* Filter ist vorhanden, auslesen *)
        buildVarName (cIoFilterInCmd, i, varname);
        ConfVars.GetConfDefStr (varname, inCmd, "");
        buildVarName (cIoFilterPast, i, varname);
        ConfVars.GetConfDefBool (varname, pastOutfile, FALSE);
        buildVarName (cIoFilterActiv, i, varname);
        ConfVars.GetConfDefBool (varname, activ, TRUE);
        *)
        (* Jetzt nachsehen, ob wir den Filter auch benutzen 
         * mssen 
         *)
        IF pastOutfile & activ
        THEN
          (* Kommandozeile zusammenbauen *)
          buildCmdLine (inCmd, cmdLine);
          (* Filter aufrufen *)
          ConfVars.GetConfDefBool (cRedirectToLog, bool, FALSE);
          CatLog.WriteStringNTime ("Rufe IO-Filter auf: ");
          CatLog.WriteLine (progName);
          CatLog.WriteStringNTime ("Kommandozeile:      ");
          MagicStrings.Assign (cmdLine, varname);
          MagicStrings.Delete (varname, 0, 1);
          CatLog.WriteLine (varname);
          ARCStarter.ExecProgram (progName, cmdLine, bool, FALSE, TRUE, res);
          IF res # 0
          THEN
            CatLog.WriteString ("Returncode ");
            CatLog.WriteInt (res);
            CatLog.WriteString (" bei Aufruf von ");
            CatLog.WriteString (progName);
            CatLog.WriteLn;
          END;
          CatLog.WriteLineNTime ("IO-Filter beendet");
          CatLog.WriteLn;
        END;
      END;
    END;
  END;
END IoFilterPastOutfile;

END IoFilter.
