IMPLEMENTATION MODULE ListDl;

FROM SYSTEM     IMPORT ADDRESS, CAST, LONGWORD, CALLSYS, CADR, ADR;

FROM Storage    IMPORT ALLOCATE, DEALLOCATE;

FROM Strings    IMPORT Space, Append, Pos, Upper, Assign, DelTrailingBlanks,
                       DelLeadingBlanks, StrEqual;

FROM BinOps     IMPORT LowerInt, HigherInt;

IMPORT GrafBase;

FROM MagicSys   IMPORT sBITSET, sINTEGER, lINTEGER;

IMPORT MagicAES, mtUtils, mtAppl, mtAlerts;

FROM mtDials    IMPORT DSTART, DFINISH, CSCREEN,
                       DialCenter, DialDraw, DialDo, DialForm, 
                       SetUserkey, GetKbdState, DialConfig, 
                       InstallHandler, RemoveHandler, CallByHandling;

FROM ListUtils  IMPORT tPoint, TwoRects, WordArray, FrameRects, DiffVertRect,
                       CalcAbsArea, formSetSlider;

IMPORT ListMan, RectFuncs;

IMPORT MTE;

IMPORT WinDials;

(*$H+*)

  TYPE  DIAL    = POINTER TO DialInfo;

        DialInfo    = RECORD
                        (* externe Parameter *)
                        tree            : ADDRESS;
                        listInfo        : ListMan.LISTINFO;
                        env             : ADDRESS;
                        boxIndex,
                        backIndex       : sINTEGER;
                        okBut, 
                        cancelBut       : sINTEGER;
                        doubleClickBut  : sINTEGER;
                        flags           : ldElemSet;
                        autoEdField     : sINTEGER;
                        lastSel         : lINTEGER;
                        hdl             : ldHandler;
                       
                        (* Fr Fensterdialoge: *)
                        listSetValues   : wdSetProc;
                        listGetValues   : wdSetProc;
                        listCheckExit   : lsCheckExitProc;
                        lastElemPtr     : POINTER TO ADDRESS;
                        
                        listSpecials    : POINTER TO ARRAY [0..511] OF specialButHdler;
                        listSpecialCount: INTEGER;
                        autoEdit        : INTEGER;
                        
                        inLoop          : inLoopProc;
                        exitBut         : INTEGER;
                        initial         : BOOLEAN;
                        
                        next            : DIAL;
                      END;


TYPE    CharSet = SET OF CHAR;

CONST   numChars        = CharSet{'0','1'..'9'};
        numSpaceChars   = CharSet{' ','0','1'..'9'};
        specialChars    = CharSet{10C, '|'};

VAR 
        dialList        : DIAL;
        
        voidO           : BOOLEAN;
        voidI           : sINTEGER;
        
        autoDial        : DIAL;

PROCEDURE BuildLdHandler (listStart   : ADDRESS; (* Zeiger auf die Liste *)
                          resetList   : ResetProc; 
                          nextEntry,
                          prevEntry   : MoveProc;
                          listElems   : CountProc;
                          isEnabled   : entryIsSelectableProc;
                          selEntry    : selectEntryProc;
                          entryToStr  : entryToStringProc;
                          isSelected  : isSelectedProc;
                          drawEntry   : drawEntryProc;
                          voffset     : sINTEGER;
                          vsize       : sINTEGER;
                          hStep       : sINTEGER;
                          maxWidth    : sINTEGER;
                          VAR handler: ldHandler);
(* Baut aus den bergebenen Daten einen ldHandler zusammen 
 *)
BEGIN
  handler := ldHandler{listStart, resetList, nextEntry, prevEntry,
                       listElems, isEnabled, selEntry, entryToStr, 
                       isSelected, drawEntry, voffset, vsize, hStep, 
                       maxWidth};
END BuildLdHandler;


PROCEDURE PrepareListDial (tree : ADDRESS;
                           elems : ldElemSet;
                           VAR handler : ldHandler;
                           boxIndex, 
                           boxBack   : sINTEGER;
                           okBut, 
                           cancelBut : sINTEGER;
                           env       : ADDRESS;
                           doubleClickButton : sINTEGER;
                           VAR d : DIAL) : BOOLEAN;
  (* Bereitet eine scrollbare Liste vor. 
   * Dabei mu ein DialInit schon erfolgt sein.
   *)
  VAR   i : sINTEGER;
        d2      : DIAL;
        selMode : ListMan.selectMode;
BEGIN
  NEW (d);
  IF d = NIL THEN RETURN FALSE END;
  d^.next := NIL;
  d^.tree       := tree;
  d^.flags      := elems;
  d^.hdl        := handler;
  d^.inLoop     := inLoopProc(NIL);
  (*
  d^.resetList  := resetList;
  d^.nextEntry  := nextEntry;
  d^.prevEntry  := prevEntry;
  d^.listElems  := listElems;
  d^.selEntry   := selEntry ;
  d^.isEnabled  := isEnabled;
  d^.isSelected := isSelected;
  *)
  
  d^.boxIndex   := boxIndex  ;
  d^.backIndex  := boxBack   ;
  d^.okBut      := okBut;
  d^.cancelBut  := cancelBut;
  d^.env        := env       ;
  (*
  d^.noSelect   := ~ select  ;
  d^.multiple   := multiple  ;
  d^.useArrows  := useArrows ;
  d^.entryToStr := entryToStr;
  d^.exitOnDoubleClick := exitOnDoubleClick;
  *)
  d^.doubleClickBut := doubleClickButton;
  (* EXCL (d^.flags, ldAutolocate); *)
  (*
  d^.autoLocator:= FALSE;
  *)
  (* Jetzt noch ein paar Sachen initialisieren *)

  WITH d^.hdl DO
    (* selMode festlegen *)
    IF ~(ldSelect IN elems)
    THEN
      selMode := ListMan.smNone
    ELSIF (ldMultiple IN elems)
    THEN
      selMode := ListMan.smMulti
    ELSE
      selMode := ListMan.smOne
    END;
    (* ListInfo erzeugen *)
    IF ~ListMan.InitList (list, env, selEntry, isSelected, 
                         isEnabled, drawEntry,                     
                         resetList, nextEntry, prevEntry, listElems, selMode,
                         d^.listInfo)
    THEN
      RETURN FALSE
    END;
    IF ~ListMan.SetListDialSpecs(d^.listInfo, tree, boxIndex, boxBack, voffset, 
                                 vsize, hStep, maxWidth)
    THEN
      voidO := ListMan.FreeList (d^.listInfo);
      RETURN FALSE
    END;
  END;
  
  (* Jetzt in Liste hngen *)
  IF dialList = NIL 
  THEN 
    dialList := d
  ELSE
    d2 := dialList;
    WHILE d2^.next # NIL DO
      d2 := d2^.next;
    END;
    d2^.next := d;
  END;
  
  WITH d^ DO
    (* Selektierten Eintrag suchen! *)
    ListMan.ListScroll2Selection (listInfo, FALSE);
    lastSel := ListMan.ListFirstSelected (listInfo);
  
    IF ldArrows IN flags THEN
     FOR i := 0 TO 4 DO
      SetUserkey (tree, -3, 72, BITSET(i), FALSE, TRUE); (* Cursor Up *)
      SetUserkey (tree, -4, 80, BITSET(i), FALSE, TRUE); (* Cursor Down *)
     END;
     SetUserkey (tree, -5, 71, {}, FALSE, TRUE);  (* Home *)
     SetUserkey (tree, -6, 71, {1}, FALSE, TRUE); (* Shift-Home *)
    END;
  
  END;  
  (*
  DialDraw (tree, 0, 999, sFrame, FALSE);
  *)
  RETURN TRUE
END PrepareListDial;  

PROCEDURE ReleaseListDial (VAR d : DIAL) : BOOLEAN;
(* gibt einen Listenhandler wieder frei. *)
  VAR d2 : DIAL;
      m  : sINTEGER;
BEGIN
  IF d = NIL THEN RETURN FALSE END;
  d2 := dialList;
  IF d2 = d
  THEN
    dialList := d^.next;
  ELSE
    WHILE (d2 # NIL) & (d2^.next # d) DO 
      d2 := d2^.next;
    END;
    IF d2 = NIL
    THEN RETURN FALSE
    END;
    (* d aushngen *)
    d2^.next := d^.next;
  END;
  (* ListMan freigeben *)
  voidO := ListMan.FreeList (d^.listInfo);
  IF ldArrows IN d^.flags
  THEN
    (* Userkeys wieder austragen *)
    FOR m := 0 TO 4 DO
      SetUserkey (d^.tree, -3, 72, BITSET(m), FALSE, FALSE); (* Cursor Up *)
      SetUserkey (d^.tree, -4, 80, BITSET(m), FALSE, FALSE); (* Cursor Down *)
    END;
    SetUserkey (d^.tree, -5, 71, {},  FALSE, FALSE); (* Home *)
    SetUserkey (d^.tree, -6, 71, {1}, FALSE, FALSE); (* Shift-Home *)
  END;
  DISPOSE (d);
  RETURN TRUE
END ReleaseListDial;

PROCEDURE NilDIAL () : DIAL;
BEGIN
  RETURN NIL
END NilDIAL;

PROCEDURE FindDial (findTree : ADDRESS; button : sINTEGER; treeOnly : BOOLEAN) : DIAL;
(* Gibt den zu dem Tree und dem Button gehrenden DIAL zurck,
 * ansonsten NIL.
 * Wenn der Button < 0 ist, wird nur nach dem Tree gesucht.
 *)
 VAR d : DIAL;
     but : sINTEGER;
     found : BOOLEAN;
     useArrows: BOOLEAN;
BEGIN
  d := dialList;
  but := INTEGER (BITSET(button)-{15});
  WHILE d # NIL DO
    WITH d^ DO 
      useArrows := ldArrows IN flags;
      IF tree=findTree 
      THEN 
        found := TRUE;
      ELSE 
        found := FALSE;
      END;
      IF treeOnly & found
      THEN
        RETURN d
      ELSIF ~treeOnly & 
         found & 
         (
         (but = boxIndex) OR
         (but = backIndex ) OR
         (button = -1) OR               (* Undo *)
         ((button = -3) & useArrows) OR (* Arrow 1 *)
         ((button = -4) & useArrows) OR (* Arrow 2 *)
         ((button = -5) & useArrows) OR (* Home *)
         ((button = -6) & useArrows) OR (* Shift-Home *)
         (but = okBut) OR
         (but = cancelBut))
      THEN
        found := TRUE;
      ELSE
        found := FALSE;
      END;
      IF found THEN RETURN d END;
    END;
    d := d^.next;
  END;
  d := NIL;
  RETURN d
END FindDial;

PROCEDURE InitAutoLocator (d: DIAL);
BEGIN
  autoDial := d;
END InitAutoLocator;

PROCEDURE FreeAutoLocator ();
BEGIN
  autoDial := NIL;
END FreeAutoLocator;

PROCEDURE AutoLocate (tree : ADDRESS; mx, my : sINTEGER; scan : sINTEGER;
                      kstate : sBITSET; event : sBITSET; edObj : sINTEGER): sBITSET;
  VAR i        : sINTEGER;
      idx      : lINTEGER;
      pt,
      ptx      : ADDRESS;
      d        : DIAL;
      search,
      lastStr,
      str       : ARRAY [0..255] OF CHAR;
      sFrame    : mtUtils.tRect;
      halfSize  : INTEGER;
      voidO     : BOOLEAN;
BEGIN
  IF ~(MagicAES.MUKEYBD IN event) THEN RETURN {} END;  (* Ist uns egal *)
  (* Cursortasten, Funktionstasten und noch ein paar Sondertasten abfangen *)
  IF ((scan >=59) & (scan <= 72)) OR (scan = 75) OR (scan = 77) OR
     ((scan >= 80) & (scan <= 98)) OR (scan = 42) OR (scan=54) OR (scan=71)
  THEN RETURN {} END;
  (* So, jetzt ist das also fr uns interessant! *)
  IF autoDial = NIL
  THEN
    (* Dial suchen *)
    d := FindDial (tree, -1, TRUE);
    IF d = NIL THEN RETURN {} END;
  ELSE
    d := autoDial
  END;
  (* Jetzt suchen *)
  WITH d^ DO
    (* Test auf Namecompletion *)
    autoEdField := edObj;
    IF autoEdField >= 0
    THEN
      (* String aus Editfeld holen *)
      mtUtils.ObjcString (tree, edObj, search);
      Upper (search);
      DelLeadingBlanks (search);
    ELSE
      (* Nur Taste auswerten *)
      search[0] := CAP(mtUtils.CharCode (scan, kstate));
      search[1] := 0C;
    END;
    (* ListBox an aktuelle Koordinaten anpassen *)
    IF (scan = 15) (* TAB *) & (autoEdField >= 0)
    THEN
      lastStr[0] := '';
      hdl.resetList (hdl.list);
      (* Namen komplettieren *)
      REPEAT 
        ptx := hdl.nextEntry (hdl.list);
        IF ptx # NIL
        THEN
          hdl.entryToStr (ptx, env, str);
          Upper (str);
          DelLeadingBlanks (str);
          IF hdl.isEnabled (ptx, env) & (Pos (search, str, 0) = 0)
          THEN
            (* Ist enthalten, also merken *)
            IF lastStr[0] # ''
            THEN
              (* vergleichen mit letztem, bis der erste Unterschied auftritt *)
              i := 0;
              WHILE (CAP(str[i]) = CAP(lastStr[i])) & (i < INTEGER(LENGTH(lastStr))) DO INC (i) END;
              hdl.entryToStr (ptx, env, lastStr);
              lastStr[i] := 0C;
            ELSE
              hdl.entryToStr (ptx, env, lastStr);
            END;
          END;
        END;
      UNTIL ptx = NIL;
      IF lastStr[0] # ''
      THEN
        DelTrailingBlanks (lastStr);
        (*
        i := INTEGER(LENGTH(lastStr))-1;
        WHILE (i > 0) & ~(lastStr[i] IN specialChars) DO DEC (i); END;
        IF lastStr[i] IN specialChars THEN lastStr[i] := 0C; END;
        DelTrailingBlanks (lastStr);
        i := INTEGER(LENGTH(lastStr))-1;
        WHILE (i > 0) & (lastStr[i] IN numChars) & (lastStr[i-1] IN numSpaceChars) DO lastStr[i] := 0C; DEC(i); END;
        (* Namen in Editfeld setzen und neu zeichnen *)
        DelTrailingBlanks (lastStr);
        *)
        mtUtils.SetObjcString (tree, edObj, lastStr);
        WinDials.WinEditChanged (tree, edObj);
        RETURN {1};
      ELSE
        RETURN {}
      END;
    ELSE
      hdl.resetList (hdl.list);
      idx := 0;
      REPEAT 
        ptx := hdl.nextEntry (hdl.list);
        INC (idx);
        IF ptx # NIL
        THEN
          hdl.entryToStr (ptx, env, str);
          Upper (str);
          DelLeadingBlanks (str);
          IF hdl.isEnabled (ptx, env) & (Pos (search, str, 0) = 0)
          THEN
            (* Jetzt auch darstellen *)
            ListMan.ListInvertEntry (listInfo, idx);
            lastSel := idx;
            IF ADDRESS(d^.inLoop) # NIL
            THEN 
              d^.inLoop (ptx, env);
            END;
            RETURN {}
          END;
        END;
      UNTIL ptx = NIL;
      (* Nichts gefunden, alten Eintrag deselektieren *)
      IF lastSel > 0 THEN ListMan.ListDeselEntry (listInfo, lastSel); END;
      lastSel := -1;
      IF ADDRESS(d^.inLoop) # NIL
      THEN 
        d^.inLoop (ptx, env);
      END;
    END;
  END;
  RETURN {};
END AutoLocate;


PROCEDURE HandleListDial (d : DIAL; 
                          VAR but : sINTEGER;     (* Der von DialDo gelieferte Button *)
                          mx, my  : sINTEGER;     (* Die Mauskoordinaten *)
                          kState  : BITSET;     (* Der Tastaturstatus  *)
                      VAR lastElem : ADDRESS) : BOOLEAN; 
(* Verwaltet die Events, wenn etwas in diesem Dial angeklickt wurde 
 *)

 VAR sFrame : mtUtils.tRect;
     doubleClick : BOOLEAN;
     button : BITSET;
     i      : sINTEGER;
     clicks : sINTEGER;
     selLast: lINTEGER;
     noSelect: BOOLEAN;
     multiple: BOOLEAN;
     oldLast: lINTEGER;
 
BEGIN
  WITH d^ DO 
    noSelect := ~(ldSelect IN flags);
    multiple := ldMultiple IN flags;
    oldLast := lastSel;
    IF but = -1 (* Userproc *) THEN
      (* Redraw von edObj *)
      WinDials.WinDialDraw (tree, autoEdField, 0, sFrame, FALSE);
      RETURN FALSE;
    (*
      IF ldAutolocate IN flags
      THEN
        (* Redraw von edObj *)
        DialDraw (tree, autoEdField, 0, sFrame, FALSE);
        RETURN FALSE;
      ELSE
        but := cancelBut;
        mtUtils.InclState (tree, but, MagicAES.SELECTED);
        DialDraw (tree, but, 0, sFrame, FALSE);
        RETURN TRUE
      END;
    *)
    END;
    IF initial & (but < 0)
    THEN
      RETURN FALSE;
    END;
    IF but = -3      (* Cursor Up *)
    THEN
      IF (2 IN kState) OR noSelect OR multiple
      THEN
        (* Ctrl *)
        IF (0 IN kState) OR (1 IN kState)
        THEN 
          (* Ctrl-Shift *)
          ListMan.ListPgUp (listInfo);
        ELSE
          ListMan.ListLnUp (listInfo);
        END;
      ELSE
        IF (0 IN kState) OR (1 IN kState)
        THEN
          (* Shift + Control *)
          lastSel := ListMan.ListSelPgUp (listInfo);
        ELSE
          lastSel := ListMan.ListSelUp (listInfo);
        END;
      END;
    ELSIF (but = -4) (* Cursor Down *)
    THEN
      IF (2 IN kState) OR noSelect OR multiple
      THEN
        (* ctrl: Scrollen *)
        IF (0 IN kState) OR (1 IN kState)
        THEN
          (* Shift *)
          ListMan.ListPgDown (listInfo);
        ELSE
          ListMan.ListLnDown (listInfo);
        END;
      ELSE
        (* Selected Entry Down   *)
        IF (0 IN kState) OR (1 IN kState)
        THEN
          lastSel := ListMan.ListSelPgDown (listInfo);
        ELSE
          lastSel := ListMan.ListSelDown (listInfo);
        END;
      END;
    ELSIF but = -5
    THEN
      (* Home: Nach oben gehen *)
      IF noSelect OR multiple
      THEN
        ListMan.ListFirst (listInfo);
      ELSE
        lastSel := ListMan.ListSelFirst (listInfo);
      END;
    ELSIF but = -6
    THEN
      (* Shift-Home: Ans Ende gehen *)
      IF noSelect OR multiple
      THEN
        ListMan.ListLast (listInfo);
      ELSE
        lastSel := ListMan.ListSelLast (listInfo);
      END;
    END;
    lastElem := ListMan.ListFindEntry (listInfo, lastSel);
    IF lastSel # oldLast
    THEN
      WinDials.WinEditChanged (tree, -1);
    END;
    doubleClick := but # (INTEGER (BITSET(but)-{15}));
    but := INTEGER(BITSET(but)-{15});
    IF but >= 0
    THEN
      IF (but = okBut) OR (but = cancelBut)
      THEN
        RETURN TRUE
      ELSIF (but = boxIndex) OR (but = backIndex)
      THEN
        clicks := 1;
        IF doubleClick THEN INC (clicks) END;
        selLast := ListMan.ListClick (listInfo, mx, my, kState, clicks);
        IF selLast >= 0
        THEN
          lastSel := selLast;
          WinDials.WinEditChanged (tree, -1);
        END;
        lastElem := ListMan.ListFindEntry (listInfo, lastSel);
        IF doubleClick & (ldDoubleExit IN flags) & (lastElem # NIL)
         & (but = boxIndex)
        THEN
          IF ADDRESS(d^.inLoop) # NIL
          THEN 
            d^.inLoop (d^.lastElemPtr^, d^.env);
          END;
          IF ~mtUtils.InState (tree, doubleClickBut, MagicAES.DISABLED)
            & mtUtils.InFlag (tree, doubleClickBut, MagicAES.SELECTABLE)
          THEN
            but := doubleClickBut;
            mtUtils.InclState (tree, but, MagicAES.SELECTED);
            WinDials.WinDialDraw (tree, but, 0, sFrame, FALSE);
            RETURN (but = okBut) OR (but = cancelBut);
          END;
        END;
      END;
    END; (* IF but *)
  END (* WITH d^ DO *);
  RETURN FALSE; 
END HandleListDial;

PROCEDURE dummyInLoop (entry, env : ADDRESS);
(* Dummy-Prozedur als InLoopProc *)
END dummyInLoop;

PROCEDURE dummyCheckExit (entry, env : ADDRESS; button: INTEGER; 
                          lastSel: ADDRESS; kstate: BITSET): BOOLEAN;
(* Dummy-Prozedur fr Exitcheck, gibt immer TRUE zurck *)
BEGIN
  RETURN TRUE;
END dummyCheckExit;


PROCEDURE DrawListDial (d: DIAL; clip: GrafBase.Rectangle);
(* Zeichnet die Liste komplett inklusive aller Elemente *)
  VAR upperFrame : GrafBase.Rectangle;
BEGIN
  IF d # NIL
  THEN
    WITH d^ DO
      initial := FALSE;
      mtUtils.CalcArea (tree, backIndex, upperFrame);
      ListMan.DrawList (listInfo, RectFuncs.ClipRect (upperFrame, clip));
    END;
  END;
END DrawListDial;

PROCEDURE doListDial (tree : ADDRESS;
                      elems : ldElemSet;
                      VAR handler : ldHandler;
                      boxIndex, boxBack,
                      okBut, cancelBut : sINTEGER;
                      env      : ADDRESS;
                      doubleClickButton : sINTEGER;
                      autoEdit   : INTEGER;
                      inLoop     : inLoopProc;
                      specials   : ARRAY OF specialButHdler;
                      setValues  : wdSetProc;
                      getValues  : wdSetProc;
                      VAR lastElem : ADDRESS) : sINTEGER;

(* Handelt eine komplette Scrollbare Liste.
 * Gibt den Index des Exit-Buttons zurck.
 * In lastElem wird die Adresse des zuletzt selektierten Elements zurckgegeben.
 * Bei multiple = TRUE knnen mehrere Eintrge selektiert werden.
 *)

  VAR org, space : mtUtils.tRect;

      but        : sINTEGER;
      buts       : BITSET;
      kstate     : BITSET;
      draw,
      exit       : BOOLEAN;
      d          : DIAL;
      d2         : DIAL;
      upperFrame,
      lowerFrame : GrafBase.Rectangle;
      mx,
      my         : sINTEGER;
      i          : sINTEGER;
      pt         : ADDRESS;
      str, str2  : ARRAY [0..255] OF CHAR;

BEGIN
        (* no grow & shrink boxes *)
  org.w := -1;

  IF ldAutolocate IN elems
  THEN
    InstallHandler (tree, AutoLocate, CallByHandling, 0, 0, upperFrame, NIL);
  END;

  DialCenter (tree, CSCREEN, 0, 0, upperFrame);
  DialForm (tree, DSTART, upperFrame, lowerFrame);
  
  IF ~ PrepareListDial (tree, elems,
                        handler, 
                        boxIndex, boxBack,
                        okBut, cancelBut,
                        env, doubleClickButton,
                        d)
  THEN 
    lastElem := NIL;   
    DialForm (tree, DFINISH, upperFrame, lowerFrame);
    RETURN cancelBut 
  END;
  MagicAES.WindUpdate (MagicAES.BEGUPDATE);
  MagicAES.WindUpdate (MagicAES.BEGMCTRL);
  DialDraw (tree, 0, 999, upperFrame, FALSE);
  mtUtils.CalcArea (tree, boxBack, upperFrame);
  ListMan.DrawList (d^.listInfo, upperFrame);
  IF ldAutolocate IN elems
  THEN
    buts := AutoLocate (tree, 0, 0, 0, {}, {MagicAES.MUKEYBD}, autoEdit);
  ELSE 
  END;
  mtAppl.MouseArrow();
  LOOP
    IF inLoop # inLoopProc(NIL) THEN
      inLoop (lastElem, env);      (* inLoopProc aufrufen *)
    END;
    
    but := DialDo (tree, 0);
    MagicAES.GrafMkstate (mx, my, buts, kstate);
    kstate := GetKbdState();
    
    d2 := FindDial (tree, but, FALSE);
    IF d2 # NIL
    THEN
      IF HandleListDial  (d2, but, mx, my, kstate, lastElem) THEN EXIT END;
    END;
    IF inLoop # inLoopProc(NIL) THEN
      inLoop (lastElem, env);      (* inLoopProc aufrufen *)
    END;

    (* specials *)
    but := INTEGER(BITSET(but)-{15});
    FOR i := 0 TO INTEGER(HIGH (specials)) DO
      IF (but = specials[i].objc)
       & ~mtUtils.InState (tree, specials[i].objc, MagicAES.DISABLED)
       &  mtUtils.InFlag (tree, specials[i].objc, MagicAES.SELECTABLE)
      THEN
        specials[i].proc (but, env, ADDRESS(d^.listInfo), draw, exit);
        IF draw THEN
          mtUtils.CalcArea (tree, boxBack, upperFrame);
          ListMan.DrawList (d^.listInfo, upperFrame);
          IF ldAutolocate IN elems
          THEN
            buts := AutoLocate (tree, 0, 0, 0, {}, {MagicAES.MUKEYBD}, autoEdit);
          END;
          d^.lastSel := ListMan.ListFirstSelected (d^.listInfo);
        END;
        IF exit THEN EXIT END;
      END;
    END; (* FOR i *)
  END (* LOOP *);
  kstate := GetKbdState();
  IF (ldAutolocate IN elems) & (d^.lastSel # -1) & (autoEdit >= 0)
  THEN 
    WITH d^ DO
      (* Ein Eintrag ist selektiert und der Dialog wurde nicht mit Shift-xxx verlassen *)
      lastElem := ListMan.ListFindEntry (listInfo, lastSel);
      (* String noch setzen *)
      IF lastElem # NIL
      THEN
        hdl.entryToStr (lastElem, env, str);
        IF ~((0 IN kstate) OR (1 IN kstate))
        THEN
          mtUtils.SetObjcString (tree, autoEdit, str);
        ELSE
          (* Vergleichen, ob der String im Editfeld und der von dem 
           * selektierten Eintrag gleich sind
           *)
          mtUtils.ObjcString (tree, autoEdit, str2);
          IF ~StrEqual (str2, str)
          THEN
            lastElem := NIL
          END;
        END;
      END;
    END;
  END;
  mtUtils.ExclState (tree, but, MagicAES.SELECTED);
  IF ReleaseListDial (d) THEN END;
  DialForm (tree, DFINISH, upperFrame, lowerFrame);
  RemoveHandler (tree);
  MagicAES.WindUpdate (MagicAES.ENDMCTRL);
  MagicAES.WindUpdate (MagicAES.ENDUPDATE);
  RETURN but
END doListDial;

PROCEDURE listSetGetValues (tree: ADDRESS; private: ADDRESS; set: BOOLEAN; exitBut: INTEGER);
  VAR d: DIAL;
      pt         : ADDRESS;
      str, str2  : ARRAY [0..255] OF CHAR;
      kstate     : BITSET;
      i          : INTEGER;
BEGIN
  d := private;
  IF set
  THEN
    (* Userkeys setzen *)
    IF ldArrows IN d^.flags THEN
     FOR i := 0 TO 4 DO
      SetUserkey (tree, -3, 72, BITSET(i), FALSE, TRUE); (* Cursor Up *)
      SetUserkey (tree, -4, 80, BITSET(i), FALSE, TRUE); (* Cursor Down *)
     END;
     SetUserkey (tree, -5, 71, {}, FALSE, TRUE);  (* Home *)
     SetUserkey (tree, -6, 71, {1}, FALSE, TRUE); (* Shift-Home *)
    END;
  
    (* Wird am Anfang aufgerufen *)
    d^.listSetValues (tree, d^.env, set, exitBut);
    
  ELSE
    (* Get wird nur am Ende aufgerufen, wir knnen jetzt den DIAL freigeben *)
    d^.listGetValues (tree, d^.env, set, d^.exitBut);
    (* Jetzt unseren Kram freigeben *)
  
    kstate := GetKbdState();
    IF (ldAutolocate IN d^.flags) & (d^.lastSel # -1) & (d^.autoEdit >= 0)
    THEN 
      WITH d^ DO
        (* Ein Eintrag ist selektiert und der Dialog wurde nicht mit Shift-xxx verlassen *)
        d^.lastElemPtr^ := ListMan.ListFindEntry (d^.listInfo, d^.lastSel);
        (* String noch setzen *)
        IF d^.lastElemPtr^ # NIL
        THEN
          hdl.entryToStr (d^.lastElemPtr^, d^.env, str);
          IF ~((0 IN kstate) OR (1 IN kstate))
          THEN
            mtUtils.SetObjcString (tree, d^.autoEdit, str);
          ELSE
            (* Vergleichen, ob der String im Editfeld und der von dem 
             * selektierten Eintrag gleich sind
             *)
            mtUtils.ObjcString (tree, d^.autoEdit, str2);
            IF ~StrEqual (str2, str)
            THEN
              d^.lastElemPtr^ := NIL
            END;
          END;
        END;
      END;
    END;
    mtUtils.ExclState (tree, d^.exitBut, MagicAES.SELECTED);
    IF ReleaseListDial (d) THEN END;
    RemoveHandler (tree);
  END;
END listSetGetValues;

PROCEDURE listDraw (tree: ADDRESS; private: ADDRESS; clip: GrafBase.Rectangle);
  VAR d: DIAL;
      bset: BITSET;
BEGIN
  d := private;
  WITH d^ DO
    ListMan.DrawList (listInfo, clip);
    IF initial 
    THEN
      initial := FALSE;
      (* Autolocator ggf. aufrufen *)
      IF (ldAutolocate IN d^.flags)
      THEN
        bset := AutoLocate (tree, 0, 0, 0, {}, {MagicAES.MUKEYBD}, d^.autoEdit);
      END;
    END;
  END;
END listDraw;

PROCEDURE listButton (tree: ADDRESS; private: ADDRESS; button: INTEGER;
                      mx, my : INTEGER; kstate: BITSET; clicks: INTEGER): BOOLEAN;
  VAR d, d2     : DIAL;
      draw,
      exit      : BOOLEAN;
      i         : INTEGER;
      buts      : BITSET;
      upperFrame: GrafBase.Rectangle;
BEGIN
  d := DIAL(private);
  IF ADDRESS(d^.inLoop) # NIL
  THEN 
    d^.inLoop (d^.lastElemPtr^, d^.env);
  END;
            
  d2 := FindDial (tree, button, FALSE);
  IF d2 # NIL
  THEN
    IF HandleListDial  (d2, button, mx, my, kstate, d^.lastElemPtr^) 
    THEN 
      (* Exit gewnscht *)
      IF ADDRESS(d^.inLoop) # NIL
      THEN 
        d^.inLoop (d^.lastElemPtr^, d^.env);
      END;
      d^.exitBut := button;
      RETURN d^.listCheckExit (tree, d^.env, button, d^.lastElemPtr^, kstate);
    END;
    IF ADDRESS(d^.inLoop) # NIL
    THEN 
      d^.inLoop (d^.lastElemPtr^, d^.env);
    END;
  END;
  (* specials *)
  button := INTEGER(BITSET(button)-{15});
  FOR i := 0 TO d^.listSpecialCount-1 DO
    IF (button = d^.listSpecials^[i].objc)
    & ~mtUtils.InState (tree, d^.listSpecials^[i].objc, MagicAES.DISABLED)
    (* &  mtUtils.InFlag (tree, d^.listSpecials^[i].objc, MagicAES.SELECTABLE) *)
    THEN
      d^.listSpecials^[i].proc (button, d^.env, ADDRESS(d^.listInfo), draw, exit);
      IF draw THEN
        mtUtils.CalcArea (tree, d^.backIndex, upperFrame);
        ListMan.DrawList (d^.listInfo, upperFrame);
        IF ldAutolocate IN d^.flags
        THEN
          buts := AutoLocate (tree, 0, 0, 0, {}, {MagicAES.MUKEYBD}, d^.autoEdit);
        END;
        d^.lastSel := ListMan.ListFirstSelected (d^.listInfo);
      END;
      IF exit THEN 
        (* Eventuell rausgehen, wenn Validierung erfolgreich *)
        d^.exitBut := button;
        IF ADDRESS(d^.inLoop) # NIL
        THEN 
          d^.inLoop (d^.lastElemPtr^, d^.env);
        END;
        RETURN d^.listCheckExit (tree, d^.env, button, d^.lastElemPtr^, kstate);
      END;
    ELSIF button = d^.listSpecials^[i].objc
    THEN
      (* Kann hier nur ankommen, wenn der Button disabled ist *)
      mtUtils.ExclState (tree, button, MagicAES.SELECTED);
      WinDials.WinDialDraw (tree, button, 0, upperFrame, FALSE);
    END;
  END; (* FOR i *)
  IF ADDRESS(d^.inLoop) # NIL
  THEN 
    d^.inLoop (d^.lastElemPtr^, d^.env);
  END;
  RETURN FALSE;
END listButton;


PROCEDURE WinListDial (tree : ADDRESS;
                        elems : ldElemSet;
                        VAR handler : ldHandler;
                        boxIndex, boxBack,
                        okBut, cancelBut : sINTEGER;
                        env      : ADDRESS;
                        doubleClickButton : sINTEGER;
                        autoEdit   : INTEGER;
                        inLoop     : inLoopProc;
                        VAR specials: ARRAY OF specialButHdler;
                        specialCount: INTEGER;
                        checkExit  : lsCheckExitProc;
                        setValues  : wdSetProc;
                        getValues  : wdSetProc;
                        VAR lastElem : ADDRESS) : BOOLEAN;

(* Handelt eine komplette Scrollbare Liste.
 * Gibt den Index des Exit-Buttons zurck.
 * In lastElem wird die Adresse des zuletzt selektierten Elements zurckgegeben.
 * Bei multiple = TRUE knnen mehrere Eintrge selektiert werden.
 *)

  VAR org, space : mtUtils.tRect;

      but        : sINTEGER;
      buts       : BITSET;
      kstate     : BITSET;
      draw,
      exit       : BOOLEAN;
      d          : DIAL;
      d2         : DIAL;
      upperFrame,
      lowerFrame : GrafBase.Rectangle;
      mx,
      my         : sINTEGER;
      i          : sINTEGER;
      pt         : ADDRESS;
      str, str2  : ARRAY [0..255] OF CHAR;

BEGIN
        (* no grow & shrink boxes *)
  org.w := -1;

  IF ~ PrepareListDial (tree, elems,
                        handler, 
                        boxIndex, boxBack,
                        okBut, cancelBut,
                        env, doubleClickButton,
                        d)
  THEN 
    lastElem := NIL;
    MTE.noMemAlert();
    RETURN FALSE;
  END;
  d^.listSetValues := setValues;
  d^.listGetValues := getValues;
  d^.listCheckExit := checkExit;
  d^.lastElemPtr := ADR (lastElem);
  d^.listSpecials := ADR (specials);
  d^.listSpecialCount := specialCount;
  d^.autoEdit := autoEdit;
  d^.inLoop   := inLoop;
  d^.initial  := TRUE;
  
  IF ~WinDials.OpenWinDial (tree, ldModal IN elems, listSetGetValues, listSetGetValues, listButton,
                            listDraw, "", d)
  THEN
    voidO := ReleaseListDial (d);
    lastElem := NIL;

    MTE.info (MTE.NoWinDial);
    
    RETURN FALSE;
  END;

  IF ldAutolocate IN elems
  THEN
    InstallHandler (tree, AutoLocate, CallByHandling, 0, 0, upperFrame, NIL);
  END;
  
  IF ldModal IN elems
  THEN
    WinDials.WinDialHandleEvents ();
  END;

  RETURN TRUE;
END WinListDial;

BEGIN
  dialList := NIL;
  autoDial     := NIL;
END ListDl.

