DEFINITION MODULE msgList;

(*==============================================================*
 * Modul:               CAT-Anzeigemodul fr Stichwortlisten    *
 * Autor:               Dirk Steins                             *
 * erstellt am:         25.07.1992                              *
 * letzte nderung am:  25.07.1992                              *
 * Version:             0.1                                     *
 * Interne Version:     V#0001                                  *
 *==============================================================*

 *----------------------------------------------------------------------------
 * Datum    Vers. Autor  nderung (Arbeitsbericht)
 *----------------------------------------------------------------------------
 * 25.07.92 0.1   DS     Erste Definition erstellt
 *----------------------------------------------------------------------------
 *)

FROM SYSTEM IMPORT ADDRESS;

IMPORT data;
IMPORT grin;

(*$H+*)
TYPE getNumberProc = PROCEDURE ((* dataHandle *) data.OneGroupHandle):CARDINAL;
 
PROCEDURE listOpen (gruppe, which : CARDINAL; grinWindow : INTEGER; 
                    mode: grin.openMode);
(* Ein Stichwortlistenfenster wird geffnet *)

PROCEDURE listOpenWithProc (gruppe : CARDINAL; whichNr : getNumberProc; 
                            grinWdw : INTEGER; mode: grin.openMode);
(* Ein Stichwortlistenfenster wird geffnet,
 * die Nummer wird ber eine Prozedur ermittelt
 *)

PROCEDURE listClose (wdw : INTEGER);
(* Ein Stichwortlistenfenster wird geschlossen *)

PROCEDURE listSavePos();
(* Sichert die Positionen der Stichwortlistenfenster *)

PROCEDURE listRestorePos();
(* Restauriert die Positionen der Stichwortlistenfenster *)

PROCEDURE listCloseAll ();
(* Alle Sichwortlistenfenster werden geschlossen *)

PROCEDURE listFreeAll();
(* Schliet alle Stichwortlistenfenster und gibt auch alle Handles wieder 
 * wieder frei 
 *)

PROCEDURE listWindowTop (wdw : INTEGER) : BOOLEAN;
(* Ist ein Stichwortlistenfenster oben? *)

PROCEDURE listSetFont (wdw: INTEGER; font, fontSize : INTEGER);
(* Setzt den Font fr das Stichwortlistenfenster *)

(* Die folgenden Funktionen sind nur fr die Interaktion 
 * mit der Anzeige relevant und werden ber ein handle angesprochen,
 * welches der Anzeige mit bergeben wird. Dadurch werden Probleme 
 * durch doppelte WdwHandles vermieden, die auftreten knnten.
 *)

PROCEDURE listUnlockWdw (hdl : LONGCARD);
(* Mu von grin aufgerufen werden, wenn die Gruppe gewechselt wurde
 * oder das Fenster geschlossen wurde. Geht leider nicht anders :-(
 *)

PROCEDURE listCloseByHandle (hdl : LONGCARD);
(* Ein Stichwortlistenfenster wird ber das assoziierte Handle 
 * geschlossen, wird von grin aufgerufen, wenn ein Fenster mit
 * geschlossen wird, an das eine Stichwortliste gelinkt ist
 *)

PROCEDURE listChangeGroup (hdl: LONGCARD; grinWdw: INTEGER; mess, newGroup: CARDINAL;
                           mode: grin.openMode): BOOLEAN;
(* Wechselt die dargestellte Gruppe in der durch wdw spezifizierten 
 * Liste und dem Fenster dazu. Ruft selber listUnlockWdw auf, mu 
 * also nicht vorher aufgerufen werden.
 * Gibt zurck, ob die Gruppe gewechselt wurde.
 *)

PROCEDURE listRemoveTree (hdl: LONGCARD; grinWdw: INTEGER; 
                          msgNum : CARDINAL; setReadFlag: BOOLEAN);
(* Entfernt den Baum mit der Nachricht msgNum aus der Liste im Fenster 
 * wdw, falls das grinWdw bereinstimmt. Wenn setReadFlag TRUE ist,
 * dann wird auch noch fr alle Nachrichten im Baum das Gelesen-Flag 
 * gesetzt.
 *)

PROCEDURE listChangeFlags (hdl: LONGCARD; grinWdw: INTEGER; 
                           msgNum: CARDINAL; newBits: BITSET);
(* grin teilt der Stichwortliste mit, da sich bei einer Nachricht
 * die Flags gendert haben 
 *)

PROCEDURE listSelectAll(wdw : INTEGER; withRedraw: BOOLEAN);
(* Alle Nachrichten in dem Fenster werden selektiert *)

PROCEDURE listSelectNew(wdw : INTEGER);
(* Alle neuen Nachrichten in dem Fenster werden selektiert *)

PROCEDURE listSaveAs (wdw: INTEGER);
(* Speichert die selektierten Nachrichten in einer Datei *)

PROCEDURE listCopy(wdw : INTEGER);
(* Alle selektierten Nachrichten werden exportiert*)

PROCEDURE listSomethingSelected (wdw: INTEGER): BOOLEAN;
(* Gibt zurck, ob in der Stichwortliste Nachrichten markiert 
 * sind
 *)

PROCEDURE listSetFlags(wdw : INTEGER; setBits, clearBits : BITSET);
(* Fr alle selektierten Nachrichten 
 * ein paar Flags setzen oder lschen 
 *)

PROCEDURE listSearch (wdw: INTEGER);
(* Nachrichten werden ber die Suchfunktion selektiert 
 *)

PROCEDURE CheckListLine (REF listline : ARRAY OF CHAR): BOOLEAN; 
(* Testet, ob die bergebene Zeile fr die Stichwortliste in Ordnung ist *)

  (* Das Format einer Zeile besteht aus einzelnen Krzeln 
   * fr einen Eintrag und darauffolgend in Klammern die Weite dieses Eintrages.
   * Beispiel: D(12),I(10),U(30),W(30),F(2)
   * Erlaubte Zeilenelemente:
   * D: Datum. Das Datum ist maximal 12 Zeichen lang und hat folgende Form:
   *           Mo, 16.11.92
   *    Optional kann man auch einen Formatstring angeben, der folgende Elemente
   *    enthalten darf:
   *     D: Tagesziffer  (d: wird gelscht, wenn nicht auffllbar)
   *     Y: Jahresziffer (y: wird gelscht, wenn nicht auffllbar)
   *     M: Monatsziffer (m: wird gelscht, wenn nicht auffllbar)
   *     G: Monatsbuchstabe Deutsch
   *     F: Monatsbuchstabe Frnzsisch
   *     E: Monatsbuchstabe English
   *     #E (Hinter Ziffern): englische Zahlfortsetzung (1st, 2nd, 3rd, 4th...)
   * T: Zeit. Default-Format fr die Zeit ist HH:SS, also 06:13.
   *    Optional kann man auch einen Formatstring angeben, der folgendes 
   *    enthalten darf:
   *
   *   HH: Stundenzahl ("hh" lscht fhrende Null)
   *   MM: Minutenzahl ("mm" lscht fhrende Null)
   *   SS: Sekundenzahl ("ss" lscht fhrende Null) 
   *
   *   #E: AM/PM-Notierung (English) ("e" schreibt "am" bzw. "pm")
   *
   *   Beispiel: Aus 'hh Uhr mm' wird '13 Uhr 5'
   *   Die Angabe von Sekunden in einem Formatstring ist allerdings unsinnig,
   *   da die Zeitangaben zu Mails keine Sekunden enthalten.
   * W: Wochentag. Auch hier kann man einen Formatstring angeben, der die Sprache
   *    spezifiziert, in der der Wochentag ausgegeben wird.
   *
   *     G: Tagesbuchstabe Deutsch
   *     F: Tagesbuchstabe Frnzsisch
   *     E: Tagesbuchstabe English
   *
   *    Es mu fr die gesamte gewnschte Lnge ein Formatbuchstabe angegeben
   *    werden.
   *    Wenn das erste Zeichen in der Maske ein # ist, dann wird die  
   *    Maske komplett mit dem zweiten Zeichen der Maske gefllt. 
   *    Beispiel: Aus #G wird intern GGGGGGGGGGGGGG
   * 
   * I: Message-Id. Im Moment ist diese maximal 10 Zeichen lang, aber ab Maus 9.0
   *                knnen Ids wesentlich lnger werden.
   * R: Reference-Id: Bei Kommentaren oder Antworten die Id der Referenznachricht.
   * U: Username 
   * A: Realname (wird ersetzt durch Username, falls nicht vorhanden)
   * W: Stichwort. Vor MAUS 9 maximal 30 Zeichen.
   * F: Flags. Die Flags der Nachrichten, die man in CAT setzen kann. 
   *           Reihenfolge wie in der Messageanzeige, also LFITDKB12
   * S: Status. Nur fr persnliche Nachrichten, andernfalls wird es ignoriert.
   * C: Kommentare: Anzahl der Kommentare auf die Nachricht, nicht fr persnliche Nachrichten.
   * #: interne Nummer der Message. Maximal 5 Stellen.
   *)

PROCEDURE SetListLine (REF listline : ARRAY OF CHAR; private : BOOLEAN);
(* Setzt die Stichwortlistenzeile fr die private oder andere Gruppen
 * Wenn die Zeile fehlerhaft ist, wird die vorhandene Zeile nicht
 * verndert.
 *)

PROCEDURE InitMsgList ();
(* Initialisiert die Messageliste *)

END msgList.