DEFINITION MODULE Printer;
(*****************************************************************************
  Projekt:
  Datei  : PRINTER.D
  Datum  : 17.1.1992
  Autor  : Uwe Mindrup, Altewiekring 13, 3300 Braunschweig; (0531) 79 78 40

  V 1.09: 17.01.92   Variable 'Proportional' zustzlich in SendInit
                     initialisiert.
  V 1.08: 30.07.91   Umstellung auf Compiler 4.00 und neue Bibliotheken
  V 1.07: 03.03.91   Die Druckerinitialisierung war nicht vollstndig.
                     Optimierung der Druckercharakteristik-Tabelle hinzugefgt.
  V 1.06: 01.03.91   'Effect' hinzugefgt.
  V 1.05: 26.02.91   Fehler in Translation-Prozedur behoben.
                     Fehler in SetMode behoben.
                     Diverse Prozeduren 'gestrafft'.
  V 1.04: 18.02.91   Proportionalschrift verbessert.
  V 1.03: 14.02.91   Ergnzung um Proportionalschrift
  V 1.02: 12.11.90   'ErrorProc' mit Rckgabewert ausgestattet; wird bei
                     '...DeviceBusy' FALSE zurckgeliefert, wird die Aktion
                     abgebrochen; bei TRUE wird weitergearbeitet.
                     'SetTabSize' eingefhrt.
  V 1.01: 07.11.90   Fehler beim Setzen von Schrifttyp und Farbauswahl
                     ausgemerzt.
  V 1.00: 04.11.90   Erste Umsetzung auf GEM+

   Druckertreiber;
   Der Druckertreiber dient dazu, Texte mit Hilfe der reichlich vorhandenen
   Druckeranpassungen fr Wordplus ausdrucken zu knnen. Zur Initialisierung
   mu eine CFG-Datei geladen werden.
   Aufbau einer CFG-Datei:
   - Header 'GST-CFG:'; dient als Kennung zur Unterscheidung von anderen
                        Formaten.
   - Es folgen sechs Bytes mit folgender Bedeutung:
     (1)   1 fr Typenrad mit horizontalem Spacing, sonst 0
     (2)   Zeichenbreite in Druckereinheiten (Typenrad)
     (3)   Zeilenabstand in Druckereinheiten
     (4)   Wagenmitte (Typenrad); Pixel/Inch (Matrixdrucker)
     (5)   Zeichenabstand bei Fettdruck (Typenrad)
     (6)   1 Pause mit Abfrage zwischen Seiten, sonst 0
     Diese Werte werden vom vorliegenden Modul ignoriert.

   - Jetzt folgen die Druckercharakteristiken.
     Diese Tabelle enthlt die Steuersequenzen, die zur Steuerung des
     Druckers benutzt werden (hoch, tief, fett, unterstrichen, kursiv,
     Draft, NLQ etc).
     Die Tabelle endet mit einem Null-Byte.
   - Zuletzt folgt die bersetzungstabelle. Hier werden einzelne Zeichen
     in andere Zeichen (-folgen) umgesetzt.
     Auch diese Tabelle endet mit einem Null-Byte.

   Bei einer Initialisierung werden alle Aussschaltsequenzen der aktuellen
   Druckqualitt sowie die horizontale und vertikale Initialisierung durch-
   laufen.
   Sind die Sequenzen der Druckercharakteristiken fr die beiden
   Druckqualitten gleich (unterscheiden sich also nur durch die Ein/Ausschalt-
   sequenz fr NLQ/Draft), dann reicht es aus, nur bei einer Ausschaltsequenz
   die entsprechende Schriftqualitt zu setzen (am besten bei einer selten
   bentigten Sequenz, z.B. Index unten).
   Mehr noch: Fr die brigen Druckercharakteristiken reicht es aus, wenn die
   Sequenz nur in der einen Schriftqualitt gesetzt wird. Der Treiber setzt die
   Sequenz automatisch auch fr die andere Schriftqualitt ein.

*******************************************************************************)

CONST cCon       = "CON:";
      cAux       = "AUX:";
      cPrn       = "PRN:";

      cPica      = 0;    (* 10 cpi *)
      cElite     = 1;    (* 12 cpi *)
      cCondensed = 2;    (* 17 cpi *)
      cExpanded  = 3;    (*  6 cpi *)

      cFat        = 0;   (* fett          *)
      cLightened  = 1;   (* hell          *)
      cSlanted    = 2;   (* kursiv        *)
      cUnderlined = 3;   (* unterstrichen *)

      (* Color-Farbbnder: *)
      cBlack     = 0;
      cMagenta   = 1;
      cCyan      = 2;
      cYellow    = 3;

TYPE  tResponse  = (done, notdone, noMemory, CFGnotFound, wrongIdent,
                    ErrorInStreams, Cancelled,
                    PrnDeviceBusy, AuxDeviceBusy, ConDeviceBusy);
      tErrorProc = PROCEDURE (tResponse): BOOLEAN;


VAR Response  : tResponse;
    ErrorProc : tErrorProc;           (* ist mit 'DummyError' initialisiert *)

PROCEDURE LoadCFGFile(VAR name: ARRAY OF CHAR);
(* Erffnung einer CFG-Datei. Der komplette Dateiname mu bergeben werden.
   Falls keine Erffnung mglich war, wird 'Response' auf 'CFGnotFound' gesetzt.
   Die Extension wird ggfs. automatisch korrigiert.
   Die Datei wird nach Erffnung auf die Dateikennung in den ersten 8 Bytes
   berprft ('GST-CFG:'). Falls die Kennung nicht bereinstimmt, wird
   die Datei wieder geschlossen und 'Response' auf 'wrongIdent' gesetzt.
   Im Erfolgsfall enthlt 'Response' 'done'.
*)

PROCEDURE FreeCFGFile();
(* Lscht eine bereits geladene CFG-Datei aus dem Speicher
 *)

PROCEDURE QueryCFGFile(VAR path, name    : ARRAY OF CHAR;
                       REF InfoText: ARRAY OF CHAR);
(* erfragt eine CFG-Datei per FileSelectBox.
   InfoText ist fr die FileSelectBox gedacht.
   Die Endung wird fest auf '*.CFG' festgeschrieben.
   Die Funktion ruft anschlieend OpenCFGFile auf.
*)
PROCEDURE GetPrnName(VAR Printername: ARRAY OF CHAR);
(* Eingetragenen Druckernamen liefern. Falls keiner vorhanden ist, dann
   wird ein Leerstring ('') geliefert.
 *)

PROCEDURE SendExit();
(*  Es wird Eintrag 21H (Allgemeiner Druckerreset am Textende) gesendet. *)

PROCEDURE SendInit(NLQ: BOOLEAN; hAndVInit: BOOLEAN);
(* Drucker-Initialisierung senden. Es werden nacheinander folgende
   Eintrge geschickt (in Klammern die NLQ-Eintrge):
   - Eintrag 20H       Horizontale Initialisierung (nur bei hAndVInit = TRUE)
   - Eintrag 1FH       Vertikale Initialisierung (nur bei hAndVInit = TRUE)
   - Eintrag  7H ( 9H) Fettschrift aus.
   - Eintrag  BH ( DH) Kursivschrift aus.
   - Eintrag  FH (11H) Light aus.
   - Eintrag 13H (15H) Superscript aus.
   - Eintrag 17H (19H) Subscript aus.
   - Eintrag 1BH (1DH) Unterstreichung aus.
   Falls vorhanden, werden noch die folgenden Sequenzn geschickt:
   - Eintrag 28H (29H) PICA (10 CPI).
   - Eintrag 30H       Color-Einstellung: Black.
   - Eintrag 35H       Proportional-Schrift wird ausgeschaltet.
 *)

PROCEDURE SetMode(Mode : SHORTINT);
(*  Zeichenbreite einstellen.
    Achtung: vorher 'SendInit' durchfhren!
    Untersttzt werden:
    - cPica      = 0;    (* 10 cpi *)
    - cElite     = 1;    (* 12 cpi *)
    - cCondensed = 2;    (* 17 cpi *)
    - cExpanded  = 3;    (*  6 cpi *)
 *)

PROCEDURE SetColor(Color : SHORTINT);
(*  Farbe einstellen. Untersttzt werden folgende Farben:
    - cBlack     = 0;
    - cMagenta   = 1;
    - cCyan      = 2;
    - cYellow    = 3;
*)
PROCEDURE SetPropSize(Prop: BOOLEAN; PropSize: SHORTINT);
(* Ein/Ausschalten der Proportionalschrift sowie Setzen einer mittleren
   Zeichenbreite (fr Tabulierung notwendig)
*)

PROCEDURE Bold(Set: BOOLEAN);         (* Fettdruck ein/ausschalten          *)

PROCEDURE Light(Set: BOOLEAN);        (* Helle Schrift ein/ausschalten      *)

PROCEDURE Italic(Set: BOOLEAN);       (* Schrgschrift ein/ausschalten      *)

PROCEDURE Underline(Set: BOOLEAN);    (* Unterstreichen ein/ausschalten     *)


PROCEDURE Superscript(Set: BOOLEAN);  (* Hochstellen ein/ausschalten        *)

PROCEDURE Subscript(Set: BOOLEAN);    (* Tiefstellen ein/ausschalten        *)


PROCEDURE Effect(Effects: BITSET; Set: BOOLEAN);
(* Texteffekte setzen/rcksetzen.
   Die Wertung im BITSET ist wie folgt:
   0 : fett,
   1 : hell,
   2 : kursiv,
   3 : unterstrichen
 *)

PROCEDURE FormFeed();                 (* Seitenvorschub ausgeben            *)

PROCEDURE SetTabSize(Tab: SHORTINT);  (* Tabulatorweite einstellen          *)

PROCEDURE WriteLn();                  (* Zeilenvorschub generieren          *)

PROCEDURE Write(ch: CHAR);

PROCEDURE WriteString(REF Str: ARRAY OF CHAR);

PROCEDURE Open(REF Name: ARRAY OF CHAR);
(* Erffnung eines Druckerkanals; wenn ein Fehler auftritt, wird 'Response'
   auf 'ErrorInStreams' gestellt; die eigentliche Fehlerursache ist dann bei
   'Streams.Response' zu suchen.
*)

PROCEDURE Close();
(* Schlieung eines Druckerkanals; wenn ein Fehler auftritt, wird 'Response'
   auf 'ErrorInStreams' gestellt; die eigentliche Fehlerursache ist dann bei
   'Streams.Response' zu suchen.
*)

PROCEDURE DummyError(ErrNr : tResponse): BOOLEAN;
(* Mit dieser Fehlerprozedur ist 'ErrorProc' initialisiert. Sie besteht nur
   aus einem 'RETURN TRUE'.
   Mit Hilfe der ProzedurVariablen 'ErrorProc' kann eine eigene Fehlerausgabe
   angegeben werden.
 *)

END Printer.
