IMPLEMENTATION MODULE Protokoll;

(*==============================================================*
 * Modul:               Protokollmodul fr Cat                  *
 * Autor:               Johannes Gttker-Schnetmann             *
 * erstellt am:         Mrz 1991                               *
 * letzte nderung am:  01.12.1992                              *
 * Version:             1.0                                     *
 * Interne Version:     V#0002                                  *
 *==============================================================*

 Dieses Modul macht Cat zu einem Xacc-Hauptprogramm der Stufe 0. Auerdem 
 werden vom AV/VA-Protokoll die sinnvollen Msgs untersttzt.

 Weiterhin ist hier das Protokoll fr den WatchDog verwirklicht:
 Ein WatchDog-Acc wird erkannt an seinem Namen "WATCHDOG", der beim AV- und
 XAcc-Protokoll erkannt wird. Sobald er erkannt wurde, bekommt er eine Msg
 [HelloWatchDog, CatGreetings, Version, (CatVersionString)]. Er kann dann 
 [HelloWatchDog, enable/disableWatchDog] verschicken und bekommt jeweils
 eine [HelloWatchDog, CatAffirm, on/off] als Besttigung. Auerdem bekommt
 er eine Msg CatTerminate am Programmende von Cat. Ist der WatchDog
 whrend des Einfgens von Msgs angeschaltet, dann wird er jeweils mit
 einem [HelloWatchDog, CatAsk, ptr] gefragt, ob diese Msg passieren darf.
 Er kann mit einem [HelloWatchDog, CatAccept/CatForget/CatFiltered] antworten.
 Im ersten Fall wird die Msg bernommen, im zweiten Fall wird sie verworfen und
 im dritten Fall wird bei ihr das "gefiltert"-Flag gesetzt.

 Dieses WatchDog-Protokoll macht unter MultiTOS mit Memoryprotection vorraussichtlich
 Probleme, da der WatchDog nach der bisherigen Implementation auf Speicher von 
 Cat zugreift! Wahrscheinlich mssen wir und dafr noch etwas einfallen lassen.
 (shared-memory..)

 Die Erluterung des erweiterten Protokolls steht weiter unten. "HelloWatchDog"
 hat jetzt die allgemeinere Bedeutung "CatMsg".
 Insbesondere die Erkennung des WatchDogs soll in einer der nchsten Versionen
 nach Mglichkeit auf die Methode des erweiterten Protokolls umgestellt werden.

 *----------------------------------------------------------------------------
 * Datum    Vers. Autor  nderung (Arbeitsbericht)                            
 *----------------------------------------------------------------------------
 * 01.02.92 0002  JGS    Protokollerweiterung fr Steueraccs
 * 24.05.93 0003  DS     Erweiterung um MultiTasking-Untersttzung
 *                       CAT XACC-Versionsnummer gendert auf 20H
 *----------------------------------------------------------------------------
 *)


FROM SYSTEM          IMPORT ADR, ADDRESS, WORD, CADR, CALLSYS, TSIZE;
FROM WdwManager      IMPORT NewWindowIsTop, WindowIsClosed, WdwDoesDragDrop,
                            WdwDDGetExts, WdwDDAcceptData, WdwDDWriteData,
                            HandleEvent;
FROM UserInformation IMPORT UserBLK;
FROM Void            IMPORT v;
FROM CatTypes        IMPORT Str255Ptr, String255, BigTextPtr;
FROM CatGlobal       IMPORT multiTask, multiTOS, magIx, magIxVer;
FROM Storage         IMPORT ALLOCATE, DEALLOCATE;
IMPORT FileNames;
IMPORT Block;
IMPORT Strings;
IMPORT StrConv;

IMPORT data;
IMPORT grin;

IMPORT MagicAES;
IMPORT MagicStrings;
IMPORT mtAppl;
IMPORT mtAlerts;
IMPORT mtCommand;
IMPORT MagicDOS;
IMPORT MagicCookie;
IMPORT Mintbind;
IMPORT ConvertDate;

CONST

  appName    = 'CAT     '+0C+'XDSC'+0C+'1MausTausch-Programm'+0C+'2DC'+0C+0C;
  watchDog   = 'WATCHDOG';
  watchFound = 0815H;
  
(* CalcClock *)
  CALCLOCK  = 100;
  CALCLOCK_OK = 101;

(* XAcc: *)

  ACC_ID    = 400H;
  ACC_OPEN  = 401H;
  ACC_CLOSE = 402H;
  ACC_ACC   = 403H;
  ACC_EXIT  = 404H;
  ACC_ACK   = 500H;
  ACC_TEXT  = 501H;
  ACC_KEY   = 502H;
  
(* PC-Help Protokoll *)
  AC_HELP   = 1025;         (* message to DA *)
  AC_REPLY  = 1026;         (* message from DA *)
  AC_VERSION= 1027;         (* Ask DA for version *)
  AC_COPY   = 1028;         (* Ask DA to copy screen to clipboard *)

(* AV/VA-Protokoll *)

  AV_PROTOKOLL     = 4700H;
  VA_PROTOSTATUS   = 4701H;

(* not used *)
  AV_GETSTATUS     = 4703H;
  AV_STATUS        = 4704H;
  VA_SETSTATUS     = 4705H;
(*----------*)

  AV_SENDKEY       = 4710H;
  VA_START         = 4711H;

(* not used *)
  AV_ASKFILEFONT   = 4712H;
  VA_FILEFONT      = 4713H;
  AV_ASKCONFONT    = 4714H;
  VA_CONFONT       = 4715H;
  AV_ASKOBJECT     = 4716H;
  VA_OBJECT        = 4717H;
  AV_OPENCONSOLE   = 4718H;
  VA_CONSOLEOPEN   = 4719H;
(*----------*)

  AV_OPENWIND      = 4720H;
  VA_WINDOPEN      = 4721H;
  AV_STARTPROG     = 4722H;
  VA_PROGSTART     = 4723H;
  AV_ACCWINDOPEN   = 4724H;

(* not used *)
  VA_DRAGACCWIND   = 4725H; (* NOCH, wichtig fr Kommandozeile an Takeoff *)
(*----------*)

  AV_ACCWINDCLOSED = 4726H;
  AV_PATH_UPDATE   = 4730H;

(* Bits *)

  xAV_SENDKEY       = 0;
  xAV_ASKFILEFONT   = 1;
  xAV_ASKCONFONT    = 2; (* auch xAV_OPENCONSOLE *)
  xAV_ASKOBJECT     = 3;
  xAV_OPENWIND      = 4;
  xAV_STARTPROG     = 5;
  xAV_ACCWINDOPEN   = 6; (* auch xAV_ACCWINDCLOSED *)
  xAV_STATUS        = 7; (* auch AV_GETSTATUS *)


(*## Cat-Protokoll ################################################*)

(* GutenTag *)
  HelloWatchDog = 8000H;
  CatMsg        = 8000H;
  ExtCatMsg     = 8001H;

  Cat2FiltMsg   = 8002H;        (* Fr Laberfilter-Anbindung *)
  CatProtoMsg   = 8003H;        (* Neues CAT-Protokoll *)

  (* Sub-Opcodes fr CatMsg *)

  enableWatchDog  = 0;
  disableWatchDog = 1;
  CatAffirm       = 2;

  CatAsk          = 3;
  CatAccept       = 4;
  CatForget       = 5;
  CatGreetings    = 6;
  CatTerminate    = 7;
  CatFiltered     = 8;


  WdwManager     = 9;
  CatWinInfo     = 10;
  CatWhatsIn     = 11;
  CatPosition    = 12;
  CatNewMess     = 13;
  CatInform      = 14;
  CatWinDirty    = 15;

  CatProtoVersion   = 1; (* Das ist Version Nummer 1 *)
  
  (* Sub-Opcodes fr Cat2FiltMsg *)
  
  FilterStandard    = 0;        (* Default *)
  
  (* Sub-Opcodes fr neues CAT-Protokoll *)
  (* Doku: Siehe NEWPROTO.TXT            *)
  
  CAT_ACK           = 0;
  CAT_ERROR         = 1;
  CAT_LOGIN         = 2;
  CAT_LOGOFF        = 3;
  CAT_SENDER        = 4;
  CAT_RNAME         = 5;
  CAT_RECEIVER      = 6;
  CAT_MAUS_ID       = 7;
  CAT_MSG_ID        = 8;
  CAT_MAUS_REF      = 9;
  CAT_REF           = 10;
  CAT_TOPIC         = 11;
  CAT_BOX           = 12;
  CAT_GATE          = 13;
  CAT_DIST          = 14;
  CAT_MIME          = 15;
  CAT_STATUS        = 16;
  CAT_FOLLOWUP      = 17;
  CAT_REPLYTO       = 18;
  CAT_TEXT          = 19;
  CAT_MSG_DATE      = 20;
  CAT_STATUS_DATE   = 21;
  CAT_GROUP         = 22;
  CAT_ORG_TEXT      = 23;
  CAT_IS_OWN        = 24;
  CAT_UNIQUE_ID     = 25;
  CAT_REAL_SENDER   = 26;
  CAT_WDW_NAME      = 27;
  
  CAT_NEW_MSG       = 100;
  CAT_TERMINATE     = 101;
  
(* 
mess[0] = 0x8002;
mess[1] = apId;  (* 0 *)
mess[2] = oversize; (* 0 *)
mess[3] = length of filter data;   (* must not be 0 *)
mess[4] + mess[5] = pointer to filter data;
mess[6] + mess[7] = unused (* must be 0 *)
*)



(* --------------------------------------------------------------------------
Also dann: 
mess[0] ist bei allen Msgs 0x8000, d.h. 8000 hexadezimal.
mess[1] ist ja einfach die Applikation-ID des Absenders;
mess[2] ist wie blich die berlnge der Nachricht, also die Gesamtlnge 
der Nachricht - 16. CB = CatBaum als Arbeitsname.

mess[3],mess[4]
CatGreetings
 6              CB->Cat Anmeldung bei Cat (evtl. kommt in mess[5] noch eine 
                Unterscheidung, ob es sich um einen WatchDog handelt, also 
                dort am besten eine Null eintragen). Fr etwas anderes als
                einen Watchdog darf in mess[5] keine 0 stehen.
 6      version Cat-CB  Anmeldung angekommen, Protokoll-Versionsnummer zum 
                berprfen als Anlage. (evtl. mess[5,6,7] Cat-Versionsnummer 
                als String? - erstmal nicht!)

WdwManager
 9      art     ACC->Cat, Anforderung der offenen Fenster, mess[5+6] 
                enthalten eine Adresse fr den Buffer, in den CAT 
                die Information hineinschreibt. mess[7] enthlt die 
                Maximalanzahl an Fenstern, die in den Buffer passen 
                (pro Eintrag ein Integer, also 2 Byte)
                "art" enthlt die Art der Fenster, die geliefert 
                werden sollen: 0: MsgFenster, 1: Editoren, 2: 
                Stichwortlisten
 9      art     Cat->ACC, offene Fenster, mess[5+6] enthlt den Pointer 
                auf den Buffer, der vorher bergeben wurde. In dem Buffer
                steht die Liste der Fensterhandles, wenn das vom AES aus 
                klappt. mess[7] enthlt die Anzahl der Fenster, die von 
                CAT geliefert werden.

Die folgenden Msgs werden aber zunchst einmal nur von MsgFenstern 
untersttzt.

CatWinInfo
10      whdl    ACC->Cat, Informationen anfordern. mess[5+6] enhlt eine 
                Adresse, die sagt, wo die Informationen sollen , Format der 
                Information wie "winInfo" (noch nicht festgelegt). mess[7] 
                enthlt die Lnge des Buffers in mess[5+6].
10      whdl    Cat->ACC, Information wurde geschrieben, Bufferadresse 
                wieder in mess[5+6].
CatWhatsIn
11      whdl    CB->Cat, Welche Informationen werden im angegebenen Fenster 
                dargestellt?
11      whdl    Cat->CB mess[5] Gruppe, mess[6] interne MsgNummer
CatPosition
12      whdl    CB->Cat, Informationen ber MsgPositionen. mess[7] gibt an, 
                um welche Art es sich handelt: 0 Neue, 1 letzte Position, 2 
                ab Datum, 3 ab Nummer. In den letzten beiden Fllen wird in 
                mess[5+6] noch eine Adresse auf einen String bergeben, der 
                das Datum bzw. die ID enthlt. Er ist nullterminiert.
12      whdl    Cat->CB, mess[5] interne Msgenummer; 0xffff heit: Nicht 
                gefunden
CatNewMess
13      whdl    CB->Cat, neuen Inhalt setzen: mess[5] Gruppe mess[6] interne 
                MsgNummer
13      whdl    Cat->CB mess[5] = 0 -> hat geklappt, mess[5] # 0 -> geht 
                nich..

Hier nun meine erste Idee zu der Sache mit den Fensterwechseln. Dabei gibt 
es jetzt ja ein paar Probleme. Ich mu dann ja den Speicher des Accs 
beschreiben, das sollte ich aber nur nach Aufforderung machen. Wenn der User 
jetzt sehr schnell durch die Msgs durchrennt, dann bekommt das Acc nicht 
jeden Wechsel mit, da es ja immer die neue Informationen anfordern mu. Aber 
den Prozewechsel kann man dann ja vielleicht doch irgendwie forcieren. Dann 
wei das Acc zwar, da es hier einen neuen Inhalt gibt und bekommt auch mit, 
wenn der User weitergeblttert hat, ohne da die Information ber die neue 
Fensterposition schon angekommen ist, es kommt aber nicht zu Verstopfungen. 
Aber vielleicht ist dieses Verhalten ja auch erwnscht, denn wenn er dann 
anhlt, wird die neue Position ja dargestellt. 

CatInform
14              CB->Cat, Acc in die Fensterwechsel/-ffnen-Informationsliste 
                ein-/austragen. mess[5] = 0 austragen, mess[5] = 1 eintragen,
                mess[6] ist das Window-Handle des ACC-Fensters
14              Cat->CB Acc ein-/ausgetragen, mess[5] wie oben.
CatWinDirty
15      whdl    Cat->CB dieses Fenster hat einen neuen Inhalt bzw. wurde 
                geffnet. mess[5] = 0: neuer Inhalt; 1: getopped; 2: neu 
                geffnet; 3: geschlossen


---- ab hier Erweitungsplanung fr spter, erstmal ist hier Ende der Erweiterung -----
Es wird etwas problematisch, wenn das Acc in die MsgBase schreiben will. 
Zunchst ist das in Cat nur dann ohne grere Probleme mglich, wenn auch 
die entsprechende Gruppe geffnet ist. Daher brauche ich das whdl, um damit 
das Gruppenhandle zu bekommen. Mir fllt auch gerade auf, da wir noch keine 
Mglichkeit vorgesehen haben, an MsgInfos von Msgs zu kommen, die nicht 
gerade in einem Fenster dargestellt werden. Das mu ja wohl auch noch rein, 
wenn Du damit Bume markieren willst. Also
16      whdl    CB->Cat; Infos zu einer anderen Msg aus dieser Gruppe 
                anfordern. mess[5+6] Pointer wie in Msg Nummer 10, mess[6] 
                interne Nummer. Leider knnen wir dann nicht unterscheiden, 
                welche Infos angefordert werden sollen. Hmm..
Die Gruppe wird hier also ber das whdl festgelegt, was natrlich nicht so 
schn ist. Wie gesagt, wahrscheinlich mssen wir uns beim Schreiben noch 
etwas neues ausdenken. 
Anwort ist dann eine Msg vom Typ 10
17      whdl    CB->Cat; Flags setzen. mess[5] interne Nummer; mess[6] neue 
                gesetzte Flags; mess[7] neu gelschte Flags.
17      whdl    Cat->CB; Flags setzen; mess[5] = 0 alles ok; mess[5] # 0 = 
                etwas ist schiefgelaufen.

-----------------------------------------------------------------------------*)

VAR     watch  : INTEGER;   (* Ist ein WatchDog installiert? *)
        isWatch : BOOLEAN;  (* Soll der auch gefragt werden? *)

(* Um sich alle installierten Accs beim XAcc-Protokoll zu merken wird hier ein *) 
(* kleines Array angelegt. Das ist natrlich statisch, aber darber wrde ich  *)
(* mir erstmal noch keine Sorgen machen.. :-)                                  *)
CONST maxAcc = 80;      (* Wegen MTOS von 6 auf 80 erweitert. Zwar kann 
                         * man unter MTOS auch mehr als 80 ACCs und Applikationen 
                         * haben, aber das wre ziemlich krank *)

CONST maxBufLen = 128;  (* Lnge des statischen Textbuffers *)

      memDescrSize  = 2048; (* 2 kB globaler Speicher *)

TYPE        (* for systems with memory protection (MultiTOS) *)
        memDescr        = RECORD
                            CASE : BOOLEAN OF
                            TRUE:
                              apName  : ARRAY [0..49] OF CHAR; (* our extended application name *)
                              textBuf : ARRAY [0..maxBufLen-1] OF CHAR;
                              dataPtr : data.PtrRecord;
                              mPath   : String255;
                              mName   : String255; |
                            FALSE:
                              allInOne: ARRAY [0..memDescrSize-1] OF CHAR;
                            END;
                          END;

VAR accCount : CARDINAL;
    acc      : ARRAY[1..maxAcc] OF INTEGER;
    protoCnt : CARDINAL;
    protoAcc : ARRAY[1..maxAcc] OF INTEGER; (* Noch ein Array fr Apps,
                                             * diesmal fr die, die am CAT-Proto 
                                             * teilnehmen 
                                             *)
    messBuff : ARRAY [0..31] OF INTEGER;    (* Platz fr Messagebuffer *)
    startApId: INTEGER;     (* Von wem kam die letzte AV_STARTPROG *)
    memPtr   : POINTER TO memDescr;         (* global buffer *)


PROCEDURE SendTextReady();
(* Das wrde ntig werden, wenn wir XAcc Stufe 1 knnen *)
BEGIN
END SendTextReady;

PROCEDURE send(msg : CARDINAL; dest : INTEGER; REF data : ARRAY OF WORD);
(* Anpassen, damit man damit demnchst lngere Msgs verschicken kann *)
VAR mess : POINTER TO RECORD
             msg : CARDINAL;
             id  : INTEGER;
             overLength : INTEGER;
             w : ARRAY[0..4] OF WORD
           END;
   z : CARDINAL;
BEGIN
  mess := ADR(messBuff);
  FOR z := 0 TO 4 DO mess^.w[z] := data[z] END;
  mess^.msg := msg;
  mess^.id  := mtAppl.ApplIdent;
  mess^.overLength := 0;
  MagicAES.ApplWrite(dest, 16, mess^);
  MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Um es auch mal abzuschicken.. *)
  MagicAES.WindUpdate(MagicAES.BEGUPDATE);
END send;

PROCEDURE SendMess(msg : CARDINAL; catType : INTEGER; l : LONGCARD; i2, i3, dest : INTEGER);
VAR data : RECORD catType : INTEGER; l : LONGCARD; i2, i3 : INTEGER END;
BEGIN
  data.catType := catType; data.l := l; data.i2 := i2; data.i3 := i3;
  send(msg, dest, data);
END SendMess;

PROCEDURE SendMessInt(msg : CARDINAL; catType, i1, i2, i3, i4, dest : INTEGER);
VAR data : RECORD catType, i1, i2, i3, i4 : INTEGER END;
BEGIN
  data.catType := catType; data.i1 := i1; data.i2 := i2; data.i3 := i3; data.i4 := i4;
  send(msg, dest, data);
END SendMessInt;

PROCEDURE SendAdr(msg : CARDINAL; catType, i1 : INTEGER; l : LONGCARD; i2, dest : INTEGER);
VAR data : RECORD catType : INTEGER; i1 : INTEGER; l : LONGCARD; i2 : INTEGER END;
BEGIN
  data.catType := catType; data.i1 := i1; data.l := l; data.i2 := i2;
  send(msg, dest, data);
END SendAdr;

PROCEDURE SendNewMsgInWdw ();
TYPE mType = RECORD
              messId  : CARDINAL;
              apId    : INTEGER;
              over    : CARDINAL;
              catType : CARDINAL;
              length  : LONGCARD;
              buf     : ADDRESS;
            END;
 VAR mPtr: POINTER TO mType;
     i   : CARDINAL;
BEGIN
  mPtr := ADR (messBuff);
  WITH mPtr^ DO
    messId  := CatProtoMsg;
    apId    := mtAppl.ApplIdent;
    over    := 0;
    catType := CAT_NEW_MSG;
    length  := 0;
    buf     := NIL;
  END;
  MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Um es auch mal abzuschicken.. *)
  FOR i := 1 TO protoCnt - 1 DO 
    MagicAES.ApplWrite (protoAcc[i], 16, mPtr^);
  END;
  MagicAES.WindUpdate(MagicAES.BEGUPDATE);
END SendNewMsgInWdw;

PROCEDURE HandleCatProtoMsg (mess: ADDRESS);

TYPE mType = RECORD
              messId  : CARDINAL;
              apId    : INTEGER;
              over    : CARDINAL;
              catType : CARDINAL;
              length  : LONGCARD;
              buf     : ADDRESS;
            END;

 VAR mPtr: POINTER TO mType;
     wdw : INTEGER;

  PROCEDURE ReplyError (mess: ADDRESS);
    VAR to: INTEGER;
  BEGIN
    to := mPtr^.apId;
    mPtr^.apId := mtAppl.ApplIdent;
    mPtr^.length := VAL (LONGCARD, mPtr^.catType) * 65536L;
    mPtr^.catType := CAT_ERROR;
    mPtr^.buf := NIL;
    MagicAES.ApplWrite(to, 16, mPtr^);
    MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Um es auch mal abzuschicken.. *)
    MagicAES.WindUpdate(MagicAES.BEGUPDATE);
  END ReplyError;
  
  PROCEDURE ReplyOk (mess: ADDRESS);
    VAR to: INTEGER;
  BEGIN
    to := mPtr^.apId;
    mPtr^.apId := mtAppl.ApplIdent;
    mPtr^.catType := CAT_ACK;
    mPtr^.length := 0;
    mPtr^.buf := NIL;
    MagicAES.ApplWrite(to, 16, mPtr^);
    MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Um es auch mal abzuschicken.. *)
    MagicAES.WindUpdate(MagicAES.BEGUPDATE);
  END ReplyOk;
  
  PROCEDURE Reply (mess: ADDRESS);
    VAR to: INTEGER;
  BEGIN
    to := mPtr^.apId;
    mPtr^.apId := mtAppl.ApplIdent;
    MagicAES.ApplWrite(to, 16, mPtr^);
    MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Um es auch mal abzuschicken.. *)
    MagicAES.WindUpdate(MagicAES.BEGUPDATE);
  END Reply;

  PROCEDURE RemoveAcc (apId : INTEGER);
    VAR i : CARDINAL;
  BEGIN
    i := 1;
    WHILE (i < protoCnt) & (protoAcc[i] # apId) DO INC (i) END;
    IF i < protoCnt THEN
      FOR i := i TO protoCnt - 2 DO
        protoAcc[i] := protoAcc[i+1];
      END;
      DEC (protoCnt);
    END;
  END RemoveAcc;

  PROCEDURE checkId (id : INTEGER): BOOLEAN;
  (* Prft, ob schon ein Prozess mit der ID registriert ist *)
    VAR i : CARDINAL;
  BEGIN
    FOR i := 1 TO protoCnt - 1 DO 
      IF protoAcc[i] = id THEN RETURN FALSE END;
    END;
    RETURN id # mtAppl.ApplIdent;
  END checkId;
  
  PROCEDURE AddAcc (apId: INTEGER) : BOOLEAN;
  BEGIN
    IF (protoCnt < maxAcc) & checkId (apId)
    THEN
      protoAcc[protoCnt] := apId; (* ID des Acc, das sich gerade anmeldet *)
      INC(protoCnt);
      RETURN TRUE;
    ELSE
      RETURN FALSE;
    END;
  END AddAcc;

TYPE    dateRecType = RECORD
                        date: CARDINAL;
                        time: CARDINAL;
                      END;

  VAR   dateRecPtr : POINTER TO dateRecType;
        strPtr     : POINTER TO ARRAY [0..32767] OF CHAR;
        lcPtr      : POINTER TO LONGCARD;
        intPtr     : POINTER TO INTEGER;
        lCard      : LONGCARD;
        sizeReq    : BOOLEAN;
        dist       : data.tDistribution;
        date       : ConvertDate.Date;
        time       : ConvertDate.Time;

BEGIN
  mPtr := mess;
  IF (mPtr^.catType > CAT_WDW_NAME) & (mPtr^.catType # CAT_NEW_MSG) & (mPtr^.catType # CAT_TERMINATE)
  THEN
    ReplyError (mess);
    RETURN;
  END;
  CASE mPtr^.catType OF
      CAT_ACK        : ;
    | CAT_ERROR      : ;
    | CAT_LOGIN      : 
        (* ACC anmelden *)
        IF AddAcc (mPtr^.apId)
        THEN
          ReplyOk (mess);
        ELSE
          ReplyError (mess);
        END;
    | CAT_LOGOFF     : 
        RemoveAcc(mPtr^.apId);
        ReplyOk (mess);
    | CAT_NEW_MSG    :
    | CAT_TERMINATE  :
  ELSE
    (* Window von grin erfragen *)
    grin.GetGrinTopWdw (wdw);
    IF wdw < 0
    THEN
      ReplyError (mess);
      RETURN;
    END;
    (* Andere Nachricht *)
    IF mPtr^.length = 0
    THEN
      lcPtr := ADR (mPtr^.length);
      sizeReq := TRUE;
    ELSE
      lcPtr := ADR (lCard);
      sizeReq := FALSE;
    END;
    (* Grenabfrage *)
    CASE mPtr^.catType OF
        CAT_SENDER     : 
          lcPtr^ := grin.ActFromSize (wdw);
      | CAT_RNAME      : 
          lcPtr^ := grin.ActNameSize (wdw);
      | CAT_RECEIVER   : 
          lcPtr^ := grin.ActReceiverSize (wdw);
      | CAT_MAUS_ID    : 
          lcPtr^ := grin.ActIdSize (wdw);
      | CAT_MSG_ID     : 
          lcPtr^ := grin.ActMIdSize (wdw);
      | CAT_MAUS_REF   : 
          lcPtr^ := grin.ActRefIdSize (wdw);
      | CAT_REF        : 
          lcPtr^ := grin.ActRIdSize (wdw);
      | CAT_TOPIC      : 
          lcPtr^ := grin.ActSubjectSize (wdw);
      | CAT_BOX        : 
          lcPtr^ := grin.ActBoxSize (wdw);
      | CAT_GATE       : 
          lcPtr^ := grin.ActGateSize (wdw);
      | CAT_DIST       : 
          lcPtr^ := grin.ActDistSize (wdw);
      | CAT_MIME       : 
          lcPtr^ := grin.ActMimeSize (wdw);
      | CAT_STATUS     : 
          lcPtr^ := grin.ActStatusSize (wdw);
      | CAT_FOLLOWUP   : 
          lcPtr^ := grin.ActFollowupSize (wdw);
      | CAT_REPLYTO    : 
          lcPtr^ := grin.ActReplyToSize (wdw);
      | CAT_TEXT       : 
          lcPtr^ := grin.ActTextSize (wdw);
      | CAT_MSG_DATE   : 
          lcPtr^ := grin.ActDateSize (wdw);
      | CAT_STATUS_DATE: 
          lcPtr^ := grin.ActStatusDateSize (wdw);
      | CAT_GROUP      : 
          lcPtr^ := grin.ActGroupSize (wdw);
      | CAT_ORG_TEXT   : 
          lcPtr^ := grin.ActOrgTextSize (wdw);
      | CAT_IS_OWN     : 
          lcPtr^ := grin.ActIsOwnSize (wdw);
      | CAT_UNIQUE_ID  : 
          lcPtr^ := grin.ActInternalIdSize (wdw);
      | CAT_REAL_SENDER: 
          lcPtr^ := grin.ActSenderSize (wdw);
      | CAT_WDW_NAME   : 
          lcPtr^ := grin.ActWdwNameSize (wdw);
    ELSE
    END;
    IF sizeReq
    THEN
      (* Antwort schicken *)
      Reply (mess);
    ELSE
      (* Datenabfrage *)
      IF (lCard > mPtr^.length)
      OR (mPtr^.buf = NIL)
      THEN
        ReplyError (mess);
        RETURN ;
      END;
      strPtr := mPtr^.buf;
      dateRecPtr := mPtr^.buf;
      CASE mPtr^.catType OF
          CAT_SENDER     : 
            grin.ActualFrom   (wdw, strPtr^);
        | CAT_RNAME      : 
            grin.ActualName   (wdw, strPtr^);
        | CAT_RECEIVER   : 
            grin.ActualReceiver (wdw, strPtr^);
        | CAT_MAUS_ID    : 
            grin.ActualID       (wdw, strPtr^);
        | CAT_MSG_ID     : 
            grin.ActualMId      (wdw, strPtr^);
        | CAT_MAUS_REF   : 
            grin.ActualRefId    (wdw, strPtr^);
        | CAT_REF        : 
            grin.ActualRId      (wdw, strPtr^);
        | CAT_TOPIC      : 
            grin.ActualSubject  (wdw, strPtr^);
        | CAT_BOX        : 
            grin.ActualBox  (wdw, strPtr^);
        | CAT_GATE       : 
            grin.ActualGate     (wdw, strPtr^);
        | CAT_DIST       : 
            grin.ActualDist     (wdw, dist);
            strPtr^[0] := CHR(ORD (dist) + ORD('K'));
            strPtr^[1] := 0C;
        | CAT_MIME       : 
            grin.ActualMime     (wdw, strPtr^);
        | CAT_STATUS     : 
            grin.ActualStatus   (wdw, strPtr^);
        | CAT_FOLLOWUP   : 
            grin.ActualFollowup (wdw, strPtr^);
        | CAT_REPLYTO    : 
            grin.ActualReplyto  (wdw, strPtr^);
        | CAT_TEXT       : 
            grin.ActualText     (wdw, mPtr^.buf);
        | CAT_MSG_DATE   : 
            grin.ActualDate     (wdw, date, time);
            dateRecPtr^.date := ConvertDate.PackDate (date);
            dateRecPtr^.time := ConvertDate.PackTime (time);
        | CAT_STATUS_DATE: 
            grin.ActualStatusDate (wdw, date, time);
            dateRecPtr^.date := ConvertDate.PackDate (date);
            dateRecPtr^.time := ConvertDate.PackTime (time);
        | CAT_GROUP      : 
            grin.ActualGroup  (wdw, strPtr^);
        | CAT_ORG_TEXT   : 
            grin.ActualOrgText (wdw, mPtr^.buf);
        | CAT_IS_OWN     : 
            grin.ActualIsOwn  (wdw, v.bool);
            intPtr := mPtr^.buf;
            IF v.bool
            THEN
              intPtr^ := 1;
            ELSE
              intPtr^ := 0;
            END; 
        | CAT_UNIQUE_ID  : 
            lcPtr := mPtr^.buf;
            grin.ActualInternalId  (wdw, lcPtr^); 
        | CAT_REAL_SENDER: 
            grin.ActualSender (wdw, strPtr^);
        | CAT_WDW_NAME   : 
            grin.ActualWdwName (wdw, strPtr^);
      ELSE
      END;
      Reply (mess);
    END;
  END;
END HandleCatProtoMsg;

PROCEDURE HandleCatMsg(m : ADDRESS);
(* Hier ist die Schaltzentrale fr das CatProtokoll *)
TYPE mType = RECORD
              messId  : CARDINAL;
              apId    : INTEGER;
              over    : INTEGER;
              catType : INTEGER;
              CASE : INTEGER OF 
               0 : data    : ARRAY[0..3] OF INTEGER; |
               1 : d0      : INTEGER;
                   adr     : ADDRESS;
                   d3      : INTEGER; |
              END;
            END;
    grInfoType = RECORD
              gr: CARDINAL;
              nr: CARDINAL;
              am: LONGCARD;
            END;
VAR mess : POINTER TO mType;
    a    : POINTER TO ADDRESS;
    wAnz : INTEGER;
    grInfo: grInfoType;
    b    : POINTER TO grInfoType;
BEGIN
  mess := m;
  CASE mess^.catType OF
    CatGreetings    : SendMessInt(CatMsg, CatGreetings, CatProtoVersion,0,0,0, mess^.apId); 
      (* "Hallo Acc, hier bin ich und das ist meine Versionsnummer!" *)
  | enableWatchDog  : 
                      IF watch < 0 THEN RETURN END;
                      isWatch := TRUE;
                      SendMessInt(CatMsg, CatAffirm,  0,0,0,1,  watch);
  | disableWatchDog : 
                      IF watch < 0 THEN RETURN END;
                      isWatch := FALSE;
                      SendMessInt(CatMsg, CatAffirm,  0,0,0,0,  watch);
  (*$? grin.grinWindowProto:
  | WdwManager      : (* Typ in data[0] *)
                      IF mess^.data[0] = 0 THEN
                        a := ADR(mess^.adr); (* Dort sollen sie hin. *)
                        grin.getAllWinds(a^, mess^.data[3], wAnz); (* Fenster besorgen *)
                        SendAdr(CatMsg, WdwManager, 0, a^, wAnz,  mess^.apId);
                      ELSE
                        (* Knnen wir noch nicht *)
                        SendAdr(CatMsg, WdwManager, mess^.data[0], a^, 0,  mess^.apId);
                      END;
  | CatWinInfo      : a := ADR(mess^.adr); (* Dort sollen sie hin. *)
                      IF grin.getInfo(mess^.data[0], a^, LONGCARD(LONG((mess^.data[3])))) THEN
                        SendAdr(CatMsg, CatWinInfo, mess^.data[0], mess^.adr, mess^.data[3], mess^.apId);
                      ELSE
                        SendAdr(CatMsg, CatWinInfo, mess^.data[0], 0, 0, mess^.apId);
                      END;
  | CatWhatsIn      : b := mess^.adr;
                      IF grin.getBasicInfo(mess^.data[0], grInfo.gr, grInfo.nr, grInfo.am) THEN
                        b^ := grInfo;
                        SendAdr(CatMsg, CatWinInfo, mess^.data[0], LONGCARD(b), 0, mess^.apId);

                        (* SendMessInt(CatMsg, CatWhatsIn, mess^.data[0], INTEGER(gr), INTEGER(nr), INTEGER(am), mess^.apId); *)
                        (* sind eigentlich CARDINAL-Werte, aber die Typkonversion ndert ja nichts.. *)
                      ELSE
                        SendAdr(CatMsg, CatWinInfo, mess^.data[0], 0, 0, mess^.apId);
                        (* SendMessInt(CatMsg, CatWhatsIn, mess^.data[0], 0, 0,0, mess^.apId); *)
                      END;
  | CatPosition     : a := ADR(mess^.adr);
                      SendMessInt(CatMsg, CatPosition, 
                                  mess^.data[0], 
                                  INTEGER(grin.getPos(a^, mess^.data[3], mess^.data[0])), 
                                  0,0,  mess^.apId);
  | CatNewMess      : IF grin.switchMess(mess^.data[0], CARDINAL(mess^.data[2]), CARDINAL(mess^.data[1])) THEN
                        wAnz := 0;
                      ELSE
                        wAnz := 1;
                      END;
                      SendMessInt(CatMsg, CatNewMess,  wAnz,0,0,0, mess^.apId);
  | CatInform       :
  (* CatWinDirty kommt hier nicht an.. :-) *)
  *)
  ELSE
  END;
END HandleCatMsg;


(* TYPE watchType = (pCatAccept, pCatForget, pCatFiltered); *)
PROCEDURE AcceptThis(VAR ptr : data.PtrRecord):watchType;
(* Diese Prozedur wird von parser.i benutzt, um herauszufinden, ob die gerade   *)
(* reinkommende Msg passieren darf                                              *)
VAR mess  : POINTER TO RECORD mess : CARDINAL; w : ARRAY[1..7] OF INTEGER; END;
    event : BITSET; dr : RECORD x,y,w,h : INTEGER END;
    retry,
    noMess : CARDINAL;
BEGIN
  mess := ADR(messBuff);
  (* Block umkopieren in globalen Speicher *)
  memPtr^.dataPtr := ptr;
  IF isWatch THEN
    SendMess(HelloWatchDog, CatAsk, LONGCARD(ADR(memPtr^.dataPtr)), 0,0, watch);
    retry := 0;
    noMess:= 0;
    REPEAT (* Eine kleine Schleife mit TimeOut und Retry-Zhler *)
      event := MagicAES.EvntMulti(
               {MagicAES.MUMESAG, MagicAES.MUTIMER},
                0,{},{}, 0,dr, 0,dr, mess^, 1000,0,
                v.int, v.int, v.bset, v.int, v.bset, v.int, v.char, v.int);

      IF (MagicAES.MUMESAG IN event) &
         (mess^.mess # HelloWatchDog) OR
         ((mess^.w[3] # CatForget) & (mess^.w[3] # CatAccept) & (mess^.w[3] # CatFiltered)) THEN

        (* Standardmessages direkt verarbeiten *)
        IF ~HandleEvent (event, mess^, {}, {}, 0C, 0C, 0, 0, 0, 0)
        THEN
          (* Nicht verarbeitet? Aufbewahren, wird vielleicht spter 
           * mal verarbeitet 
           *)
          MagicAES.ApplWrite(mtAppl.ApplIdent, 16, mess^);
        END;

        INC(noMess);
        IF noMess > 100 THEN
          event  := {MagicAES.MUTIMER};
          noMess := 0;
        END;
      END;

      IF event = {MagicAES.MUTIMER} THEN (* entweder sowieso oder von oben *)
        INC(retry);
        IF    retry > 5 THEN isWatch := FALSE
        ELSIF retry > 2 THEN
          IF 2 = mtAlerts.Alert(1,'[2][CAT:|WatchDog antwortet nicht!][[Nochmal|[Abschalten]') THEN
            isWatch := FALSE
          END;
        END;
      END;

    UNTIL (mess^.mess = HelloWatchDog) &
          ((mess^.w[3] = CatForget) OR (mess^.w[3] = CatAccept) OR (mess^.w[3] = CatFiltered)) OR ~isWatch;
    IF isWatch THEN
      IF    mess^.w[3] = CatAccept THEN
        RETURN pCatAccept
      ELSIF mess^.w[3] = CatForget THEN
        RETURN pCatForget
      ELSIF mess^.w[3] = CatFiltered THEN
        RETURN pCatFiltered
      ELSE
        (* Meckern? *)
        RETURN pCatAccept
      END;
    ELSE
      RETURN pCatAccept
    END;
  ELSE
    RETURN pCatAccept
  END;
END AcceptThis;

PROCEDURE SendTerminate(terminate: BOOLEAN);
(* Nachricht an den WatchDog schicken, da Cat beendet wurde *)
TYPE mType = RECORD
              messId  : CARDINAL;
              apId    : INTEGER;
              over    : CARDINAL;
              catType : CARDINAL;
              length  : LONGCARD;
              buf     : ADDRESS;
            END;
  VAR m : POINTER TO RECORD
            messId : CARDINAL;
            apId   : INTEGER;
            over   : INTEGER;
            r      : ARRAY[3..7] OF INTEGER;
          END;
  VAR mPtr: POINTER TO mType;
      i : CARDINAL;
BEGIN
  IF watch >= 0 THEN
    SendMess(HelloWatchDog, CatTerminate, 0,0,0, watch);
  END;
  (* Und unter MultiTOS auch noch ACC_EXIT verschicken an alle 
   * anderen Applikationen, die sich gemeldet haben
   *)
  IF terminate & multiTask
  THEN
    m := ADR(messBuff);
    WITH m^ DO
      messId := ACC_EXIT;
      apId   := mtAppl.ApplIdent;
      over   := 0;
    END;
    MagicAES.WindUpdate(MagicAES.ENDUPDATE);
    FOR i := 1 TO accCount - 1 DO
      MagicAES.ApplWrite(acc[i], 16, m^);
    END;
    MagicAES.WindUpdate(MagicAES.BEGUPDATE);
  ELSE
    accCount := 1;
  END;

  IF terminate
  THEN
    mPtr := ADR (messBuff);
    WITH mPtr^ DO
      messId  := CatProtoMsg;
      apId    := mtAppl.ApplIdent;
      over    := 0;
      catType := CAT_TERMINATE;
      length  := 0;
      buf     := NIL;
    END;
    MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Um es auch mal abzuschicken.. *)
    FOR i := 1 TO protoCnt - 1 DO 
      MagicAES.ApplWrite (protoAcc[i], 16, mPtr^);
    END;
    MagicAES.WindUpdate(MagicAES.BEGUPDATE);
  END;
END SendTerminate;

PROCEDURE SendProgstart(error : BOOLEAN);
(* dem Acc mitteilen, ob der Programmstart geklappt hat, not yet implementet *)
  VAR val : INTEGER;
BEGIN
  IF error THEN val := 0 ELSE val := 1 END;
  SendMess(VA_PROGSTART, val, 0,0,0, startApId);
END SendProgstart;

PROCEDURE SendAcknowledge (apId : INTEGER; succ: BOOLEAN);
  VAR val : INTEGER;
BEGIN
  IF succ THEN val := 1 ELSE val := 0 END;
  SendMess (ACC_ACK, val, 0,0,0, apId);
END SendAcknowledge;

PROCEDURE SendVaWindopen (apId : INTEGER; error : BOOLEAN);
(* Das ist unbenutzt, ist wohl auch nur in Gemini sinnvoll *)
BEGIN
  IF error 
  THEN
    SendMess(VA_WINDOPEN, 0, 0,0,0, apId);
  ELSE
    SendMess(VA_WINDOPEN, 1, 0,0,0, apId);
  END;
END SendVaWindopen;

PROCEDURE FindAVServer (VAR id: INTEGER): BOOLEAN;
  CONST cAVServer = "AVSERVER=";
  VAR avServer : ARRAY [0..79] OF CHAR;
      i        : INTEGER;
      apType   : INTEGER;
BEGIN
  FOR i := 0 TO 79 DO avServer[i] := 0C; END;
  id := -1;
  IF mtCommand.EnvVar (cAVServer, avServer)
  THEN
    FOR i := 0 TO 7 DO 
      IF avServer[i] = 0C
      THEN
        avServer[i] := ' ';
      END;
    END;
    avServer[8] := 0C;
    id := MagicAES.ApplFind (ADR(avServer));
  END;
  IF id < 0
  THEN
    (* Nach Applikation AVSERVER suchen *)
    MagicStrings.Assign (cAVServer, avServer);
    avServer[8] := 0C;
    id := MagicAES.ApplFind (ADR(avServer));
  END;
  IF id < 0
  THEN
    (* Shell mit appl_search suchen *)
    IF multiTOS OR (magIx & (magIxVer >= $200))
    THEN
      IF MagicAES.ApplSearch (MagicAES.APSHELL, avServer, apType, id)
      THEN
        IF apType # MagicAES.APTAPP
        THEN
          id := -1;
        END;
      END;
    END;
  END;
  RETURN id >= 0;
END FindAVServer;

PROCEDURE SendPathUpdate (REF path: ARRAY OF CHAR);
(* Schickt ein AV_PATH_UPDATE an den AV_SERVER *)
  VAR mess : POINTER TO RECORD
               messId : INTEGER;
               apId   : INTEGER;
               over   : INTEGER;
               fPath              : ADDRESS;
               fName              : ADDRESS;
               e15                : INTEGER;
             END;
     avId : INTEGER;
BEGIN
  IF ~FindAVServer (avId)
  THEN
    (* Wir schicken das an ID 0 *)
    avId := 0;
  END;
  (* Wir schicken nichts an uns selbst *)
  IF (avId = mtAppl.ApplIdent) THEN RETURN END;
  mess := ADR (messBuff);
  MagicStrings.Assign (FileNames.FilePath(path), memPtr^.mPath);
  WITH mess^ DO
    messId := AV_PATH_UPDATE;
    apId   := mtAppl.ApplIdent;
    over   := 0;
    fPath  := ADR (memPtr^.mPath);
    fName  := NIL;
    e15    := 0;
  END;
  MagicAES.ApplWrite(avId, 16, mess^);
END SendPathUpdate;

PROCEDURE SendOpenwind (id : INTEGER; path, name : ADDRESS);
(* einer anderen Applikation sagen, das sie ein Fenster ffnen soll
 * ist fr den Fall, das CAT oder Eddix schon geffnet sind, dann wird 
 * diese Message verschickt.
 *)
  VAR mess : POINTER TO RECORD
               messId : INTEGER;
               apId   : INTEGER;
               over   : INTEGER;
               fPath              : ADDRESS;
               fName              : ADDRESS;
               e15                : INTEGER;
             END;
     fPtr  : Str255Ptr;
BEGIN
  mess := ADR (messBuff);
  (* Filenamen und Pfad kopieren *)
  fPtr := path;
  MagicStrings.Assign (fPtr^, memPtr^.mPath);
  fPtr := name;
  MagicStrings.Assign (fPtr^, memPtr^.mName);
  WITH mess^ DO
    messId := AV_OPENWIND;
    apId   := mtAppl.ApplIdent;
    over   := 0;
    fPath  := ADR (memPtr^.mPath);
    fName  := ADR (memPtr^.mName);
    e15    := 0;
  END;
  MagicAES.ApplWrite(id, 16, mess^);
  MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Um es auch mal abzuschicken.. *)
  MagicAES.WindUpdate(MagicAES.BEGUPDATE);
END SendOpenwind;

PROCEDURE SendHelp (VAR str : ARRAY OF CHAR);
  (* Request Help via PC_HELP Protokoll *)
  VAR accId : INTEGER;
      mess  : POINTER TO RECORD
               messId : INTEGER;
               apId   : INTEGER;
               over   : INTEGER;
               data   : ADDRESS;
               fName  : ADDRESS;
               e15    : INTEGER;
             END;
     helpApp: ARRAY [0..9] OF CHAR; 
     i      : INTEGER;
BEGIN
  mess := ADR(messBuff);
  (* String kopieren in statischen Buffer (igitt) *)
  i := 0;
  WHILE (i < maxBufLen-1) & (str[i] # 0C) DO
    memPtr^.textBuf[i] := str[i];
    INC (i);
  END;
  memPtr^.textBuf [i] := 0C;
  (* Search for PC/TC-HELP Accessory *)
  MagicStrings.Assign (UserBLK.helpServer, helpApp);
  accId := MagicStrings.Length (helpApp);
  (* Extend name with Spaces *)
  WHILE accId < 8 DO 
    helpApp[accId] := ' '; 
    helpApp[accId+1] := ''; 
    INC (accId);
  END;
  accId := MagicAES.ApplFind (ADR(helpApp));
  IF accId >= 0
  THEN
    WITH mess^ DO
      messId := AC_HELP;
      apId   := mtAppl.ApplIdent;
      over   := 0;
      data   := ADR(memPtr^.textBuf);
    END;
    MagicAES.ApplWrite (accId, 16, mess^);
  END;
END SendHelp;

PROCEDURE Send2Filter (to: INTEGER; length: CARDINAL; what: ADDRESS): BOOLEAN;
  (* Request Help via PC_HELP Protokoll *)
  VAR mess  : POINTER TO RECORD
               messId : CARDINAL;
               apId   : INTEGER;
               over   : INTEGER;
               len    : CARDINAL;
               data   : ADDRESS;
               res1,
               res2   : INTEGER;
             END;
     helpApp: ARRAY [0..9] OF CHAR; 
     i      : INTEGER;
BEGIN
  mess := ADR(messBuff);
  (* Daten kopieren in globalen Buffer *)
  IF length > memDescrSize THEN RETURN FALSE END;
  Block.Copy (what, length, memPtr);
  WITH mess^ DO
    messId := Cat2FiltMsg;
    apId   := mtAppl.ApplIdent;
    over   := 0;
    len    := length;
    data   := ADR(memPtr^.allInOne);
    res1   := 0;
    res2   := 0;
  END;
  MagicAES.ApplWrite (to, 16, mess^);
  MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Um es auch mal abzuschicken.. *)
  MagicAES.WindUpdate(MagicAES.BEGUPDATE);
  RETURN TRUE;
END Send2Filter;

PROCEDURE SendExtendedHelp (VAR str : ARRAY OF CHAR; VAR text: ARRAY OF CHAR): BOOLEAN;
  (* Request Help via VA_START from ST-GUIDE *)
  CONST theGuide    = 'ST-GUIDE';
  
  VAR accId : INTEGER;
      vaMess: POINTER TO RECORD
               messId : INTEGER;
               apId   : INTEGER;
               over   : INTEGER;
               cmd    : ADDRESS;
               res    : LONGINT;
               res2   : INTEGER;
             END;
      helpApp: ARRAY [0..9] OF CHAR; 
BEGIN
  (* Search for ST-Guide Accessory/Applikation *)
  MagicStrings.Assign (theGuide, helpApp);
  accId := MagicAES.ApplFind (ADR(helpApp));
  IF accId < 0
  THEN
    (* ST-Guide nicht gefunden, interne Hilfe aufrufen *)
    RETURN FALSE
  END;
  
  (* Kommandozeile fr VA_START aufbauen *)
  (* hier kann ich das per Assign machen, da die internen
   * Hilfeberschriften nie den statischen Buffer berschreiten 
   * werden 
   *)
  MagicStrings.Assign ("*:\", memPtr^.allInOne);
  MagicStrings.Append (text, memPtr^.allInOne);
  MagicStrings.Append (" ", memPtr^.allInOne);
  MagicStrings.Append (str, memPtr^.allInOne);
  
  vaMess := ADR(messBuff);
  WITH vaMess^ DO
    messId := VA_START;
    apId   := mtAppl.ApplIdent;
    over   := 0;
    cmd    := ADR(memPtr^.allInOne);
    res    := 0;
    res2   := 0;
  END;
  MagicAES.ApplWrite (accId, 16, vaMess^);

  MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Um es auch mal abzuschicken.. *)
  MagicAES.WindUpdate(MagicAES.BEGUPDATE);

  RETURN TRUE;
END SendExtendedHelp;

PROCEDURE SendPrintFile (REF path, name: ARRAY OF CHAR; withEffects: BOOLEAN): BOOLEAN;
(* Druckt die angegebene Datei ber Idealist bzw. das CALCLOCK-Protokoll
 *)
  CONST ideaList    = 'IDEALIST';
        ideaAcc     = 'ILIST   ';
  
  VAR accId : INTEGER;
      vaMess: POINTER TO RECORD
               messId : INTEGER;
               apId   : INTEGER;
               over   : INTEGER;
               len    : INTEGER;
               ptr    : ADDRESS;
               copies : INTEGER;
               bits   : BITSET;
             END;
      messy : POINTER TO RECORD
               messId : INTEGER;
               apId   : INTEGER;
               over   : INTEGER;
               wdw    : INTEGER;
               magic  : ADDRESS;
               fName  : ADDRESS;
             END;
      helpApp: ARRAY [0..9] OF CHAR; 
      i      : INTEGER;
      event  : BITSET; 
      dr : RECORD x,y,w,h : INTEGER END;
      retry,
      noMess : CARDINAL;
BEGIN
  (* Erstmal nach Idealist suchen *)
  MagicStrings.Assign (ideaList, helpApp);
  accId := MagicAES.ApplFind (ADR(helpApp));
  IF accId < 0
  THEN
    (* Idealist nicht gefunden, IList suchen *)
    MagicStrings.Assign (ideaAcc, helpApp);
    accId := MagicAES.ApplFind (ADR(helpApp));
    IF accId < 0
    THEN
      (* Idealist nicht gefunden, IList suchen *)
      RETURN FALSE
    END;
  END;
  
  (* Idealist gefunden, Calclock-Message aufbauen *)
  MagicStrings.Assign (path, memPtr^.allInOne);
  MagicStrings.Append (name, memPtr^.allInOne);

  vaMess := ADR(messBuff);
  WITH vaMess^ DO
    messId := CALCLOCK;
    apId   := mtAppl.ApplIdent;
    over   := 0;
    len    := LENGTH (memPtr^.allInOne);
    ptr    := ADR(memPtr^.allInOne);
    copies := 1;
    bits   := {};
    INCL (bits, 0); (* Bit 0 setzen, tenmporre Datei lschen *)
    IF withEffects
    THEN
      INCL (bits, 8);
    ELSE
      INCL (bits, 9);
    END;
  END;

  MagicAES.ApplWrite (accId, 16, vaMess^);

  MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Um es auch mal abzuschicken.. *)

  (* Wait for Event *)

  messy := ADR(messBuff);
  retry := 0;
  REPEAT (* Eine kleine Schleife mit TimeOut und Retry-Zhler *)
    event := MagicAES.EvntMulti(
             {MagicAES.MUMESAG, MagicAES.MUTIMER},
              0,{},{}, 0,dr, 0,dr, messy^, 250,0,
              v.int, v.int, v.bset, v.int, v.bset, v.int, v.char, v.int);

    IF (MagicAES.MUMESAG IN event) &
       (messy^.messId # CALCLOCK_OK) 
    THEN
      MagicAES.WindUpdate(MagicAES.BEGUPDATE);
      (* Standardmessages direkt verarbeiten *)
      IF ~HandleEvent (event, messy^, {}, {}, 0C, 0C, 0, 0, 0, 0)
      THEN
        (* Nicht verarbeitet? Aufbewahren, wird vielleicht spter 
         * mal verarbeitet 
         *)
        MagicAES.ApplWrite(mtAppl.ApplIdent, 16, messy^);
      END;
      MagicAES.WindUpdate(MagicAES.ENDUPDATE);

      INC(noMess);
      IF noMess > 20 THEN
        event  := {MagicAES.MUTIMER};
      END;
    END;
    
    IF (MagicAES.MUTIMER IN event)
    THEN
      INC (retry);
    END;
  UNTIL (messy^.messId = CALCLOCK_OK) OR (retry > 3);
  MagicAES.WindUpdate(MagicAES.BEGUPDATE);

  RETURN messy^.messId = CALCLOCK_OK;
END SendPrintFile;

PROCEDURE IsProtokoll(mess : ADDRESS;
                     VAR kReturn : INTEGER; VAR mokState : BITSET;
                     VAR pName, pCmd : Str255Ptr):React;

  PROCEDURE SendToWatchDog();
  (* WatchDog-Initialisierungsmsg von Cat *)
  VAR m : POINTER TO RECORD
            messId : CARDINAL;
            apId   : INTEGER;
            over   : INTEGER;
            r      : ARRAY[3..7] OF INTEGER;
          END;
  BEGIN
    m := ADR(messBuff);
    WITH m^ DO
      messId := HelloWatchDog;
      apId   := mtAppl.ApplIdent;
      over   := 0;
      r[3]   := CatGreetings;
      r[4]   := CatProtoVersion;
      r[5]   := 0; r[6]   := 0; r[7] := 0;
    END;
    MagicAES.ApplWrite(watch, 16, m^);
    MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Abschicken *)
    MagicAES.WindUpdate(MagicAES.BEGUPDATE);
    isWatch  := FALSE;
  END SendToWatchDog;

  PROCEDURE CheckName(VAR str : ARRAY OF CHAR; id : INTEGER);
  (* Ist das ein WatchDog? Ja -> Begrung losschicken *)
  VAR i : INTEGER;
  BEGIN
    IF MagicStrings.Equal(str, watchDog) THEN
      watch := id;
      IF UserBLK.watchBox THEN
        i := mtAlerts.Alert(1, '[1][CAT:|WatchDog found][[Jippie]');
      END;
      SendToWatchDog();
    END;
  END CheckName;

  PROCEDURE CheckNameXACC(mess : ADDRESS);
  VAR access : POINTER TO RECORD
                 messId : INTEGER;
                 apId   : INTEGER;
                 over   : INTEGER;
                 version, protoStep : CHAR;
                 name               : Str255Ptr;
                 menuId             : INTEGER;
                 accId              : INTEGER;
               END;
  BEGIN
    access := mess;
    IF (access^.name > ADDRESS($F000))
    THEN
      CheckName(access^.name^, access^.apId);
    END;
  END CheckNameXACC;

  PROCEDURE CheckNameVA(mess : ADDRESS);
  VAR access : POINTER TO RECORD
                 messId : INTEGER;
                 apId   : INTEGER;
                 over   : INTEGER;
                 b11, b12, b13      : BITSET;
                 accName            : Str255Ptr;
               END;
  BEGIN
    access := mess;
    IF (access^.accName > ADDRESS($F000))
    THEN
      CheckName(access^.accName^, access^.apId);
    END;
  END CheckNameVA;

  PROCEDURE accInit(mess : ADDRESS);
  (* XAcc-Protokoll, accInit behandeln *)
  VAR i : CARDINAL;
      access : POINTER TO RECORD
                 messId : INTEGER;
                 apId   : INTEGER;
                 over   : INTEGER;
                 version, protoStep : CHAR;
                 name               : Str255Ptr;
                 menuId             : INTEGER;
                 accId              : INTEGER
               END;

    PROCEDURE checkId (id : INTEGER): BOOLEAN;
    (* Prft, ob schon ein Prozess mit der ID registriert ist *)
      VAR i : CARDINAL;
    BEGIN
      FOR i := 1 TO accCount - 1 DO 
        IF acc[i] = id THEN RETURN FALSE END;
      END;
      RETURN id # mtAppl.ApplIdent;
    END checkId;
    
  BEGIN
    access := mess;
    IF multiTask
    THEN
      IF (accCount < maxAcc) & checkId (access^.apId)
      THEN
        acc[accCount]     := access^.apId; (* ID des Acc, das sich gerade anmeldet *)
        IF access^.messId # ACC_ACC
        THEN
          access^.messId    := ACC_ACC;
          access^.apId      := mtAppl.ApplIdent;
          access^.version   := CHR(20H);
          access^.protoStep := 1C;
          access^.name      := ADR(memPtr^.apName);
          access^.menuId    := 0;
          access^.accId     := 0;
  
          MagicAES.ApplWrite(acc[accCount], 16, access^);
          MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Abschicken *)
          MagicAES.WindUpdate(MagicAES.BEGUPDATE);
        END;
        INC(accCount);
      END;
    ELSE (* not MultiTOS *)
      IF (accCount < maxAcc) & checkId (access^.apId) THEN
  
        acc[accCount]     := access^.apId; (* ID des Acc, das sich gerade anmeldet *)
        access^.messId    := ACC_ACC;
        access^.accId     := access^.apId;
        access^.apId      := mtAppl.ApplIdent;
        access^.over      := 0;
        (* Die anderen Daten stimmen zu diesem Zeitpunkt schon.. *)
  
        FOR i := 1 TO accCount-1 DO
          MagicAES.ApplWrite(acc[i], 16, access^);
          MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Abschicken *)
          MagicAES.WindUpdate(MagicAES.BEGUPDATE);
        END; (* Alle anderen Accs benachrichtigen *)
                                   
        access^.messId    := ACC_ID;
        access^.version   := CHR(20H);
        access^.protoStep := 1C;
        access^.name      := ADR(memPtr^.apName);
        access^.menuId    := 0;
        access^.accId     := 0;
  
        MagicAES.ApplWrite(acc[accCount], 16, access^);
        INC(accCount);
  
        MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Abschicken *)
        MagicAES.WindUpdate(MagicAES.BEGUPDATE);
      END; 
    END;
  END accInit;

  PROCEDURE SendVaProtostatus(mess : ADDRESS);
  VAR id : INTEGER;
      access : POINTER TO RECORD
                 messId : INTEGER;
                 apId   : INTEGER;
                 over   : INTEGER;
                 b21, b22, b23      : BITSET;
                 prgName            : ADDRESS;
               END;
  BEGIN
    access := mess;
    id             := access^.apId;
    access^.messId := VA_PROTOSTATUS;
    access^.apId   := mtAppl.ApplIdent;
    access^.over   := 0;
    access^.b21    := {xAV_SENDKEY, xAV_ACCWINDOPEN, xAV_STARTPROG, xAV_OPENWIND};
    access^.b22    := {};
    access^.b23    := {};
    access^.prgName:= ADR(memPtr^.apName);
    MagicAES.ApplWrite(id, 16, access^);
    MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Abschicken *)
    MagicAES.WindUpdate(MagicAES.BEGUPDATE);
  END SendVaProtostatus;
  
  PROCEDURE RemoveAcc (apId : INTEGER);
    VAR i : CARDINAL;
  BEGIN
    i := 1;
    WHILE (i < accCount) & (acc[i] # apId) DO INC (i) END;
    IF i < accCount THEN
      FOR i := i TO accCount - 2 DO
        acc[i] := acc[i+1];
      END;
      DEC (accCount);
    END;
  END RemoveAcc;

VAR access : POINTER TO RECORD
               messId : INTEGER;
               apId   : INTEGER;
               over   : INTEGER;
               CASE :CARDINAL OF
               (* AV_SENDKEY *)
                 1 : mokState       : INTEGER;
                     kReturn        : INTEGER;
                     e11, e12, e13  : INTEGER|
               (* AV_STARTPROG *)
                 2 : pName, pCmd    : ADDRESS;
                     e14            : INTEGER|
               (* AV_OPENWIND  *)
                 3 : fPath          : ADDRESS;
                     fName          : ADDRESS;
                     e15            : INTEGER |
                 4 : i              : ARRAY[0..4] OF INTEGER|
                 5 : e16            : INTEGER;
                     text           : ADDRESS; |
               END;
             END;

(* TYPE React        = (rNoproto, rNone, rChar, rText); *)
BEGIN
  access := mess;
  (* Zunchst einmal die Initialisierungsmsgs fr XAcc und VA *)
  IF    access^.messId = ACC_ID   THEN             (* hier behandelt, WatchDog suchen *)
    CheckNameXACC(mess);
    accInit(mess);      (* Xacc-Protokoll behandeln *)
    RETURN rNone;
  (* AccOpen und AccClose werden einfach berhrt! *)
  ELSIF (access^.messId = ACC_ACC) & multiTask THEN  (* hier behandelt, WatchDog suchen *)
    CheckNameXACC(mess);
    accInit (mess);
    RETURN rNone;
  (* AccOpen und AccClose werden einfach berhrt! *)
  ELSIF access^.messId = AV_PROTOKOLL THEN         (* hier behandelt, WatchDog suchen *)
    CheckNameVA(mess);
    SendVaProtostatus(mess);                  (* .. und dem Acc sagen, was wir knnen *)
    RETURN rNone;
  
  ELSIF access^.messId = ACC_EXIT THEN                    (* XACC Acc_Exit *)
    (* remove Accessorie from acc list *)
    RemoveAcc (access^.apId);
    RETURN rNone;
  ELSIF access^.messId = ACC_KEY THEN                     (* Tastendruck durchreichen *)
    (* Genau umgedreht zu AV_SENDKEY *)
    kReturn := access^.mokState;
    mokState:= BITSET(access^.kReturn);
    SendAcknowledge (access^.apId, TRUE);
    RETURN rChar;
  ELSIF access^.messId = ACC_TEXT THEN
    (* get text via XACC *)
    kReturn := access^.apId;    (* apId merken fr Acknowledge *)
    pName   := access^.text;    (* Pointer auf Text *)
    RETURN rText;
    
  (* AV-Msgs, die direkt hier behandelt werden *)
  ELSIF access^.messId = AV_SENDKEY   THEN                (* Tastendruck durchreichen *)
    kReturn := access^.kReturn;
    mokState:= BITSET(access^.mokState);
    RETURN rChar;
  ELSIF access^.messId = AV_ACCWINDOPEN  THEN  (* AccFenster auf -> rein in die Queue *)
    NewWindowIsTop(access^.i[0], access^.apId);
    RETURN rNone;
  ELSIF access^.messId = AV_ACCWINDCLOSED THEN (* AccFenster zu -> raus aus der Queue *)
    WindowIsClosed(access^.i[0]);
    RETURN rNone;
  ELSIF access^.messId = AV_STARTPROG THEN   (* Acc sagt: Folgendes Programm starten! *)
    pName := access^.pName; pCmd := access^.pCmd;
    startApId := access^.apId;
    RETURN rStart;
  ELSIF access^.messId = AV_OPENWIND THEN                              (* Woasn doas? *)
    pName := access^.fPath; pCmd := access^.fName;
    SendVaWindopen(access^.apId, FALSE);
    RETURN rOpen;
  ELSIF access^.messId = VA_START THEN
    (* Unter Mag!X: Neue Dokumente auf CAT gezogen oder so *)
    pName := NIL; pCmd := access^.fPath;
    (* Besttigung gibt es keine *)
    RETURN rOpen;

  ELSIF CARDINAL(access^.messId) = CatMsg THEN                           (* CatMsg *)
    HandleCatMsg(mess);
    RETURN rNone;
  ELSIF CARDINAL(access^.messId) = ExtCatMsg THEN                        (* CatMsg *)
    kReturn := access^.mokState;
    RETURN rExtCatMsg;
  ELSIF CARDINAL(access^.messId) = CatProtoMsg THEN                      (* neues CAT-Protokoll *)
    (* wird komplett hier behandelt: *)
    HandleCatProtoMsg (mess);
    RETURN rNone;
  ELSE (* Kein bekanntes Protokoll *)
    RETURN rNoproto;
  END;
  RETURN rNoproto;
END IsProtokoll;

PROCEDURE ProtoInit ();
(* XACC-Initialisierung unter MTOS *)
  VAR i : INTEGER;
      access : POINTER TO RECORD
                 messId : INTEGER;
                 apId   : INTEGER;
                 over   : INTEGER;
                 version, protoStep : CHAR;
                 name               : Str255Ptr;
                 menuId             : INTEGER;
                 accId              : INTEGER
               END;
     apName : String255;
     apMode,
     apType,
     apId   : INTEGER;
BEGIN
  IF multiTOS OR (magIx & (magIxVer >= $200))
  THEN
    access := ADR(messBuff);
    apMode := MagicAES.APFIRST;
    WHILE MagicAES.ApplSearch (apMode, apName, apType, apId) DO
      apMode := MagicAES.APNEXT;
      IF (apType # 1) & (apId # mtAppl.ApplIdent)
      THEN
        access^.messId    := ACC_ID;
        access^.apId      := mtAppl.ApplIdent;
        access^.version   := CHR(20H);
        access^.protoStep := 1C;
        access^.name      := ADR(memPtr^.apName);
        access^.menuId    := 0;
        access^.accId     := 0;
    
        MagicAES.ApplWrite(apId, 16, access^);
    
      END;
    END;
  ELSIF magIx & (magIxVer < $200)
  THEN
    FOR i := 0 TO MagicAES.AESGlobal.apCount - 1 DO
      apName[0] := '?';
      apName[1] := 0C;
      apName[2] := CHR(i);
      apName[3] := 0C;
      apId := MagicAES.ApplFind (ADR(apName));
      IF (apId # 0) & (apId # mtAppl.ApplIdent)
      THEN 
        access^.messId    := ACC_ID;
        access^.apId      := mtAppl.ApplIdent;
        access^.version   := CHR(20H);
        access^.protoStep := 1C;
        access^.name      := ADR(memPtr^.apName);
        access^.menuId    := 0;
        access^.accId     := 0;
      
        MagicAES.ApplWrite(apId, 16, access^);
      END;
    END;
  END;
END ProtoInit;

PROCEDURE isWatchDog(VAR installed, active : BOOLEAN);
(* Ist der WatchDog angemeldet und aktiv? *)
BEGIN
  installed := watch >= 0;
  active    := isWatch;
END isWatchDog;

VAR hasMxAlloc  : BOOLEAN;
    i           : CARDINAL;
    tPtr        : Str255Ptr;

BEGIN
  isWatch  := FALSE;
  watch    := -1;
  accCount := 1;
  protoCnt := 1;
  hasMxAlloc := MagicDOS.Mxalloc ($FFFFFFFF,3) # ADDRESS(-32L);
  IF multiTask & hasMxAlloc
  THEN
    (* Mxalloc: Readable, TT-RAM preferred *)
    memPtr := MagicDOS.Mxalloc (TSIZE (memDescr), $43);
  ELSE
    ALLOCATE (memPtr, TSIZE (memDescr));
  END;
  IF memPtr = NIL THEN HALT END;
  tPtr := CADR (appName);
  FOR i := 0 TO 49 DO
    memPtr^.apName[i]   := tPtr^[i];
  END;
END Protokoll.
