IMPLEMENTATION MODULE QuickSort;

FROM SYSTEM IMPORT ADDRESS, LOC, ADR, ASSEMBLER;

FROM Storage IMPORT ALLOCATE, DEALLOCATE;

IMPORT Lists;

IMPORT Block;

(*$H+*)

TYPE    copyProc    = PROCEDURE ((* from: *)  ADDRESS, 
                                 (* count: *) LONGCARD, 
                                 (* to:   *)  ADDRESS);

VAR memCopy:    copyProc;

(*$L-*)
PROCEDURE copyByte (from: ADDRESS; count: LONGCARD; to: ADDRESS);
BEGIN
  ASSEMBLER
    MOVE.L  -(A3),A1
    SUBA.W  #4,A3
    MOVE.L  -(A3),A0
    MOVE.B  (A0),(A1)
  END;
END copyByte;

PROCEDURE copyWord (from: ADDRESS; count: LONGCARD; to: ADDRESS);
BEGIN
  ASSEMBLER
    MOVE.L  -(A3),A1
    SUBA.W  #4,A3
    MOVE.L  -(A3),A0
    MOVE.W  (A0),(A1)
  END;
END copyWord;

PROCEDURE copyLong (from: ADDRESS; count: LONGCARD; to: ADDRESS);
BEGIN
  ASSEMBLER
    MOVE.L  -(A3),A1
    SUBA.W  #4,A3
    MOVE.L  -(A3),A0
    MOVE.L  (A0),(A1)
  END;
END copyLong;

PROCEDURE copy6 (from: ADDRESS; count: LONGCARD; to: ADDRESS);
BEGIN
  ASSEMBLER
    MOVE.L  -(A3),A1
    SUBA.W  #4,A3
    MOVE.L  -(A3),A0
    MOVE.L  (A0)+,(A1)+
    MOVE.W  (A0),(A1)
  END;
END copy6;

PROCEDURE copyDLong (from: ADDRESS; count: LONGCARD; to: ADDRESS);
BEGIN
  ASSEMBLER
    MOVE.L  -(A3),A1
    SUBA.W  #4,A3
    MOVE.L  -(A3),A0
    MOVE.L  (A0)+,(A1)+
    MOVE.L  (A0),(A1)
  END;
END copyDLong;

PROCEDURE copy10 (from: ADDRESS; count: LONGCARD; to: ADDRESS);
BEGIN
  ASSEMBLER
    MOVE.L  -(A3),A1
    SUBA.W  #4,A3
    MOVE.L  -(A3),A0
    MOVE.L  (A0)+,(A1)+
    MOVE.L  (A0)+,(A1)+
    MOVE.W  (A0),(A1)
  END;
END copy10;

PROCEDURE copy12 (from: ADDRESS; count: LONGCARD; to: ADDRESS);
BEGIN
  ASSEMBLER
    MOVE.L  -(A3),A1
    SUBA.W  #4,A3
    MOVE.L  -(A3),A0
    MOVE.L  (A0)+,(A1)+
    MOVE.L  (A0)+,(A1)+
    MOVE.L  (A0),(A1)
  END;
END copy12;
(*$L=*)

VAR  breakJN: BOOLEAN;

PROCEDURE sortIt (low, high : LONGINT; VAR ptrelem : LONGARRAY OF LOC;
                  comp : compProc; entrySize : LONGINT;
                  break: breakProc): BOOLEAN;

  CONST   Limit = 20;   (* Ab Limit wird auf InsertSort umgeschaltet. *)
  
  VAR tmpBuf : ADDRESS;
      valueBuf : ADDRESS;

    PROCEDURE getIdx (idx: LONGINT) : ADDRESS;
    BEGIN
      RETURN (ADR(ptrelem)+ADDRESS (idx*entrySize));
    END getIdx;

    PROCEDURE assign (i, j: LONGINT);
      (* ptrelem[i] := ptrelem [j] *)
    BEGIN
      Block.Copy (getIdx (j), entrySize, getIdx(i));
    END assign; 
    
    PROCEDURE Push (i : LONGINT);
    BEGIN
      Block.Copy (getIdx (i), entrySize, tmpBuf);
    END Push;
    
    PROCEDURE PushValue (i : LONGINT);
    BEGIN
      Block.Copy (getIdx (i), entrySize, valueBuf);
    END PushValue;
    
    PROCEDURE Pop (j: LONGINT);
    BEGIN
      Block.Copy (tmpBuf, entrySize, getIdx (j));
    END Pop;
    
    PROCEDURE Swap (i, j: LONGINT); 
    BEGIN
      Block.Copy (getIdx (i), entrySize, tmpBuf);
      Block.Copy (getIdx (j), entrySize, getIdx(i));
      Block.Copy (tmpBuf, entrySize, getIdx (j));
    END Swap;

    PROCEDURE InsertSort (l, r : LONGINT);
      VAR  i, j  : LONGINT;
           b     : BOOLEAN;

    BEGIN
      FOR i := (l + 1) TO r DO
        Push (i);
        j := i - 1;
        b := comp (tmpBuf,getIdx(j));

        WHILE b DO
          assign (j+1, j);
          DEC (j);
          b := (j >= l);

          IF b THEN
            b := comp (tmpBuf, getIdx(j));
          END;
        END;
        Pop (j+1);
        (* ptrelem[j + 1] := a; *)
      END;
    END InsertSort;
    
    FORWARD ChooseSort (l, r : LONGINT);

    PROCEDURE QuickSort (l, r : LONGINT);
      VAR i, j : LONGINT;
          p    : ADDRESS;
    BEGIN
      i := l;
      j := r;
      PushValue ((l+r) DIV 2);
      REPEAT
        WHILE comp (getIdx(i), valueBuf) DO INC (i); END;
        WHILE comp (valueBuf, getIdx(j)) DO DEC (j); END;
        IF (i <= j) 
        THEN
          Swap (i, j);
          INC (i);
          DEC (j);
        END;
      UNTIL (i > j);
      IF (l < j) THEN ChooseSort (l, j); END;
      IF breakJN THEN RETURN END;
      IF (i < r) THEN ChooseSort (i, r); END;
    END QuickSort;

    PROCEDURE ChooseSort (l, r : LONGINT);
    BEGIN
      IF breakJN THEN RETURN END;
      IF ((r - l) > Limit) THEN
        QuickSort (l, r);
        IF break() THEN 
          breakJN := TRUE;
          RETURN 
        END;
      ELSE
        InsertSort (l, r);
        IF break() THEN 
          breakJN := TRUE;
          RETURN 
        END;
      END;
    END ChooseSort;

  BEGIN
    ALLOCATE (tmpBuf, entrySize);
    IF tmpBuf = NIL THEN RETURN FALSE END;
    ALLOCATE (valueBuf, entrySize);
    IF valueBuf = NIL THEN 
      DEALLOCATE (tmpBuf, 0);
      RETURN FALSE 
    END;
    breakJN := FALSE;
    CASE entrySize OF
       1    : memCopy := copyByte; |
       2    : memCopy := copyWord; |
       4    : memCopy := copyLong; |
       6    : memCopy := copy6;    |
       8    : memCopy := copyDLong; |
       10   : memCopy := copy10; |
       12   : memCopy := copy12; |
    ELSE
      memCopy := Block.Copy;
    END;
    ChooseSort (low, high);
    DEALLOCATE (valueBuf, 0);
    DEALLOCATE (tmpBuf, 0);
    RETURN TRUE;
  END sortIt;

(* Interne Prozedur fr SortFonts *)
TYPE PCarrier = POINTER TO Carrier;
     Carrier  = RECORD
                  next, prev: PCarrier;
                  data: ADDRESS
                END;
     MList = RECORD
               current: PCarrier;
               root   : PCarrier;
               user   : PCarrier
             END;

PROCEDURE ZipSort (VAR list: MList; nodes: CARDINAL; 
                   comp : compProc; break: breakProc);
(* Achtung: Diese Prozedur geht direkt auf die internen Datenstrukturen
 * von Lists drauf. 
 *)
  VAR
    list1, list2: MList;
    I           : CARDINAL;
    result,last1: MList;

BEGIN
  IF break() THEN 
    breakJN := TRUE;
    RETURN 
  END;
  
  IF nodes<=1 THEN RETURN END; (* Eine Liste aus einem Element ist schon sortiert *)

  (* Liste in etwa gleich groe Teile aufteilen: *)
  (* list1 erhlt nodes DIV 2; list2 den rest *)
  list1 := list;
  list2 := list;
  FOR I := 1 TO (nodes DIV 2) DO
    last1 := list2;
    list2.current := list2.current^.next
  END;
  last1.current^.next := NIL; (* Listen auseinanderschneiden *)
  list2.root := list2.current;
  (* List1 zurcksetzen *)
  list1.current := list1.root;
  (*WriteList (list1);
    WriteList (list2); *)
  (* list1 und list2 enthalten garantiert mindestens 1 Element *)
  (* Einzellisten Sortieren *)
  ZipSort (list1,nodes DIV 2, comp, break);
  ZipSort (list2,nodes - (nodes DIV 2), comp, break);

  (* Jetzt kommts: *)
  (* Erstes Listenelement festlegen: *)
  (* Die folgende IF-Bedingung macht das selbe wie die Zeile
   * >IF Strings.Compare (list2^.info.name, list1^.info.name)=Strings.less THEN
   * Ist aber statistisch wesentlich schneller.
   * Die IF-Bedingung bentigt sonst mehr Rechenzeit als alle anderen
   * Befehle zusammen.
   *)
  IF comp(list2.current^.data, list1.current^.data)
  THEN
    list := list2;
    list2.current := list2.current^.next;
    list2.root := list2.current;
  ELSE
    list := list1; 
    list1.current := list1.current^.next;
    list1.root := list1.current;
  END;

  LOOP
    IF (list1.current=NIL) OR (list1.current=list1.root) OR (list1.current^.data=NIL) THEN
      list.current^.next := list2.current; EXIT
    END;
    IF (list2.current=NIL) OR (list2.current=list2.root)OR (list2.current^.data=NIL) THEN
      list.current^.next := list1.current; EXIT
    END;
    IF comp(list2.current^.data, list1.current^.data)
    THEN
      list.current^.next := list2.current; 
      list2.current := list2.current^.next;
    ELSE
      list.current^.next := list1.current; 
      list1.current := list1.current^.next;
    END;
    list.current := list.current^.next;
  END;
END ZipSort;


(* Gc: *)
PROCEDURE ListSort (list: Lists.List; comp : compProc; break: breakProc): BOOLEAN;
(* Sortiert eine Liste mittels Zip-Sort (c) by Gerd Castan
 *)
  VAR prev, 
      oRoot,
      entry : PCarrier;
      rList : MList;
BEGIN
  breakJN := FALSE;
  Lists.ResetList (list);
  rList := MList (list);
  rList.current := rList.root^.next;
  oRoot := rList.root;
  rList.root := rList.current;
  (* Liste als einfach verkettete Liste interpretieren und sortieren: *)
  ZipSort (rList, Lists.NoOfEntries (list), comp, break);

  IF breakJN THEN RETURN FALSE; END;
  
  (* Zerstrte Rckwrtsverkettung restaurieren: *)
  rList.current := rList.root^.next;
  rList.root := oRoot;
  rList.root^.next := rList.current;
  prev := NIL;
  entry := rList.current;
  WHILE entry#NIL DO
    entry^.prev := prev;
    prev := entry;
    entry := entry^.next
  END;
  list := Lists.List (rList);
  RETURN TRUE;
END ListSort;

PROCEDURE noBreak(): BOOLEAN;
BEGIN
  RETURN FALSE
END noBreak;

BEGIN
  memCopy := Block.Copy;
END QuickSort.
