DEFINITION MODULE WinDials;

FROM SYSTEM IMPORT ADDRESS, LOC;

FROM GrafBase IMPORT Rectangle;


(*$Z+*)
(*$H+*)
TYPE    wdSetProc   = PROCEDURE ( (* tree: *)    ADDRESS, 
                                  (* private: *) ADDRESS,
                                  (* setJN:   *) BOOLEAN,
                                  (* exitBut: *) INTEGER);
        (* Eine Prozedur dieses Typs wird einmal
         * bei der Initialisierung aufgerufen mit
         * setJN = TRUE und einmal beim Verlassen
         * mit setJN = FALSE. Beim Verlassen ist der
         * Exitbutton in exitBut angegeben. Dieser ist
         * -1, falls der Dialog ber den Closer oder 
         * beim Verlassen geschlossen wird.
         *)
        wdButtonProc= PROCEDURE ( (* tree: *)    ADDRESS, 
                                  (* private: *) ADDRESS,
                                  (* button: *)  INTEGER,
                                  (* mx:     *)  INTEGER,
                                  (* my:     *)  INTEGER,
                                  (* kstate: *)  BITSET,
                                  (* clicks: *)  INTEGER): BOOLEAN;
        (* Wird bei jedem Exitbutton aufgerufen.
         * Wird mit gesetztem WindUpdate aufgerufen 
         * Gibt zurck, ob der Dialog beendet werden soll 
         *)
        wdDrawProc  = PROCEDURE ( (* tree: *)    ADDRESS, 
                                  (* private: *) ADDRESS,
                                  (* clip: *)    Rectangle);
        (* Private Redrawprozedur fr Objekte im Dialog,
         * die nicht automatisch bei DialDraw mitgezeichnet 
         * werden
         *)

PROCEDURE defDraw (tree, private: ADDRESS; clip: Rectangle);

PROCEDURE defSet (tree, private: ADDRESS; set: BOOLEAN; exitBut: INTEGER);

PROCEDURE WinDialDraw (tree: ADDRESS; entry, dpth: INTEGER;
                       clip: ARRAY OF LOC; clipping: BOOLEAN);
(* Zeichnet ein oder mehrere Objekte in einem Fensterdialog.
 * Wenn der Dialog nicht gefunden wird, wird nichts gezeichnet
 *)

PROCEDURE OpenWinDial (tree: ADDRESS;
                       modal: BOOLEAN;
                       setValues: wdSetProc;
                       getValues: wdSetProc;
                       button   : wdButtonProc;
                       draw     : wdDrawProc;
                       REF title    : ARRAY OF CHAR;
                       private: ADDRESS): BOOLEAN;
(* Erzeugt einen Dialog im Fenster. 
 * Wenn der Dialog schon in einem Fenster ist, dann
 * wird er getoppt.
 * Ansonsten wird ein neues Fenster angelegt, setValues aufgerufen,
 * um die Werte zu setzen, und dann wird das Fenster geffnet und
 * angezeigt. Falls der Dialog modal ist, wird das Menu
 * disablet.
 * setValues wird aufgerufen, wenn der Dialog initialisiert wird.
 * getValues wird aufgerufen, wenn der Dialog beendet wird.
 * button wird jedesmal dann aufgerufen, wenn ein Button angeklickt 
 * wurde, der ein Exit-Flag gesetzt hat
 * draw ist eine zustzliche Zeichenprozedur, die aufgerufen wird,
 * wenn Redraws notwendig sind. Man kann auch NILDRAW bergeben
 * Rckgabewert:
 * False: Kein Fenster mehr frei
 *)

PROCEDURE CloseWinDial (tree: ADDRESS);
(* schliet einen Fensterdialog. getValues wird aufgerufen.
 *)

PROCEDURE WinDialHandleEvents ();
(* Durchluft die Eventloop, bis das Topwindow mit dem aktuellen Dialog
 * geschlossen wird
 *)

PROCEDURE WinDialDo (tree: ADDRESS; edit: INTEGER): INTEGER;
(* Fhrt einen kompletten modalen Fensterdialog durch. Es wird
 * der Button zurckgegeben, mit dem der Dialog verlassen wurde.
 * Wenn kein Fenster mehr frei ist, wird ein normaler Dialog gezeichnet
 *)

PROCEDURE WinDialIsTop (wdw: INTEGER): BOOLEAN;
(* Fragt nach, ob das Fenster wdw ein Fensterdialog ist 
 *)

PROCEDURE SetApplicationName (REF name : ARRAY OF CHAR);
(* Setzt den Namen der Applikation, der in Dialogen ohne Titel
 * per Default eingesetzt wird.
 *)

PROCEDURE WinEditChanged (tree: ADDRESS; edit: INTEGER);
(* Setzt den Cursor in ein neues Editfeld bzw. teilt dem Modul
 * mit, da sich der Text gendert hat und da der Cursor neu plaziert 
 * werden soll. edit = -1: Nur Cursor neu plazieren.
 *)

PROCEDURE WinUpdateTitle (tree: ADDRESS);
(* Dialog soll seinen Titel aktualisieren 
 *)

PROCEDURE IsWinDial (tree: ADDRESS): BOOLEAN;
(* Fragt nach, ob der Dialog schon in einem Fenster ist
 *)

PROCEDURE ActivateWinDial (tree: ADDRESS);
(* Aktiviert den Window-Dialog, falls vorhanden 
 *)

PROCEDURE AlertsInWindow (doIt: BOOLEAN);
(* Aktiviert oder deaktiviert die Darstellung der
 * Alerts im Fenster 
 *)

END WinDials.
