DEFINITION MODULE Mintbind;

(*==============================================================*
 *                                                              *
 * Mint-Binding fr Megamax Modula-2, unter Benutzung der       *
 * MagicLib von Peter Hellinger.                                *
 *                                                              *
 * (c) 1992 Dirk Steins                                         *
 *                                                              *
 * Dieses Modul ist Freeware!                                   *
 *                                                              *
 *==============================================================*
 * Autor:               Dirk Steins                             *
 * erstellt am:         13.8.1992                               *
 * letzte nderung am:  13.8.1992                               *
 * Version:             1.0                                     *
 *==============================================================*
 * Datum    Vers. Autor     nderung (Arbeitsbericht)           *
 *==============================================================*
 *
 * 13.8.92  1.0   DS        Modul erstellt
 * 20.1.93  1.1   DS        Signals hinzugefgt
 *
 *==============================================================*)

FROM SYSTEM     IMPORT ADDRESS;
FROM MagicSys   IMPORT sINTEGER, lINTEGER, sBITSET, lBITSET, 
                       sCARDINAL, lCARDINAL;

FROM FileSys    IMPORT XATTR;

CONST
    NSIG        = 31;      (* number of signals recognized *)

    SIGNULL     = 0 ;      (* not really a signal *)
    SIGHUP      = 1 ;      (* hangup signal *)
    SIGINT      = 2 ;      (* sent by ^C *)
    SIGQUIT     = 3 ;      (* quit signal *)
    SIGILL      = 4 ;      (* illegal instruction *)
    SIGTRAP     = 5 ;      (* trace trap *)
    SIGABRT     = 6 ;      (* abort signal *)
    SIGPRIV     = 7 ;      (* privilege violation *)
    SIGFPE      = 8 ;      (* divide by zero *)
    SIGKILL     = 9 ;      (* cannot be ignored *)
    SIGBUS      = 10;      (* bus error *)
    SIGSEGV     = 11;      (* illegal memory reference *)
    SIGSYS      = 12;      (* bad argument to a system call *)
    SIGPIPE     = 13;      (* broken pipe *)
    SIGALRM     = 14;      (* alarm clock *)
    SIGTERM     = 15;      (* software termination signal *)

    SIGURG      = 16;      (* urgent condition on I/O channel *)
    SIGSTOP     = 17;      (* stop signal not from terminal *)
    SIGTSTP     = 18;      (* stop signal from terminal *)
    SIGCONT     = 19;      (* continue stopped process *)
    SIGCHLD     = 20;      (* child stopped or exited *)
    SIGTTIN     = 21;      (* read by background process *)
    SIGTTOU     = 22;      (* write by background process *)
    SIGIO       = 23;      (* I/O possible on a descriptor *)
    SIGXCPU     = 24;      (* CPU time exhausted *)
    SIGXFSZ     = 25;      (* file size limited exceeded *)
    SIGVTALRM   = 26;      (* virtual timer alarm *)
    SIGPROF     = 27;      (* profiling timer expired *)
    SIGWINCH    = 28;      (* window size changed *)
    SIGUSR1     = 29;      (* user signal 1 *)
    SIGUSR2     = 30;      (* user signal 2 *)

    SIG_DFL     = 0 ;
    SIG_IGN     = 1 ;

PROCEDURE Syield() : sINTEGER;

PROCEDURE Fpipe (VAR usrh : ARRAY OF sINTEGER): sINTEGER;

PROCEDURE Fcntl (f : sINTEGER; arg : ADDRESS; cmd : INTEGER): lINTEGER;

PROCEDURE Finstat (f : sINTEGER): lINTEGER;

PROCEDURE Foutstat (f : sINTEGER) : lINTEGER;

PROCEDURE Fgetchar (f : sINTEGER; mode : sINTEGER) : lINTEGER;

PROCEDURE Fputchar (f : sINTEGER; char: lINTEGER; mode: sINTEGER): lINTEGER;

PROCEDURE Pwait (): lINTEGER;

PROCEDURE Pnice (delta : sINTEGER) : sINTEGER;

PROCEDURE Pgetpid () : sINTEGER;

PROCEDURE Pgetppid () : sINTEGER;

PROCEDURE Pgetpgrp () : sINTEGER;

PROCEDURE Psetpgrp (pid, grp : sINTEGER) : sINTEGER;

PROCEDURE Pgetuid () : sINTEGER;

PROCEDURE Psetuid (id : sINTEGER) : sINTEGER;

PROCEDURE Pkill (pid, sig : sINTEGER) : sINTEGER;

PROCEDURE Psignal (sig : sINTEGER; handler : ADDRESS) : ADDRESS;

PROCEDURE Pvfork () : sINTEGER;

PROCEDURE Pgetgid () : sINTEGER;

PROCEDURE Psetgid (id : sINTEGER) : sINTEGER;

PROCEDURE Psigblock (mask : lBITSET) : lBITSET;

PROCEDURE Psigsetmask (mask : lBITSET) : lBITSET;

PROCEDURE Pusrval (val : lINTEGER) : lINTEGER;

PROCEDURE Pdomain (domain : sINTEGER) : sINTEGER;

PROCEDURE Psigreturn ();

PROCEDURE Pfork () : sINTEGER;

PROCEDURE Pwait3 (flag : sINTEGER; VAR rusage : ARRAY OF lINTEGER): lINTEGER;

PROCEDURE Fselect (time : sINTEGER; VAR rfd, wfd : lBITSET; xfd : ADDRESS) : sINTEGER;

PROCEDURE Prusage (VAR r : ARRAY OF LONGINT);

PROCEDURE Psetlimit (limit : sINTEGER; val : lINTEGER) : lINTEGER;

PROCEDURE Talarm (time : lINTEGER) : lINTEGER;

PROCEDURE Pause ();

PROCEDURE Sysconf (n : sINTEGER): lINTEGER;

PROCEDURE Psigpending () : lBITSET;

PROCEDURE Dpathconf (REF name : ARRAY OF CHAR; mode : sINTEGER) : lINTEGER;

TYPE Msg = RECORD 
             msg1, msg2 : lINTEGER;
             pid        : sINTEGER;
           END;

PROCEDURE Pmsg (mode : sINTEGER; mboxid : lINTEGER; VAR msg : Msg) : lINTEGER;

PROCEDURE Fmidipipe (pid, in, out : sINTEGER) : lINTEGER;

PROCEDURE Prenice (pid, delta : sINTEGER) : lINTEGER;

PROCEDURE Dopendir (REF name : ARRAY OF CHAR; flag : sINTEGER) : lINTEGER;

PROCEDURE Dreaddir (handle : lINTEGER; VAR name : ARRAY OF CHAR) : lINTEGER;

PROCEDURE Drewinddir (handle : lINTEGER) : lINTEGER;

PROCEDURE Dclosedir (handle : lINTEGER) : lINTEGER;

PROCEDURE Dxreaddir (handle : lINTEGER; VAR name : ARRAY OF CHAR; 
                     VAR xattr : XATTR; VAR xr: lINTEGER) : lINTEGER;

PROCEDURE Fxattr (flag : sINTEGER; REF name : ARRAY OF CHAR; VAR xattr : XATTR) : lINTEGER;

PROCEDURE Flink (REF oldname, newname : ARRAY OF CHAR) : lINTEGER;

PROCEDURE Fsymlink (REF oldname, newname : ARRAY OF CHAR) : lINTEGER;

PROCEDURE Freadlink (REF name : ARRAY OF CHAR; VAR buf : ARRAY OF CHAR) : lINTEGER;

PROCEDURE Dcntl (cmd : sINTEGER; REF name : ARRAY OF CHAR; arg : lINTEGER) : lINTEGER;

PROCEDURE Fchown (REF name : ARRAY OF CHAR; uid, gid : sINTEGER) : lINTEGER;

PROCEDURE Fchmod (REF name : ARRAY OF CHAR; mode : sBITSET) : lINTEGER;

PROCEDURE Pumask (mode : sBITSET): sBITSET;

PROCEDURE Psemaphore (mode : sINTEGER; id : lCARDINAL; timeout : lINTEGER) : lINTEGER;

PROCEDURE Dlock (mode : sINTEGER; drv : sINTEGER) : lINTEGER;

PROCEDURE Psigpause (mask : lBITSET);

TYPE    SigAction   = RECORD
                        sa_handler  : lINTEGER;
                        sa_mask     : lBITSET;
                        sa_flags    : sINTEGER;
                      END;

PROCEDURE Psigaction (sig : sINTEGER; VAR act : SigAction; VAR oact : SigAction) : lINTEGER;

PROCEDURE Pgeteuid() : sINTEGER;

PROCEDURE Pgetegid() : sINTEGER;

END Mintbind.
