IMPLEMENTATION MODULE PutzGr;

FROM SYSTEM IMPORT ADDRESS, ADR, CADR, TSIZE, BYTE, ASSEMBLER, CALLSYS;


(* MM2 *)
FROM Storage            IMPORT ALLOCATE, DEALLOCATE;

FROM Directory          IMPORT  GetDefaultPath, SetDefaultPath, Delete, Rename;

FROM FileNames          IMPORT  ValidatePath, SplitPath, FileName;

FROM Lists      IMPORT  List,           CreateList,     DeleteList,
                        AppendEntry,    ListEmpty,      CurrentEntry,
                        NextEntry,      ResetList,      RemoveEntry,
                        ScanEntries,    LDir,           LCondProc,
                        NoOfEntries,    PrevEntry;


IMPORT  Strings, StrConv, BinOps;

(* MagicLib *)
IMPORT MagicSys, MagicStrings, MagicTypes, MagicDOS, MagicAES;
FROM MagicSys IMPORT lINTEGER;

        (* magic tools *)
IMPORT mtAppl, mtDir, mtAlerts, FileSys, Mintbind;

(* CatPutz *)

FROM PutzTypes          IMPORT ptrGrEntry, defaultOpts, totalEntry, 
                               putzList, groupList, displayList,
                               listEntryType, displayListEntry, 
                               ptrDispEntry;

IMPORT PutzLog, PutzTypes, PutzGroup, PutzRsc;
IMPORT PutzInit, PutzDo;

(* CAT *)
IMPORT CatGlobal, dataSys, ConvertDate, GroupSelect;
IMPORT MTPaths, Infofiles, WildCards;

FROM dataSys IMPORT posType;

FROM Void   IMPORT v;

  CONST  noInfFile = '[3][CATPUTZ:|GRUPPEN.INF nicht|gefunden][[Abbruch]';
         noMemAlrt = '[3][CATPUTZ:|Konnte Gruppenliste|nicht anlegen!|(Zuwenig Speicher)][[Abbruch]';

VAR     maxNameLength : CARDINAL;

PROCEDURE MaxNameLength(): CARDINAL;
BEGIN
  RETURN maxNameLength;
END MaxNameLength;

PROCEDURE FreeGroupList ();
  VAR entry : ptrGrEntry;
BEGIN
  ResetList (putzList);
  entry := NextEntry (putzList);
  WHILE (entry # NIL) DO
    DISPOSE (entry);
    RemoveEntry (putzList, v.bool);
    entry := NextEntry (putzList);
  END;
  DeleteList (putzList, v.bool);
END FreeGroupList;

PROCEDURE BuildDisplayList (VAR l : List);
  VAR entry : ptrDispEntry;
      data  : ptrGrEntry;

  PROCEDURE appEntry (typ: listEntryType; adr : ADDRESS);
  BEGIN
    NEW (entry);
    IF entry # NIL
    THEN
      WITH entry^ DO 
        type := typ;
        data := adr;
      END;
      AppendEntry (l, entry, v.bool);
    END;
  END appEntry;

BEGIN
  CreateList (l, v.bool);
  appEntry (listHead1, NIL);
  appEntry (listHead2, NIL);
  appEntry (separatorLine, NIL);
  ResetList (putzList);
  data := NextEntry (putzList);
  WHILE data # NIL DO
    appEntry (dataEntry, data);
    data := NextEntry (putzList);
  END;
  appEntry (separatorLine, NIL);
  appEntry (dataEntry, ADR(totalEntry));
END BuildDisplayList;

PROCEDURE ClearDisplayList;
  VAR entry : ptrDispEntry;
BEGIN
  ResetList (displayList);
  entry := NextEntry (displayList);
  WHILE entry # NIL DO
    DISPOSE (entry);
    entry := NextEntry (displayList);
  END;
  DeleteList (displayList, v.bool);
END ClearDisplayList;

(*$Z-*)
PROCEDURE findNumber (entry: ADDRESS; info: ADDRESS): BOOLEAN;
  VAR intPtr : POINTER TO INTEGER;
      gr      : ptrGrEntry;
BEGIN
  gr := entry;
  intPtr := info;
  RETURN gr^.info^.catNumber = intPtr^;
END findNumber;
(*$Z=*)

PROCEDURE FindGroupByNumber (number : INTEGER; VAR group: ptrGrEntry): BOOLEAN;
  VAR p : CARDINAL;
      found: BOOLEAN;
BEGIN
  ResetList (putzList);
  ScanEntries (putzList, forward, findNumber, ADR(number), found);
  IF found
  THEN
    group := CurrentEntry (putzList);
    RETURN group # NIL
  END;
  RETURN FALSE
END FindGroupByNumber;

CONST   privatePat  = "private.*";
        groupPat    = "grupp*.*";
        allWildPat  = "*.*";
        datPat      = "*.dat";
        parPat      = "*.par";
        tabPat      = "*.tab";

PROCEDURE FindGroup (REF name : ARRAY OF CHAR; ignoreCase: BOOLEAN; VAR group: ptrGrEntry): BOOLEAN;
  TYPE CharSet = SET OF CHAR;
  VAR pos: CARDINAL;
      groupNumber: INTEGER;
BEGIN
  IF WildCards.MatchName (name, privatePat, ignoreCase)
  THEN
    (* Private *)
    groupNumber := 0;
  ELSE
    pos := 0;
    WHILE (pos < LENGTH (name)) AND NOT (name[pos] IN CharSet{'0'..'9'}) DO 
      INC (pos) 
    END;
    (* Jetzt die Zahl bilden *)
    groupNumber := StrConv.StrToInt (name, pos, v.bool);
    IF ~v.bool 
    THEN
      RETURN FALSE;
    END;
    INC (groupNumber);
  END;
  IF group # NIL
  THEN
    (* gltiger Zeiger, wird auch nicht illegal verndert *)
    IF findNumber (group, ADR(groupNumber))
    THEN
      RETURN TRUE;
    END;
  END;
  (* Gruppe zu Nummer suchen *)
  RETURN FindGroupByNumber (groupNumber, group);
END FindGroup;

PROCEDURE Dxreaddir (handle : lINTEGER; VAR name : ARRAY OF CHAR; 
                     VAR xattr : FileSys.XATTR; VAR xr: lINTEGER) : lINTEGER;
  VAR res: lINTEGER;
BEGIN
  res := Mintbind.Dxreaddir (handle, name, xattr, xr);
  IF res = LONG(MagicDOS.EInvFN)
  THEN
    res := Mintbind.Dreaddir (handle, name);
    xr := Mintbind.Fxattr (0, name, xattr);
  END;
  RETURN res;
END Dxreaddir;

PROCEDURE SearchGroups();
  VAR
    oDta       : ADDRESS;
    dta        : MagicDOS.DTA;
    res        : INTEGER;
    ext        : ARRAY [0..7] OF CHAR;
    tmp,
    mask       : ARRAY [0..1023] OF CHAR;
    toLower    : BOOLEAN;
    firstFound : BOOLEAN;
    nr         : CARDINAL;
    dHandle    : LONGINT;
    nPtr       : POINTER TO ARRAY [0..1023] OF CHAR;
    xattr      : FileSys.XATTR;
    err        : LONGINT;
    xattrErr   : LONGINT;
    grPtr      : ptrGrEntry;
    groupCount : CARDINAL;
    oneFound   : BOOLEAN;
BEGIN
  groupCount := 0;
  grPtr := NIL;
  IF CatGlobal.isMintDomain
  THEN
    (* case-sensitivity vom filesystem feststellen *)
    (* Zugriff ber Dopendir, Dreaddir *)
    dHandle := Mintbind.Dopendir (MTPaths.DataPath, 0);
    IF dHandle >= 0
    THEN
      nPtr := ADR (mask[4]);
      firstFound := TRUE;
      toLower    := TRUE;
      REPEAT
        CatGlobal.busyMouse();
        err := Mintbind.Dxreaddir (dHandle, mask, xattr, xattrErr);
        oneFound := FALSE;
        IF err = 0
        THEN
          IF firstFound 
          THEN
            (* Get attributes of file, follow link *)
            MagicStrings.Assign (MTPaths.DataPath, tmp);
            MagicStrings.Append (nPtr^, tmp);
            toLower := 0 # (Mintbind.Dpathconf (tmp, 6));
            firstFound := FALSE;
          END;
          (* Jetzt feststellen, ob es ein Gruppenname ist *)
          IF WildCards.MatchName (nPtr^, groupPat, toLower)
          OR WildCards.MatchName (nPtr^, privatePat, toLower)
          THEN
            (* Gruppe zu Nummer suchen *)
            IF FindGroup (nPtr^, toLower, grPtr)
            THEN
              oneFound := TRUE;
              (* Typ feststellen *)
              IF WildCards.MatchName (nPtr^, datPat, toLower)
              THEN
                (* Ist Datendatei *)
                (* Get attributes of file, follow link *)
                (*
                MagicStrings.Assign (MTPaths.DataPath, tmp);
                MagicStrings.Append (nPtr^, tmp);
                err := Mintbind.Fxattr (0, tmp, xattr);
                *)
                grPtr^.datDa := xattrErr = 0;
                IF grPtr^.datDa
                THEN
                  grPtr^.preBytes := LONGCARD(xattr.size);
                END;
              ELSIF WildCards.MatchName (nPtr^, parPat, toLower)
              THEN
                (* Ist Parameterdatei *)

                (* Get attributes of file, follow link *)
                (*
                MagicStrings.Assign (MTPaths.DataPath, tmp);
                MagicStrings.Append (nPtr^, tmp);
                err := Mintbind.Fxattr (0, tmp, xattr);
                *)
                grPtr^.parDa := xattrErr = 0;
                IF grPtr^.parDa
                THEN 
                  grPtr^.preMsgs := (LONGCARD(xattr.size) - dataSys.dbHeaderLength) DIV TSIZE(dataSys.pBlock);
                END;
              ELSIF WildCards.MatchName (nPtr^, tabPat, toLower)
              THEN
                (* Ist TAB-Datei *)
                (*
                MagicStrings.Assign (MTPaths.DataPath, tmp);
                MagicStrings.Append (nPtr^, tmp);
                err := Mintbind.Fxattr (0, tmp, xattr);
                *)
                grPtr^.tabDa := xattrErr = 0;
              ELSE
                oneFound := FALSE;
              END;
            END;
          END;
        END;
        IF oneFound
        THEN
          INC (groupCount);
          PutzInit.SetCurrentValue (groupCount);
        END;
      UNTIL err # 0;
      IF err # MagicDOS.ENMFil
      THEN
      END;
      v.lint := Mintbind.Dclosedir (dHandle);
    END;
  ELSE
    (* Suche per Fsfirst und Fsnext *)
    oDta := MagicDOS.Fgetdta ();
    MagicDOS.Fsetdta (ADR(dta));

    res := MagicDOS.Fsfirst (allWildPat, {});
    WHILE res = 0 DO

      (*
      (* Build normal filename from DTA *)
      MagicStrings.Copy (dta.dFname, 0, 8, tmp);
      MagicStrings.Copy (dta.dFname, 8, 3, ext);
      Strings.EatSpaces (ext);
      IF LENGTH (ext) > 0
      THEN
        MagicStrings.Append ('.', tmp);
        MagicStrings.Append (ext, tmp);
      END;
      Strings.EatSpaces (tmp);
      *)
      MagicStrings.Assign (dta.dFname, tmp);
      
      oneFound := FALSE;
      IF WildCards.MatchName (tmp, groupPat, TRUE)
      OR WildCards.MatchName (tmp, privatePat, TRUE)
      THEN
        (* Gruppe zu Nummer suchen *)
        IF FindGroup (tmp, TRUE, grPtr)
        THEN
          oneFound := TRUE;
          (* Typ feststellen *)
          IF WildCards.MatchName (tmp, datPat, TRUE)
          THEN
            (* Ist Datendatei *)
            grPtr^.datDa := TRUE;
            grPtr^.preBytes := dta.dLength;
          ELSIF WildCards.MatchName (tmp, parPat, TRUE)
          THEN
            (* Ist Parameterdatei *)
            grPtr^.parDa := TRUE;
            grPtr^.preMsgs := (dta.dLength- dataSys.dbHeaderLength) DIV TSIZE(dataSys.pBlock);
          ELSIF WildCards.MatchName (tmp, tabPat, TRUE)
          THEN
            (* Ist TAB-Datei *)
            grPtr^.tabDa := TRUE;
          ELSE
            oneFound := FALSE;
          END;
        END;
      END;
      IF oneFound
      THEN
        INC (groupCount);
        PutzInit.SetCurrentValue (groupCount);
      END;
      (* Nchsten Eintrag nehmen *)
      res := MagicDOS.Fsnext();
      CatGlobal.busyMouse();
    END;

    MagicDOS.Fsetdta (oDta);
  END;
END SearchGroups;

PROCEDURE InitGroups() : BOOLEAN;
  VAR tmpFileName : PutzTypes.FileStr;
      i           : CARDINAL;

  VAR datName,
      tabName,
      parName : PutzTypes.FileStr;
      gName   : PutzTypes.PathStr;
      oDta: ADDRESS;
      dta : MagicDOS.DTA;
      res : INTEGER;
      entry : GroupSelect.groupPtr;
      grEntry   : ptrGrEntry;
      groupCount: LONGINT;


BEGIN
  PutzInit.InitInitBox();
  (* Jetzt erstmal die Gruppenliste einlesen *)
  IF ~GroupSelect.ReadGroupList()
  THEN
    v.int := mtAlerts.Alert (1,"[3][CATPUTZ:|Kann Gruppenliste|nicht lesen!][[Abbruch]");
    PutzInit.CloseInitBox();
    RETURN FALSE
  END;
  
  CatGlobal.busyMouse();
  
  i := 1;
  totalEntry.info := NIL;
  totalEntry.preMsgs := 0;
  totalEntry.preBytes := 0;
  totalEntry.badDel := 0;
  totalEntry.totalDel := 0;
  totalEntry.partDel := 0;
  totalEntry.postMsgs := 0;
  totalEntry.postBytes := 0;

  (* Gruppenliste holen *)
  IF ~GroupSelect.GetReadList (groupList, FALSE, FALSE, FALSE)
  THEN
    v.int := mtAlerts.Alert (1, noMemAlrt);
    PutzInit.CloseInitBox();
    RETURN FALSE
  END;
  
  (* Anzahl Gruppen * 3 als Maximalwert *)
  PutzInit.SetMaxValue(VAL(LONGINT, NoOfEntries(groupList) * 3));
  
  CatGlobal.busyMouse();
 
  (* Jetzt alle Eintrge durchscannen, und die neue Liste anlegen mit
   * den Infopointern 
   *)
  CreateList (putzList, v.bool);
  IF v.bool THEN
    GroupSelect.DeleteList (groupList);
    v.int := mtAlerts.Alert (1,noMemAlrt);
    PutzInit.CloseInitBox();
  END;
 
  ResetList (groupList);
  groupCount := 1;
  maxNameLength := 0;
  entry := NextEntry (groupList);
  WHILE entry # NIL DO
    NEW (grEntry);
    IF grEntry = NIL
    THEN
      (* Listen lschen *)
      FreeGroupList ();
      GroupSelect.DeleteList (groupList);
      v.int := mtAlerts.Alert (1,noMemAlrt);
      PutzInit.CloseInitBox();
      RETURN FALSE
    END;
    WITH grEntry^ DO
      maxNameLength := BinOps.HigherCard (maxNameLength, LENGTH (entry^.name^));
      info := entry;
      selected  := FALSE;
      preMsgs   := 0;
      preBytes  := 0;
      badDel    := 0;
      totalDel  := 0;
      partDel   := 0;
      postMsgs  := 0;
      postBytes := 0;
      tabDa     := FALSE;
      parDa     := FALSE;
      datDa     := FALSE;
    END;
    CatGlobal.busyMouse();
    AppendEntry (putzList, grEntry, v.bool);
    IF v.bool
    THEN
      (* Listen lschen *)
      FreeGroupList ();
      GroupSelect.DeleteList (groupList);
      v.int := mtAlerts.Alert (1,noMemAlrt);
      PutzInit.CloseInitBox();
      RETURN FALSE
    END;
    INC (groupCount);
    entry := NextEntry (groupList);
  END;
  
  (* Jetzt haben wir die Liste, jetzt gehen wir durch das Datenbank-
   * Verzeichnis und holen uns die Daten der ganzen Gruppen
   *)
  SearchGroups();
  
  (* Und jetzt noch die Liste fr die Anzeige aufbauen *)
  ClearDisplayList();
  BuildDisplayList(displayList);
  
  PutzDo.CalcTotals();

  PutzInit.CloseInitBox();
  RETURN TRUE
END InitGroups;

PROCEDURE SelectFile (REF title : ARRAY OF CHAR;
                      REF mask  : ARRAY OF CHAR;
                      VAR path  : ARRAY OF CHAR;
                      VAR fName : ARRAY OF CHAR;
                      makeFullName: BOOLEAN) : BOOLEAN;
  VAR void128 : PutzTypes.PathStr;
      name    : PutzTypes.FileStr;
      success : BOOLEAN;
      targetPath : PutzTypes.PathStr;
BEGIN
  IF Strings.Empty (path)
  THEN
    GetDefaultPath(targetPath);
  ELSE
    Strings.Assign (path, targetPath, v.bool);
  END;
  IF ~Strings.Empty (mask)
  THEN
    Strings.Append (mask, targetPath, v.bool);
  ELSE
    Strings.Append ('*.*', targetPath, v.bool);
  END;
  void128 := 0C;
  Strings.Assign (fName, name, v.bool);
  success := mtDir.GetDir (targetPath, name, title);
  IF ~success THEN RETURN FALSE END;
  SplitPath(targetPath, targetPath, void128);
  ValidatePath (targetPath);
  SetDefaultPath (targetPath, v.int);
  Strings.Assign (targetPath, path, v.bool);
  IF makeFullName
  THEN
    Strings.Assign (targetPath, fName, v.bool);
    Strings.Append (name, fName, v.bool);
  ELSE
    Strings.Assign (name, fName, v.bool);
  END;
  RETURN TRUE
END SelectFile;

PROCEDURE GetTempPath(env: ADDRESS);
  VAR name, value : Strings.String;
      tmpAvail    : BOOLEAN;
  
  PROCEDURE EnvSearch (VAR val : ARRAY OF CHAR; REF name: ARRAY OF CHAR;
                       env : ADDRESS);
                       
    PROCEDURE search (str, name : ARRAY OF CHAR;
                      VAR val : ARRAY OF CHAR) : BOOLEAN;
      VAR i : INTEGER;
    BEGIN
      Strings.Append ('=',name,v.bool);
      i := Strings.Pos (name, str, 0);
      IF i < 0 THEN RETURN FALSE END;
      Strings.Copy (str, i+INTEGER(Strings.Length(name)),
                    INTEGER(Strings.Length (str)-Strings.Length(name))-i, val, v.bool);
      RETURN TRUE
    END search;
    
    VAR ch : POINTER TO CHAR;
        lastCh : CHAR;
        i      : INTEGER;
        str    : ARRAY [0..255] OF CHAR;
  BEGIN
    ch := env;
    lastCh := ch^;
    i := 0;
    (* Solange noch env da ist *)
    WHILE (ch^ # 0C) OR (lastCh # 0C) DO
      str[i] := ch^;
      INC(i);
      lastCh := ch^;
      IF lastCh = 0C
      THEN (* string complete, search *)
        IF search(str, name, val) THEN RETURN
        ELSE
          str := "";
          i := 0;
        END;
      END;
      INC (ch);
    END;
  END EnvSearch;
  
  PROCEDURE TestTmpPath() : BOOLEAN;
  BEGIN
    ValidatePath (PutzTypes.tmpPath);
    SetDefaultPath (PutzTypes.tmpPath, v.int);
    RETURN v.int = 0;
  END TestTmpPath;
  
BEGIN
  name := "TMP";
  value := "";
  tmpAvail := FALSE;
  EnvSearch (value, name, env);
  IF ~Strings.Empty (value)
  THEN
    Strings.Assign (value, PutzTypes.tmpPath, v.bool);
    IF v.bool
    THEN
      tmpAvail := TestTmpPath() ;
    END;
  END;
  IF ~tmpAvail
  THEN
    name := "TMPDIR";
    value := "";
    EnvSearch (value, name, env);
    IF ~Strings.Empty (value)
    THEN
      Strings.Assign (value, PutzTypes.tmpPath, v.bool);
      IF v.bool
      THEN
        tmpAvail := TestTmpPath() ;
      END;
    END;
  END;
  (* kein tmp-Pfad vorhanden, Db-Path nehmen *)
  IF ~tmpAvail THEN Strings.Assign (MTPaths.DataPath, PutzTypes.tmpPath, v.bool); END;
  IF ~defaultOpts.useTmp THEN Strings.Assign (MTPaths.DataPath, PutzTypes.tmpPath, v.bool); END;
  IF defaultOpts.tmpLock THEN Strings.Assign (defaultOpts.tmpDir, PutzTypes.tmpPath, v.bool) END;
  IF ~TestTmpPath()
  THEN
    Strings.Assign (MTPaths.DataPath, PutzTypes.tmpPath, v.bool);
  END;
END GetTempPath;

CONST RscName = "CATPUTZ.RSC";

PROCEDURE ReadCatInf (): BOOLEAN;
  VAR i : INTEGER;
     basePage    : MagicTypes.PtrPD;
     env         : ADDRESS;
BEGIN
  IF ~MTPaths.ReadInf (CatGlobal.infPath, CatGlobal.infName, TRUE, TRUE)
  THEN
    RETURN FALSE
  END;
  CatGlobal.busyMouse;
  SetDefaultPath (MTPaths.DataPath, v.int);
  PutzRsc.LoadParameter (MTPaths.DataPath);

  Infofiles.LoadInfoInf (TRUE, FALSE);
  IF ~InitGroups() THEN 
    mtAppl.MouseArrow(); 
    RETURN FALSE 
  END;

  (* suchen nach Environment-Variable TMP oder TEMP *)
  basePage := MagicSys.Basepage();
  env := basePage^.pEnv;
  GetTempPath(env);

  mtAppl.MouseArrow();
  RETURN TRUE
END ReadCatInf;

PROCEDURE ReadNewInf (): BOOLEAN;
VAR   oldDir: PutzTypes.MaxStr;
      infName: Strings.String;
      test  : PutzTypes.MaxStr;
BEGIN
 Strings.Assign (CatGlobal.infPath, oldDir, v.bool);
 Strings.Assign (CatGlobal.infName, infName, v.bool);
 IF ~SelectFile ('CAT.INF laden','*.INF', CatGlobal.infPath, CatGlobal.infName, FALSE)
 THEN
   RETURN FALSE
 END;
 (* Gruppenliste lschen *)
 FreeGroupList ();
 Strings.Assign (CatGlobal.infPath, test, v.bool);
 Strings.Assign (CatGlobal.infName, test, v.bool);
 GroupSelect.DeleteList (groupList);
 IF ~ReadCatInf ()
 THEN
   (* Altes CAT.INF wieder lesen *)
   Strings.Assign (oldDir, CatGlobal.infPath, v.bool);
   Strings.Assign (infName, CatGlobal.infName, v.bool);
   RETURN ReadCatInf ();
 END;
 RETURN TRUE;
END ReadNewInf;

BEGIN
  maxNameLength := 10;
END PutzGr.
