IMPLEMENTATION MODULE PutzHelp;

FROM SYSTEM     IMPORT ADDRESS, ADR, CADR, TSIZE;

(* MM2 Module *)
IMPORT Lists;
FROM Storage    IMPORT ALLOCATE, DEALLOCATE;
IMPORT GrafBase;
IMPORT FileNames;

(* CAT Module *)
IMPORT ListDl;
IMPORT CatTypes;
IMPORT ListHelp;
IMPORT PutzRsc;
IMPORT MTE;
IMPORT CatGlobal;
FROM CatGlobal       IMPORT multiTask, multiTOS, magIx, magIxVer;
IMPORT CatFiles;
IMPORT data;
IMPORT dataSys;
IMPORT Varnames;
IMPORT ConfVars;

FROM Void IMPORT v;

(* MagicLib *)
IMPORT MagicAES, MagicDOS, MagicStrings, MagicConvert;
IMPORT mtAppl;
IMPORT mtDials;
IMPORT mtTextfiles;
IMPORT mtUtils;

IMPORT Catputz2;

TYPE    tHelpKeys = ARRAY [general..window] OF ARRAY [0..39] OF CHAR;

CONST   cHelpFile = 'catputz.hlp';
        cHelpIdx  = 'putzhelp.idx';
        cHelpCode = '#(&)';
        cHelpTable= tHelpKeys{'#(&)CATPUTZ',
                              '#(&)FLAGS',
                              '#(&)OPTIONS',
                              '#(&)WINDOW'};
    cGuideHelpTable= tHelpKeys{
                              'CatPutz - Das Putzprogramm',
                              'Lschflags einstellen',
                              'Einstellungen',
                              'Das Hauptfenster'};

        cHelpMagic          = 43415448H; (* 'CATH' *)
        cHelpVersion        = 0100H;
        cHelpVersionMagic   = 0243H;

TYPE    tHelpIdx    = RECORD
                        helpSize    : LONGCARD;
                        helpOffset  : ARRAY [general..window] OF LONGCARD;
                      END;
                      
        tHelpInfo   = RECORD
                        l           : Lists.List;
                      END;
                      
        tHelpPtr    = POINTER TO tHelpInfo;

(* Fr Protokoll *)
CONST
  appName    = 'CATPUTZ '+0C;

CONST maxBufLen = 128;  (* Lnge des statischen Textbuffers *)

      memDescrSize  = 2048; (* 2 kB globaler Speicher *)

TYPE        (* for systems with memory protection (MultiTOS) *)
        memDescr        = RECORD
                            CASE : BOOLEAN OF
                            TRUE:
                              apName  : ARRAY [0..49] OF CHAR; (* our extended application name *)
                              textBuf : ARRAY [0..maxBufLen-1] OF CHAR;
                              dataPtr : data.PtrRecord;
                              mPath   : CatTypes.String255;
                              mName   : CatTypes.String255; |
                            FALSE:
                              allInOne: ARRAY [0..memDescrSize-1] OF CHAR;
                            END;
                          END;

VAR   
    memPtr   : POINTER TO memDescr;         (* global buffer *)

PROCEDURE buildIdx (REF path: ARRAY OF CHAR; size: LONGCARD; 
                    f : mtTextfiles.TEXTFILE;
                    VAR helpIdx: tHelpIdx): BOOLEAN;
  VAR helpName        : CatTypes.String255;
      skipString      : CatTypes.String127;
      helpKey         : CatTypes.String127;
      str             : CatTypes.Str127Ptr;
      i               : tHelp;
      found           : BOOLEAN;
      header          : dataSys.FileHeaderType;
      fi              : INTEGER;
BEGIN
  helpIdx.helpSize := size;
  FOR i := general TO window DO
    helpIdx.helpOffset[i] := $FFFFFFFF;
  END;
  (* Jetzt alle Zeilen nach dem Key durchsuchen *)
  LOOP
    (* Zeilen bis zum Anfang des Helptextes berspringen *)
    mtTextfiles.ReadLine(f, skipString);
    IF mtTextfiles.EndofText (f) THEN EXIT END;
    found := FALSE;
    IF MagicStrings.Pos (cHelpCode, skipString, 0, FALSE) = 0
    THEN
      i := general;
      LOOP
        IF found THEN EXIT END;
        IF MagicStrings.Pos (cHelpTable[i], skipString, 0, FALSE) < LENGTH (skipString)
        THEN 
          (* Gefunden. In der nchsten Zeile geht es los *)
          mtTextfiles.ReadLn (f);
          (* Offset merken *)
          helpIdx.helpOffset[i] := mtTextfiles.Textpos (f);
          found := TRUE;
        ELSE
          IF i < MAX (tHelp)
          THEN 
            INC (i);
          ELSE 
            EXIT
          END;
        END;
      END;
    END;
    IF ~found
    THEN
      mtTextfiles.ReadLn (f);
    END;
    IF mtTextfiles.EndofText (f) THEN EXIT END;
  END;
  mtTextfiles.Reset (f);
  (* Jetzt den Hilfeindex schreiben *)
  fi := CatFiles.CreateFile (path, cHelpIdx);
  IF fi < 0 
  THEN 
    RETURN FALSE
  END;
  (* Fileheader aufbauen *)
  header := dataSys.FileHeaderType {cHelpMagic, cHelpVersion, cHelpVersionMagic};
  CatFiles.WriteMuch (dataSys.dbHeaderLength, fi, ADR(header));
  CatFiles.WriteMuch (TSIZE (tHelpIdx), fi, ADR(helpIdx));
  CatFiles.CloseFile (fi);
  RETURN TRUE;
END buildIdx;

PROCEDURE readHelp (VAR text : Lists.List; which: tHelp): BOOLEAN;
  VAR helpName        : CatTypes.String255;
      skipString      : CatTypes.String127;
      helpKey         : CatTypes.String127;
      str             : CatTypes.Str127Ptr;
      f               : mtTextfiles.TEXTFILE;
      header          : dataSys.FileHeaderType;
      fi              : INTEGER;
      helpSize        : LONGCARD;
      helpIdx         : tHelpIdx;
      path            : CatTypes.String255;
      name            : CatTypes.nameStrType;
      succ            : BOOLEAN;
BEGIN
  (*  Lies Hilfe-Datei ein.
   *)
  CatGlobal.ApplPath (helpName);
  MagicStrings.Append (cHelpFile, helpName);
  mtAppl.MouseBusy();
  IF ~mtTextfiles.OpenTextfile (helpName, mtTextfiles.READ, 8192, f)
  THEN
    MagicStrings.Assign (cHelpFile, helpName);
    MagicAES.ShelFind (helpName);
    IF ~mtTextfiles.OpenTextfile (helpName, mtTextfiles.READ, 8192, f)
    THEN
      mtAppl.MouseArrow();
      MTE.InfoAlert (MTE.noFile1, cHelpFile, MTE.noFile4);
      RETURN FALSE;
    END;
  END;
  
  (* Gre der Hilfedatei feststellen *)
  helpSize := CatFiles.FileSize (helpName, v.bool);
  (* Jetzt nachsehen, ob wir auch eine Indexdatei haben. Wenn nein, 
   * dann wird die jetzt aufgebaut
   *)
  FileNames.SplitPath (helpName, path, name);
  fi := CatFiles.OpenFile (path, cHelpIdx, CatFiles.readFile);
  IF fi < 0
  THEN
    (* Kein Index vorhanden, also aufbauen *)
    succ := buildIdx (path, helpSize, f, helpIdx);
  ELSE
    (* Erstmal Header lesen und prfen *)
    CatFiles.ReadMuch (dataSys.dbHeaderLength, fi, ADR(header));
    IF (header.CatMagic # cHelpMagic) OR
       (header.Version  # cHelpVersion) OR 
       (header.VersionMagic # cHelpVersionMagic)
    THEN
      (* Index ist ungltig, neu aufbauen *)
      CatFiles.CloseFile (fi);
      succ := buildIdx (path, helpSize, f, helpIdx);
    ELSE
      CatFiles.ReadMuch (TSIZE (tHelpIdx), fi, ADR(helpIdx));
      IF CatFiles.FileError < 0 
      THEN 
        CatFiles.ErrorAlert (CatFiles.FileError);
        CatFiles.CloseFile (fi);
        succ := buildIdx (path, helpSize, f, helpIdx);
      ELSIF helpIdx.helpSize # helpSize
      THEN
        (* Index veraltet, neu aufbauen *)
        CatFiles.CloseFile (fi);
        succ := buildIdx (path, helpSize, f, helpIdx);
      ELSE
        CatFiles.CloseFile (fi);
        succ := TRUE;
      END;
    END;
  END;
  (* Jetzt haben wir einen gltigen Index, nun mssen wir nur noch den 
   * Schlssel suchen
   *)
  IF succ & (helpIdx.helpOffset[which] < helpIdx.helpSize)
  THEN
    (* Auf richtige Position setzen *)
    mtTextfiles.SetTextpos (f, helpIdx.helpOffset[which]);
  ELSE
    (* Schade, Schlssel kann nicht stimmen, also doch einfach suchen *)
  
    (* Schlsselwort fr Suche bestimmen *)
    MagicStrings.Assign (cHelpTable[which], helpKey);
    (* Jetzt alle Zeilen nach dem Key durchsuchen *)
    LOOP
      (* Zeilen bis zum Anfang des Helptextes berspringen *)
      mtTextfiles.ReadLine(f, skipString);
      IF mtTextfiles.EndofText (f) THEN EXIT END;
      IF MagicStrings.Pos (helpKey, skipString, 0, FALSE) < LENGTH (skipString)
      THEN 
        (* Gefunden. In der nchsten Zeile geht es los *)
        mtTextfiles.ReadLn (f);
        EXIT 
      END;
      mtTextfiles.ReadLn (f);
      IF mtTextfiles.EndofText (f) THEN EXIT END;
    END;
  END;
  
  MagicStrings.Assign (cHelpCode, helpKey);
  
  Lists.CreateList (text, v.bool);
  IF v.bool THEN       
    mtAppl.MouseArrow();
    MTE.noMemAlert (); 
    RETURN FALSE 
  END;
  LOOP
    mtTextfiles.ReadLine(f, skipString);
    IF MagicStrings.Pos (helpKey, skipString, 0, FALSE) < LENGTH (skipString)
    THEN 
      (* Gefunden. Hilfetext ist nun zu Ende *)
      EXIT 
    END;
    NEW (str);
    IF str = NIL THEN MTE.noMemAlert (); EXIT END;
    MagicStrings.Assign (skipString, str^);
    Lists.AppendEntry (text, str, v.bool);
    IF v.bool THEN MTE.noMemAlert (); DISPOSE (str); EXIT END;
    IF mtTextfiles.EndofText (f) THEN EXIT END;
    mtTextfiles.ReadLn (f);
    IF mtTextfiles.EndofText (f) THEN EXIT END;
  END;
   
  mtTextfiles.CloseTextfile (f);
  mtAppl.MouseArrow();
  RETURN TRUE;
END readHelp;

PROCEDURE lineToStr (entry, env : ADDRESS; VAR str : ARRAY OF CHAR);
  VAR inf : CatTypes.Str127Ptr;
BEGIN
  inf := CatTypes.Str127Ptr (entry);
  IF inf = NIL THEN RETURN END;
  MagicStrings.Assign (inf^, str);
END lineToStr;

PROCEDURE selEntry (entry, env: ADDRESS; line : INTEGER): BOOLEAN;
BEGIN
  RETURN FALSE
END selEntry;

PROCEDURE isSelected (entry, env : ADDRESS) : BOOLEAN;
BEGIN
  RETURN FALSE;
END isSelected;

PROCEDURE isEnabled (entry, env : ADDRESS) : BOOLEAN;
BEGIN
  RETURN TRUE
END isEnabled;

PROCEDURE countEntries ( l : ADDRESS; VAR ll: LONGINT; VAR ww: INTEGER);
 VAR lp : POINTER TO Lists.List;
BEGIN
 lp := l;
 ww := 0;
 ll := VAL (LONGINT, Lists.NoOfEntries (lp^));
END countEntries;

PROCEDURE DeleteSimpleList (VAR l: Lists.List; killCarrier: BOOLEAN);

VAR   entry: ADDRESS;

BEGIN
  Lists.ResetList (l);
  entry := Lists.PrevEntry (l);
  WHILE entry # NIL DO
    Lists.RemoveEntry (l, v.bool);
    DEALLOCATE (entry, 0L);
    entry := Lists.CurrentEntry (l);
  END;
  IF killCarrier THEN Lists.DeleteList (l, v.bool) END;
END DeleteSimpleList;

VAR  tr     : mtUtils.tObjcTree;

PROCEDURE drawHelpEntry (entry, env : ADDRESS; x, y : INTEGER;
                         offset : INTEGER; clip   : GrafBase.Rectangle);

  CONST spaceString = "          ";
  BEGIN
    tr^[0].obX := x;
    tr^[0].obY := y;
    IF entry # NIL
    THEN
      (* Jetzt hier den String fr den Gruppennamen zusammenbauen *)
      mtUtils.SetObjcStringAdr (tr, 0, ADDRESS(entry));
    ELSE
      mtUtils.SetObjcStringAdr (tr, 0, CADR(spaceString));
    END;
    MagicAES.ObjcDraw (tr, 0, 8, clip);
  END drawHelpEntry;

PROCEDURE setAndGetHelpValues (tree: ADDRESS; private: ADDRESS; set: BOOLEAN; exitBut: INTEGER);
  VAR l : tHelpPtr;
BEGIN
  IF ~set
  THEN
    l := private;
    DeleteSimpleList (l^.l, TRUE);
    DISPOSE (l);
   END;
END setAndGetHelpValues;

PROCEDURE internalDoHelp (topic: tHelp; doReadHelp: BOOLEAN; VAR externalHelp: Lists.List);
  VAR l : tHelpPtr;
      exit : INTEGER;
      specials : ARRAY [0..0] OF ListDl.specialButHdler;
      helpBox  : ADDRESS;
      maxLength: INTEGER;
      r        : GrafBase.Rectangle;
      dialHandler: ListDl.ldHandler;
BEGIN
  NEW (l);
  IF l = NIL
  THEN
    MTE.noMemAlert();
    RETURN
  END;
  
  helpBox := PutzRsc.GetTreeAddr (Helpbox);
  specials[0].objc := -99;

  IF doReadHelp & ~readHelp (l^.l, topic)
  THEN 
    RETURN 
  END;
  IF ~doReadHelp
  THEN
    l^.l := externalHelp;
  END;
  
  (* Jetzt das Object zusammenbauen *)
  mtUtils.CalcArea (helpBox, Hhelpbox, r);
  maxLength := r.w DIV mtAppl.CharWidth;

  (* Objectadresse holen *)
  tr := PutzRsc.GetTreeAddr (Listtext);
  tr^[0].obSpec.TedPtr^.teTxtlen := maxLength;
  tr^[0].obWidth := maxLength*mtAppl.CharWidth;
  mtUtils.SetState (tr, 0, MagicAES.SELECTED, FALSE);
  mtUtils.SetState (tr, 0, MagicAES.DISABLED, FALSE);

  (* WindowDialog vorbereiten *)
  ListDl.BuildLdHandler (ADR(l^.l), 
                         ListHelp.resetList,
                         ListHelp.nextEntry,
                         ListHelp.prevEntry,
                         countEntries, 
                         isEnabled, selEntry,
                         lineToStr, isSelected, 
                         drawHelpEntry,
                         0, mtAppl.CharHeight, 
                         0, 0,
                         dialHandler);

  (* ListWinDialog ffnen *)
  IF ListDl.WinListDial (helpBox, ListDl.ldElemSet{ListDl.ldArrows, ListDl.ldModal},
                       dialHandler,
                       Hhelpbox,
                       Hhelpback,
                       Hok,
                       -1,
                       l, -1,
                       -1,
                       ListDl.dummyInLoop,
                       specials,
                       0,
                       ListDl.dummyCheckExit,
                       setAndGetHelpValues,
                       setAndGetHelpValues,
                       v.a)
  THEN
  END;
END internalDoHelp;

PROCEDURE SendExtendedHelp (VAR str : ARRAY OF CHAR; VAR text: ARRAY OF CHAR): BOOLEAN;
  (* Request Help via VA_START from ST-GUIDE *)
  CONST theGuide    = 'ST-GUIDE';
        VA_START    = 4711H;
  
  VAR accId : INTEGER;
      vaMess: RECORD
               messId : INTEGER;
               apId   : INTEGER;
               over   : INTEGER;
               cmd    : ADDRESS;
               res    : LONGINT;
               res2   : INTEGER;
             END;
      helpApp: ARRAY [0..9] OF CHAR; 
BEGIN
  (* Search for ST-Guide Accessory/Applikation *)
  MagicStrings.Assign (theGuide, helpApp);
  accId := MagicAES.ApplFind (ADR(helpApp));
  IF accId < 0
  THEN
    (* ST-Guide nicht gefunden, interne Hilfe aufrufen *)
    RETURN FALSE
  END;
  
  (* Kommandozeile fr VA_START aufbauen *)
  (* hier kann ich das per Assign machen, da die internen
   * Hilfeberschriften nie den statischen Buffer berschreiten 
   * werden 
   *)
  MagicStrings.Assign ("*:\", memPtr^.allInOne);
  MagicStrings.Append (text, memPtr^.allInOne);
  MagicStrings.Append (" ", memPtr^.allInOne);
  MagicStrings.Append (str, memPtr^.allInOne);
  
  WITH vaMess DO
    messId := VA_START;
    apId   := mtAppl.ApplIdent;
    over   := 0;
    cmd    := ADR(memPtr^.allInOne);
    res    := 0;
    res2   := 0;
  END;
  MagicAES.ApplWrite (accId, 16, vaMess);

  MagicAES.WindUpdate(MagicAES.ENDUPDATE); (* Um es auch mal abzuschicken.. *)
  MagicAES.WindUpdate(MagicAES.BEGUPDATE);

  RETURN TRUE;
END SendExtendedHelp;

CONST cHyperTxt = "CatPutzHelpText";

PROCEDURE DoHelp (topic: tHelp);
  VAR externalHelp : BOOLEAN;
      helpText     : CatTypes.String255;
      helpTopic    : CatTypes.String255;
      voidList     : Lists.List;
BEGIN
  (* Hilfe an Hilfesystem schicken *)
  ConfVars.GetConfDefStr (cHyperTxt, helpText, "catputz.hyp");
  MagicStrings.Assign (cGuideHelpTable[topic], helpTopic);
  (* SendExtendedHelp liefert FALSE zurck, falls das Hilfesystem
   * ST-Guide nicht gefunden wurde
   *)
  externalHelp := SendExtendedHelp (helpTopic, helpText);

  (* Wenn der externe Aufruf nicht geklappt hat, dann wird hier
   * noch die interne Hilfe aufgerufen 
   *)
  IF ~externalHelp
  THEN
    internalDoHelp (topic, TRUE, voidList);
  END;
END DoHelp;


VAR hasMxAlloc  : BOOLEAN;
    i           : CARDINAL;
    tPtr        : CatTypes.Str255Ptr;

BEGIN
  hasMxAlloc := MagicDOS.Mxalloc ($FFFFFFFF,3) # ADDRESS(-32L);
  IF multiTask & hasMxAlloc
  THEN
    (* Mxalloc: Readable, TT-RAM preferred *)
    memPtr := MagicDOS.Mxalloc (TSIZE (memDescr), $43);
  ELSE
    ALLOCATE (memPtr, TSIZE (memDescr));
  END;
  IF memPtr = NIL THEN HALT END;
  tPtr := CADR (appName);
  FOR i := 0 TO 49 DO
    memPtr^.apName[i]   := tPtr^[i];
  END;
END PutzHelp.
